
stu void game_update( game_state_t* state, r32 delta_time, gl_state_t* gl, window_state_t* window ) {
    
    gl_batch_t* textured = gl->batches + batch_id_textured;
    gl_batch_t* rectangles = gl->batches + batch_id_rectangle;
    // state->chapter = chapter_5;
    
    if ( state->chapter == chapter_1 ) {
        
        if ( !state->initialized ) {
            state->c1.x = 1280 / 3;
            state->c1.y = 720 / 3;
            state->c1.scale = 1;
            state->c1.image_index = 0;
            state->c1.blink_time = random_range_r32( &state->seed, 1.0f, 5.0f );
            
            state->initialized = true;
        }
        
        if ( state->c1.image_index != 2 ) {
            if ( input_key_is_pressed( window, vk_left ) ) {
                state->c1.x -= ( ( state->c1.image_index + 1 ) * 200 * delta_time );
            } else if ( input_key_is_pressed( window, vk_right ) ) {
                state->c1.x += ( ( state->c1.image_index + 1 ) * 200 * delta_time );
            } else if ( input_key_is_pressed( window, vk_up ) ) {
                state->c1.scale += ( 1 * delta_time );
            } else if ( input_key_is_pressed( window, vk_down ) ) {
                state->c1.scale -= ( 1 * delta_time );
            }
        } else {
            r32 x_target = 640  + 32;
            r32 y_target = 360 - 280;
            
            state->c1.x += ( x_target - state->c1.x ) * delta_time * 0.5f;
            state->c1.y += ( y_target - state->c1.y  ) * delta_time * 0.5f;
            state->c1.scale += delta_time * 0.5f;
        }
        
        state->c1.x = math_clamp( 0, state->c1.x, ( r32 ) window->width );
        
        if ( state->c1.scale < 0.75f ) {
            state->c1.scale = 0.75f;
        } else if ( state->c1.scale < 1.5f ) {
            state->c1.image_index = 0;
        } else if ( state->c1.scale < 2.25f ) {
            state->c1.image_index = 1;
        } else if ( state->c1.scale < 7.0f ) {
            state->c1.image_index = 2;
        } else {
            state->chapter = chapter_2;
            state->initialized = false;
        }
        
        gl_texture_t* texture = g_textures + state->c1.image_index;
        
        rect2 mug_rect = r2_center_size( ( r32 ) state->c1.x, ( r32 ) state->c1.y, ( r32 ) texture->width * state->c1.scale, ( r32 ) texture->height * state->c1.scale );
        
        gl_batch_rectangle_textured( textured, mug_rect.e, 0, 0, white, texture->handle, texture->bytes_per_pixel );
        
        if ( state->c1.blink_time <= 0.0f ) {
            
            r32 half_height = window->height * 0.5f;
            r32 percentage = state->c1.blink_time / -0.1f;
            
            if ( state->c1.blink_time < -0.1f ) {
                percentage = 1.0f - ( ( state->c1.blink_time + 0.1f ) / -0.1f );
            }
            
            rect2 rect_top = r2_min_max( 0, ( r32) window->height - half_height * percentage, ( r32 ) window->width, ( r32 ) window->height );
            rect2 rect_bottom = r2_min_max( 0, 0, ( r32 ) window->width, half_height * percentage );
            
            gl_batch_rectangle( rectangles, rect_top.e, 0, 0, v4( 0.05f, 0.05f, 0.05f, 1.0f ) );
            gl_batch_rectangle( rectangles, rect_bottom.e, 0, 0, v4( 0.05f, 0.05f, 0.05f, 1.0f ) );
            
            if ( state->c1.blink_time <= -0.2f ) {
                state->c1.blink_time = random_range_r32( &state->seed, 1.0f, 5.0f );
            }
        }
        
        state->c1.blink_time -= delta_time;
        
    } else if ( state->chapter == chapter_2 ) {
        
        vec2 boat_hot_spot = v2( 115, 15 );
        vec2 boat_size = v2( 165, 25 );
        // vec2 sugar_hot_spot = v2( 64, 64 );
        vec2 sugar_size = v2( 45, 20 );
        
        if ( !state->initialized ) {
            
            state->c2.x = - 150;
            state->c2.y = 245;
            state->c2.time = 0;
            state->c2.next_sugar = 0;
            
            for ( umm i = 0; i < 5; i++ ) {
                state->c2.sugar_target[ i ] = random_range_r32( &state->seed, 170, 320 );
                state->c2.sugar_positions[ i ].x = random_range_r32( &state->seed, ( r32 ) ( 200 + ( i * 200 ) ), ( r32 ) ( 200 + ( ( i + 1 ) * 200 ) ) );
                state->c2.sugar_positions[ i ].y = state->c2.sugar_target[ i ];
                state->c2.next_sugar++;
                state->c2.time = 0.0f;
            }
            
            state->initialized = true;
        }
        
        if ( input_key_is_pressed( window, vk_up ) ) {
            state->c2.y += 30 * delta_time;
        } else if ( input_key_is_pressed( window, vk_down ) ) {
            state->c2.y -= 30 * delta_time;
        }
        
        state->c2.y = math_clamp( 170, state->c2.y, 320 );
        state->c2.x += delta_time * 50;
        
        rect2 boat_hit_box = r2_min_size( state->c2.x - ( boat_size.x / 2 ), state->c2.y, boat_size.x, boat_size.y );
        _assert( &boat_hit_box );
        
        state->c2.time += delta_time;
        
        if ( state->c2.time > 4.0f && state->c2.next_sugar < array_count( state->c2.sugar_target ) ) {
            umm i = state->c2.next_sugar;
            state->c2.sugar_target[ i ] = random_range_r32( &state->seed, 170, 320 );
            state->c2.sugar_positions[ i ].x = random_range_r32( &state->seed, state->c2.x + 300, state->c2.x + 350 );
            state->c2.sugar_positions[ i ].y = -50;
            state->c2.next_sugar++;
            state->c2.time = 0.0f;
            
            if ( state->c2.next_sugar == array_count( state->c2.sugar_target ) ) {
                state->c2.sugar_positions[ i ].x = state->c2.x + 20;
                state->c2.sugar_target[ i ] = state->c2.y;
            }
        }
        
        for ( umm i = 0; i < state->c2.next_sugar; i++ ) {
            if ( state->c2.sugar_target[ i ] && state->c2.sugar_positions[ i ].y < state->c2.sugar_target[ i ] ) {
                state->c2.sugar_positions[ i ].y += delta_time * 400;
                if ( state->c2.sugar_positions[ i ].y >= state->c2.sugar_target[ i ] ) {
                    state->c2.sugar_positions[ i ].y = state->c2.sugar_target[ i ];
                    state->c2.sugar_timer[ i ] = 0.0f;
                }
            }
        }
        
        gl_texture_t* boat = g_textures + texture_id_boat;
        gl_texture_t* sugar_under = g_textures + texture_id_sugar_under;
        gl_texture_t* sugar_over = g_textures + texture_id_sugar_over;
        gl_texture_t* sugar_wave_01 = g_textures + texture_id_sugar_wave_01;
        gl_texture_t* sugar_wave_02 = g_textures + texture_id_sugar_wave_02;
        
        for ( umm i = 0; i < state->c2.next_sugar; i++ ) {
            
            vec2 position = state->c2.sugar_positions[ i ];
            r32 target = state->c2.sugar_target[ i ];
            rect2 sugar_rect = r2_center_size( position.x, position.y, ( r32 ) sugar_under->width, ( r32 ) sugar_under->height );
            
            state->c2.sugar_timer[ i ] += delta_time;
            
            if ( position.y >= target ) {
                
                if ( position.y >= state->c2.y ) {
                    
                    r32 t = state->c2.sugar_timer[ i ];
                    
                    if ( t - ( r32 ) ( ( u32 ) t ) < 0.5f ) {
                        gl_batch_rectangle_textured( textured, sugar_rect.e, 0, 0, white, sugar_over->handle, sugar_under->bytes_per_pixel );
                        gl_batch_rectangle_textured( textured, sugar_rect.e, 0, 0, white, sugar_wave_01->handle, sugar_under->bytes_per_pixel );
                    } else {
                        r2_translate( &sugar_rect, 0, -1 );
                        gl_batch_rectangle_textured( textured, sugar_rect.e, 0, 0, white, sugar_over->handle, sugar_under->bytes_per_pixel );
                        r2_translate( &sugar_rect, 0, 1 );
                        gl_batch_rectangle_textured( textured, sugar_rect.e, 0, 0, white, sugar_wave_02->handle, sugar_under->bytes_per_pixel );
                    }
                }
                
            } else {
                gl_batch_rectangle_textured( textured, sugar_rect.e, 0, 0, white, sugar_under->handle, sugar_under->bytes_per_pixel );
            }
        }
        
        rect2 boat_rect = r2_min_size( state->c2.x - boat_hot_spot.x, state->c2.y - boat_hot_spot.y, ( r32 ) boat->width, ( r32 ) boat->height );
        gl_batch_rectangle_textured( textured, boat_rect.e, 0, 0, white, boat->handle, boat->bytes_per_pixel );
        
        for ( umm i = 0; i < state->c2.next_sugar; i++ ) {
            
            vec2 position = state->c2.sugar_positions[ i ];
            r32 target = state->c2.sugar_target[ i ];
            rect2 sugar_rect = r2_center_size( position.x, position.y, ( r32 ) sugar_under->width, ( r32 ) sugar_under->height );
            
            state->c2.sugar_timer[ i ] += delta_time;
            
            if ( position.y >= target ) {
                
                if ( position.y < state->c2.y ) {
                    r32 t = state->c2.sugar_timer[ i ];
                    
                    if ( t - ( r32 ) ( ( u32 ) t ) < 0.5f ) {
                        gl_batch_rectangle_textured( textured, sugar_rect.e, 0, 0, white, sugar_over->handle, sugar_under->bytes_per_pixel );
                        gl_batch_rectangle_textured( textured, sugar_rect.e, 0, 0, white, sugar_wave_01->handle, sugar_under->bytes_per_pixel );
                    } else {
                        r2_translate( &sugar_rect, 0, -1 );
                        gl_batch_rectangle_textured( textured, sugar_rect.e, 0, 0, white, sugar_over->handle, sugar_under->bytes_per_pixel );
                        r2_translate( &sugar_rect, 0, 1 );
                        gl_batch_rectangle_textured( textured, sugar_rect.e, 0, 0, white, sugar_wave_02->handle, sugar_under->bytes_per_pixel );
                    }
                }
            }
        }
        
        for ( umm i = 0; i < state->c2.next_sugar; i++ ) {
            
            vec2 position = state->c2.sugar_positions[ i ];
            
            if ( position.y >= state->c2.sugar_target[ i ] ) {
                
                rect2 sugar_hit_box = r2_center_size( position.x, position.y, sugar_size.x, sugar_size.y );
                
                if ( r2_overlap( boat_hit_box, sugar_hit_box ) ) {
                    
                    state->chapter = chapter_3;
                    state->initialized = false;
                    break;
                }
            }
        }
        
        if ( state->c2.x > ( r32 ) window->width ) {
            state->chapter = chapter_3;
            state->initialized = false;
        }
        
    } else if ( state->chapter == chapter_3 ) {
        
        if ( !state->initialized ) {
            state->c3.x = 640;
            state->c3.y = 450;
            state->c3.time = 0.0f;
            state->c3.scale = 8.0f;
            state->c3.started = false;
            state->initialized = true;
        }
        
        gl_texture_t* boat_01 = g_textures + texture_id_boat_64_01;
        gl_texture_t* boat_02 = g_textures + texture_id_boat_64_02;
        gl_texture_t* boat_03 = g_textures + texture_id_boat_64_03;
        gl_texture_t* boat_04 = g_textures + texture_id_boat_64_04;
        gl_texture_t* boat_05 = g_textures + texture_id_boat_64_05;
        
        state->c3.time += delta_time;
        gl_texture_t* to_display = boat_01;
        static r32 surface = 0;
        
        if ( state->c3.time < 1.5f ) {
            to_display = boat_01;
        } else if ( state->c3.time < 3.0f ) {
            to_display = boat_02;
        } else if ( state->c3.time < 4.5f ) {
            to_display = boat_03;
        } else if ( state->c3.time < 6.0f ) {
            to_display = boat_04;
            state->c3.y = 450 - 17 * state->c3.scale;
            surface = 450 - 30 * state->c3.scale;
        } else if ( state->c3.time < 12.0f ) {
            to_display = boat_05;
            state->c3.y -= delta_time * 50;
            rect2 surface_rect = r2_min_size( 300, surface, ( r32 ) window->width - 600, 1 * state->c3.scale );
            gl_batch_rectangle( rectangles, surface_rect.e, 0, 0, white );
        } else {
            
            to_display = boat_05;
            
            if ( !state->c3.started ) {
                state->c3.y = 720 - 32;
                state->c3.started = true;
                state->c3.particle_time = 0;
                state->c3.particle_count = 0;
                state->c3.particle_time_min = 0.25f;
                state->c3.particle_time_max = 0.5f;
                state->c3.switch_timer = 3.0f;
            }
            
            if ( state->c3.y > 100 ) {
                
                if ( input_key_is_pressed( window, vk_left ) ) {
                    state->c3.x -= delta_time * 10;
                } else if ( input_key_is_pressed( window, vk_right ) ) {
                    state->c3.x += delta_time * 10;
                }
                
                state->c3.y -= delta_time * 20;
                
                state->c3.x = math_clamp( 32 * state->c3.scale, state->c3.x, window->width - 32 * state->c3.scale );
                
            } else {
                state->c3.particle_time_min = 1.0f;
                state->c3.particle_time_max = 2.0f;
                state->c3.switch_timer -= delta_time;
                
                if ( state->c3.switch_timer <= 0.0f ) {
                    state->chapter = chapter_4;
                    state->initialized = false;
                }
            }
            
            state->c3.particle_time -= delta_time;
            
            if ( state->c3.particle_time <= 0.0f ) {
                
                if ( state->c3.particle_count < array_count( state->c3.particle_positions ) ) {
                    vec2 position;
                    position.x = state->c3.x + random_range_r32( &state->seed, -10 * state->c3.scale, 22 * state->c3.scale );
                    position.y = state->c3.y;
                    state->c3.particle_positions[ state->c3.particle_count ] = position;
                    state->c3.particle_count++;
                } else {
                    
                    for ( umm i = 0; i < array_count( state->c3.particle_positions ); i++ ) {
                        
                        vec2* p = state->c3.particle_positions + i;
                        
                        if ( p->y > window->height ) {
                            p->x = state->c3.x + random_range_r32( &state->seed, -10 * state->c3.scale, 22 * state->c3.scale );
                            p->y = state->c3.y;
                            break;
                        }
                    }
                }
                
                state->c3.particle_time = random_range_r32( &state->seed, state->c3.particle_time_min, state->c3.particle_time_max );
            }
            
            for ( umm i = 0; i < array_count( state->c3.particle_positions ); i++ ) {
                
                vec2* p = state->c3.particle_positions + i;
                
                if ( p->x > 0 ) {
                    p->y += delta_time * 150;
                    
                    rect2 particle_rect = r2_center_size( p->x, p->y, state->c3.scale, state->c3.scale );
                    gl_batch_rectangle( rectangles, particle_rect.e, 0, 0, white );
                }
            }
            
            rect2 surface_rect = r2_min_size( 0, 50, ( r32 ) window->width, 1 * state->c3.scale );
            gl_batch_rectangle( rectangles, surface_rect.e, 0, 0, white );
            rect2 ground_rect = r2_min_size( 0, 0, ( r32 ) window->width, 50 );
            gl_batch_rectangle( rectangles, ground_rect.e, 0, 0, black );
        }
        
        rect2 boat_rect = r2_center_size( state->c3.x, state->c3.y, ( r32 ) to_display->width * state->c3.scale, ( r32 ) to_display->height * state->c3.scale );
        
        gl_batch_rectangle_textured( textured, boat_rect.e, 0, 0, white, to_display->handle, to_display->bytes_per_pixel );
        
    } else if ( state->chapter == chapter_4 ) {
        
        if ( !state->initialized ) {
            state->c4.x = 640;
            state->c4.y = 360;
            state->c4.scale = 8;
            state->c4.target_scale = 20;
            state->c4.target_x = state->c4.x - 24 * state->c4.target_scale;
            state->c4.target_y = state->c4.y - 8 * state->c4.target_scale;
            state->c4.blink_time = 1.0f;
            
            state->initialized = true;
        }
        
        if ( state->c4.x - state->c4.target_x > 15.f ) {
            state->c4.x += ( state->c4.target_x - state->c4.x ) * delta_time * 0.5f;
            state->c4.y += ( state->c4.target_y - state->c4.y ) * delta_time * 0.5f;
            state->c4.scale += ( state->c4.target_scale - state->c4.scale ) * delta_time * 0.5f;
        } else {
            
            state->c4.blink_time -= delta_time;
            
            if ( state->c4.blink_time <= 0.0f ) {
                rect2 eye_rect = r2_min_size( state->c4.x + 24 * state->c4.scale, state->c4.y + 10 * state->c4.scale, state->c4.scale, state->c4.scale );
                
                if ( state->c4.blink_time >= -2.0f || state->c4.blink_time < -2.2f ) {
                    gl_batch_rectangle( rectangles, eye_rect.e, 0, 0, white );
                }
                
                if ( state->c4.blink_time < -4.0f ) {
                    state->initialized = false;
                    state->chapter = chapter_5;
                }
            }
        }
        
        gl_texture_t* boat = g_textures + texture_id_boat_64_05;
        rect2 boat_rect = r2_center_size( state->c4.x, state->c4.y, ( r32 ) boat->width * state->c4.scale, ( r32 ) boat->width * state->c4.scale );
        
        gl_batch_rectangle_textured( textured, boat_rect.e, 0, 0, white, boat->handle, boat->bytes_per_pixel );
        
        rect2 surface_rect = r2_min_size( 0, state->c4.y - 7 * state->c4.scale, ( r32 ) window->width, state->c4.scale );
        gl_batch_rectangle( rectangles, surface_rect.e, 0, 0, white );
        rect2 ground_rect = r2_min_size( 0, 0, ( r32 ) window->width, state->c4.y - 7 * state->c4.scale );
        gl_batch_rectangle( rectangles, ground_rect.e, 0, 0, black );
        
    } else if ( state->chapter == chapter_5 ) {
        
        if ( !state->initialized ) {
            
            g_render_as_ascii = true;
            state->c5.x = 95;
            state->c5.y = 0;
            state->c5.use_left = false;
            state->c5.screen_index = 0;
            
            state->initialized = true;
        }
        
        b32 moving = false;
        
        if ( input_key_is_pressed( window, vk_left ) ) {
            state->c5.x -= delta_time * 300;
            state->c5.use_left = true;
            moving = true;
        } else if ( input_key_is_pressed( window, vk_right ) ) {
            state->c5.x += delta_time * 300;
            state->c5.use_left = false;
            moving = true;
        }
        
        if ( input_key_is_pressed( window, vk_up ) ) {
            state->c5.y += delta_time * 300;
        } else if ( input_key_is_pressed( window, vk_down ) ) {
            state->c5.y -= delta_time * 300;
        }
        
        state->c5.y = math_clamp( 0, state->c5.y, 720 - 256 );
        
        gl_texture_t* boat = g_textures + texture_id_boat_ascii;
        gl_texture_t* idle_right = g_textures + texture_id_squid_idle_ascii_right;
        gl_texture_t* idle_left = g_textures + texture_id_squid_idle_ascii_left;
        gl_texture_t* move_left = g_textures + texture_id_squid_move_ascii_left;
        gl_texture_t* move_right = g_textures + texture_id_squid_move_ascii_right;
        gl_texture_t* mug = g_textures + texture_id_mug_01;
        
        if ( state->c5.screen_index == 0 ) {
            
            render_texture_rgba( software_framebuffer.bytes, window->width, window->height, window->width,
                                boat->pixels, boat->width, boat->height, boat->width,
                                0, 0 );
            
            if ( state->c5.just_switched ) {
                state->c5.just_switched = false;
                state->c5.x = -256;
            }
            
            if ( state->c5.x > 1280 - 256 ) {
                state->c5.x = 1280 - 256;
            }
            
            if ( state->c5.x < -256 ) {
                state->c5.screen_index = 1;
                state->c5.just_switched = true;
            }
            
        } else if ( state->c5.screen_index == 1 ) {
            
            if ( state->c5.just_switched ) {
                
                state->c5.just_switched = false;
                
                if ( state->c5.x > 1000 ) {
                    state->c5.x = -255;
                } else {
                    state->c5.x = 1280 - 256;
                }
            }
            
            if ( state->c5.x > 1280 - 256 ) {
                state->c5.screen_index = 0;
                state->c5.just_switched = true;
            }
            
            if ( state->c5.x < -256 ) {
                state->c5.screen_index = 2;
                state->c5.just_switched = true;
            }
            
        } else if ( state->c5.screen_index == 2 ) {
            
            render_texture_rgba( software_framebuffer.bytes, window->width, window->height, window->width,
                                mug->pixels, mug->width, mug->height, mug->width,
                                0, 0 );
            
            if ( state->c5.just_switched ) {
                state->c5.just_switched = false;
                state->c5.x = 1280 - 256;
            }
            
            if ( state->c5.x > 1280 - 256 ) {
                state->c5.screen_index = 1;
                state->c5.just_switched = true;
            }
            
            if ( state->c5.x < -256 ) {
                state->c5.x = -256;
            }
            
            if ( state->c5.x < 640 ) {
                state->chapter = chapter_1;
                g_render_as_ascii = false;
                state->initialized = false;
            }
        }
        
        gl_texture_t* squid = idle_right;
        
        if ( moving ) {
            
            if ( state->c5.use_left ) {
                squid = move_left;
            } else {
                squid = move_right;
            }
        } else {
            
            if ( state->c5.use_left ) {
                squid = idle_left;
            } else {
                squid = idle_right;
            }
        }
        
        render_texture_rgba( software_framebuffer.bytes, window->width, window->height, window->width,
                            squid->pixels, squid->width, squid->height, squid->width,
                            ( s32 ) state->c5.x, ( s32 ) state->c5.y );
    }
}