
#ifndef FONT_H

#define DEBUG_FONT_BUILD_ATLAS 0

#if DEBUG_FONT_BUILD_ATLAS
#include "../lib/tga.h"
#endif

#define STBRP_ASSERT( x ) _assert( x )
#define STB_RECT_PACK_IMPLEMENTATION
#include "../lib/stb_rect_pack.h"

#define STBTT_assert( x ) _assert( x )
#define STB_TRUETYPE_IMPLEMENTATION
#include "../lib/stb_truetype.h"

// NOTE simon: ISO/IEC_8859-15 + 0xfffd + 0x2026.
// NOTE simon: U+FFFD is the codepoint for the replacement character.
// NOTE simon: U+2026 is horizontal elipsis.
u32 font_codepoint_set[ ] = {
    
    0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
    0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
    0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
    0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
    0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e,
    0xa0, 0xa1, 0xa2, 0xa3, 0xa5, 0xa7, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
    0xb0, 0xb1, 0xb2, 0xb3, 0xb5, 0xb6, 0xb7, 0xb9, 0xba, 0xbb, 0xbf,
    0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
    0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
    0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
    0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff,
    
    0x20ac, 0x160, 0x161, 0x17d, 0x17e, 0x152, 0x153, 0x178, 0xfffd, 0x2026
};

u32 font_replacement_character = 0xfffd;
u32 font_replacement_character_index = array_count( font_codepoint_set ) - 2;

typedef struct codepoint_metrics_t {
    
    u16 u0, v0, u1, v1;
    r32 x0, y0, x1, y1;
    r32 advance;
    u32 kerning_count;
    u32 kerning_offset;
} codepoint_metrics_t;

#define codepoint_metrics_t_alignment 4

typedef struct font_t {
    
    stbtt_fontinfo info;
    u32* codepoint_set;
    codepoint_metrics_t* cp_metrics;
    s32* kerning;
    u32 codepoint_count;
    r32 size;
    r32 scale;
    s32 ascent;
    s32 descent;
    s32 line_gap;
    u8 cache[ 256 ];
    u32 cache_offset;
    u32 first_index_out_of_cache;
    b32 is_monospaced;
} font_t;

#define font_t_alignment 8

typedef struct font_collection_t {
    
    data_t memory;
    data_t atlas;
    font_t* fonts;
    u16 atlas_width;
    u16 atlas_height;
    u8 font_count;
    
} font_collection_t;

#define font_collection_t_alignment sizeof( umm )

typedef struct font_metrics_t {
    font_t* font;
    u32 name;
    r32 line_height;
    r32 baseline;
    r32 top_to_baseline;
} font_metrics_t;

stu void font_build_atlas( font_collection_t* collection ) {
    
    collection->atlas_width = 256;
    collection->atlas_height = 256;
    
    b32 success = false;
    
    while ( !success ) {
        
        collection->atlas.reserved = collection->atlas.used = collection->atlas_width * collection->atlas_height;
        collection->atlas.raw = memory_push_size( &collection->memory, collection->atlas.used, 1 );
        
        stbtt_pack_context context;
        stbtt_PackBegin( &context, collection->atlas.bytes, collection->atlas_width, collection->atlas_height, 0, 1, 0 );
        success = true;
        
        for ( umm font_index = 0; success && font_index < collection->font_count; font_index++ ) {
            
            font_t* f = collection->fonts + font_index;
            
            stbtt_packedchar char_datas[ 1024 ]; /* TODO simon: change this. */
            _assert( f->codepoint_count < array_count( char_datas ) );
            stbtt_pack_range range = { ( r32 ) f->size, 0, ( int* ) f->codepoint_set, ( int ) f->codepoint_count, 0, 0, 0 };
            range.chardata_for_range = char_datas;
            
            stbtt_PackSetOversampling( &context, 2, 2 );
            success = stbtt_PackFontRanges( &context, f->info.data, 0, &range, 1 );
            
            if ( success ) {
                
                for ( umm codepoint_index = 0; codepoint_index < f->codepoint_count; codepoint_index++ ) {
                    
                    stbtt_packedchar* packed_char = char_datas + codepoint_index;
                    codepoint_metrics_t* cm = f->cp_metrics + codepoint_index;
                    
                    cm->x0 = packed_char->xoff;
                    cm->y0 = -packed_char->yoff2;
                    cm->x1 = packed_char->xoff2;
                    cm->y1 = -packed_char->yoff;
                    
                    u16 height = packed_char->y1 - packed_char->y0;
                    
                    cm->u0 = packed_char->x0;
                    cm->v0 = collection->atlas_height - packed_char->y0 - height;
                    cm->u1 = packed_char->x1;
                    cm->v1 = cm->v0 + height;
                    cm->advance = packed_char->xadvance;
                }
            }
        }
        
        stbtt_PackEnd( &context );
        
        if ( !success ) {
            
            memory_pop_size( &collection->memory, collection->atlas.used );
            collection->atlas.reserved = collection->atlas.used = 0;
            
            if ( collection->atlas_width <= collection->atlas_height ) {
                collection->atlas_width *= 2;
            } else {
                collection->atlas_height *= 2;
            }
        }
    }
    
    u8* line_up = collection->atlas.bytes + collection->atlas_width * ( collection->atlas_height - 1 );
    u8* line_down = collection->atlas.bytes;
    
    for ( umm y = 0; y < collection->atlas_height / 2; y++ ) {
        
        u32* word_up = ( u32* ) line_up;
        u32* word_down = ( u32* ) line_down;
        
        for ( umm x = 0; x < collection->atlas_width / sizeof( u32 ); x++ ) {
            
            u32 temp = *word_up;
            *word_up = *word_down;
            *word_down = temp;
            word_up++;
            word_down++;
        }
        
        line_up -= collection->atlas_width;
        line_down += collection->atlas_width;
    }
    
#if DEBUG_FONT_BUILD_ATLAS
    u8 header[ 18 ] = { 0 };
    tga_set_image_type( header, tga_image_type_black_and_white );
    tga_set_width( header, ( u16 ) collection->atlas_width );
    tga_set_height( header, ( u16 ) collection->atlas_height );
    tga_set_pixel_depth( header, 8 );
    u32 error;
    file_handle_t h = file_open( string_arg_l( "test.tga" ), file_mode_write, true, &error );
    file_write_handle( h, 0, 18, header, &error );
    file_write_handle( h, 18, collection->atlas.used, collection->atlas.bytes, &error );
    file_close( h, &error );
#endif
    
}

stu font_t* font_collection_get_font( font_collection_t* collection, umm font_id ) {
    font_t* font = collection->fonts + font_id;
    return font;
}

stu void font_add_font( font_collection_t* collection, data_t* file, r32 size, u32* codepoint_set, u32 codepoint_count, b32 is_monospaced ) {
    
    _assert( collection );
    _assert( file );
    _assert( codepoint_set );
    
    font_t* f = memory_push_struct( &collection->memory, font_t );
    
    if ( !collection->fonts ) {
        collection->fonts = f;
    }
    
    collection->font_count++;
    
    f->codepoint_count = codepoint_count;
    f->codepoint_set = codepoint_set;
    f->size = size;
    f->is_monospaced = is_monospaced;
    
    s32 offset = stbtt_GetFontOffsetForIndex( file->bytes, 0 );
    stbtt_InitFont( &f->info, file->bytes, offset );
    f->scale = stbtt_ScaleForPixelHeight( &f->info, size );
    stbtt_GetFontVMetrics( &f->info, &f->ascent, &f->descent, &f->line_gap );
}

stu void font_set_codepoints( font_collection_t* collection, u32* codepoint_set, u32 codepoint_count ) {
    
    for ( umm font_index = 0; font_index < collection->font_count; font_index++ ) {
        font_t* f = collection->fonts + font_index;
        f->codepoint_count = codepoint_count;
        f->codepoint_set = codepoint_set;
    }
}

stu void font_build_cache( font_t* font ) {
    
    for ( umm i = 0; i < font->codepoint_count - 1; i++ ) {
        
        for ( umm j = i; j < font->codepoint_count; j++ ) {
            
            if ( font->codepoint_set[ i ] > font->codepoint_set[ j ] ) {
                swap_type( font->codepoint_set[ i ], font->codepoint_set[ j ], u32 );
            }
        }
    }
    
    memory_zero( font->cache, sizeof( font->cache )  );
    
    font->cache_offset = font->codepoint_set[ 0 ];
    font->first_index_out_of_cache = 0;
    
    for ( umm i = 0; i < font->codepoint_count; i++ ) {
        
        umm index = font->codepoint_set[ i ] - font->cache_offset;
        
        if ( index < array_count( font->cache ) ) {
            font->cache[ index ] = ( u8 ) i;
        } else {
            font->first_index_out_of_cache = ( u32 ) i;
            break;
        }
    }
}

stu void font_build_collection( font_collection_t* collection ) {
    
    for ( umm font_index = 0; font_index < collection->font_count; font_index++ ) {
        
        font_t* f = collection->fonts + font_index;
        f->cp_metrics = memory_push_array( &collection->memory, codepoint_metrics_t, f->codepoint_count );
        f->kerning = memory_current_aligned( &collection->memory, s32, sizeof( s32 ) );
        
        font_build_cache( f );
        
        for ( umm codepoint_index = 0; codepoint_index < f->codepoint_count; codepoint_index++ ) {
            
            u32 codepoint = f->codepoint_set[ codepoint_index ];
            codepoint_metrics_t* cm = f->cp_metrics + codepoint_index;
            memory_zero_struct( cm );
            s32 left, right, bottom, top;
            stbtt_GetCodepointBitmapBox( &f->info, codepoint, f->scale, f->scale, &left, &bottom, &right, &top );
            cm->x0 = ( r32 ) left;
            cm->y0 = ( r32 ) bottom;
            cm->x1 = ( r32 ) right;
            cm->y1 = ( r32 ) top;
            
            for ( umm kerning_index = 0; kerning_index < f->codepoint_count; kerning_index++ ) {
                
                s32 kerning_advance = stbtt_GetCodepointKernAdvance( &f->info, codepoint, f->codepoint_set[ kerning_index ] );
                
                if ( kerning_advance ) {
                    s32* kerning = memory_push_array_aligned( &collection->memory, s32, 2, sizeof( s32 ) );
                    kerning[ 0 ] = f->codepoint_set[ kerning_index ];
                    kerning[ 1 ] = kerning_advance;
                    cm->kerning_count++;
                }
            }
        }
    }
    
    font_build_atlas( collection );
}

/* TODO simon: better cache by building it up with things that are realy used. */
stu u32 font_get_codepoint_metrics_index( font_t* font, u32 codepoint ) {
    
    u32 result;
    u32 cache_index = codepoint - font->cache_offset;
    
    if ( cache_index < array_count( font->cache ) ) {
        result = font->cache[ cache_index ];
    } else {
        
        result = font->codepoint_count;
        
        smm left = font->first_index_out_of_cache;
        smm right = font->codepoint_count - 1;
        
        while ( left <= right ) {
            
            smm midle = ( left + right ) / 2;
            
            if ( font->codepoint_set[ midle ] < codepoint ) {
                left = midle + 1;
            } else if ( font->codepoint_set[ midle ] > codepoint ) {
                right = midle - 1;
            } else {
                result = ( u32 ) midle;
                break;
            }
        }
        
        if ( left > right ) {
            result = font_replacement_character_index;
        }
    }
    
    return result;
}

stu s32 font_get_kerning( font_t* f, codepoint_metrics_t* cm, u32 codepoint ) {
    
    s32 result = 0;
    
    if ( cm->kerning_count > 0 ) {
        
        s32* kerning = ( s32* ) ( ( ( u8* ) f->kerning ) + cm->kerning_offset );
        
        for ( u32 i = 0; i < cm->kerning_count; i++ ) {
            
            u32 kerning_codepoint_offset = i * 2;
            
            if ( ( u32 ) kerning[ kerning_codepoint_offset ] == codepoint ) {
                result = kerning[ kerning_codepoint_offset + 1 ];
                break;
            }
        }
    }
    
    return result;
}

stu r32 font_get_line_height( font_t* font, b32 include_line_gap ) {
    
    r32 height = ( -font->descent + ( ( include_line_gap ) ? font->line_gap : 0 ) + font->ascent ) * font->scale;
    return height;
}

stu r32 font_get_baseline( font_t* font ) {
    
    r32 result = -font->descent * font->scale;
    return result;
}

font_metrics_t font_get_metrics( font_collection_t* collection, u32 font_name ) {
    
    _assert( collection );
    font_metrics_t result = { 0 };
    
    if ( font_name < collection->font_count ) {
        result.font = collection->fonts + font_name;
        result.name = font_name;
        result.line_height = font_get_line_height( result.font, false );
        result.baseline = font_get_baseline( result.font );
        result.top_to_baseline = result.line_height - result.baseline;
    }
    
    return result;
}

#define font_first_character_offset( font, cm ) ( ( ( font )->is_monospaced ) ? 0 : ( cm )->x0 )

#define FONT_H

#endif
