
#ifndef GL_HARDWARE_H

#ifdef GL_SOFTWARE_H
# error "GL_HARDWARE_H defined while GL_SOFTWARE_H is already defined."
#endif

#include "gl.h"

#include "glcorearb.h"

#ifdef WINDOWS

PFNGLGETSTRINGPROC glGetString;
PFNGLGETINTEGERVPROC glGetIntegerv;
PFNGLGETERRORPROC glGetError;
PFNGLCLEARCOLORPROC glClearColor;
PFNGLGENTEXTURESPROC glGenTextures;
PFNGLBINDTEXTUREPROC glBindTexture;
PFNGLTEXPARAMETERIPROC glTexParameteri;
PFNGLTEXIMAGE2DPROC glTexImage2D;
PFNGLTEXSUBIMAGE2DPROC glTexSubImage2D;
PFNGLCLEARPROC glClear;
PFNGLDRAWARRAYSPROC glDrawArrays;
PFNGLFLUSHPROC glFlush;
PFNGLPIXELSTOREIPROC glPixelStorei;
PFNGLDELETETEXTURESPROC glDeleteTextures;
PFNGLFINISHPROC glFinish;
PFNGLENABLEPROC glEnable;
PFNGLBLENDFUNCPROC glBlendFunc;
PFNGLDRAWELEMENTSPROC glDrawElements;
PFNGLDISABLEPROC glDisable;
PFNGLVIEWPORTPROC glViewport;
PFNGLISENABLEDPROC glIsEnabled;
PFNGLACTIVETEXTUREPROC glActiveTexture;
PFNGLSCISSORPROC glScissor;
PFNGLREADPIXELSPROC glReadPixels;

#endif

PFNGLGETSHADERIVPROC glGetShaderiv;
PFNGLGETSHADERINFOLOGPROC glGetShaderInfoLog;
PFNGLGETSHADERSOURCEPROC glGetShaderSource;
PFNGLATTACHSHADERPROC glAttachShader;
PFNGLBINDATTRIBLOCATIONPROC glBindAttribLocation;
PFNGLCOMPILESHADERPROC glCompileShader;
PFNGLCREATEPROGRAMPROC glCreateProgram;
PFNGLCREATESHADERPROC glCreateShader;
PFNGLDELETEPROGRAMPROC glDeleteProgram;
PFNGLDELETESHADERPROC glDeleteShader;
PFNGLDETACHSHADERPROC glDetachShader;
PFNGLDISABLEVERTEXATTRIBARRAYPROC glDisableVertexAttribArray;
PFNGLENABLEVERTEXATTRIBARRAYPROC glEnableVertexAttribArray;
PFNGLGETPROGRAMIVPROC glGetProgramiv;
PFNGLGETPROGRAMINFOLOGPROC glGetProgramInfoLog;
PFNGLLINKPROGRAMPROC glLinkProgram;
PFNGLSHADERSOURCEPROC glShaderSource;
PFNGLUSEPROGRAMPROC glUseProgram;
PFNGLBINDBUFFERPROC glBindBuffer;
PFNGLGENBUFFERSPROC glGenBuffers;
PFNGLBUFFERDATAPROC glBufferData;
PFNGLBUFFERSUBDATAPROC glBufferSubData;
PFNGLVERTEXATTRIBPOINTERPROC glVertexAttribPointer;
PFNGLVERTEXATTRIBIPOINTERPROC glVertexAttribIPointer;
PFNGLBINDFRAGDATALOCATIONPROC glBindFragDataLocation;
PFNGLGETATTRIBLOCATIONPROC glGetAttribLocation;
PFNGLDELETEBUFFERSPROC glDeleteBuffers;
PFNGLDELETEVERTEXARRAYSPROC glDeleteVertexArrays;
PFNGLGENVERTEXARRAYSPROC glGenVertexArrays;
PFNGLBINDVERTEXARRAYPROC glBindVertexArray;
PFNGLGENERATEMIPMAPPROC glGenerateMipmap;
PFNGLGETSTRINGIPROC glGetStringi;
PFNGLGETUNIFORMLOCATIONPROC glGetUniformLocation;
PFNGLUNIFORM1UIPROC glUniform1ui;
PFNGLUNIFORMMATRIX2X3FVPROC glUniformMatrix2x3fv;
PFNGLUNIFORMMATRIX3X4FVPROC glUniformMatrix3x4fv;
PFNGLUNIFORMMATRIX4FVPROC glUniformMatrix4fv;
PFNGLUNIFORM4FVPROC glUniform4fv;
PFNGLPRIMITIVERESTARTINDEXPROC glPrimitiveRestartIndex;

#if DEBUG
PFNGLDEBUGMESSAGECALLBACKARBPROC glDebugMessageCallbackARB;
#endif

stu void gl_handle_error( void* message );

#ifdef DEBUG
#define gl_check( x ) do { \
x; \
GLenum __open_gl_error__ = glGetError( ); \
_assert( __open_gl_error__ == GL_NO_ERROR ); \
} while ( 0 )

#else
#define gl_check( x ) x
#endif

#ifdef DEBUG
stu void gl_debug_callback( GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const GLchar *message, const void* user ) {
    
    log_s( log_name_gl, ( void* ) message );
    log_l( log_name_gl, "\n" );
    
    if ( severity == GL_DEBUG_SEVERITY_LOW_ARB || severity == GL_DEBUG_SEVERITY_MEDIUM_ARB || severity == GL_DEBUG_SEVERITY_HIGH_ARB ){
        debug_break( );
    }
}
#endif

#if WINDOWS
// TODO simon: retrieve wglGetProcAddress... from opengl32.dll manually ?
// #include <Wingdi.h> // NOTE simon: wglGetProcAddress, wglCreateContext, wglDeleteContext, wglMakeCurrent. Allready included by windows.h

stu void gl_get_functions( ) {
    
    HMODULE dll = LoadLibrary( "opengl32.dll" );
    
    glGetString = ( PFNGLGETSTRINGPROC ) GetProcAddress( dll, "glGetString" );
    glGetIntegerv = ( PFNGLGETINTEGERVPROC ) GetProcAddress( dll, "glGetIntegerv" );
    glGetError = ( PFNGLGETERRORPROC ) GetProcAddress( dll, "glGetError" );
    glClearColor = ( PFNGLCLEARCOLORPROC ) GetProcAddress( dll, "glClearColor" );
    glGenTextures = ( PFNGLGENTEXTURESPROC ) GetProcAddress( dll, "glGenTextures" );
    glBindTexture = ( PFNGLBINDTEXTUREPROC ) GetProcAddress( dll, "glBindTexture" );
    glTexParameteri = ( PFNGLTEXPARAMETERIPROC ) GetProcAddress( dll, "glTexParameteri" );
    glTexImage2D = ( PFNGLTEXIMAGE2DPROC ) GetProcAddress( dll, "glTexImage2D" );
    glTexSubImage2D = ( PFNGLTEXSUBIMAGE2DPROC ) GetProcAddress( dll, "glTexSubImage2D" );
    glClear = ( PFNGLCLEARPROC ) GetProcAddress( dll, "glClear" );
    glDrawArrays = ( PFNGLDRAWARRAYSPROC ) GetProcAddress( dll, "glDrawArrays" );
    glFlush = ( PFNGLFLUSHPROC ) GetProcAddress( dll, "glFlush" );
    glPixelStorei = ( PFNGLPIXELSTOREIPROC ) GetProcAddress( dll, "glPixelStorei" );
    glDeleteTextures = ( PFNGLDELETETEXTURESPROC ) GetProcAddress( dll, "glDeleteTextures" );
    glFinish = ( PFNGLFINISHPROC ) GetProcAddress( dll, "glFinish" );
    glEnable = ( PFNGLENABLEPROC ) GetProcAddress( dll, "glEnable" );
    glBlendFunc = ( PFNGLBLENDFUNCPROC ) GetProcAddress( dll, "glBlendFunc" );
    glDrawElements = ( PFNGLDRAWELEMENTSPROC ) GetProcAddress( dll, "glDrawElements" );
    glDisable = ( PFNGLDISABLEPROC ) GetProcAddress( dll, "glDisable" );
    glViewport = ( PFNGLVIEWPORTPROC ) GetProcAddress( dll, "glViewport" );
    glScissor = ( PFNGLSCISSORPROC ) GetProcAddress( dll, "glScissor" );
    glReadPixels = ( PFNGLREADPIXELSPROC ) GetProcAddress( dll, "glReadPixels" );
    glIsEnabled = ( PFNGLISENABLEDPROC ) GetProcAddress( dll, "glIsEnabled" );
    glActiveTexture = ( PFNGLACTIVETEXTUREPROC ) wglGetProcAddress( "glActiveTexture" );
    
    FreeLibrary( dll );
    
    glGetShaderiv = ( PFNGLGETSHADERIVPROC ) wglGetProcAddress( "glGetShaderiv" );
    glGetShaderInfoLog = ( PFNGLGETSHADERINFOLOGPROC ) wglGetProcAddress( "glGetShaderInfoLog" );
    glGetShaderSource = ( PFNGLGETSHADERSOURCEPROC ) wglGetProcAddress( "glGetShaderSource" );
    glAttachShader = ( PFNGLATTACHSHADERPROC ) wglGetProcAddress( "glAttachShader" );
    glBindAttribLocation = ( PFNGLBINDATTRIBLOCATIONPROC ) wglGetProcAddress( "glBindAttribLocation" );
    glCompileShader = ( PFNGLCOMPILESHADERPROC ) wglGetProcAddress( "glCompileShader" );
    glCreateProgram = ( PFNGLCREATEPROGRAMPROC ) wglGetProcAddress( "glCreateProgram" );
    glCreateShader = ( PFNGLCREATESHADERPROC ) wglGetProcAddress( "glCreateShader" );
    glDeleteProgram = ( PFNGLDELETEPROGRAMPROC ) wglGetProcAddress( "glDeleteProgram" );
    glDeleteShader = ( PFNGLDELETESHADERPROC ) wglGetProcAddress( "glDeleteShader" );
    glDetachShader = ( PFNGLDETACHSHADERPROC ) wglGetProcAddress( "glDetachShader" );
    glDisableVertexAttribArray = ( PFNGLDISABLEVERTEXATTRIBARRAYPROC ) wglGetProcAddress( "glDisableVertexAttribArray" );
    glEnableVertexAttribArray = ( PFNGLENABLEVERTEXATTRIBARRAYPROC ) wglGetProcAddress( "glEnableVertexAttribArray" );
    glGetProgramiv = ( PFNGLGETPROGRAMIVPROC ) wglGetProcAddress( "glGetProgramiv" );
    glGetProgramInfoLog = ( PFNGLGETPROGRAMINFOLOGPROC ) wglGetProcAddress( "glGetProgramInfoLog" );
    glLinkProgram = ( PFNGLLINKPROGRAMPROC ) wglGetProcAddress( "glLinkProgram" );
    glShaderSource = ( PFNGLSHADERSOURCEPROC ) wglGetProcAddress( "glShaderSource" );
    glUseProgram = ( PFNGLUSEPROGRAMPROC ) wglGetProcAddress( "glUseProgram" );
    glBindBuffer = ( PFNGLBINDBUFFERPROC ) wglGetProcAddress( "glBindBuffer" );
    glGenBuffers = ( PFNGLGENBUFFERSPROC ) wglGetProcAddress( "glGenBuffers" );
    glBufferData = ( PFNGLBUFFERDATAPROC ) wglGetProcAddress( "glBufferData" );
    glBufferSubData = ( PFNGLBUFFERSUBDATAPROC ) wglGetProcAddress( "glBufferSubData" );
    glVertexAttribPointer = ( PFNGLVERTEXATTRIBPOINTERPROC ) wglGetProcAddress( "glVertexAttribPointer" );
    glVertexAttribIPointer = ( PFNGLVERTEXATTRIBIPOINTERPROC ) wglGetProcAddress( "glVertexAttribIPointer" );
    glBindFragDataLocation = ( PFNGLBINDFRAGDATALOCATIONPROC ) wglGetProcAddress( "glBindFragDataLocation" );
    glGetAttribLocation = ( PFNGLGETATTRIBLOCATIONPROC ) wglGetProcAddress( "glGetAttribLocation" );
    glDeleteBuffers = ( PFNGLDELETEBUFFERSPROC ) wglGetProcAddress( "glDeleteBuffers" );
    glDeleteVertexArrays = ( PFNGLDELETEVERTEXARRAYSPROC ) wglGetProcAddress( "glDeleteVertexArrays" );
    glGenVertexArrays = ( PFNGLGENVERTEXARRAYSPROC ) wglGetProcAddress( "glGenVertexArrays" );
    glBindVertexArray = ( PFNGLBINDVERTEXARRAYPROC ) wglGetProcAddress( "glBindVertexArray" );
    glGenerateMipmap = ( PFNGLGENERATEMIPMAPPROC ) wglGetProcAddress( "glGenerateMipmap" );
    glGetStringi = ( PFNGLGETSTRINGIPROC ) wglGetProcAddress( "glGetStringi" );
    glGetUniformLocation = ( PFNGLGETUNIFORMLOCATIONPROC ) wglGetProcAddress( "glGetUniformLocation" );
    glUniform1ui = ( PFNGLUNIFORM1UIPROC ) wglGetProcAddress( "glUniform1ui" );
    glUniformMatrix2x3fv = ( PFNGLUNIFORMMATRIX2X3FVPROC ) wglGetProcAddress( "glUniformMatrix2x3fv" );
    glUniformMatrix3x4fv = ( PFNGLUNIFORMMATRIX3X4FVPROC ) wglGetProcAddress( "glUniformMatrix3x4fv" );
    glUniformMatrix4fv = ( PFNGLUNIFORMMATRIX4FVPROC ) wglGetProcAddress( "glUniformMatrix4fv" );
    glUniform4fv = ( PFNGLUNIFORM4FVPROC ) wglGetProcAddress( "glUniform4fv" );
    glPrimitiveRestartIndex = ( PFNGLPRIMITIVERESTARTINDEXPROC ) wglGetProcAddress( "glPrimitiveRestartIndex" );
    
#ifdef DEBUG
    glDebugMessageCallbackARB = ( PFNGLDEBUGMESSAGECALLBACKARBPROC ) wglGetProcAddress( "glDebugMessageCallbackARB" );
    gl_check( glDebugMessageCallbackARB( gl_debug_callback, NULL ) );
    gl_check( glEnable( GL_DEBUG_OUTPUT_SYNCHRONOUS_ARB ) );
#endif
}

#define WGL_NUMBER_PIXEL_FORMATS_ARB            0x2000
#define WGL_DRAW_TO_WINDOW_ARB                  0x2001
#define WGL_DRAW_TO_BITMAP_ARB                  0x2002
#define WGL_ACCELERATION_ARB                    0x2003
#define WGL_NEED_PALETTE_ARB                    0x2004
#define WGL_NEED_SYSTEM_PALETTE_ARB             0x2005
#define WGL_SWAP_LAYER_BUFFERS_ARB              0x2006
#define WGL_SWAP_METHOD_ARB                     0x2007
#define WGL_NUMBER_OVERLAYS_ARB                 0x2008
#define WGL_NUMBER_UNDERLAYS_ARB                0x2009
#define WGL_TRANSPARENT_ARB                     0x200A
#define WGL_TRANSPARENT_RED_VALUE_ARB           0x2037
#define WGL_TRANSPARENT_GREEN_VALUE_ARB         0x2038
#define WGL_TRANSPARENT_BLUE_VALUE_ARB          0x2039
#define WGL_TRANSPARENT_ALPHA_VALUE_ARB         0x203A
#define WGL_TRANSPARENT_INDEX_VALUE_ARB         0x203B
#define WGL_SHARE_DEPTH_ARB                     0x200C
#define WGL_SHARE_STENCIL_ARB                   0x200D
#define WGL_SHARE_ACCUM_ARB                     0x200E
#define WGL_SUPPORT_GDI_ARB                     0x200F
#define WGL_SUPPORT_OPENGL_ARB                  0x2010
#define WGL_DOUBLE_BUFFER_ARB                   0x2011
#define WGL_STEREO_ARB                          0x2012
#define WGL_PIXEL_TYPE_ARB                      0x2013
#define WGL_COLOR_BITS_ARB                      0x2014
#define WGL_RED_BITS_ARB                        0x2015
#define WGL_RED_SHIFT_ARB                       0x2016
#define WGL_GREEN_BITS_ARB                      0x2017
#define WGL_GREEN_SHIFT_ARB                     0x2018
#define WGL_BLUE_BITS_ARB                       0x2019
#define WGL_BLUE_SHIFT_ARB                      0x201A
#define WGL_ALPHA_BITS_ARB                      0x201B
#define WGL_ALPHA_SHIFT_ARB                     0x201C
#define WGL_ACCUM_BITS_ARB                      0x201D
#define WGL_ACCUM_RED_BITS_ARB                  0x201E
#define WGL_ACCUM_GREEN_BITS_ARB                0x201F
#define WGL_ACCUM_BLUE_BITS_ARB                 0x2020
#define WGL_ACCUM_ALPHA_BITS_ARB                0x2021
#define WGL_DEPTH_BITS_ARB                      0x2022
#define WGL_STENCIL_BITS_ARB                    0x2023
#define WGL_AUX_BUFFERS_ARB                     0x2024

#define WGL_NO_ACCELERATION_ARB                 0x2025
#define WGL_GENERIC_ACCELERATION_ARB            0x2026
#define WGL_FULL_ACCELERATION_ARB               0x2027

#define WGL_SWAP_EXCHANGE_ARB                   0x2028
#define WGL_SWAP_COPY_ARB                       0x2029
#define WGL_SWAP_UNDEFINED_ARB                  0x202A

#define WGL_TYPE_RGBA_ARB                       0x202B
#define WGL_TYPE_COLORINDEX_ARB                 0x202C

#define WGL_CONTEXT_MAJOR_VERSION_ARB           0x2091
#define WGL_CONTEXT_MINOR_VERSION_ARB           0x2092
#define WGL_CONTEXT_LAYER_PLANE_ARB             0x2093
#define WGL_CONTEXT_FLAGS_ARB                   0x2094
#define WGL_CONTEXT_PROFILE_MASK_ARB            0x9126

#define WGL_CONTEXT_DEBUG_BIT_ARB               0x0001
#define WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB  0x0002

#define WGL_CONTEXT_CORE_PROFILE_BIT_ARB        0x00000001
#define WGL_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB 0x00000002

#define ERROR_INVALID_VERSION_ARB               0x2095
#define ERROR_INVALID_PROFILE_ARB               0x2096

typedef BOOL ( * wglChoosePixelFormatARB_t )( HDC hdc, const int *piAttribIList, const FLOAT *pfAttribFList, UINT nMaxFormats, int *piFormats, UINT *nNumFormats );
typedef HGLRC ( * wglCreateContextAttribsARB_t )( HDC hDC, HGLRC hshareContext, const int *attribList );
typedef const char * ( * wglGetExtensionsStringARB_t )( HDC hdc );
typedef void ( * wglSwapIntervalEXT_t )( int intervalCount );

wglGetExtensionsStringARB_t wglGetExtensionsStringARB = 0;
wglChoosePixelFormatARB_t wglChoosePixelFormatARB = 0;
wglCreateContextAttribsARB_t wglCreateContextAttribsARB = 0;
wglSwapIntervalEXT_t wglSwapIntervalEXT = 0;

stu void gl_get_platform_functions( void ) {
    
    wglGetExtensionsStringARB = ( wglGetExtensionsStringARB_t ) wglGetProcAddress( "wglGetExtensionsStringARB" );
    wglChoosePixelFormatARB = ( wglChoosePixelFormatARB_t ) wglGetProcAddress( "wglChoosePixelFormatARB" );
    wglCreateContextAttribsARB = ( wglCreateContextAttribsARB_t ) wglGetProcAddress( "wglCreateContextAttribsARB" );
    wglSwapIntervalEXT = ( wglSwapIntervalEXT_t ) wglGetProcAddress( "wglSwapIntervalEXT" );
    
    _assert( ( void* ) wglGetExtensionsStringARB > ( void* ) 3 );
    _assert( ( void* ) wglChoosePixelFormatARB > ( void* ) 3 );
    _assert( ( void* ) wglCreateContextAttribsARB > ( void* ) 3 );
    _assert( ( void* ) wglSwapIntervalEXT > ( void* ) 3 );
}

#elif LINUX

stu void gl_get_functions( void ) {
    _assert( !"Not implemented." );
}

stu void gl_get_platform_functions( void ) {
    _assert( !"Not implemented." );
}

#else

# error Unsuported platorm.

#endif

stu void gl_log_infos( void ) {
    
    log_l( log_name_gl, "GL_VERSION: " );
    char* version = ( char* ) glGetString( GL_VERSION );
    log_s( log_name_gl, version );
    log_l( log_name_gl, "\n" );
    
    if ( ( version[ 0 ] == '3' && version[ 2 ] >= '2' ) || version[ 0 ] > '3' ) {
        
        s32 major, minor;
        glGetIntegerv( GL_MAJOR_VERSION, &major );
        glGetIntegerv( GL_MINOR_VERSION, &minor );
        log_l( log_name_gl, "GL_MAJOR_VERSION: " );
        log_s64( log_name_gl, major, 0, 0, 0 );
        log_l( log_name_gl, " \nGL_MINOR_VERSION: " );
        log_s64( log_name_gl, minor, 0, 0, 0 );
        log_l( log_name_gl, "\n" );
        
        s32 mask = 0;
        glGetIntegerv( GL_CONTEXT_PROFILE_MASK, &mask );
        log_l( log_name_gl, "GL_CONTEXT_PROFILE_MASK: " );
        
        if ( mask & GL_CONTEXT_CORE_PROFILE_BIT ) {
            log_l( log_name_gl, "GL_CONTEXT_CORE_PROFILE_BIT\n" );
        } else if ( mask & GL_CONTEXT_COMPATIBILITY_PROFILE_BIT ) {
            log_l( log_name_gl, "GL_CONTEXT_COMPATIBILITY_PROFILE_BIT\n" );
        }
        
        s32 flags = 0;
        glGetIntegerv( GL_CONTEXT_FLAGS, &flags );
        log_l( log_name_gl, "GL_CONTEXT_FLAGS: " );
        
        if ( flags & GL_CONTEXT_FLAG_FORWARD_COMPATIBLE_BIT ){
            log_l( log_name_gl, "GL_CONTEXT_FLAG_FORWARD_COMPATIBLE_BIT " );
        }
        
        if ( flags & GL_CONTEXT_FLAG_DEBUG_BIT ) {
            log_l( log_name_gl, "GL_CONTEXT_FLAG_DEBUG_BIT " );
        }
        
        log_l( log_name_gl, "\n\nExtensions ( >= 3.2 ):\n" );
        GLint extensionCount;
        glGetIntegerv( GL_NUM_EXTENSIONS, &extensionCount );
        
        for ( s32 extensionIndex = 0; extensionIndex < extensionCount; extensionIndex++ ) {
            log_s( log_name_gl, ( char* ) glGetStringi( GL_EXTENSIONS, extensionIndex ) );
            log_l( log_name_gl, "\n" );
        }
        
        log_l( log_name_gl, "\n" );
        
    } else {
        
        log_l( log_name_gl, "Extensions ( <= 3.1 ):\n" );
        log_s( log_name_gl, ( char * ) glGetString( GL_EXTENSIONS ) );
        log_l( log_name_gl, "\n\n" );
    }
}

stu void gl_handle_error( void* message ) {
    
#if DEBUG
    if ( message ) {
        log_s( log_name_gl, message );
        log_l( log_name_gl,  "\n" );
        log_increment_indent( log_name_gl );
    }
    
    GLuint error = glGetError( );
    
    switch ( error ) {
        
        case GL_NO_ERROR: {
            if ( message ) {
                log_indent( log_name_gl );
                log_l( log_name_gl, "No error.\n" );
            }
        } break;
        
        case GL_INVALID_ENUM: {
            log_indent( log_name_gl );
            log_l( log_name_gl, "GL_INVALID_ENUM\n" );
        } break;
        
        case GL_INVALID_VALUE: {
            log_indent( log_name_gl );
            log_l( log_name_gl, "GL_INVALID_VALUE\n" );
        } break;
        
        case GL_INVALID_OPERATION: {
            log_indent( log_name_gl );
            log_l( log_name_gl, "GL_INVALID_OPERATION\n" );
        } break;
        
        case GL_INVALID_FRAMEBUFFER_OPERATION: {
            log_indent( log_name_gl );
            log_l( log_name_gl, "GL_INVALID_FRAMEBUFFER_OPERATION\n" );
        } break;
        
        case GL_OUT_OF_MEMORY: {
            log_indent( log_name_gl );
            log_l( log_name_gl, "GL_OUT_OF_MEMORY\n" );
        } break;
        
        default: {
            log_indent( log_name_gl );
            log_l( log_name_gl, "Unknown error.\n" );
        }
    }
    
    if ( message != 0 ) {
        log_decrement_indent( log_name_gl );
    }
#endif
}

// MARK shader

stu void gl_handle_shader_error( GLuint* shader_handle ) {
    
    GLint compiled = 0;
    glGetShaderiv( *shader_handle, GL_COMPILE_STATUS, &compiled );
    
    if( compiled == GL_FALSE ) {
        
        log_l( log_name_gl, "Shader error:\n" );
        log_increment_indent( log_name_gl );
        
        log_indent( log_name_gl );
        log_l( log_name_gl, "Handle: " );
        log_u64( log_name_gl, *shader_handle, 0, 0, 0 );
        log_l( log_name_gl, "\n" );
        
        GLint type = 0;
        glGetShaderiv( *shader_handle, GL_SHADER_TYPE, &type );
        log_indent( log_name_gl );
        log_l( log_name_gl, "Type: " );
        log_l( log_name_gl, ( type == GL_VERTEX_SHADER ) ? "vertex shader" : ( ( type == GL_FRAGMENT_SHADER ) ? "fragment shader" : "geometry shader" ) );
        log_l( log_name_gl, "\n" );
        
        log_indent( log_name_gl );
        log_l( log_name_gl, "Error:\n" );
        
        {
            GLint length = 0;
            glGetShaderiv( *shader_handle, GL_INFO_LOG_LENGTH, &length );
            data_t log = log_get_buffer( log_name_gl, length );
            glGetShaderInfoLog( *shader_handle, length, 0, log.text );
            log.text[ length - 1 ] = '\n';
        }
        
        log_indent( log_name_gl );
        log_l( log_name_gl, "Source:\n" );
        
        {
            GLint source_length = 0;
            glGetShaderiv( *shader_handle, GL_SHADER_SOURCE_LENGTH, &source_length );
            data_t log = log_get_buffer( log_name_gl, source_length );
            glGetShaderSource( *shader_handle, source_length, 0, log.text );
            log.text[ source_length - 1 ] = '\n';
        }
        
        log_l( log_name_gl, "\n" );
        log_decrement_indent( log_name_gl );
        
        glDeleteShader( *shader_handle );
        *shader_handle = 0;
    }
}

stu void gl_handle_program_error( GLuint* program_handle ) {
    
    GLint linked = 0;
    glGetProgramiv( *program_handle, GL_LINK_STATUS, &linked );
    
    if( linked == GL_FALSE ) {
        
        log_s( log_name_gl, "Shader program error:\n" );
        log_increment_indent( log_name_gl );
        log_indent( log_name_gl );
        log_s( log_name_gl, "Handle: " );
        log_u64( log_name_gl, *program_handle, 0, 0, 0 );
        log_s( log_name_gl, "\nError:\n" );
        log_indent( log_name_gl );
        
        GLint length = 0;
        glGetProgramiv( *program_handle, GL_INFO_LOG_LENGTH, &length );
        data_t log = log_get_buffer( log_name_gl, length );
        glGetProgramInfoLog( *program_handle, length, &length, log.text );
        log.text[ length - 1 ] = '\n';
        
        log_decrement_indent( log_name_gl );
        
        glDeleteProgram( *program_handle );
        *program_handle = 0;
    }
}

stu u32 gl_create_shader_program( void* vertex_shader_code, void* fragment_shader_code ) {
    
    u32 result = { 0 };
    u32 vertex_handle = glCreateShader( GL_VERTEX_SHADER );
    
    u8* vs_sources[ ] = { ( u8* ) "#version 150\n", vs_lib, vertex_shader_code };
    glShaderSource( vertex_handle, array_count( vs_sources ), ( const char* const* ) vs_sources, 0 );
    glCompileShader( vertex_handle );
    gl_handle_shader_error( &vertex_handle );
    
    u32 fragment_handle = glCreateShader( GL_FRAGMENT_SHADER );
    
    u8* fs_sources[ ] = { ( u8* ) "#version 150\n", fragment_shader_code };
    glShaderSource( fragment_handle, array_count( fs_sources ), ( const char* const* ) fs_sources, 0 );
    glCompileShader( fragment_handle );
    gl_handle_shader_error( &fragment_handle );
    
    if ( vertex_handle != 0 && fragment_handle != 0 ) {
        
        result = glCreateProgram( );
        glAttachShader( result, vertex_handle );
        glAttachShader( result, fragment_handle );
        glBindFragDataLocation( result, 0, "fragment_out_color" );
        glLinkProgram( result );
        gl_handle_program_error( &result );
        
        if ( result == 0 ) {
            glDeleteShader( vertex_handle );
            vertex_handle = 0;
            glDeleteShader( fragment_handle );
            fragment_handle = 0;
        }
    }
    
    return result;
}

/* MARK texture */

/* NOTE simon: http://docs.gl/gl3/glIsTexture specifies that 0 is an invalid texture handle. */

typedef enum gl_texture_flag_t {
    
    gl_texture_flag_none = 0 << 0,
    
    gl_texture_flag_inverse_input_format = 1 << 0,
    gl_texture_flag_bilinear = 1 << 1,
    gl_texture_flag_generate_mipmaps = 1 << 2,
    
} gl_texture_flag_t;

/* NOTE simon: reminder that opengl function can't be called in a different thread that the one that created them. */
stu gl_texture_t gl_create_texture( u8* pixels, u32 width, u32 height, u8 bytes_per_pixel, u32 flags ) {
    
    b32 inverse_input_format = flag_is_set_u32( flags, gl_texture_flag_inverse_input_format );
    b32 bilinear = flag_is_set_u32( flags, gl_texture_flag_bilinear );
    b32 generate_mipmaps = flag_is_set_u32( flags, gl_texture_flag_generate_mipmaps );
    
    gl_texture_t result = { 0 };
    result.pixels = pixels;
    result.width = width;
    result.height = height;
    result.bytes_per_pixel = bytes_per_pixel;
    
    glGenTextures( 1, &result.handle );
    glBindTexture( GL_TEXTURE_2D, result.handle );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0 );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0 );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
    
    if ( bilinear ) {
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
        if ( generate_mipmaps ) {
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
        } else {
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        }
    } else {
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
    }
    
    u8 mipmap_level = 0;
    u32 internal_format = GL_RGBA;
    u32 format = GL_RGBA;
    u32 type = GL_UNSIGNED_BYTE;
    
    if ( bytes_per_pixel == 1 ) {
        
        internal_format = GL_RED;
        format = GL_RED;
        
    } else if ( bytes_per_pixel == 2 ) {
        
        internal_format = GL_RG;
        format = GL_RG;
        
    } else if ( bytes_per_pixel == 3 ) {
        
        internal_format = GL_RGB;
        
        if ( inverse_input_format ) {
            format = GL_BGR;
        } else {
            format = GL_RGB;
        }
        
    } else if ( inverse_input_format ) {
        
        format = GL_BGRA;
    }
    
    b32 change_alignment = ( ( width * bytes_per_pixel ) % 4 ) != 0;
    
    s32 alignment_backup = 4;
    
    if ( change_alignment ) {
        glGetIntegerv( GL_UNPACK_ALIGNMENT, &alignment_backup );
        glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
    }
    
    gl_check( glTexImage2D( GL_TEXTURE_2D, mipmap_level, internal_format, width, height, 0, format, type, pixels ) );
    
    if ( change_alignment ) {
        glPixelStorei( GL_UNPACK_ALIGNMENT, alignment_backup );
    }
    
    if ( generate_mipmaps ) {
        glGenerateMipmap( GL_TEXTURE_2D );
    }
    
    /* TODO simon: this is core 3.3 and I should probably use it for 1 channel textures.
https://hero.handmade.network/forums/code-discussion/t/7207-stb_truetype_generates_red_text#22260
    GLint swizzle[] = { GL_ONE, GL_ONE, GL_ONE, GL_RED };
    glTexParameteriv(GL_TEXTURE_2D, GL_TEXTURE_SWIZZLE_RGBA, swizzle);
    */
    
    return result;
}

stu void gl_update_texture_content( gl_texture_t* texture, u32 flags ) {
    
    profiler_event_start( gl_update_texture_content );
    
    b32 inverse_input_format = flag_is_set_u32( flags, gl_texture_flag_inverse_input_format );
    b32 generate_mipmaps = flag_is_set_u32( flags, gl_texture_flag_generate_mipmaps );
    
    _assert( texture->handle );
    glBindTexture( GL_TEXTURE_2D, texture->handle );
    
    u8 mipmap_level = 0;
    u32 format = GL_RGBA;
    u32 type = GL_UNSIGNED_BYTE;
    
    if ( texture->bytes_per_pixel == 1 ) {
        
        format = GL_RED;
        
    } else if ( texture->bytes_per_pixel == 2 ) {
        
        format = GL_RG;
        
    } else if ( texture->bytes_per_pixel == 3 ) {
        
        if ( inverse_input_format ) {
            format = GL_BGR;
        } else {
            format = GL_RGB;
        }
        
    } else if ( inverse_input_format ) {
        
        format = GL_BGRA;
    }
    
    b32 change_alignment = ( ( texture->width * texture->bytes_per_pixel ) % 4 ) != 0;
    
    s32 alignment_backup = 4;
    
    if ( change_alignment ) {
        glGetIntegerv( GL_UNPACK_ALIGNMENT, &alignment_backup );
        glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
    }
    
    gl_check( glTexSubImage2D( GL_TEXTURE_2D, mipmap_level, 0, 0, texture->width, texture->height, format, type, texture->pixels ); );
    
    if ( change_alignment ) {
        glPixelStorei( GL_UNPACK_ALIGNMENT, alignment_backup );
    }
    
    if ( generate_mipmaps ) {
        glGenerateMipmap( GL_TEXTURE_2D );
    }
    
    profiler_event_end( gl_update_texture_content );
}

stu void gl_free_texture( gl_texture_t* texture ) {
    _assert( texture->handle );
    glDeleteTextures( 1, &texture->handle );
    texture->handle = 0;
}

stu void internal_gl_do_scissor( gl_command_t* command ) {
    
    if ( command->scissor.enable ) {
        glEnable( GL_SCISSOR_TEST );
        glScissor( command->scissor.x, command->scissor.y, command->scissor.width, command->scissor.height );
    } else {
        glDisable( GL_SCISSOR_TEST );
    }
}

/* TODO simon: currently most of the index buffers are not needed (I think) (need to fill the vertex buffer in the correct order ?). Remove them. */
/* TODO simon: Try to make batches that share storage to have only one upload of data pervertex format.
It would allow to have one upload for several draw calls and allow to have batch -> scissor -> batch -> sicssor -> batch... */
/* TODO simon: make sure that if two scissor calls are one after the other to only send one ? Would it help ? Investigate. */
stu void gl_render( gl_command_buffer_t* command_buffer ) {
    
    gl_command_t* buffer_commands = ( gl_command_t* ) command_buffer->commands.bytes;
    
    for ( u32 global_command_index = 0; global_command_index < command_buffer->command_count; global_command_index++ ) {
        
        gl_command_t* buffer_command = buffer_commands + global_command_index;
        
        switch ( buffer_command->type ) {
            
            case gl_command_type_viewport: {
                
                glViewport( ( s32 ) buffer_command->viewport.position.x, ( s32 ) buffer_command->viewport.position.y, ( u32 ) buffer_command->viewport.size.x, ( u32 ) buffer_command->viewport.size.y );
                
            } break;
            
            case gl_command_type_clear: {
                
                vec4 color = v4_from_u32_color( buffer_command->clear.color );
                glClearColor( color.r, color.g, color.b, color.a );
                glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
                
            } break;
            
            case gl_command_type_scissor: {
                
                if ( buffer_command->scissor.enable ) {
                    glEnable( GL_SCISSOR_TEST );
                    glScissor( buffer_command->scissor.x, buffer_command->scissor.y, buffer_command->scissor.width, buffer_command->scissor.height );
                } else {
                    glDisable( GL_SCISSOR_TEST );
                }
                
            } break;
            
            case gl_command_type_batch: {
                
                gl_batch_t* batch = buffer_command->batch.batch;
                gl_command_t* batch_commands = ( gl_command_t* ) batch->commands.bytes;
                
                if ( batch->command_count > 0 ) {
                    
                    u8 position_component_count = gl_vertex_format_get_position_component_count( batch->vertex_format );
                    u8 has_color = gl_vertex_format_has_color( batch->vertex_format );
                    u8 has_uv = gl_vertex_format_has_uv( batch->vertex_format );
                    u8 index_size = gl_vertex_format_get_index_size_byte( batch->vertex_format );
                    
                    gl_check( glUseProgram( batch->shader_handle ) );
                    GLint matrix_location = glGetUniformLocation( batch->shader_handle, "projection" );
                    gl_check( glUniformMatrix4fv( matrix_location, 1, GL_FALSE, batch->matrix ) );
                    
                    GLint bytes_per_pixel_location = -1;
                    
                    /* TODO simon: not all shader using textures have a bytes_per_pixel uniform. What should I do ? */
                    if ( has_uv ) {
                        bytes_per_pixel_location = glGetUniformLocation( batch->shader_handle, "bytes_per_pixel" );
                        gl_check( glUniform1ui( bytes_per_pixel_location, 4 ) );
                    }
                    
                    if ( batch->vertex_handle == 0 ) {
                        gl_check( glGenBuffers( 1, &batch->vertex_handle ) );
                        gl_check( glGenBuffers( 1, &batch->index_handle ) );
                    }
                    
                    gl_check( glBindBuffer( GL_ARRAY_BUFFER, batch->vertex_handle ) );
                    gl_check( glBufferData( GL_ARRAY_BUFFER, batch->v_buffer_size, 0, GL_DYNAMIC_DRAW ) );
                    gl_check( glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, batch->index_handle ) );
                    gl_check( glBufferData( GL_ELEMENT_ARRAY_BUFFER, batch->i_buffer_size, 0, GL_DYNAMIC_DRAW ) );
                    
                    u32 index_type = GL_UNSIGNED_INT;
                    u32 primitive_restart_index = gl_primitive_restart_index_u32;
                    
                    if ( index_size == 1 ) {
                        index_type = GL_UNSIGNED_BYTE;
                        primitive_restart_index = gl_primitive_restart_index_u8;
                    } else if ( index_size == 2 ) {
                        index_type = GL_UNSIGNED_SHORT;
                        primitive_restart_index = gl_primitive_restart_index_u16;
                    }
                    
                    u32 vertex_format_size_byte = gl_vertex_format_compute_vertex_size_byte( batch->vertex_format );
                    umm vertex_advance = 0;
                    
                    u32 position = glGetAttribLocation( batch->shader_handle, "position" );
                    gl_check( glEnableVertexAttribArray( position ) );
                    gl_check( glVertexAttribPointer( position, position_component_count, GL_FLOAT, GL_FALSE, vertex_format_size_byte, ( void* ) vertex_advance ) );
                    vertex_advance += position_component_count * sizeof( r32 );
                    
                    if ( has_color ) {
                        
                        u32 color = glGetAttribLocation( batch->shader_handle, "color" );
                        gl_check( glEnableVertexAttribArray( color ) );
                        gl_check( glVertexAttribIPointer( color, 1, GL_UNSIGNED_INT, vertex_format_size_byte, ( void * ) vertex_advance ) );
                        vertex_advance += sizeof( u32 );
                    }
                    
                    if ( has_uv ) {
                        
                        u32 uvs = glGetAttribLocation( batch->shader_handle, "in_uv" );
                        gl_check( glEnableVertexAttribArray( uvs ) );
                        gl_check( glVertexAttribPointer( uvs, 2, GL_FLOAT, GL_FALSE, vertex_format_size_byte, ( void * ) vertex_advance ) );
                        /* vertex_advance += 2 * sizeof( r32 ); // NOTE simon: If another property is added, re-enable this. */
                    }
                    
                    switch ( batch->type ) {
                        
                        case gl_command_type_point: {
                            
                            umm vertex_format_size = vertex_format_size_byte / 4;
                            
                            for ( umm batch_command_index = 0; batch_command_index < batch->command_count; batch_command_index++ ) {
                                
                                gl_command_t* command = batch_commands + batch_command_index;
                                
                                if ( command->type == gl_command_type_point ) {
                                    
                                    data_t backup = command_buffer->datas;
                                    
                                    u32 v_array_size = vertex_format_size_byte * ( ( position_component_count == 2 ) ? 4 : 8 ) * command->point.point_count;
                                    u32 i_array_size = command->point.point_count * ( ( position_component_count == 2 ) ? 5 : 15 ) * index_size;
                                    
                                    r32* v_array = ( r32* ) memory_push_size( &command_buffer->datas, v_array_size, 4 );
                                    u32* c_array = ( u32* ) v_array;
                                    u32* c_input_array = ( u32* ) command->point.vertex_array;
                                    void* i_array = memory_push_size( &command_buffer->datas, i_array_size, index_size );
                                    umm i_array_index = 0;
                                    umm v_array_index = 0;
                                    
                                    r32 radius = command->point.size * 0.5f;
                                    
                                    vec4 offsets[ 8 ] = {
                                        v4( -radius, -radius, -radius, 1.0f ),
                                        v4( -radius, radius, -radius, 1.0f ),
                                        v4( radius, -radius, -radius, 1.0f ),
                                        v4( radius, radius, -radius, 1.0f ),
                                        
                                        v4( -radius, -radius, radius, 1.0f ),
                                        v4( -radius, radius, radius, 1.0f ),
                                        v4( radius, -radius, radius, 1.0f ),
                                        v4( radius, radius, radius, 1.0f )
                                    };
                                    
                                    if ( position_component_count == 2 ) {
                                        
                                        umm input_index = 0;
                                        
                                        for ( umm point_index = 0; point_index < command->point.point_count; point_index++ ) {
                                            
                                            for ( umm offset_index = 0; offset_index < 4; offset_index++ ) {
                                                
                                                for ( umm i = 0; i < 2; i++ ) {
                                                    v_array[ v_array_index++ ] = command->point.vertex_array[ input_index + i ] + offsets[ offset_index ].e[ i ];
                                                }
                                                
                                                if ( has_color ) {
                                                    c_array[ v_array_index++ ] = c_input_array[ input_index + 2 ];
                                                }
                                                
                                                internal_gl_push_index( i_array, &i_array_index, batch->vertex_index++, index_size );
                                            }
                                            
                                            input_index += vertex_format_size;
                                            
                                            internal_gl_push_primitive_restart( i_array, &i_array_index, index_size );
                                        }
                                        
                                    } else {
                                        
                                        umm input_index = 0;
                                        
                                        for ( umm point_index = 0; point_index < command->point.point_count; point_index++ ) {
                                            
                                            for ( umm offset_index = 0; offset_index < array_count( offsets ); offset_index++ ) {
                                                
                                                for ( umm i = 0; i < position_component_count; i++ ) {
                                                    v_array[ v_array_index++ ] = command->point.vertex_array[ input_index + i ] + offsets[ offset_index ].e[ i ];
                                                }
                                                
                                                if ( has_color ) {
                                                    c_array[ v_array_index++ ] = c_input_array[ input_index + position_component_count ];
                                                }
                                            }
                                            
                                            input_index += vertex_format_size;
                                            
                                            internal_gl_push_index( i_array, &i_array_index, batch->vertex_index + 0, index_size );
                                            internal_gl_push_index( i_array, &i_array_index, batch->vertex_index + 1, index_size );
                                            internal_gl_push_index( i_array, &i_array_index, batch->vertex_index + 2, index_size );
                                            internal_gl_push_index( i_array, &i_array_index, batch->vertex_index + 3, index_size );
                                            internal_gl_push_index( i_array, &i_array_index, batch->vertex_index + 7, index_size );
                                            internal_gl_push_index( i_array, &i_array_index, batch->vertex_index + 1, index_size );
                                            internal_gl_push_index( i_array, &i_array_index, batch->vertex_index + 5, index_size );
                                            internal_gl_push_index( i_array, &i_array_index, batch->vertex_index + 0, index_size );
                                            internal_gl_push_index( i_array, &i_array_index, batch->vertex_index + 4, index_size );
                                            internal_gl_push_index( i_array, &i_array_index, batch->vertex_index + 2, index_size );
                                            internal_gl_push_index( i_array, &i_array_index, batch->vertex_index + 6, index_size );
                                            internal_gl_push_index( i_array, &i_array_index, batch->vertex_index + 7, index_size );
                                            internal_gl_push_index( i_array, &i_array_index, batch->vertex_index + 4, index_size );
                                            internal_gl_push_index( i_array, &i_array_index, batch->vertex_index + 5, index_size );
                                            
                                            batch->vertex_index += 8;
                                            
                                            internal_gl_push_primitive_restart( i_array, &i_array_index, index_size );
                                        }
                                    }
                                    
                                    glBufferSubData( GL_ARRAY_BUFFER, batch->vertex_used, v_array_size, v_array );
                                    glBufferSubData( GL_ELEMENT_ARRAY_BUFFER, batch->index_used, i_array_size, i_array );
                                    batch->vertex_used += v_array_size;
                                    batch->index_used += i_array_size;
                                    _assert( batch->vertex_used <= batch->v_buffer_size );
                                    _assert( batch->index_used <= batch->i_buffer_size );
                                    
                                    command_buffer->datas = backup;
                                }
                            }
                            
                            glEnable( GL_PRIMITIVE_RESTART );
                            glPrimitiveRestartIndex( primitive_restart_index );
                            
                            u32 index_count_per_point = 5;
                            
                            if ( position_component_count == 3 ) {
                                index_count_per_point = 15;
                            }
                            
                            u32 index_count_to_draw = 0;
                            umm index_byte_offset = 0;
                            
                            for ( umm i = 0; i < batch->command_count; i++ ) {
                                
                                gl_command_t* command = batch_commands + i;
                                
                                if ( command->type == gl_command_type_point ) {
                                    
                                    index_count_to_draw += command->point.point_count * index_count_per_point;
                                    
                                } else {
                                    
                                    _assert( command->type == gl_command_type_scissor );
                                    
                                    if ( index_count_to_draw ) {
                                        glDrawElements( GL_TRIANGLE_STRIP, index_count_to_draw, index_type, ( void* ) index_byte_offset );
                                        index_byte_offset += index_count_to_draw * index_size;
                                        index_count_to_draw = 0;
                                    }
                                    
                                    internal_gl_do_scissor( command );
                                }
                            }
                            
                            if ( index_count_to_draw ) {
                                glDrawElements( GL_TRIANGLE_STRIP, index_count_to_draw, index_type, ( void* ) index_byte_offset );
                            }
                            
                            glDisable( GL_PRIMITIVE_RESTART );
                            
                        } break;
                        
                        case gl_command_type_line: {
                            
                            data_t backup = command_buffer->datas;
                            
                            u32 v_array_size = 0;
                            u32 i_array_size = 0;
                            void* index_array = memory_current_aligned( &command_buffer->datas, void, index_size );
                            
                            for ( umm batch_command_index = 0; batch_command_index < batch->command_count; batch_command_index++ ) {
                                
                                gl_command_t* command = batch_commands + batch_command_index;
                                
                                if ( command->type == gl_command_type_line ) {
                                    
                                    u32 strip_i_array_size = ( command->line.point_count + 1 ) * index_size;
                                    v_array_size += vertex_format_size_byte * command->line.point_count;
                                    i_array_size += strip_i_array_size;
                                    void* i_array = memory_push_size( &command_buffer->datas, strip_i_array_size, index_size );
                                    umm i_array_index = 0;
                                    
                                    for ( umm point_index = 0; point_index < command->line.point_count; point_index++ ) {
                                        internal_gl_push_index( i_array, &i_array_index, batch->vertex_index++, index_size );
                                    }
                                    
                                    internal_gl_push_primitive_restart( i_array, &i_array_index, index_size );
                                }
                            }
                            
                            glBufferSubData( GL_ARRAY_BUFFER, batch->vertex_used, v_array_size, batch->datas.bytes );
                            glBufferSubData( GL_ELEMENT_ARRAY_BUFFER, batch->index_used, i_array_size, index_array );
                            
                            batch->vertex_used += v_array_size;
                            batch->index_used += i_array_size;
                            _assert( batch->vertex_used <= batch->v_buffer_size );
                            _assert( batch->index_used <= batch->i_buffer_size );
                            
                            glEnable( GL_PRIMITIVE_RESTART );
                            glPrimitiveRestartIndex( primitive_restart_index );
                            
                            u32 index_count_to_draw = 0;
                            umm index_byte_offset = 0;
                            
                            for ( umm i = 0; i < batch->command_count; i++ ) {
                                
                                gl_command_t* command = batch_commands + i;
                                
                                if ( command->type == gl_command_type_line ) {
                                    index_count_to_draw += command->line.point_count + 1;
                                } else {
                                    
                                    _assert( command->type == gl_command_type_scissor );
                                    
                                    if ( index_count_to_draw ) {
                                        glDrawElements( GL_LINE_STRIP, index_count_to_draw, index_type, ( void* ) index_byte_offset );
                                        index_byte_offset += index_count_to_draw * index_size;
                                        index_count_to_draw = 0;
                                    }
                                    
                                    internal_gl_do_scissor( command );
                                }
                            }
                            
                            if ( index_count_to_draw ) {
                                glDrawElements( GL_LINE_STRIP, index_count_to_draw, index_type, ( void* ) index_byte_offset );
                            }
                            
                            glDisable( GL_PRIMITIVE_RESTART );
                            
                            command_buffer->datas = backup;
                            
                        } break;
                        
                        /* TODO simon: gl_command_type_triangle. */
                        
                        case gl_command_type_triangle_strip: {
                            
                            data_t backup = command_buffer->datas;
                            
                            u32 v_array_size = 0;
                            u32 i_array_size = 0;
                            void* index_array = memory_current_aligned( &command_buffer->datas, void, index_size );
                            
                            for ( umm batch_command_index = 0; batch_command_index < batch->command_count; batch_command_index++ ) {
                                
                                gl_command_t* command = batch_commands + batch_command_index;
                                
                                if ( command->type == gl_command_type_triangle_strip ) {
                                    
                                    u32 strip_i_array_size = ( 3 + command->triangle_strip.triangle_count ) * index_size;
                                    v_array_size += ( 2 + command->triangle_strip.triangle_count ) * vertex_format_size_byte;
                                    i_array_size += strip_i_array_size;
                                    
                                    void* i_array = memory_push_size( &command_buffer->datas, strip_i_array_size, index_size );
                                    umm i_array_index = 0;
                                    
                                    umm vertex_count = command->triangle_strip.triangle_count + 2;
                                    
                                    for ( umm vertex_index = 0; vertex_index < vertex_count; vertex_index++ ) {
                                        internal_gl_push_index( i_array, &i_array_index, batch->vertex_index++, index_size );
                                    }
                                    
                                    internal_gl_push_primitive_restart( i_array, &i_array_index, index_size );
                                }
                            }
                            
                            glBufferSubData( GL_ARRAY_BUFFER, batch->vertex_used, v_array_size, batch->datas.bytes );
                            glBufferSubData( GL_ELEMENT_ARRAY_BUFFER, batch->index_used, i_array_size, index_array );
                            batch->vertex_used += v_array_size;
                            batch->index_used += i_array_size;
                            _assert( batch->vertex_used <= batch->v_buffer_size );
                            _assert( batch->index_used <= batch->i_buffer_size );
                            
                            if ( has_uv ) {
                                glActiveTexture( GL_TEXTURE0 + 0 );
                                glBindTexture( GL_TEXTURE_2D, batch->texture );
                            }
                            
                            glEnable( GL_PRIMITIVE_RESTART );
                            glPrimitiveRestartIndex( primitive_restart_index );
                            
                            u32 index_count_to_draw = 0;
                            umm index_byte_offset = 0;
                            
                            for ( umm i = 0; i < batch->command_count; i++ ) {
                                
                                gl_command_t* command = batch_commands + i;
                                
                                if ( command->type == gl_command_type_triangle_strip ) {
                                    index_count_to_draw += command->triangle_strip.triangle_count + 3;
                                } else {
                                    
                                    _assert( command->type == gl_command_type_scissor );
                                    
                                    if ( index_count_to_draw ) {
                                        glDrawElements( GL_TRIANGLE_STRIP, index_count_to_draw, index_type, ( void* ) index_byte_offset );
                                        index_byte_offset += index_count_to_draw * index_size;
                                        index_count_to_draw = 0;
                                    }
                                    
                                    internal_gl_do_scissor( command );
                                }
                            }
                            
                            if ( index_count_to_draw ) {
                                glDrawElements( GL_TRIANGLE_STRIP, index_count_to_draw, index_type, ( void* ) index_byte_offset );
                            }
                            
                            glDisable( GL_PRIMITIVE_RESTART );
                            
                            command_buffer->datas = backup;
                            
                        } break;
                        
                        case gl_command_type_triangle_strip_textured: {
                            
                            data_t backup = command_buffer->datas;
                            
                            u32 v_array_size = 0;
                            u32 i_array_size = 0;
                            void* index_array = memory_current_aligned( &command_buffer->datas, void, index_size );
                            
                            for ( umm batch_command_index = 0; batch_command_index < batch->command_count; batch_command_index++ ) {
                                
                                gl_command_t* command = batch_commands + batch_command_index;
                                
                                if ( command->type == gl_command_type_triangle_strip_textured ) {
                                    
                                    u32 strip_i_array_size = ( 3 + command->triangle_strip_textured.triangle_count ) * index_size;
                                    v_array_size += ( 2 + command->triangle_strip_textured.triangle_count ) * vertex_format_size_byte;
                                    i_array_size += strip_i_array_size;
                                    
                                    void* i_array = memory_push_size( &command_buffer->datas, strip_i_array_size, index_size );
                                    umm i_array_index = 0;
                                    
                                    umm vertex_count = command->triangle_strip_textured.triangle_count + 2;
                                    
                                    for ( umm vertex_index = 0; vertex_index < vertex_count; vertex_index++ ) {
                                        internal_gl_push_index( i_array, &i_array_index, batch->vertex_index++, index_size );
                                    }
                                    
                                    internal_gl_push_primitive_restart( i_array, &i_array_index, index_size );
                                }
                            }
                            
                            glBufferSubData( GL_ARRAY_BUFFER, batch->vertex_used, v_array_size, batch->datas.bytes );
                            glBufferSubData( GL_ELEMENT_ARRAY_BUFFER, batch->index_used, i_array_size, index_array );
                            batch->vertex_used += v_array_size;
                            batch->index_used += i_array_size;
                            _assert( batch->vertex_used <= batch->v_buffer_size );
                            _assert( batch->index_used <= batch->i_buffer_size );
                            
                            glActiveTexture( GL_TEXTURE0 + 0 );
                            glEnable( GL_PRIMITIVE_RESTART );
                            glPrimitiveRestartIndex( primitive_restart_index );
                            
                            umm index_byte_offset = 0;
                            
                            for ( umm i = 0; i < batch->command_count; i++ ) {
                                
                                gl_command_t* command = batch_commands + i;
                                
                                if ( command->type == gl_command_type_triangle_strip_textured ) {
                                    
                                    u32 index_count_to_draw = command->triangle_strip_textured.triangle_count + 3;
                                    glBindTexture( GL_TEXTURE_2D, command->triangle_strip_textured.texture );
                                    glUniform1ui( bytes_per_pixel_location, command->triangle_strip_textured.bytes_per_pixel );
                                    glDrawElements( GL_TRIANGLE_STRIP, index_count_to_draw, index_type, ( void* ) index_byte_offset );
                                    index_byte_offset += index_count_to_draw * index_size;
                                    
                                } else {
                                    
                                    _assert( command->type == gl_command_type_scissor );
                                    internal_gl_do_scissor( command );
                                }
                            }
                            
                            glDisable( GL_PRIMITIVE_RESTART );
                            
                            command_buffer->datas = backup;
                            
                        } break;
                        
                        case gl_command_type_text: {
                            
                            if ( batch->texture == 0 ) {
                                gl_texture_t texture = gl_create_texture( batch->collection->atlas.bytes, batch->collection->atlas_width, batch->collection->atlas_height, 1, gl_texture_flag_bilinear );
                                batch->texture = texture.handle;
                            }
                            
                            for ( umm batch_command_index = 0; batch_command_index < batch->command_count; batch_command_index++ ) {
                                
                                gl_command_t* command = batch_commands + batch_command_index;
                                
                                if ( command->type == gl_command_type_text ) {
                                    
                                    data_t backup = command_buffer->datas;
                                    
                                    u32 codepoint_count = command->text.codepoint_count;
                                    u32 index_count = codepoint_count * 5;
                                    u32 vertex_sub_buffer_size = vertex_format_size_byte * 4 * codepoint_count;
                                    u32 index_sub_buffer_size = index_count * index_size;
                                    
                                    r32* v_array = ( r32* ) memory_push_size( &command_buffer->datas, vertex_sub_buffer_size, 4 );
                                    void* i_array = memory_push_size( &command_buffer->datas, index_sub_buffer_size, index_size );
                                    
                                    command->text.outputed_codepoint_count = gl_text_to_vertex_array( command->text.string.text, command->text.string.used, command->text.position.x, command->text.position.y, command->text.do_first_line_character_offset, batch->collection, command->text.font_index, v_array, i_array, batch->vertex_used / vertex_format_size_byte, command->text.color, batch->vertex_format );
                                    _assert( command->text.outputed_codepoint_count <= codepoint_count );
                                    
                                    vertex_sub_buffer_size = vertex_format_size_byte * 4 * command->text.outputed_codepoint_count;
                                    index_sub_buffer_size = command->text.outputed_codepoint_count * 5 * index_size;
                                    
                                    glBufferSubData( GL_ARRAY_BUFFER, batch->vertex_used, vertex_sub_buffer_size, v_array );
                                    glBufferSubData( GL_ELEMENT_ARRAY_BUFFER, batch->index_used, index_sub_buffer_size, i_array );
                                    batch->vertex_used += vertex_sub_buffer_size;
                                    batch->index_used += index_sub_buffer_size;
                                    _assert( batch->vertex_used <= batch->v_buffer_size );
                                    _assert( batch->index_used <= batch->i_buffer_size );
                                    
                                    command_buffer->datas = backup;
                                }
                            }
                            
                            glActiveTexture( GL_TEXTURE0 + 0 );
                            glBindTexture( GL_TEXTURE_2D, batch->texture );
                            
                            glEnable( GL_PRIMITIVE_RESTART );
                            glPrimitiveRestartIndex( primitive_restart_index );
                            
                            u32 index_count_to_draw = 0;
                            umm index_byte_offset = 0;
                            
                            for ( umm i = 0; i < batch->command_count; i++ ) {
                                
                                gl_command_t* command = batch_commands + i;
                                
                                if ( command->type == gl_command_type_text ) {
                                    index_count_to_draw += command->text.outputed_codepoint_count * 5;
                                } else {
                                    
                                    _assert( command->type == gl_command_type_scissor );
                                    
                                    if ( index_count_to_draw ) {
                                        glDrawElements( GL_TRIANGLE_STRIP, index_count_to_draw, index_type, ( void* ) index_byte_offset );
                                        index_byte_offset += index_count_to_draw * index_size;
                                        index_count_to_draw = 0;
                                    }
                                    
                                    internal_gl_do_scissor( command );
                                }
                            }
                            
                            if ( index_count_to_draw ) {
                                glDrawElements( GL_TRIANGLE_STRIP, index_count_to_draw, index_type, ( void* ) index_byte_offset );
                            }
                            
                            glDisable( GL_PRIMITIVE_RESTART );
                            
                        } break;
                        
                        default: {
                            debug_break( );
                        } break;
                    }
                    
                    /* TODO simon: Is this ok to do here ? Problem was to keep commands in the batch intact so I could draw it without doing gl_batchXXX each frame if nothing changes. */
                    batch->vertex_used = 0;
                    batch->vertex_index = 0;
                    batch->index_used = 0;
                }
            } break;
            
            default: {
                debug_break( );
            } break;
        }
    }
}

stu void gl_frame_buffer_to_memory( s32 x, s32 y, s32  width, s32 height, void* memory, umm memory_size ) {
    
    _assert( memory_size >= width * height * 4 );
    
    struct pixel_store_pack_t {
        GLenum name;
        GLint backup, value;
    };
    
    struct pixel_store_pack_t pixel_store_packs[ 8 ] = {
        GL_PACK_SWAP_BYTES, 0, GL_FALSE,
        GL_PACK_LSB_FIRST, 0, GL_FALSE,
        GL_PACK_ROW_LENGTH, 0, 0,
        GL_PACK_IMAGE_HEIGHT, 0, 0,
        GL_PACK_SKIP_PIXELS, 0, 0,
        GL_PACK_SKIP_ROWS, 0, 0,
        GL_PACK_SKIP_IMAGES, 0, 0,
        GL_PACK_ALIGNMENT, 0, 1
    };
    
    GLint pixel_pack_buffer;
    
    glGetIntegerv( GL_PIXEL_PACK_BUFFER_BINDING, &pixel_pack_buffer );
    glBindBuffer( GL_PIXEL_PACK_BUFFER, 0 );
    
    for ( umm i = 0; i < array_count( pixel_store_packs ); i++ ) {
        glGetIntegerv( pixel_store_packs[ i ].name, &pixel_store_packs[ i ].backup );
        glPixelStorei( pixel_store_packs[ i ].name, pixel_store_packs[ i ].value );
    }
    
    glReadPixels( x, y, width, height, GL_RGBA, GL_UNSIGNED_BYTE, memory );
    
    glBindBuffer( GL_PIXEL_PACK_BUFFER, pixel_pack_buffer );
    
    for ( umm i = 0; i < array_count( pixel_store_packs ); i++ ) {
        glPixelStorei( pixel_store_packs[ i ].name, pixel_store_packs[ i ].backup );
    }
}

#define GL_HARDWARE_H 1

#include "gl_state.h"

stu gl_state_t gl_start( s32 window_width, s32 window_height, u32 flags ) {
    
    u32 vao;
    glGenVertexArrays( 1, &vao );
    glBindVertexArray( vao );
    
    if ( flags & gl_state_flag_alpha_blending ) {
        glEnable( GL_BLEND );
        glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    }
    
    if ( flags & gl_state_flag_depth_testing ) {
        glEnable( GL_DEPTH_TEST );
    }
    
    if ( flags & gl_state_flag_backface_culling ) {
        glEnable( GL_CULL_FACE );
    }
    
    gl_state_t result = { 0 };
    
    data_t memory = memory_get_virtual( mebioctet( 1 ), 0 );
    
    result.commands = gl_create_command_buffer( &memory, kibioctet( 512 ), kibioctet( 512 ) );
    matrix_create_screen_space( result.matrix, window_width, window_height );
    
    return result;
}

#endif
