
#ifndef GL_STATE_H

typedef struct gl_state_t {
    gl_command_buffer_t commands;
    gl_batch_t batches[ 16 ];
    u32 shaders[ 16 ];
    r32 matrix[ 16 ];
    u8 batch_count;
} gl_state_t;

typedef enum gl_state_flag_t {
    gl_state_flag_none = 0x0,
    gl_state_flag_depth_testing = 0x1 << 0,
    gl_state_flag_backface_culling = 0x1 << 1,
    gl_state_flag_alpha_blending = 0x1 << 2,
} gl_state_flags_t;

stu u8 gl_add_point_extended( gl_state_t* state, umm point_count, umm command_count, u8 vertex_format, u32 shader_handle, r32* matrix ) {
    
    u8 result = state->batch_count;
    _assert( result < array_count( state->batches ) );
    
    if ( !vertex_format ) {
        vertex_format = gl_vertex_format_create( 2, 1, 0, 2 );
    }
    
    if ( !shader_handle ) {
        
        if ( !state->shaders[ result ] ) {
#ifdef GL_HARDWARE_H
            state->shaders[ result ] = gl_create_shader_program( vs_screen_space_color, fs_default_vertex_color );
#endif
        }
        
        shader_handle = state->shaders[ state->batch_count ];
    }
    
    if ( !matrix ) {
        matrix = state->matrix;
    }
    
    u8 vertex_size_byte = gl_vertex_format_compute_vertex_size_byte( vertex_format );
    
    umm memory_size = ( sizeof( gl_command_t ) * command_count ) + ( vertex_size_byte * point_count );
    data_t memory = memory_get_virtual( memory_size, 0 );
    state->batches[ state->batch_count ] = gl_batch_create_point( &memory, point_count, command_count, vertex_format, shader_handle, matrix );
    
    state->batch_count++;
    
    return result;
}

stu u8 gl_add_point( gl_state_t* state ) {
    u8 result = gl_add_point_extended( state, 100, 100, 0, 0, 0 );
    return result;
}

stu u8 gl_add_line_extended( gl_state_t* state, umm point_count, umm command_count, u8 vertex_format, u32 shader_handle, r32* matrix ) {
    
    u8 result = state->batch_count;
    _assert( result < array_count( state->batches ) );
    
    if ( !vertex_format ) {
        vertex_format = gl_vertex_format_create( 2, 1, 0, 2 );
    }
    
    if ( !shader_handle ) {
        
        if ( !state->shaders[ result ] ) {
#ifdef GL_HARDWARE_H
            state->shaders[ result ] = gl_create_shader_program( vs_screen_space_color, fs_default_vertex_color );
#endif
        }
        
        shader_handle = state->shaders[ state->batch_count ];
    }
    
    if ( !matrix ) {
        matrix = state->matrix;
    }
    
    u8 vertex_size_byte = gl_vertex_format_compute_vertex_size_byte( vertex_format );
    
    umm memory_size = ( sizeof( gl_command_t ) * command_count ) + ( vertex_size_byte * point_count );
    data_t memory = memory_get_virtual( memory_size, 0 );
    state->batches[ state->batch_count ] = gl_batch_create_line( &memory, point_count, command_count, vertex_format, shader_handle, matrix );
    
    state->batch_count++;
    
    return result;
}

stu u8 gl_add_line( gl_state_t* state ) {
    u8 result = gl_add_line_extended( state, 100, 100, 0, 0, 0 );
    return result;
}

stu u8 gl_add_rectangle_extended( gl_state_t* gl_state, umm rectangle_count, umm command_count, u8 vertex_format, u32 shader_handle, r32* matrix ) {
    
    u8 result = gl_state->batch_count;
    _assert( result < array_count( gl_state->batches ) );
    
    if ( !vertex_format ) {
        vertex_format = gl_vertex_format_create( 2, 1, 0, 2 );
    }
    
    if ( !shader_handle ) {
        
        if ( !gl_state->shaders[ result ] ) {
#ifdef GL_HARDWARE_H
            gl_state->shaders[ result ] = gl_create_shader_program( vs_screen_space_color, fs_default_vertex_color );
#endif
        }
        
        shader_handle = gl_state->shaders[ gl_state->batch_count ];
    }
    
    if ( !matrix ) {
        matrix = gl_state->matrix;
    }
    
    u8 vertex_size_byte = gl_vertex_format_compute_vertex_size_byte( vertex_format );
    umm triangle_count = rectangle_count * 2;
    umm vertex_count = triangle_count + ( 2 * command_count );
    
    umm memory_size = ( sizeof( gl_command_t ) * command_count ) + ( vertex_size_byte * vertex_count );
    data_t memory = memory_get_virtual( memory_size, 0 );
    gl_state->batches[ gl_state->batch_count ] = gl_batch_create_triangle_strip( &memory, triangle_count, command_count, vertex_format, shader_handle, matrix );
    gl_state->batch_count++;
    
    return result;
}

stu u8 gl_add_rectangle( gl_state_t* gl_state ) {
    u8 result = gl_add_rectangle_extended( gl_state, 1000, 1000, 0, 0, 0 );
    return result;
}

stu u8 gl_add_text_extended( gl_state_t* gl_state, umm letter_count, umm command_count, u8 vertex_format, u32 shader_handle, r32* matrix, font_collection_t* collection, umm font_size ) {
    
    u8 result = gl_state->batch_count;
    
    _assert( result < array_count( gl_state->batches ) );
    
    if ( !vertex_format ) {
        vertex_format = gl_vertex_format_create( 2, 1, 2, 4 );
    }
    
    if ( !shader_handle ) {
        if ( !gl_state->shaders[ result ]) {
#ifdef GL_HARDWARE_H
            gl_state->shaders[ result ] = gl_create_shader_program( vs_screen_space_texture_color, fs_default_u8_color );
#endif
        }
        shader_handle = gl_state->shaders[ result ];
    }
    
    if ( !matrix ) {
        matrix = gl_state->matrix;
    }
    
    if ( !collection ) {
        
        if ( !font_size ) {
            font_size = 14;
        }
        
        data_t verdana_file = file_read_entire_s( "C:\\Windows\\Fonts\\Verdana.ttf", false ); /* TODO simon: Should I free this memory ? */
        data_t memory = memory_get_virtual( mebioctet( 1 ), 0 );
        collection = memory_push_struct( &memory, font_collection_t );
        collection->memory = memory;
        font_add_font( collection, &verdana_file, ( r32 ) font_size, font_codepoint_set, array_count( font_codepoint_set ), false );
        font_build_collection( collection );
    }
    
    umm memory_size = ( sizeof( gl_command_t ) * command_count ) + ( letter_count ); /* TODO simon: should I multiply by 4 for utf 8 ? */
    data_t memory = memory_get_virtual( memory_size, 0 );
    gl_state->batches[ result ] = gl_batch_create_text( &memory, letter_count, command_count, vertex_format, shader_handle, matrix, collection );
    gl_state->batch_count++;
    
    return result;
}

stu u8 gl_add_text( gl_state_t* gl_state ) {
    u8 result = gl_add_text_extended( gl_state, 10000, 1000, 0, 0, 0, 0, 0 );
    return result;
}

stu u8 gl_add_textured_quad_extended( gl_state_t* gl_state, umm quad_count, umm command_count, u8 vertex_format, u32 shader_handle, r32* matrix ) {
    
    u8 result = gl_state->batch_count;
    
    _assert( result < array_count( gl_state->batches ) );
    
    if ( !vertex_format ) {
        vertex_format = gl_vertex_format_create( 2, 1, 2, 2 );
    }
    
    if ( !shader_handle ) {
        if ( !gl_state->shaders[ result ] ) {
#ifdef GL_HARDWARE_H
            gl_state->shaders[ result ] = gl_create_shader_program( vs_screen_space_texture_color, fs_textured_color );
#endif
        }
        shader_handle = gl_state->shaders[ result ];
    }
    
    if ( !matrix ) {
        matrix = gl_state->matrix;
    }
    
    u8 vertex_size_byte = gl_vertex_format_compute_vertex_size_byte( vertex_format );
    umm triangle_count = quad_count * 2;
    umm vertex_count = triangle_count + ( 2 * command_count );
    
    umm memory_size = ( sizeof( gl_command_t ) * command_count ) + ( vertex_size_byte * vertex_count );
    data_t memory = memory_get_virtual( memory_size, 0 );
    gl_state->batches[ result ] = gl_batch_create_triangle_strip_textured( &memory, quad_count, command_count, vertex_format, shader_handle, matrix );
    gl_state->batch_count++;
    
    return result;
}

stu u8 gl_add_textured_quad( gl_state_t* gl_state ) {
    u8 result = gl_add_textured_quad_extended( gl_state, 1000, 1000, 0, 0, 0 );
    return result;
}

stu u8 gl_add_triangle_strip_extended( gl_state_t* gl_state, umm triangle_count, umm command_count, u8 vertex_format, u32 shader_handle, r32* matrix ) {
    
    u8 result = gl_state->batch_count;
    _assert( result < array_count( gl_state->batches ) );
    
    if ( !vertex_format ) {
        vertex_format = gl_vertex_format_create( 2, 1, 2, 2 );
    }
    
    if ( !shader_handle ) {
        
        if ( !gl_state->shaders[ result ] ) {
#ifdef GL_HARDWARE_H
            gl_state->shaders[ result ] = gl_create_shader_program( vs_screen_space_color, fs_default_vertex_color );
#endif
        }
        
        shader_handle = gl_state->shaders[ gl_state->batch_count ];
    }
    
    if ( !matrix ) {
        matrix = gl_state->matrix;
    }
    
    u8 vertex_size_byte = gl_vertex_format_compute_vertex_size_byte( vertex_format );
    umm vertex_count = triangle_count + ( 2 * command_count );
    
    umm memory_size = ( sizeof( gl_command_t ) * command_count ) + ( vertex_count * vertex_size_byte );
    data_t memory = memory_get_virtual( memory_size, 0 );
    gl_state->batches[ gl_state->batch_count ] = gl_batch_create_triangle_strip( &memory, triangle_count, command_count, vertex_format, shader_handle, matrix );
    gl_state->batch_count++;
    
    return result;
}

stu u8 gl_add_triangle_strip( gl_state_t* gl_state ) {
    u8 result = gl_add_triangle_strip_extended( gl_state, 100, 100, 0, 0, 0 );
    return result;
}

stu u8 gl_add_atlas_extended( gl_state_t* gl_state, umm triangle_count, umm command_count, u8 vertex_format, u32 shader_handle, r32* matrix, u32 texture_handle ) {
    
    u8 result = gl_state->batch_count;
    _assert( result < array_count( gl_state->batches ) );
    
    if ( !vertex_format ) {
        vertex_format = gl_vertex_format_create( 2, 1, 1, 2 );
    }
    
    if ( !shader_handle ) {
        
        if ( !gl_state->shaders[ result ] ) {
#ifdef GL_HARDWARE_H
            gl_state->shaders[ result ] = gl_create_shader_program( vs_screen_space_texture_color, fs_textured_color );
#endif
        }
        
        shader_handle = gl_state->shaders[ gl_state->batch_count ];
    }
    
    if ( !matrix ) {
        matrix = gl_state->matrix;
    }
    
    u8 vertex_size_byte = gl_vertex_format_compute_vertex_size_byte( vertex_format );
    umm vertex_count = triangle_count + ( 2 * command_count );
    
    umm memory_size = ( sizeof( gl_command_t ) * command_count ) + ( vertex_count * vertex_size_byte );
    data_t memory = memory_get_virtual( memory_size, 0 );
    gl_state->batches[ gl_state->batch_count ] = gl_batch_create_triangle_strip( &memory, triangle_count, command_count, vertex_format, shader_handle, matrix );
    gl_state->batches[ gl_state->batch_count ].texture = texture_handle;
    gl_state->batch_count++;
    
    return result;
}

stu u8 gl_add_atlas( gl_state_t* gl_state, u32 texture_handle ) {
    u8 result = gl_add_atlas_extended( gl_state, 100, 100, 0, 0, 0, texture_handle );
    return result;
}

stu gl_batch_t* gl_get_batch( gl_state_t* gl_state, umm id, b32 allow_unsafe ) {
    
    _assert( id < array_count( gl_state->batches ) );
    
    gl_batch_t* result = 0;
    
    if ( id < gl_state->batch_count || allow_unsafe ) {
        result = gl_state->batches + id;
    }
    
    return result;
}

stu void gl_frame_start( gl_state_t* gl_state ) {
    
    gl_reset_command_buffer( &gl_state->commands );
    
    for ( umm i = 0; i < gl_state->batch_count; i++ ) {
        gl_batch_frame_init( gl_state->batches + i );
    }
}

stu void gl_resize_viewport( gl_state_t* gl_state, r32 x, r32 y, s32 width, s32 height ) {
    gl_viewport( &gl_state->commands, v2( x, y ), v2_s32( width, height ) );
    matrix_create_screen_space( gl_state->matrix, width, height );
}

stu void gl_scissor_batches( gl_state_t* gl, b32 enable, r32 x, r32 y, r32 width, r32 height, u8* batches, umm batch_count ) {
    
    _assert( batch_count < array_count( gl->batches ) );
    
    u8 batch_indicies[ array_count( gl->batches ) ];
    
    if ( batches == 0 ) {
        
        batch_count = gl->batch_count;
        
        for ( umm i = 0; i < batch_count; i++ ) {
            batch_indicies[ i ] = ( u8 ) i;
        }
        
        batches = batch_indicies;
    }
    
    for ( umm i = 0; i < batch_count; i++ ) {
        _assert( batches[ i ] < array_count( gl->batches ) );
        gl_batch_scissor( gl->batches + batches[ i ], enable, ( s32 ) x, ( s32 ) y, ( s32 ) width, ( s32 ) height );
    }
}

#define gl_scissor_batches_disable( gl, batches, batch_count ) gl_scissor_batches( ( gl ), false, 0, 0, 0, 0, ( batches ), ( batch_count ) )
#define gl_scissor_batches_rect( gl, rect, batches, batch_count ) gl_scissor_batches( ( gl ), true, \
( rect ).left, \
( rect ).bottom, \
( rect ).right - ( rect ).left, \
( rect ).top - ( rect ).bottom, \
( batches ), \
( batch_count ) )

stu void gl_render( gl_command_buffer_t* commands );

stu void gl_render_state( gl_state_t* gl_state ) {
    
    for ( umm i = 0; i < gl_state->batch_count; i++ ) {
        gl_push_batch( &gl_state->commands, gl_state->batches + i );
    }
    
    gl_render( &gl_state->commands );
}

#define GL_STATE_H 1

#endif