u8 vs_lib[ ] =
"vec4 color_unpack( unsigned int color ) {\n"
"    unsigned int mask_r = uint( 0x000000ff );\n"
"    unsigned int mask_g = uint( 0x0000ff00 );\n"
"    unsigned int mask_b = uint( 0x00ff0000 );\n"
"    unsigned int mask_a = uint( 0xff000000 );\n"
"    float r = float( ( color & mask_r ) >> 0 );\n"
"    float g = float( ( color & mask_g ) >> 8 );\n"
"    float b = float( ( color & mask_b ) >> 16 );\n"
"    float a = float( ( color & mask_a ) >> 24 );\n"
"    float r_255 = 1.0 / 255.0;\n"
"    vec4 result = vec4( r * r_255 , g * r_255, b * r_255, a * r_255 );\n"
"    return result;\n"
"}\n"
"\n";

u8 vs_default[ ] =
"in vec2 position;\n"
"in vec2 in_uv;\n"
"out vec2 out_uv;\n"
"\n"
"void main( ) {\n"
"    gl_Position = vec4( position, 0.0, 1.0 );\n"
"    out_uv = in_uv;\n"
"}\n"
"\n";

u8 vs_screen_space_texture[ ] =
"in vec3 position;\n"
"in vec2 in_uv;\n"
"out vec2 out_uv;\n"
"\n"
"uniform mat4x4 screen_to_normalized;\n"
"\n"
"void main( ) {\n"
"    vec4 tPos = vec4( position, 1.0 ) * screen_to_normalized;\n"
"    gl_Position = tPos;\n"
"    out_uv = in_uv;\n"
"}\n"
"\n";

u8 vs_screen_space_texture_color[ ] =
"vec4 color_unpack( unsigned int color );\n"
"\n"
"in vec3 position;\n"
"in vec2 in_uv;\n"
"in uint color;\n"
"out vec2 out_uv;\n"
"out vec4 out_color;\n"
"\n"
"uniform mat4x4 projection;\n"
"\n"
"void main( ) {\n"
"    gl_Position = projection * vec4( position, 1.0 );\n"
"    out_uv = in_uv;\n"
"    out_color = color_unpack( color );\n"
"}\n"
"\n";

u8 vs_screen_space_color[ ] =
"in vec3 position;\n"
"in uint color;\n"
"out vec4 out_color;\n"
"\n"
"uniform mat4x4 projection;\n"
"\n"
"void main( ) {\n"
"    out_color = color_unpack( color );\n"
"    gl_Position = projection * vec4( position, 1.0 );\n"
"}\n"
"\n";

u8 vs_orthographic[ ] =
"in vec3 position;\n"
"in uint color;\n"
"out vec4 out_color;\n"
"\n"
"uniform mat4x4 projection;\n"
"\n"
"void main( ) {\n"
"    out_color = color_unpack( color );\n"
"    vec4 transformed_position = projection * vec4( position, 1.0 );\n"
"    gl_Position = transformed_position;\n"
"}\n"
"\n";

u8 fs_default[ ] =
"out vec4 fragment_out_color;\n"
"in vec2 out_uv;\n"
"\n"
"uniform sampler2D albedo;\n"
"\n"
"void main( ) {\n"
"    fragment_out_color = texture( albedo, out_uv );\n"
"}\n"
"\n";

u8 fs_default_color[ ] =
"out vec4 fragment_out_color;\n"
"uniform vec4 in_color;\n"
"\n"
"void main( ) {\n"
"    fragment_out_color = in_color;\n"
"}\n"
"\n";

u8 fs_default_vertex_color[ ] =
"out vec4 fragment_out_color;\n"
"in vec4 out_color;\n"
"\n"
"void main( ) {\n"
"    fragment_out_color = out_color;\n"
"}\n"
"\n";

u8 fs_default_u8[ ] =
"out vec4 fragment_out_color;\n"
"in vec2 out_uv;\n"
"\n"
"uniform sampler2D albedo;\n"
"\n"
"void main( ) {\n"
"    float c = texture( albedo, out_uv ).r;\n"
"    fragment_out_color = vec4( 1.0, 1.0, 1.0, c );\n"
"}\n"
"\n";

u8 fs_default_u8_color[ ] =
"out vec4 fragment_out_color;\n"
"in vec2 out_uv;\n"
"in vec4 out_color;\n"
"\n"
"uniform sampler2D albedo;\n"
"\n"
"void main( ) {\n"
"    float c = texture( albedo, out_uv ).r;\n"
"    fragment_out_color = vec4( out_color.r, out_color.g, out_color.b, out_color.a * c );\n"
"}\n"
"\n";

u8 fs_textured_color[ ] =
"out vec4 fragment_out_color;\n"
"in vec2 out_uv;\n"
"in vec4 out_color;\n"
"\n"
"uniform sampler2D albedo;\n"
"uniform uint bytes_per_pixel;\n"
"\n"
"void main( ) {\n"
"    vec4 tex;\n"
"    if ( bytes_per_pixel == uint( 1 ) ) {\n"
"        tex = vec4( texture( albedo, out_uv ).rrr, 1.0 );\n"
"    } else {\n"
"        tex = texture( albedo, out_uv );\n"
"    }\n"
"    fragment_out_color = tex * out_color;\n"
"}";

