
#if WINDOWS

#define interface struct

/* NOTE simon: from rpcsal.h */
#define __RPC__in
#define __RPC__out
#define __RPC__deref_out
/* NOTE simon: end of rpcsal.h */

/* NOTE simon: from shlwapi.h */
//
//  Association APIs
//
//  these APIs are to assist in accessing the data in HKCR
//  getting the Command strings and exe paths
//  for different verbs and extensions are simplified this way
//

enum
{
    ASSOCF_NONE                        = 0x00000000,
    ASSOCF_INIT_NOREMAPCLSID           = 0x00000001,  //  do not remap clsids to progids
    ASSOCF_INIT_BYEXENAME              = 0x00000002,  //  executable is being passed in
    ASSOCF_OPEN_BYEXENAME              = 0x00000002,  //  executable is being passed in
    ASSOCF_INIT_DEFAULTTOSTAR          = 0x00000004,  //  treat "*" as the BaseClass
    ASSOCF_INIT_DEFAULTTOFOLDER        = 0x00000008,  //  treat "Folder" as the BaseClass
    ASSOCF_NOUSERSETTINGS              = 0x00000010,  //  dont use HKCU
    ASSOCF_NOTRUNCATE                  = 0x00000020,  //  dont truncate the return string
    ASSOCF_VERIFY                      = 0x00000040,  //  verify data is accurate (DISK HITS)
    ASSOCF_REMAPRUNDLL                 = 0x00000080,  //  actually gets info about rundlls target if applicable
    ASSOCF_NOFIXUPS                    = 0x00000100,  //  attempt to fix errors if found
    ASSOCF_IGNOREBASECLASS             = 0x00000200,  //  dont recurse into the baseclass
    ASSOCF_INIT_IGNOREUNKNOWN          = 0x00000400,  //  dont use the "Unknown" progid, instead fail
#if (NTDDI_VERSION >= NTDDI_WIN8)
    ASSOCF_INIT_FIXED_PROGID           = 0x00000800,  //  the Init() pszAssoc value is a ProgId that should not be mapped using the current user defaults
    ASSOCF_IS_PROTOCOL                 = 0x00001000,  //  the Init() pszAssoc value is an uri scheme (not including the ":") that should be mapped using the current user defaults
    ASSOCF_INIT_FOR_FILE               = 0x00002000,  //  use this flag when specifying ASSOCF_INIT_FIXED_PROGID if the ProgId corresponds with a file extension based association
#endif
#if (NTDDI_VERSION >= NTDDI_WIN10_RS1)
    ASSOCF_IS_FULL_URI                 = 0x00004000,  //  Used to specify that full http/https URI is being passed for target resolution
    //  Only one of ASSOCF_INIT_FIXED_PROGID, ASSOCF_IS_PROTOCOL or ASSOCF_IS_FULL_URI can be specified at a time.
    ASSOCF_PER_MACHINE_ONLY            = 0x00008000,  //  Enforces per-machine association look-up only and avoid HKCU.
#endif
};
typedef DWORD ASSOCF;

typedef enum
{
    ASSOCSTR_COMMAND      = 1,  //  shell\verb\command string
    ASSOCSTR_EXECUTABLE,        //  the executable part of command string
    ASSOCSTR_FRIENDLYDOCNAME,   //  friendly name of the document type
    ASSOCSTR_FRIENDLYAPPNAME,   //  friendly name of executable
    ASSOCSTR_NOOPEN,            //  noopen value
    ASSOCSTR_SHELLNEWVALUE,     //  query values under the shellnew key
    ASSOCSTR_DDECOMMAND,        //  template for DDE commands
    ASSOCSTR_DDEIFEXEC,         //  DDECOMMAND to use if just create a process
    ASSOCSTR_DDEAPPLICATION,    //  Application name in DDE broadcast
    ASSOCSTR_DDETOPIC,          //  Topic Name in DDE broadcast
    ASSOCSTR_INFOTIP,           //  info tip for an item, or list of properties to create info tip from
#if (_WIN32_IE >= _WIN32_IE_IE60)
    ASSOCSTR_QUICKTIP,          //  same as ASSOCSTR_INFOTIP, except, this list contains only quickly retrievable properties
    ASSOCSTR_TILEINFO,          //  similar to ASSOCSTR_INFOTIP - lists important properties for tileview
    ASSOCSTR_CONTENTTYPE,       //  MIME Content type
    ASSOCSTR_DEFAULTICON,       //  Default icon source
    ASSOCSTR_SHELLEXTENSION,    //  Guid string pointing to the Shellex\Shellextensionhandler value.
#endif // _WIN32_IE_IE60
#if (_WIN32_IE >= _WIN32_IE_IE80)
    ASSOCSTR_DROPTARGET,        //  The CLSID of DropTarget
    ASSOCSTR_DELEGATEEXECUTE,   //  The CLSID of DelegateExecute
#endif // _WIN32_IE_IE80
    // a string value of the uri protocol schemes, for example "http:https:ftp:file:" or "*" indicating all
    ASSOCSTR_SUPPORTED_URI_PROTOCOLS,
#if (NTDDI_VERSION >= NTDDI_WIN10)
    ASSOCSTR_PROGID,            // The ProgId provided by the app associated with the file type or uri scheme based on user default settings.
    ASSOCSTR_APPID,             // The AppUserModelID of the app associated with the file type or uri scheme based on user default settings.
    ASSOCSTR_APPPUBLISHER,      // THe publisher of the app associated with the file type or uri scheme based on user default settings.
    ASSOCSTR_APPICONREFERENCE,  // The icon reference of the app associated with the file type or uri scheme based on user default settings.
#endif // NTDDI_WIN10
    ASSOCSTR_MAX                //  last item in enum...
} ASSOCSTR;

/*LWSTDAPI*/ HRESULT AssocQueryStringA( ASSOCF flags, ASSOCSTR str, LPCSTR pszAssoc, LPCSTR pszExtra, LPSTR pszOut, DWORD *pcchOut );
/*LWSTDAPI*/ HRESULT AssocQueryStringW( ASSOCF flags, ASSOCSTR str, LPCWSTR pszAssoc, LPCWSTR pszExtra, LPWSTR pszOut, DWORD *pcchOut);

/* NOTE simon: end of shlwapi.h */


/* NOTE simon: from rpcndr.h */
#define CONST_VTBL
/* NOTE simon: end of rpcndr.h */

/* NOTE simon: from wtypesbase.h */
typedef
enum tagCLSCTX
{
    CLSCTX_INPROC_SERVER	= 0x1,
    CLSCTX_INPROC_HANDLER	= 0x2,
    CLSCTX_LOCAL_SERVER	= 0x4,
    CLSCTX_INPROC_SERVER16	= 0x8,
    CLSCTX_REMOTE_SERVER	= 0x10,
    CLSCTX_INPROC_HANDLER16	= 0x20,
    CLSCTX_RESERVED1	= 0x40,
    CLSCTX_RESERVED2	= 0x80,
    CLSCTX_RESERVED3	= 0x100,
    CLSCTX_RESERVED4	= 0x200,
    CLSCTX_NO_CODE_DOWNLOAD	= 0x400,
    CLSCTX_RESERVED5	= 0x800,
    CLSCTX_NO_CUSTOM_MARSHAL	= 0x1000,
    CLSCTX_ENABLE_CODE_DOWNLOAD	= 0x2000,
    CLSCTX_NO_FAILURE_LOG	= 0x4000,
    CLSCTX_DISABLE_AAA	= 0x8000,
    CLSCTX_ENABLE_AAA	= 0x10000,
    CLSCTX_FROM_DEFAULT_CONTEXT	= 0x20000,
    CLSCTX_ACTIVATE_32_BIT_SERVER	= 0x40000,
    CLSCTX_ACTIVATE_64_BIT_SERVER	= 0x80000,
    CLSCTX_ENABLE_CLOAKING	= 0x100000,
    CLSCTX_APPCONTAINER	= 0x400000,
    CLSCTX_ACTIVATE_AAA_AS_IU	= 0x800000,
    CLSCTX_PS_DLL	= ( int  )0x80000000
} 	CLSCTX;
/* NOTE simon: end of wtypesbase.h */

/* NOTE simon: from combaseapi.h */
#define BEGIN_INTERFACE
#define END_INTERFACE
typedef struct tagPROPVARIANT PROPVARIANT;
#define CLSCTX_ALL              (CLSCTX_INPROC_SERVER| \
CLSCTX_INPROC_HANDLER| \
CLSCTX_LOCAL_SERVER| \
CLSCTX_REMOTE_SERVER)
/* NOTE simon: end of combaseapi.h */

/* NOTE simon: From mmreg.h */

#pragma pack(push, 1)
typedef struct tWAVEFORMATEX
{
    WORD    wFormatTag;        /* format type */
    WORD    nChannels;         /* number of channels (i.e. mono, stereo...) */
    DWORD   nSamplesPerSec;    /* sample rate */
    DWORD   nAvgBytesPerSec;   /* for buffer estimation */
    WORD    nBlockAlign;       /* block size of data */
    WORD    wBitsPerSample;    /* Number of bits per sample of mono data */
    WORD    cbSize;            /* The count in bytes of the size of
extra information (after cbSize) */
    
} WAVEFORMATEX;
typedef WAVEFORMATEX       *PWAVEFORMATEX;
typedef WAVEFORMATEX NEAR *NPWAVEFORMATEX;
typedef WAVEFORMATEX FAR  *LPWAVEFORMATEX;

typedef struct {
    WAVEFORMATEX    Format;
    union {
        WORD wValidBitsPerSample;       /* bits of precision  */
        WORD wSamplesPerBlock;          /* valid if wBitsPerSample==0 */
        WORD wReserved;                 /* If neither applies, set to zero. */
    } Samples;
    DWORD           dwChannelMask;      /* which channels are */
    /* present in stream  */
    GUID            SubFormat;
} WAVEFORMATEXTENSIBLE, *PWAVEFORMATEXTENSIBLE;
#pragma pack(pop)  /* Assume byte packing throughout */

/* NOTE simon: End of mmreg.h*/

/* NOTE simon: audiosessiontypes.h */
typedef enum _AUDCLNT_SHAREMODE
{
    AUDCLNT_SHAREMODE_SHARED,
    AUDCLNT_SHAREMODE_EXCLUSIVE
} AUDCLNT_SHAREMODE;
/* NOTE simon: end of audiosessiontypes.h */

/* NOTE simon: from propsys.h */
typedef struct IPropertyStore IPropertyStore;
/* NOTE simon: end of propsys.h */

/* NOTE simon: How to get GUID https://hero.handmade.network/forums/code-discussion/t/7350-waveout_sound_api#22896

__uiidof(X) returns GUID from type X.
If you look how MMDeviceEnumerator is defined, you'll see this:

[code]
MIDL_INTERFACE("A95664D2-9614-4F35-A746-DE8DB63617E6")
IMMDeviceEnumerator : public IUnknown
{
...
}
[/code]

__uiidof(MMDeviceEnumerator) will produce A95664D2-9614-4F35-A746-DE8DB63617E6 as GUID type.

To fix this in C, you should include #include <initguid.h> header first (before all the Windows stuff) or #define INITGUID before all header includes. This will make CLSID/IID guids to be defined. Without it they are only declared extern. 
*/

/* NOTE simon: from guiddef.h */
#ifdef DEFINE_GUID
#undef DEFINE_GUID
#endif

#define DEFINE_GUID(name, l, w1, w2, b1, b2, b3, b4, b5, b6, b7, b8) \
EXTERN_C const GUID DECLSPEC_SELECTANY name \
= { l, w1, w2, { b1, b2,  b3,  b4,  b5,  b6,  b7,  b8 } }
/* NOTE simon: end of guiddef.h */

/* NOTE simon: from mmdeviceapi.h */

DEFINE_GUID( CLSID_MMDeviceEnumerator, 0xBCDE0395, 0xE52F, 0x467C, 0x8E, 0x3D, 0xC4, 0x57, 0x92, 0x91, 0x69, 0x2E );
DEFINE_GUID( IID_IMMDeviceEnumerator, 0xA95664D2, 0x9614, 0x4F35, 0xA7, 0x46, 0xDE, 0x8D, 0xB6, 0x36, 0x17, 0xE6 );

typedef struct IMMNotificationClient IMMNotificationClient;
typedef struct IMMDeviceEnumerator IMMDeviceEnumerator;
typedef struct IMMDeviceCollection IMMDeviceCollection;
typedef struct IMMDevice IMMDevice;

typedef /* [public][public][public] */
enum __MIDL___MIDL_itf_mmdeviceapi_0000_0000_0002
{
    eConsole	= 0,
    eMultimedia	= ( eConsole + 1 ) ,
    eCommunications	= ( eMultimedia + 1 ) ,
    ERole_enum_count	= ( eCommunications + 1 )
} 	ERole;

typedef /* [public][public][public][public][public] */
enum __MIDL___MIDL_itf_mmdeviceapi_0000_0000_0001
{
    eRender	= 0,
    eCapture	= ( eRender + 1 ) ,
    eAll	= ( eCapture + 1 ) ,
    EDataFlow_enum_count	= ( eAll + 1 )
} 	EDataFlow;

typedef struct IMMDeviceVtbl
{
    BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )(
                                                      IMMDevice * This,
                                                      /* [in] */ REFIID riid,
                                                      /* [annotation][iid_is][out] */
                                                      _COM_Outptr_  void **ppvObject);
    
    ULONG ( STDMETHODCALLTYPE *AddRef )(
                                        IMMDevice * This);
    
    ULONG ( STDMETHODCALLTYPE *Release )(
                                         IMMDevice * This);
    
    /* [helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *Activate )(
                                                                   IMMDevice * This,
                                                                   /* [annotation][in] */
                                                                   _In_ const IID* iid, /* NOTE simon: was _In_  REFIID iid, */
                                                                   /* [annotation][in] */
                                                                   _In_  DWORD dwClsCtx,
                                                                   /* [annotation][unique][in] */
                                                                   _In_opt_  PROPVARIANT *pActivationParams,
                                                                   /* [annotation][iid_is][out] */
                                                                   _Out_  void **ppInterface);
    
    /* [helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *OpenPropertyStore )(
                                                                            IMMDevice * This,
                                                                            /* [annotation][in] */
                                                                            _In_  DWORD stgmAccess,
                                                                            /* [annotation][out] */
                                                                            _Out_  IPropertyStore **ppProperties);
    
    /* [helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *GetId )(
                                                                IMMDevice * This,
                                                                /* [annotation][out] */
                                                                _Outptr_  LPWSTR *ppstrId);
    
    /* [helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *GetState )(
                                                                   IMMDevice * This,
                                                                   /* [annotation][out] */
                                                                   _Out_  DWORD *pdwState);
    
    END_INTERFACE
} IMMDeviceVtbl;

interface IMMDevice
{
    CONST_VTBL struct IMMDeviceVtbl *lpVtbl;
};

#define IMMDevice_QueryInterface(This,riid,ppvObject)	\
( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) )

#define IMMDevice_AddRef(This)	\
( (This)->lpVtbl -> AddRef(This) )

#define IMMDevice_Release(This)	\
( (This)->lpVtbl -> Release(This) )

#define IMMDevice_Activate(This,iid,dwClsCtx,pActivationParams,ppInterface)	\
( (This)->lpVtbl -> Activate(This,iid,dwClsCtx,pActivationParams,ppInterface) )

#define IMMDevice_OpenPropertyStore(This,stgmAccess,ppProperties)	\
( (This)->lpVtbl -> OpenPropertyStore(This,stgmAccess,ppProperties) )

#define IMMDevice_GetId(This,ppstrId)	\
( (This)->lpVtbl -> GetId(This,ppstrId) )

#define IMMDevice_GetState(This,pdwState)	\
( (This)->lpVtbl -> GetState(This,pdwState) )

typedef struct IMMDeviceEnumeratorVtbl
{
    BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )(
                                                      IMMDeviceEnumerator * This,
                                                      /* [in] */ REFIID riid,
                                                      /* [annotation][iid_is][out] */
                                                      _COM_Outptr_  void **ppvObject);
    
    ULONG ( STDMETHODCALLTYPE *AddRef )(
                                        IMMDeviceEnumerator * This);
    
    ULONG ( STDMETHODCALLTYPE *Release )(
                                         IMMDeviceEnumerator * This);
    
    /* [helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *EnumAudioEndpoints )(
                                                                             IMMDeviceEnumerator * This,
                                                                             /* [annotation][in] */
                                                                             _In_  EDataFlow dataFlow,
                                                                             /* [annotation][in] */
                                                                             _In_  DWORD dwStateMask,
                                                                             /* [annotation][out] */
                                                                             _Out_  IMMDeviceCollection **ppDevices);
    
    /* [helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *GetDefaultAudioEndpoint )(
                                                                                  IMMDeviceEnumerator * This,
                                                                                  /* [annotation][in] */
                                                                                  _In_  EDataFlow dataFlow,
                                                                                  /* [annotation][in] */
                                                                                  _In_  ERole role,
                                                                                  /* [annotation][out] */
                                                                                  _Out_  IMMDevice **ppEndpoint);
    
    /* [helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *GetDevice )(
                                                                    IMMDeviceEnumerator * This,
                                                                    /* [annotation][in] */
                                                                    _In_  LPCWSTR pwstrId,
                                                                    /* [annotation][out] */
                                                                    _Out_  IMMDevice **ppDevice);
    
    /* [helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *RegisterEndpointNotificationCallback )(
                                                                                               IMMDeviceEnumerator * This,
                                                                                               /* [annotation][in] */
                                                                                               _In_  IMMNotificationClient *pClient);
    
    /* [helpstring][id] */ HRESULT ( STDMETHODCALLTYPE *UnregisterEndpointNotificationCallback )(
                                                                                                 IMMDeviceEnumerator * This,
                                                                                                 /* [annotation][in] */
                                                                                                 _In_  IMMNotificationClient *pClient);
    
    END_INTERFACE
} IMMDeviceEnumeratorVtbl;

interface IMMDeviceEnumerator
{
    CONST_VTBL struct IMMDeviceEnumeratorVtbl *lpVtbl;
};

#define IMMDeviceEnumerator_QueryInterface(This,riid,ppvObject)	\
( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) )

#define IMMDeviceEnumerator_AddRef(This)	\
( (This)->lpVtbl -> AddRef(This) )

#define IMMDeviceEnumerator_Release(This)	\
( (This)->lpVtbl -> Release(This) )

#define IMMDeviceEnumerator_EnumAudioEndpoints(This,dataFlow,dwStateMask,ppDevices)	\
( (This)->lpVtbl -> EnumAudioEndpoints(This,dataFlow,dwStateMask,ppDevices) )

#define IMMDeviceEnumerator_GetDefaultAudioEndpoint(This,dataFlow,role,ppEndpoint)	\
( (This)->lpVtbl -> GetDefaultAudioEndpoint(This,dataFlow,role,ppEndpoint) )

#define IMMDeviceEnumerator_GetDevice(This,pwstrId,ppDevice)	\
( (This)->lpVtbl -> GetDevice(This,pwstrId,ppDevice) )

#define IMMDeviceEnumerator_RegisterEndpointNotificationCallback(This,pClient)	\
( (This)->lpVtbl -> RegisterEndpointNotificationCallback(This,pClient) )

#define IMMDeviceEnumerator_UnregisterEndpointNotificationCallback(This,pClient)	\
( (This)->lpVtbl -> UnregisterEndpointNotificationCallback(This,pClient) )

/* NOTE simon: end of mmdeviceapi.h */

/* NOTE simon: audioclient.h */
DEFINE_GUID( IID_IAudioClient, 0x1CB9AD4C, 0xDBFA, 0x4c32, 0xB1, 0x78, 0xC2, 0xF5, 0x68, 0xA7, 0x03, 0xB2 );
DEFINE_GUID( IID_IAudioRenderClient, 0xF294ACFC, 0x3146, 0x4483, 0xA7, 0xBF, 0xAD, 0xDC, 0xA7, 0xC2, 0x60, 0xE2 );

typedef u64 REFERENCE_TIME;
typedef struct IAudioClient IAudioClient;

typedef struct IAudioClientVtbl
{
    BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )(
                                                      IAudioClient * This,
                                                      /* [in] */ REFIID riid,
                                                      /* [annotation][iid_is][out] */
                                                      __RPC__deref_out  void **ppvObject);
    
    ULONG ( STDMETHODCALLTYPE *AddRef )(
                                        IAudioClient * This);
    
    ULONG ( STDMETHODCALLTYPE *Release )(
                                         IAudioClient * This);
    
    HRESULT ( STDMETHODCALLTYPE *Initialize )(
                                              IAudioClient * This,
                                              /* [annotation][in] */
                                              __in  AUDCLNT_SHAREMODE ShareMode,
                                              /* [annotation][in] */
                                              __in  DWORD StreamFlags,
                                              /* [annotation][in] */
                                              __in  REFERENCE_TIME hnsBufferDuration,
                                              /* [annotation][in] */
                                              __in  REFERENCE_TIME hnsPeriodicity,
                                              /* [annotation][in] */
                                              __in  const WAVEFORMATEX *pFormat,
                                              /* [annotation][in] */
                                              __in_opt  LPCGUID AudioSessionGuid);
    
    HRESULT ( STDMETHODCALLTYPE *GetBufferSize )(
                                                 IAudioClient * This,
                                                 /* [annotation][out] */
                                                 __out  UINT32 *pNumBufferFrames);
    
    HRESULT ( STDMETHODCALLTYPE *GetStreamLatency )(
                                                    IAudioClient * This,
                                                    /* [annotation][out] */
                                                    __out  REFERENCE_TIME *phnsLatency);
    
    HRESULT ( STDMETHODCALLTYPE *GetCurrentPadding )(
                                                     IAudioClient * This,
                                                     /* [annotation][out] */
                                                     __out  UINT32 *pNumPaddingFrames);
    
    HRESULT ( STDMETHODCALLTYPE *IsFormatSupported )(
                                                     IAudioClient * This,
                                                     /* [annotation][in] */
                                                     __in  AUDCLNT_SHAREMODE ShareMode,
                                                     /* [annotation][in] */
                                                     __in  const WAVEFORMATEX *pFormat,
                                                     /* [unique][annotation][out] */
                                                     __out_opt  WAVEFORMATEX **ppClosestMatch);
    
    HRESULT ( STDMETHODCALLTYPE *GetMixFormat )(
                                                IAudioClient * This,
                                                /* [annotation][out] */
                                                __out  WAVEFORMATEX **ppDeviceFormat);
    
    HRESULT ( STDMETHODCALLTYPE *GetDevicePeriod )(
                                                   IAudioClient * This,
                                                   /* [annotation][out] */
                                                   __out_opt  REFERENCE_TIME *phnsDefaultDevicePeriod,
                                                   /* [annotation][out] */
                                                   __out_opt  REFERENCE_TIME *phnsMinimumDevicePeriod);
    
    HRESULT ( STDMETHODCALLTYPE *Start )(
                                         IAudioClient * This);
    
    HRESULT ( STDMETHODCALLTYPE *Stop )(
                                        IAudioClient * This);
    
    HRESULT ( STDMETHODCALLTYPE *Reset )(
                                         IAudioClient * This);
    
    HRESULT ( STDMETHODCALLTYPE *SetEventHandle )(
                                                  IAudioClient * This,
                                                  /* [in] */ HANDLE eventHandle);
    
    HRESULT ( STDMETHODCALLTYPE *GetService )(
                                              IAudioClient * This,
                                              /* [annotation][in] */
                                              __in const IID* riid, /* NOTE simon: was __in  REFIID riid, */
                                              /* [annotation][iid_is][out] */
                                              __out  void **ppv);
    
    END_INTERFACE
} IAudioClientVtbl;

interface IAudioClient
{
    CONST_VTBL struct IAudioClientVtbl *lpVtbl;
};

#define IAudioClient_QueryInterface(This,riid,ppvObject)	\
( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) )

#define IAudioClient_AddRef(This)	\
( (This)->lpVtbl -> AddRef(This) )

#define IAudioClient_Release(This)	\
( (This)->lpVtbl -> Release(This) )

#define IAudioClient_Initialize(This,ShareMode,StreamFlags,hnsBufferDuration,hnsPeriodicity,pFormat,AudioSessionGuid)	\
( (This)->lpVtbl -> Initialize(This,ShareMode,StreamFlags,hnsBufferDuration,hnsPeriodicity,pFormat,AudioSessionGuid) )

#define IAudioClient_GetBufferSize(This,pNumBufferFrames)	\
( (This)->lpVtbl -> GetBufferSize(This,pNumBufferFrames) )

#define IAudioClient_GetStreamLatency(This,phnsLatency)	\
( (This)->lpVtbl -> GetStreamLatency(This,phnsLatency) )

#define IAudioClient_GetCurrentPadding(This,pNumPaddingFrames)	\
( (This)->lpVtbl -> GetCurrentPadding(This,pNumPaddingFrames) )

#define IAudioClient_IsFormatSupported(This,ShareMode,pFormat,ppClosestMatch)	\
( (This)->lpVtbl -> IsFormatSupported(This,ShareMode,pFormat,ppClosestMatch) )

#define IAudioClient_GetMixFormat(This,ppDeviceFormat)	\
( (This)->lpVtbl -> GetMixFormat(This,ppDeviceFormat) )

#define IAudioClient_GetDevicePeriod(This,phnsDefaultDevicePeriod,phnsMinimumDevicePeriod)	\
( (This)->lpVtbl -> GetDevicePeriod(This,phnsDefaultDevicePeriod,phnsMinimumDevicePeriod) )

#define IAudioClient_Start(This)	\
( (This)->lpVtbl -> Start(This) )

#define IAudioClient_Stop(This)	\
( (This)->lpVtbl -> Stop(This) )

#define IAudioClient_Reset(This)	\
( (This)->lpVtbl -> Reset(This) )

#define IAudioClient_SetEventHandle(This,eventHandle)	\
( (This)->lpVtbl -> SetEventHandle(This,eventHandle) )

#define IAudioClient_GetService(This,riid,ppv)	\
( (This)->lpVtbl -> GetService(This,riid,ppv) )

typedef struct IAudioRenderClient IAudioRenderClient;

typedef struct IAudioRenderClientVtbl
{
    BEGIN_INTERFACE
        
        HRESULT ( STDMETHODCALLTYPE *QueryInterface )(
                                                      IAudioRenderClient * This,
                                                      /* [in] */ REFIID riid,
                                                      /* [annotation][iid_is][out] */
                                                      __RPC__deref_out  void **ppvObject);
    
    ULONG ( STDMETHODCALLTYPE *AddRef )(
                                        IAudioRenderClient * This);
    
    ULONG ( STDMETHODCALLTYPE *Release )(
                                         IAudioRenderClient * This);
    
    HRESULT ( STDMETHODCALLTYPE *GetBuffer )(
                                             IAudioRenderClient * This,
                                             /* [annotation][in] */
                                             __in  UINT32 NumFramesRequested,
                                             /* [annotation][out] */
                                             __out  BYTE **ppData);
    
    HRESULT ( STDMETHODCALLTYPE *ReleaseBuffer )(
                                                 IAudioRenderClient * This,
                                                 /* [annotation][in] */
                                                 __in  UINT32 NumFramesWritten,
                                                 /* [annotation][in] */
                                                 __in  DWORD dwFlags);
    
    END_INTERFACE
} IAudioRenderClientVtbl;

interface IAudioRenderClient
{
    CONST_VTBL struct IAudioRenderClientVtbl *lpVtbl;
};

#define IAudioRenderClient_QueryInterface(This,riid,ppvObject)	\
( (This)->lpVtbl -> QueryInterface(This,riid,ppvObject) )

#define IAudioRenderClient_AddRef(This)	\
( (This)->lpVtbl -> AddRef(This) )

#define IAudioRenderClient_Release(This)	\
( (This)->lpVtbl -> Release(This) )

#define IAudioRenderClient_GetBuffer(This,NumFramesRequested,ppData)	\
( (This)->lpVtbl -> GetBuffer(This,NumFramesRequested,ppData) )

#define IAudioRenderClient_ReleaseBuffer(This,NumFramesWritten,dwFlags)	\
( (This)->lpVtbl -> ReleaseBuffer(This,NumFramesWritten,dwFlags) )

/* NOTE simon: end of audioclient.h */

/* NOTE simon: From dwmapi.h */
EXTERN_C DECLSPEC_IMPORT HRESULT STDAPICALLTYPE DwmIsCompositionEnabled( BOOL* pfEnabled );
/* NOTE simon: end of dwmapi.h */

#endif
