#ifndef TGA_H
#define TGA_H

/* NOTE simon; TGA are encoded as ARGB, A beeing the most significant bit, B beeing the least significant bit. */

#include <stdint.h>

#ifndef TGA_ASSERT_OFF
#define tga_assert( expression ) if ( !( expression ) ) { *( ( volatile uint32_t* ) 0 ) = 0xa553a553; }
#else
#define tga_assert( expression )
#endif

#define tga_header_size 18
#define tga_footer_size 26

uint8_t tga_footer[] = "\0\0\0\0\0\0\0\0TRUEVISION-XFILE.\0";

typedef enum tga_colormap_type {
    tga_colormap_type_no_color_map = 0,
    tga_colormap_type_color_map = 1,
} tga_colormap_type;

typedef enum tga_image_type {
    tga_image_type_no_data = 0,
    tga_image_type_color_mapped = 1,
    tga_image_type_true_color = 2,
    tga_image_type_black_and_white = 3,
    tga_image_type_color_maped_rle = 9,
    tga_image_type_true_color_rle = 10,
    tga_image_type_black_and_white_rle = 11,
} tga_image_type;

typedef enum tga_pixel_destination {
    tga_pixel_destination_bottom_left = 0,
    tga_pixel_destination_bottom_right = 1,
    tga_pixel_destination_top_left = 2,
    tga_pixel_destination_top_right = 3,
} tga_pixel_destination;

uint8_t tga_get_id_length( void* header ) {
    uint8_t* f = ( uint8_t* ) header;
    uint8_t result = f[ 0 ];
    return result;
}

void tga_set_id_length( void* header, uintptr_t length ) {
    tga_assert( length <= 0xff );
    uint8_t* f = ( uint8_t* ) header;
    f[ 0 ] = ( uint8_t ) length;
}

uint8_t tga_get_colormap_type( void* header ) {
    uint8_t* f = ( uint8_t* ) header;
    uint8_t result = f[ 1 ];
    return result;
}

void tga_set_colormap_type( void* header, uintptr_t type ) {
    tga_assert( type <= 0xff );
    uint8_t* f = ( uint8_t* ) header;
    f[ 1 ] = ( uint8_t ) type;
}

uint8_t tga_get_image_type( void* header ) {
    uint8_t* f = ( uint8_t* ) header;
    uint8_t result = f[ 2 ];
    return result;
}

void tga_set_image_type( void* header, uintptr_t type ) {
    tga_assert( type <= 0xff );
    uint8_t* f = ( uint8_t* ) header;
    f[ 2 ] = ( uint8_t ) type;
}

uint32_t tga_is_color_maped( void* header ) {
    uint8_t* f = ( uint8_t* ) header;
    uint32_t result = ( f[ 2 ] & 0x03 ) == 0x01;
    return result;
}

uint32_t tga_is_true_color( void* header ) {
    uint8_t* f = ( uint8_t* ) header;
    uint32_t result = ( f[ 2 ] & 0x03 ) == 0x02;
    return result;
}

uint32_t tga_is_black_and_white( uint8_t* header ) {
    uint8_t* f = ( uint8_t* ) header;
    uint32_t result = ( f[ 2 ] & 0x03 ) == 0x03;
    return result;
}

uint32_t tga_is_rle( void* header ) {
    uint8_t* f = ( uint8_t* ) header;
    uint32_t result = ( f[ 2 ] & 0x08 ) == 0x08;
    return result;
}

uint16_t tga_get_color_map_first_entry_index( void* header ) {
    uint8_t* f = ( uint8_t* ) header;
    uint16_t result = *( ( uint16_t* ) ( f + 3 ) );
    return result;
}

void tga_set_color_map_first_entry_index( void* header, uintptr_t index ) {
    tga_assert( index <= 0xffff );
    uint8_t* f = ( uint8_t* ) header;
    *( ( uint16_t* ) ( f + 3 ) ) = ( uint16_t ) index;
}

uint16_t tga_get_color_map_length( void* header ) {
    uint8_t* f = ( uint8_t* ) header;
    uint16_t result = *( ( uint16_t* ) ( f + 5 ) );
    return result;
}

void tga_set_color_map_length( void* header, uintptr_t length ) {
    tga_assert( length <= 0xffff );
    uint8_t* f = ( uint8_t* ) header;
    *( ( uint16_t* ) ( f + 5 ) ) = ( uint16_t ) length;
}

uint8_t tga_get_color_map_entry_size( void* header ) {
    uint8_t* f = ( uint8_t* ) header;
    uint8_t result = f[ 7 ];
    return result;
}

void tga_set_color_map_entry_size( void* header, uintptr_t size ) {
    tga_assert( size <= 0xff );
    uint8_t* f = ( uint8_t* ) header;
    f[ 7 ] = ( uint8_t ) size;
}

uint16_t tga_get_x_origin( void* header ) {
    uint8_t* f = ( uint8_t* ) header;
    uint16_t result = *( ( uint16_t* ) ( f + 8 ) );
    return result;
}

void tga_set_x_origin( void* header, uintptr_t x_origin ) {
    tga_assert( x_origin <= 0xffff );
    uint8_t* f = ( uint8_t* ) header;
    *( ( uint16_t* ) ( f + 8 ) ) = ( uint16_t ) x_origin;
}

uint16_t tga_get_y_origin( void* header ) {
    uint8_t* f = ( uint8_t* ) header;
    uint16_t result = *( ( uint16_t* ) ( f + 10 ) );
    return result;
}

void tga_set_y_origin( void* header, uintptr_t y_origin ) {
    tga_assert( y_origin <= 0xffff );
    uint8_t* f = ( uint8_t* ) header;
    *( ( uint16_t* ) ( f + 10 ) ) = ( uint16_t ) y_origin;
}

uint16_t tga_get_width( void* header ) {
    uint8_t* f = ( uint8_t* ) header;
    uint16_t result = *( ( uint16_t* ) ( f + 12 ) );
    return result;
}

void tga_set_width( void* header, uintptr_t width ) {
    tga_assert( width <= 0xffff );
    uint8_t* f = ( uint8_t* ) header;
    *( ( uint16_t* ) ( f + 12 ) ) = ( uint16_t ) width;
}

uint16_t tga_get_height( void* header ) {
    uint8_t* f = ( uint8_t* ) header;
    uint16_t result = *( ( uint16_t* ) ( f + 14 ) );
    return result;
}

void tga_set_height( void* header, uintptr_t height ) {
    tga_assert( height <= 0xffff );
    uint8_t* f = ( uint8_t* ) header;
    *( ( uint16_t* ) ( f + 14 ) ) = ( uint16_t ) height;
}

uint8_t tga_get_pixel_depth( void* header ) {
    uint8_t* f = ( uint8_t* ) header;
    uint8_t result = f[ 16 ];
    return result;
}

void tga_set_pixel_depth( void* header, uintptr_t pixel_depth ) {
    tga_assert( pixel_depth <= 0xff );
    uint8_t* f = ( uint8_t* ) header;
    f[ 16 ] = ( uint8_t ) pixel_depth;
}

uint8_t tga_get_bytes_per_pixel( void* header ) {
    uint8_t* f = ( uint8_t* ) header;
    uint8_t result = f[ 16 ] / 8;
    return result;
}

void tga_set_bytes_per_pixel( void* header, uintptr_t bytes_per_pixel ) {
    tga_assert( bytes_per_pixel <= 0xff );
    uint8_t* f = ( uint8_t* ) header;
    f[ 16 ] = ( uint8_t ) bytes_per_pixel * 8;
}

uint8_t tga_get_image_descriptor( void* header ) {
    uint8_t* f = ( uint8_t* ) header;
    uint8_t result = f[ 17 ];
    return result;
}

void tga_set_alpha_bit_count( void* header, uintptr_t alpha_bit_count ) {
    tga_assert( alpha_bit_count <= 0xff );
    uint8_t* f = ( uint8_t* ) header;
    f[ 17 ] &= 0xf0;
    f[ 17 ] |= ( ( ( uint8_t ) alpha_bit_count ) & 0x0f );
}

void tga_set_first_pixel_destination( void* header, uintptr_t pixel_destination ) {
    tga_assert( pixel_destination <= 0xff );
    uint8_t* f = ( uint8_t* ) header;
    f[ 17 ] &= 0x30;
    f[ 17 ] |= ( ( ( uint8_t ) pixel_destination ) & 0x03 ) << 4;
}

uintptr_t tga_get_file_size( uintptr_t width, uintptr_t height, uintptr_t bytes_per_pixel ) {
    uintptr_t result = tga_header_size + width * height * bytes_per_pixel + tga_footer_size;
    return result;
}

uintptr_t tga_get_file_size_from_header( void* header ) {
    uintptr_t result = tga_header_size + tga_get_width( header ) * tga_get_height( header ) * tga_get_bytes_per_pixel( header ) + tga_footer_size;
    return result;
}

void tga_initialize_header( void* header, uintptr_t width, uintptr_t height, uintptr_t bytes_per_pixel, tga_image_type type ) {
    
    tga_set_width( header, width );
    tga_set_height( header, height );
    tga_set_bytes_per_pixel( header, bytes_per_pixel );
    tga_set_image_type( header, type );
    
    if ( bytes_per_pixel == 4 ) {
        tga_set_alpha_bit_count( header, 8 );
    } else {
        tga_set_alpha_bit_count( header, 0 );
    }
}

uint8_t* tga_get_image_id( void* file ) {
    uint8_t* f = ( uint8_t* ) file;
    uint8_t* result = ( f + tga_header_size );
    return result;
}

uint8_t* tga_get_color_map( void* file ) {
    uint8_t* f = ( uint8_t* ) file;
    uint8_t* result = ( f + tga_header_size + tga_get_id_length( file ) );
    return result;
}

uint8_t* tga_get_image_data( void* file ) {
    
    uint8_t* f = ( uint8_t* ) file;
    uint8_t* result;
    tga_assert( ( tga_get_color_map_entry_size( file ) % 8 ) == 0 );
    result = ( f + tga_header_size + tga_get_id_length( file ) + ( tga_get_color_map_entry_size( file ) / 8 ) * tga_get_color_map_length( file ) );
    return result;
}

uint8_t* tga_get_footer( void* file, uintptr_t file_size ) {
    uint8_t* f = ( uint8_t* ) file;
    uint8_t* result = ( f + file_size - tga_footer_size );
    return result;
}

void tga_initialize_footer( void* footer ) {
    
    uint8_t* f = ( uint8_t* ) footer;
    
    for ( uintptr_t i = 0; i < tga_footer_size; i++ ) {
        f[ i ] = tga_footer[ i ];
    }
}

void tga_initialize( void* file, uintptr_t width, uintptr_t height, uintptr_t bytes_per_pixel, tga_image_type type ) {
    
    uint8_t* f = ( uint8_t* ) file;
    tga_initialize_header( f, width, height, bytes_per_pixel, type );
    tga_initialize_footer( f + tga_header_size + width * height * bytes_per_pixel );
}

uint32_t tga_get_extension_area_offset( void* footer ) {
    uint32_t result = *( ( uint32_t* ) footer );
    return result;
}

void tga_set_extension_area_offset( void* footer, uintptr_t offset_from_begining ) {
    tga_assert( offset_from_begining <= 0xffffffff );
    uint8_t* f = ( uint8_t* ) footer;
    *( ( uint32_t* ) f ) = ( uint32_t ) offset_from_begining;
}

uint32_t tga_get_developper_directory_offset( void* footer ) {
    uint8_t* f = ( uint8_t* ) footer;
    uint32_t result = *( ( uint32_t* ) ( f + 4 ) );
    return result;
}

void tga_set_developper_directory_offset( void* footer, uintptr_t offset_from_begining ) {
    tga_assert( offset_from_begining <= 0xffffffff );
    uint8_t* f = ( uint8_t* ) footer;
    *( ( uint32_t* ) ( f + 4 ) ) = ( uint32_t ) offset_from_begining;
}

uint32_t tga_footer_contains_signature( void* footer ) {
    
    uint32_t result = 1;
    uint8_t* signature = ( ( uint8_t* ) footer ) + 8;
    
    for ( uintptr_t i = 0; result && i < 16; i++ ) {
        result = ( signature[ i ] == tga_footer[ i + 8 ] );
    }
    
    return result;
}

uint32_t tga_footer_is_valid( void* footer ) {
    
    uint8_t* f = ( uint8_t* ) footer;
    uint32_t result = *( ( int8_t* ) ( f + 24 ) ) == '.';
    return result;
}

uint8_t* tga_get_extension_area( void* file, intptr_t file_size ) {
    uint8_t* f = ( uint8_t* ) file;
    uint8_t* result = f + *( ( int32_t* ) tga_get_footer( file, file_size ) );
    return result;
}

uint8_t* tga_get_developper_directory( void* file, intptr_t file_size ) {
    uint8_t* f = ( uint8_t* ) file;
    uint8_t* result = f + *( ( int32_t* ) ( tga_get_footer( file, file_size ) + 4 ) );
    return result;
}

#if 0
void tga_copy_data( void* imageData, void* destination, int32_t width, int32_t height, int32_t bytesPerPixel, uint32_t convert3to4BytesPerPixel ) {
    
    int32_t destinationBytesPerPixel = bytesPerPixel;
    uint8_t* source;
    uint8_t* data;
    int32_t sourcePitch;
    int32_t destinationPitch;
    int32_t y;
    
    if ( convert3to4BytesPerPixel ) {
        destinationBytesPerPixel = 4;
    }
    
    source = ( uint8_t* ) imageData;
    data = ( uint8_t* ) destination;
    
    sourcePitch = width * bytesPerPixel;
    destinationPitch = width * destinationBytesPerPixel;
    
    for ( y = 0; y < height; y++ ) {
        
        uint8_t* dstData = data;
        uint8_t* srcData = source;
        int32_t x;
        
        for ( x = 0; x < width; x++ ) {
            
            int32_t byteIndex = 0;
            
            while( byteIndex < bytesPerPixel ) {
                *dstData++ = *srcData++;
                byteIndex++;
            }
            
            if ( byteIndex < destinationBytesPerPixel ) {
                *dstData++ = 0xff;
            }
        }
        
        source += sourcePitch;
        data += destinationPitch;
    }
}
#endif

void tga_rle_decode( void* image_data, void* destination, uintptr_t width, uintptr_t height, uintptr_t bytes_per_pixel, uint32_t convert_3_to_4_bytes_per_pixel ) {
    
    uint8_t* data = ( uint8_t* ) image_data;
    uint8_t* dest = ( uint8_t* ) destination;
    uintptr_t destination_bytes_per_pixel = bytes_per_pixel;
    
    if ( bytes_per_pixel == 3 && convert_3_to_4_bytes_per_pixel ) {
        destination_bytes_per_pixel = 4;
    }
    
    uint64_t copied_pixel_count = 0;
    uint64_t total_pixel_count = width * height;
    
    while ( copied_pixel_count < total_pixel_count ) {
        
        uint8_t repetition_count = ( ( *data ) & 0x7f ) + 1;
        
        if ( ( ( *data ) & 0x80 ) == 0x80 ) {
            // NOTE simon: Run length packet.
            data++;
            
            uint8_t source[ 4 ] = { 0 };
            
            for ( uintptr_t i = 0; i < bytes_per_pixel; i++ ) {
                source[ i ] = data[ i ];
            }
            
            if ( bytes_per_pixel < destination_bytes_per_pixel ) {
                source[ 3 ] = 0xff;
            }
            
            for ( uintptr_t i = 0; i < repetition_count; i++ ) {
                
                for ( uintptr_t byte_index = 0; byte_index < destination_bytes_per_pixel; byte_index++ ) {
                    dest[ byte_index ] = source[ byte_index ];
                }
                
                dest += destination_bytes_per_pixel;
            }
            
            data += bytes_per_pixel;
            
        } else {
            // NOTE simon: Raw packet.
            data++;
            
            for ( uintptr_t i = 0; i < repetition_count; i++ ) {
                
                for ( uintptr_t byte_index = 0; byte_index < bytes_per_pixel; byte_index++ ) {
                    dest[ byte_index ] = data[ byte_index ];
                }
                
                if ( bytes_per_pixel < destination_bytes_per_pixel ){
                    dest[ 3 ] = 0xff;
                }
                
                data += bytes_per_pixel;
                dest += destination_bytes_per_pixel;
            }
        }
        
        copied_pixel_count += repetition_count;
    }
}

uint32_t tga_are_pixel_equal_8( void* a, void* b ) {
    return ( *( uint8_t* ) a ) == ( *( uint8_t* ) b );
}

uint32_t tga_are_pixel_equal_16( void* a, void* b ) {
    return ( *( uint16_t* ) a ) == ( *( uint16_t* ) b );
}

uint32_t tga_are_pixel_equal_24( void* a, void* b ) {
    
    int32_t aValue = ( *( int32_t * ) a ) & 0x00ffffff;
    int32_t bValue = ( *( int32_t * ) b ) & 0x00ffffff;
    return aValue == bValue;
}

uint32_t tga_are_pixel_equal_32( void* a, void* b ) {
    return ( *( int32_t* ) a ) == ( *( int32_t* ) b );
}

typedef uint32_t tga_are_pixel_equal_fnc( void* a, void* b );

uintptr_t tga_rle_encode( void* src, void* dst, uintptr_t width, uintptr_t height, uintptr_t bytes_per_pixel ) {
    
    uint8_t* source = ( uint8_t* ) src;
    uint8_t* destination = ( uint8_t* ) dst;
    
    tga_are_pixel_equal_fnc* functions[ 4 ] = {
        tga_are_pixel_equal_8,
        tga_are_pixel_equal_16,
        tga_are_pixel_equal_24,
        tga_are_pixel_equal_32
    };
    
    tga_are_pixel_equal_fnc* are_pixel_equal = functions[ bytes_per_pixel - 1 ];
    
    uintptr_t data_size = width * height * bytes_per_pixel;
    
    uint8_t* current = source;
    uint8_t* end = source + data_size;
    uintptr_t x = 0;
    
    uint8_t* packet = destination;
    
    while ( current < end ) {
        
        uint8_t* pixel_data = packet + 1;
        
        for ( uintptr_t i = 0; dst && i < bytes_per_pixel; i++ ) {
            pixel_data[ i ] = current[ i ];
        }
        
        pixel_data += bytes_per_pixel;
        
        uint8_t* pixel_a = current;
        uint8_t* pixel_b = pixel_a + bytes_per_pixel;
        
        uint32_t rle = ( ( x + 1 < width ) && are_pixel_equal( pixel_a, pixel_b ) );
        uint8_t count = 0;
        
        if ( rle ) {
            
            count = 2;
            
            for ( uintptr_t i = x + 2; count < 128 && i < width; i++ ) {
                
                pixel_b += bytes_per_pixel;
                
                if ( are_pixel_equal( pixel_a, pixel_b ) ) {
                    count++;
                } else {
                    break;
                }
            }
            
            if ( dst ) {
                *packet = ( 0x80 | ( count - 1 ) );
            }
            
            packet += ( 1 + bytes_per_pixel );
            
        } else {
            
            count = 1;
            
            for ( uintptr_t i = x + 1; count < 128 && i < width; i++ ) {
                
                pixel_a = pixel_b;
                pixel_b += bytes_per_pixel;
                
                if ( i + 1 >= width || !are_pixel_equal( pixel_a, pixel_b ) ) { /* NOTE simon: if pixel b is outside the line or pixel_b is different. */
                    
                    count++;
                    
                    for ( uintptr_t byte_index = 0; dst && byte_index < bytes_per_pixel; byte_index++ ) {
                        pixel_data[ byte_index ] = pixel_a[ byte_index ];
                    }
                    
                    pixel_data += bytes_per_pixel;
                    
                } else {
                    break;
                }
            }
            
            if ( dst ) {
                *packet = ( 0x00 | ( count - 1 ) );
            }
            
            packet += ( 1 + count * bytes_per_pixel );
        }
        
        current += count * bytes_per_pixel;
        x += count;
        
        if ( x == width ) {
            x = 0;
        }
        
        tga_assert( x < width );
    }
    
    uintptr_t result = packet - destination;
    
    return result;
}

#endif