#ifndef WINDOW_H

#include "sound.h"

typedef struct platform_specific_t {
    
#if WINDOWS
    
    HWND window_handle;
    HDC device_context;
    HGLRC render_context;
    
    struct IAudioClient* audio_client;
    struct IAudioRenderClient* audio_render_client;
    
#elif LINUX
    
    Display* display;
    Window window;
    GLXContext render_context;
    
#else
# error Unsuported platform.
#endif
} platform_specific_t;

typedef struct window_state_t {
    
    b32 running;
    b32 close_requested;
    u32 has_focus;
    s32 x;
    s32 y;
    u32 width;
    u32 height;
    b32 resized;
    u32 screen_width;
    u32 screen_height;
    u16 screen_refresh_rate;
    u8 keys[ 256 ];
    u8 scancodes[ 256 ];
    u32 codepoints[ 32 ];
    u32 codepoint_count;
    u8 codepoints_byte_count_utf8[ 32 ];
    u8 codepoints_utf8[ 32 * 4 ];
    u8 codepoints_total_byte_count_utf8;
    s32 mouse_x;
    s32 mouse_y;
    r32 scroll_increment;
    platform_specific_t platform;
} window_state_t;

typedef enum input_virtual_key_t {
    
    vk_none,
    
    vk_mouse_left,
    vk_mouse_right,
    vk_mouse_midle,
    vk_mouse_4,
    vk_mouse_5,
    vk_cancel,
    vk_backspace,
    vk_tab,
    vk_clear,
    vk_return,
    vk_shift,
    vk_control,
    vk_alt,
    vk_pause,
    vk_caps_lock,
    vk_kana,
    vk_hangul,
    vk_junja,
    vk_final,
    vk_hanja,
    vk_kanji,
    vk_escape,
    vk_convert,
    vk_non_convert,
    vk_accept,
    vk_mode_change,
    vk_space,
    vk_page_up,
    vk_page_down,
    vk_end,
    vk_home,
    vk_left,
    vk_up,
    vk_right,
    vk_down,
    vk_select,
    vk_print,
    vk_execute,
    vk_print_screen,
    vk_insert,
    vk_delete,
    vk_help,
    
    vk_0,
    vk_1,
    vk_2,
    vk_3,
    vk_4,
    vk_5,
    vk_6,
    vk_7,
    vk_8,
    vk_9,
    
    vk_a,
    vk_b,
    vk_c,
    vk_d,
    vk_e,
    vk_f,
    vk_g,
    vk_h,
    vk_i,
    vk_j,
    vk_k,
    vk_l,
    vk_m,
    vk_n,
    vk_o,
    vk_p,
    vk_q,
    vk_r,
    vk_s,
    vk_t,
    vk_u,
    vk_v,
    vk_w,
    vk_x,
    vk_y,
    vk_z,
    
    vk_left_meta,
    vk_right_meta,
    vk_application,
    vk_sleep,
    vk_numpad_0,
    vk_numpad_1,
    vk_numpad_2,
    vk_numpad_3,
    vk_numpad_4,
    vk_numpad_5,
    vk_numpad_6,
    vk_numpad_7,
    vk_numpad_8,
    vk_numpad_9,
    
    vk_multiply,
    vk_divide,
    vk_add,
    vk_subtract,
    vk_separator,
    vk_decimal,
    
    vk_f1,
    vk_f2,
    vk_f3,
    vk_f4,
    vk_f5,
    vk_f6,
    vk_f7,
    vk_f8,
    vk_f9,
    vk_f10,
    vk_f11,
    vk_f12,
    vk_f13,
    vk_f14,
    vk_f15,
    vk_f16,
    vk_f17,
    vk_f18,
    vk_f19,
    vk_f20,
    vk_f21,
    vk_f22,
    vk_f23,
    vk_f24,
    
    vk_num_lock,
    vk_scroll_lock,
    vk_left_shift,
    vk_right_shift,
    vk_left_control,
    vk_right_control,
    vk_left_alt,
    vk_right_alt,
    
    vk_browser_back,
    vk_browser_forward,
    vk_browser_refresh,
    vk_browser_stop,
    vk_browser_search,
    vk_browser_favorites,
    vk_browser_home,
    vk_volume_mute,
    vk_volume_down,
    vk_volume_up,
    vk_media_next,
    vk_media_previous,
    vk_media_stop,
    vk_media_play,
    vk_launch_mail,
    vk_launch_media,
    vk_launch_application_1,
    vk_launch_application_2,
    
    vk_oem_plus,
    vk_oem_minus,
    vk_oem_comma,
    vk_oem_period,
    vk_oem_1,
    vk_oem_2,
    vk_oem_3,
    vk_oem_4,
    vk_oem_5,
    vk_oem_6,
    vk_oem_7,
    vk_oem_8,
    vk_oem_102,
    vk_oem_clear,
    vk_process_key,
    vk_packet,
    vk_attn,
    vk_crsel,
    vk_exsel,
    vk_ereof,
    vk_play,
    vk_zoom,
    vk_noname,
    vk_pa1,
    
    vk_count
} input_virtual_key_t;

/* NOTE simon: Keyboard input article: https://www.sisyphe.be/donotdelete/keyboard_input_article.txt . */
typedef enum input_scancode_t {
    
    sc_escape = 0x01,
    sc_1 = 0x02,
    sc_2 = 0x03,
    sc_3 = 0x04,
    sc_4 = 0x05,
    sc_5 = 0x06,
    sc_6 = 0x07,
    sc_7 = 0x08,
    sc_8 = 0x09,
    sc_9 = 0x0A,
    sc_0 = 0x0B,
    sc_minus = 0x0C,
    sc_equals = 0x0D,
    sc_backspace = 0x0E,
    sc_tab = 0x0F,
    sc_q = 0x10,
    sc_w = 0x11,
    sc_e = 0x12,
    sc_r = 0x13,
    sc_t = 0x14,
    sc_y = 0x15,
    sc_u = 0x16,
    sc_i = 0x17,
    sc_o = 0x18,
    sc_p = 0x19,
    sc_bracket_left = 0x1A,
    sc_bracket_right = 0x1B,
    sc_enter = 0x1C,
    sc_control_left = 0x1D,
    sc_a = 0x1E,
    sc_s =0x1F,
    sc_d = 0x20,
    sc_f = 0x21,
    sc_g = 0x22,
    sc_h = 0x23,
    sc_j = 0x24,
    sc_k = 0x25,
    sc_l = 0x26,
    sc_semicolon = 0x27,
    sc_apostrophe = 0x28,
    sc_grave = 0x29,
    sc_shift_left = 0x2A,
    sc_backslash = 0x2B,
    sc_z = 0x2C,
    sc_x = 0x2D,
    sc_c = 0x2E,
    sc_v = 0x2F,
    sc_b = 0x30,
    sc_n = 0x31,
    sc_m = 0x32,
    sc_comma = 0x33,
    sc_preiod = 0x34,
    sc_slash = 0x35,
    sc_shift_right = 0x36,
    sc_numpad_multiply = 0x37,
    sc_alt_left = 0x38,
    sc_space = 0x39,
    sc_caps_lock = 0x3A,
    sc_f1 = 0x3B,
    sc_f2 = 0x3C,
    sc_f3 = 0x3D,
    sc_f4 = 0x3E,
    sc_f5 = 0x3F,
    sc_f6 = 0x40,
    sc_f7 = 0x41,
    sc_f8 = 0x42,
    sc_f9 = 0x43,
    sc_f10 = 0x44,
    sc_num_lock = 0x45,
    sc_scroll_lock = 0x46,
    sc_numpad_7 = 0x47,
    sc_numpad_8 = 0x48,
    sc_numpad_9 = 0x49,
    sc_numpad_minus = 0x4A,
    sc_numpad_4 = 0x4B,
    sc_numpad_5 = 0x4C,
    sc_numpad_6 = 0x4D,
    sc_numpad_plus = 0x4E,
    sc_numpad_1 = 0x4F,
    sc_numpad_2 = 0x50,
    sc_numpad_3 = 0x51,
    sc_numpad_0 = 0x52,
    sc_numpad_period = 0x53,
    sc_alt_print_screen = 0x54, /* Alt + print screen. MapVirtualKeyEx( VK_SNAPSHOT, MAPVK_VK_TO_VSC_EX, 0 ) returns scancode 0x54. */
    sc_bracket_angle = 0x56, /* Key between the left shift and Z. */
    sc_f11 = 0x57,
    sc_f12 = 0x58,
    sc_oem_1 = 0x5a, /* VK_OEM_WSCTRL */
    sc_oem_2 = 0x5b, /* VK_OEM_FINISH */
    sc_oem_3 = 0x5c, /* VK_OEM_JUMP */
    sc_erase_eof = 0x5d,
    sc_oem_4 = 0x5e, /* VK_OEM_BACKTAB */
    sc_oem_5 = 0x5f, /* VK_OEM_AUTO */
    sc_zoom = 0x62,
    sc_help = 0x63,
    sc_f13 = 0x64,
    sc_f14 = 0x65,
    sc_f15 = 0x66,
    sc_f16 = 0x67,
    sc_f17 = 0x68,
    sc_f18 = 0x69,
    sc_f19 = 0x6a,
    sc_f20 = 0x6b,
    sc_f21 = 0x6c,
    sc_f22 = 0x6d,
    sc_f23 = 0x6e,
    sc_oem_6 = 0x6f, /* VK_OEM_PA3 */
    sc_katakana = 0x70,
    sc_oem_7 = 0x71, /* VK_OEM_RESET */
    sc_f24 = 0x76,
    sc_sbcschar = 0x77,
    sc_convert = 0x79,
    sc_nonconvert = 0x7B, /* VK_OEM_PA1 */
    
    sc_media_previous = 0xE010,
    sc_media_next = 0xE019,
    sc_numpad_enter = 0xE01C,
    sc_control_right = 0xE01D,
    sc_volume_mute = 0xE020,
    sc_launch_app2 = 0xE021,
    sc_media_play = 0xE022,
    sc_media_stop = 0xE024,
    sc_volume_down = 0xE02E,
    sc_volume_up = 0xE030,
    sc_browser_home = 0xE032,
    sc_numpad_divide = 0xE035,
    sc_print_screen = 0xE037,
    /*
    sc_printScreen:
    - make: 0xE02A 0xE037
    - break: 0xE0B7 0xE0AA
    - MapVirtualKeyEx( VK_SNAPSHOT, MAPVK_VK_TO_VSC_EX, 0 ) returns scancode 0x54;
    - There is no VK_KEYDOWN with VK_SNAPSHOT.
    */
    sc_alt_right = 0xE038,
    sc_cancel = 0xE046, /* CTRL + Pause */
    sc_home = 0xE047,
    sc_arrow_up = 0xE048,
    sc_page_up = 0xE049,
    sc_arrow_left = 0xE04B,
    sc_arrow_right = 0xE04D,
    sc_end = 0xE04F,
    sc_arrow_down = 0xE050,
    sc_page_down = 0xE051,
    sc_insert = 0xE052,
    sc_delete = 0xE053,
    sc_meta_left = 0xE05B,
    sc_meta_right = 0xE05C,
    sc_application = 0xE05D,
    sc_power = 0xE05E,
    sc_sleep = 0xE05F,
    sc_wake = 0xE063,
    sc_browser_search = 0xE065,
    sc_browser_favorites = 0xE066,
    sc_browser_refresh = 0xE067,
    sc_browser_stop = 0xE068,
    sc_browser_forward = 0xE069,
    sc_browser_back = 0xE06A,
    sc_launch_app1 = 0xE06B,
    sc_launch_email = 0xE06C,
    sc_launch_media = 0xE06D,
    
    sc_pause = 0xE11D45,
    /*
    sc_pause:
    - make: 0xE11D 45 0xE19D C5
    - make in raw input: 0xE11D 0x45
    - break: none
    - No repeat when you hold the key down
    - There are no break so I don't know how the key down/up is expected to work. Raw input sends "keydown" and "keyup" messages, and it appears that the keyup message is sent directly after the keydown message (you can't hold the key down) so depending on when GetMessage or PeekMessage will return messages, you may get both a keydown and keyup message "at the same time". If you use VK messages most of the time you only get keydown messages, but some times you get keyup messages too.
    - when pressed at the same time as one or both control keys, generates a 0xE046 (sc_cancel) and the string for that scancode is "break".
    */
} input_scancode_t;

u32 input_scancode_list[ ] = {
    
    sc_escape,
    sc_1,
    sc_2,
    sc_3,
    sc_4,
    sc_5,
    sc_6,
    sc_7,
    sc_8,
    sc_9,
    sc_0,
    sc_minus,
    sc_equals,
    sc_backspace,
    sc_tab,
    sc_q,
    sc_w,
    sc_e,
    sc_r,
    sc_t,
    sc_y,
    sc_u,
    sc_i,
    sc_o,
    sc_p,
    sc_bracket_left,
    sc_bracket_right,
    sc_enter,
    sc_control_left,
    sc_a,
    sc_s,
    sc_d,
    sc_f,
    sc_g,
    sc_h,
    sc_j,
    sc_k,
    sc_l,
    sc_semicolon,
    sc_apostrophe,
    sc_grave,
    sc_shift_left,
    sc_backslash,
    sc_z,
    sc_x,
    sc_c,
    sc_v,
    sc_b,
    sc_n,
    sc_m,
    sc_comma,
    sc_preiod,
    sc_slash,
    sc_shift_right,
    sc_numpad_multiply,
    sc_alt_left,
    sc_space,
    sc_caps_lock,
    sc_f1,
    sc_f2,
    sc_f3,
    sc_f4,
    sc_f5,
    sc_f6,
    sc_f7,
    sc_f8,
    sc_f9,
    sc_f10,
    sc_num_lock,
    sc_scroll_lock,
    sc_numpad_7,
    sc_numpad_8,
    sc_numpad_9,
    sc_numpad_minus,
    sc_numpad_4,
    sc_numpad_5,
    sc_numpad_6,
    sc_numpad_plus,
    sc_numpad_1,
    sc_numpad_2,
    sc_numpad_3,
    sc_numpad_0,
    sc_numpad_period,
    sc_alt_print_screen,
    sc_bracket_angle,
    sc_f11,
    sc_f12,
    sc_oem_1,
    sc_oem_2,
    sc_oem_3,
    sc_erase_eof,
    sc_oem_4,
    sc_oem_5,
    sc_zoom,
    sc_help,
    sc_f13,
    sc_f14,
    sc_f15,
    sc_f16,
    sc_f17,
    sc_f18,
    sc_f19,
    sc_f20,
    sc_f21,
    sc_f22,
    sc_f23,
    sc_oem_6,
    sc_katakana,
    sc_oem_7,
    sc_f24,
    sc_sbcschar,
    sc_convert,
    sc_nonconvert,
    
    sc_media_previous,
    sc_media_next,
    sc_numpad_enter,
    sc_control_right,
    sc_volume_mute,
    sc_launch_app2,
    sc_media_play,
    sc_media_stop,
    sc_volume_down,
    sc_volume_up,
    sc_browser_home,
    sc_numpad_divide,
    sc_print_screen,
    sc_alt_right,
    sc_cancel,
    sc_home,
    sc_arrow_up,
    sc_page_up,
    sc_arrow_left,
    sc_arrow_right,
    sc_end,
    sc_arrow_down,
    sc_page_down,
    sc_insert,
    sc_delete,
    sc_meta_left,
    sc_meta_right,
    sc_application,
    sc_power,
    sc_sleep,
    sc_wake,
    sc_browser_search,
    sc_browser_favorites,
    sc_browser_refresh,
    sc_browser_stop,
    sc_browser_forward,
    sc_browser_back,
    sc_launch_app1,
    sc_launch_email,
    sc_launch_media,
    
    sc_pause
};

u8 input_vk_to_key[ 256 ] = {
#if WINDOWS
    /* 0x00 */ vk_none, /* 0x01 */ vk_mouse_left, /* 0x02 */ vk_mouse_right, /* 0x03 */ vk_cancel, /* 0x04 */ vk_mouse_midle, /* 0x05 */ vk_mouse_4, /* 0x06 */ vk_mouse_5, /* 0x07 */ 0, /* 0x08 */ vk_backspace, /* 0x09 */ vk_tab, /* 0x0a */ 0, /* 0x0b */ 0, /* 0x0c */ vk_clear, /* 0x0d */ vk_return, /* 0x0e */ 0, /* 0x0f */ 0,
    /* 0x10 */ vk_shift, /* 0x11 */ vk_control, /* 0x12 */ vk_alt, /* 0x13 */ vk_pause, /* 0x14 */ vk_caps_lock, /* 0x15 */ vk_kana, /* 0x15 vk_hangul, *//* 0x16 */ 0, /* 0x17 */ vk_junja, /* 0x18 */ vk_final, /* 0x19 */ vk_hanja, /* 0x19 vk_kanji, */ /* 0x1a */ 0, /* 0x1b */ vk_escape, /* 0x1c */ vk_convert, /* 0x1d */ vk_non_convert, /* 0x1e */ vk_accept, /* 0x1f */ vk_mode_change,
    /* 0x20 */ vk_space, /* 0x21 */ vk_page_up, /* 0x22 */ vk_page_down, /* 0x23 */ vk_end, /* 0x24 */ vk_home, /* 0x25 */ vk_left, /* 0x26 */ vk_up, /* 0x27 */ vk_right, /* 0x28 */ vk_down, /* 0x29 */ vk_select, /* 0x2a */ vk_print, /* 0x2b */ vk_execute, /* 0x2c */ vk_print_screen, /* 0x2d */ vk_insert, /* 0x2e */ vk_delete, /* 0x2f */ vk_help,
    /* 0x30 */ vk_0, /* 0x31 */ vk_1, /* 0x32 */ vk_2, /* 0x33 */ vk_3, /* 0x34 */ vk_4, /* 0x35 */ vk_5, /* 0x36 */ vk_6, /* 0x37 */ vk_7, /* 0x38 */ vk_8, /* 0x39 */ vk_9, /* 0x3a */ 0, /* 0x3b */ 0, /* 0x3c */ 0, /* 0x3d */ 0, /* 0x3e */ 0, /* 0x3f */ 0,
    /* 0x40 */ 0, /* 0x41 */ vk_a, /* 0x42 */ vk_b, /* 0x43 */ vk_c, /* 0x44 */ vk_d, /* 0x45 */ vk_e, /* 0x46 */ vk_f, /* 0x47 */ vk_g, /* 0x48 */ vk_h, /* 0x49 */ vk_i, /* 0x4a */ vk_j, /* 0x4b */ vk_k, /* 0x4c */ vk_l, /* 0x4d */ vk_m, /* 0x4e */ vk_n, /* 0x4f */ vk_o,
    /* 0x50 */ vk_p, /* 0x51 */ vk_q, /* 0x52 */ vk_r, /* 0x53 */ vk_s, /* 0x54 */ vk_t, /* 0x55 */ vk_u, /* 0x56 */ vk_v, /* 0x57 */ vk_w, /* 0x58 */ vk_x, /* 0x59 */ vk_y, /* 0x5a */ vk_z, /* 0x5b */ vk_left_meta, /* 0x5c */ vk_right_meta, /* 0x5d */ vk_application, /* 0x5e */ 0,/* 0x5f */ vk_sleep,
    /* 0x60 */ vk_numpad_0, /* 0x61 */ vk_numpad_1, /* 0x62 */ vk_numpad_2, /* 0x63 */ vk_numpad_3, /* 0x64 */ vk_numpad_4, /* 0x65 */ vk_numpad_5,/* 0x66 */ vk_numpad_6, /* 0x67 */ vk_numpad_7, /* 0x68 */ vk_numpad_8, /* 0x69 */ vk_numpad_9, /* 0x6a */ vk_multiply, /* 0x6b */ vk_add, /* 0x6c */ vk_separator, /* 0x6d */ vk_subtract, /* 0x6e */ vk_decimal,/* 0x6f */ vk_divide,
    /* 0x70 */ vk_f1, /* 0x71 */ vk_f2, /* 0x72 */ vk_f3, /* 0x73 */ vk_f4, /* 0x74 */ vk_f5, /* 0x75 */ vk_f6, /* 0x76 */ vk_f7, /* 0x77 */ vk_f8, /* 0x78 */ vk_f9, /* 0x79 */ vk_f10, /* 0x7a */ vk_f11, /* 0x7b */ vk_f12, /* 0x7c */ vk_f13, /* 0x7d */ vk_f14, /* 0x7e */ vk_f15, /* 0x7f */ vk_f16,
    /* 0x80 */ vk_f17, /* 0x81 */ vk_f18, /* 0x82 */ vk_f19, /* 0x83 */ vk_f20, /* 0x84 */ vk_f21, /* 0x85 */ vk_f22, /* 0x86 */ vk_f23, /* 0x87 */ vk_f24, /* 0x88 */ 0, /* 0x89 */ 0, /* 0x8a */ 0, /* 0x8b */ 0, /* 0x8c */ 0, /* 0x8d */ 0, /* 0x8e */ 0, /* 0x8f */ 0,
    /* 0x90 */ vk_num_lock, /* 0x91 */ vk_scroll_lock, /* 0x92 */ 0, /* 0x93 */ 0, /* 0x94 */ 0, /* 0x95 */ 0, /* 0x96 */ 0, /* 0x97 */ 0, /* 0x98 */ 0, /* 0x99 */ 0, /* 0x9a */ 0, /* 0x9b */ 0, /* 0x9c */ 0, /* 0x9d */ 0, /* 0x9e */ 0, /* 0x9f */ 0,
    /* 0xa0 */ vk_left_shift, /* 0xa1 */ vk_right_shift, /* 0xa2 */ vk_left_control, /* 0xa3 */ vk_right_control, /* 0xa4 */ vk_left_alt, /* 0xa5 */ vk_right_alt, /* 0xa6 */ vk_browser_back, /* 0xa7 */ vk_browser_forward, /* 0xa8 */ vk_browser_refresh, /* 0xa9 */ vk_browser_stop, /* 0xaa */ vk_browser_search, /* 0xab */ vk_browser_favorites, /* 0xac */ vk_browser_home, /* 0xad */ vk_volume_mute, /* 0xae */ vk_volume_down, /* 0xaf */ vk_volume_up,
    /* 0xb0 */ vk_media_next, /* 0xb1 */ vk_media_previous, /* 0xb2 */ vk_media_stop, /* 0xb3 */ vk_media_play, /* 0xb4 */ vk_launch_mail, /* 0xb5 */ vk_launch_media, /* 0xb6 */ vk_launch_application_1, /* 0xb7 */ vk_launch_application_2, /* 0xb8 */ 0, /* 0xb9 */ 0, /* 0xba */ vk_oem_1, /* 0xbb */ vk_oem_plus, /* 0xbc */ vk_oem_comma, /* 0xbd */ vk_oem_minus, /* 0xbe */ vk_oem_period, /* 0xbf */ vk_oem_2,
    /* 0xc0 */ vk_oem_3, /* 0xc1 */ 0, /* 0xc2 */ 0, /* 0xc3 */ 0, /* 0xc4 */ 0, /* 0xc5 */ 0, /* 0xc6 */ 0, /* 0xc7 */ 0, /* 0xc8 */ 0, /* 0xc9 */ 0, /* 0xca */ 0, /* 0xcb */ 0, /* 0xcc */ 0, /* 0xcd */ 0, /* 0xce */ 0, /* 0xcf */ 0,
    /* 0xd0 */ 0, /* 0xd1 */ 0, /* 0xd2 */ 0, /* 0xd3 */ 0, /* 0xd4 */ 0, /* 0xd5 */ 0, /* 0xd6 */ 0, /* 0xd7 */ 0, /* 0xd8 */ 0, /* 0xd9 */ 0, /* 0xda */ 0, /* 0xdb */ vk_oem_4, /* 0xdc */ vk_oem_5, /* 0xdd */ vk_oem_6, /* 0xde */ vk_oem_7, /* 0xdf */ vk_oem_8,
    /* 0xe0 */ 0, /* 0xe1 */ 0, /* 0xe2 */ vk_oem_102, /* 0xe3 */ 0, /* 0xe4 */ 0, /* 0xe5 */ vk_process_key, /* 0xe6 */ 0, /* 0xe7 */ vk_packet, /* 0xe8 */ 0, /* 0xe9 */ 0, /* 0xea */ 0, /* 0xeb */ 0, /* 0xec */ 0, /* 0xed */ 0, /* 0xee */ 0, /* 0xef */ 0, /* 0xf0 */ 0, /* 0xf1 */ 0, /* 0xf2 */ 0, /* 0xf3 */ 0, /* 0xf4 */ 0, /* 0xf5 */ 0, /* 0xf6 */ vk_attn, /* 0xf7 */ vk_crsel, /* 0xf8 */ vk_exsel, /* 0xf9 */ vk_ereof, /* 0xfa */ vk_play, /* 0xfb */ vk_zoom, /* 0xfc */ vk_noname, /* 0xfd */ vk_pa1, /* 0xfe */ vk_oem_clear, /* 0xff */ 0
#endif
};

u8 input_key_to_vk[ 256 ] = {
#if WINDOWS
    /* 0x00 */ 0, /* 0x01 */ VK_LBUTTON, /* 0x02 */ VK_RBUTTON, /* 0x03 */ VK_MBUTTON, /* 0x04 */ VK_XBUTTON1, /* 0x05 */ VK_XBUTTON2, /* 0x06 */ VK_CANCEL, /* 0x07 */ VK_BACK, /* 0x08 */ VK_TAB, /* 0x09 */ VK_CLEAR, /* 0x0a */ VK_RETURN, /* 0x0b */ VK_SHIFT, /* 0x0c */ VK_CONTROL, /* 0x0d */ VK_MENU, /* 0x0e */ VK_PAUSE, /* 0x0f */ VK_CAPITAL,
    /* 0x10 */ VK_KANA, /* 0x11 */ VK_HANGUL, /* 0x12 */ VK_JUNJA, /* 0x13 */ VK_FINAL, /* 0x14 */ VK_HANJA, /* 0x15 */ VK_KANJI, /* 0x16 */ VK_ESCAPE, /* 0x17 */ VK_CONVERT,  /* 0x18 */ VK_NONCONVERT, /* 0x19 */ VK_ACCEPT, /* 0x1a */ VK_MODECHANGE, /* 0x1b */ VK_SPACE, /* 0x1c */ VK_PRIOR, /* 0x1d */ VK_NEXT, /* 0x1e */ VK_END, /* 0x1f */ VK_HOME,
    /* 0x20 */ VK_LEFT, /* 0x21 */ VK_UP, /* 0x22 */ VK_RIGHT, /* 0x23 */ VK_DOWN, /* 0x24 */ VK_SELECT, /* 0x25 */ VK_PRINT, /* 0x26 */ VK_EXECUTE, /* 0x27 */ VK_SNAPSHOT, /* 0x28 */ VK_INSERT, /* 0x29 */ VK_DELETE, /* 0x2a */ VK_HELP, /* 0x2b */ '0', /* 0x2c */ '1', /* 0x2d */ '2', /* 0x2e */ '3', /* 0x2f */ '4',
    /* 0x30 */ '5', /* 0x31 */ '6', /* 0x32 */ '7', /* 0x33 */ '8', /* 0x34 */ '9', /* 0x35 */ 'A', /* 0x36 */ 'B', /* 0x37 */ 'C', /* 0x38 */ 'D', /* 0x39 */ 'E', /* 0x3a */ 'F', /* 0x3b */ 'G', /* 0x3c */ 'H', /* 0x3d */ 'I', /* 0x3e */ 'J', /* 0x3f */ 'K',
    /* 0x40 */ 'L', /* 0x41 */ 'M', /* 0x42 */ 'N', /* 0x43 */ 'O', /* 0x44 */ 'P', /* 0x45 */ 'Q', /* 0x46 */ 'R', /* 0x47 */ 'S', /* 0x48 */ 'T', /* 0x49 */ 'U', /* 0x4a */ 'V', /* 0x4b */ 'W', /* 0x4c */ 'X', /* 0x4d */ 'Y', /* 0x4e */ 'Z', /* 0x4f */ VK_LWIN,
    /* 0x50 */ VK_RWIN, /* 0x51 */ VK_APPS, /* 0x52 */ VK_SLEEP, /* 0x53 */ VK_NUMPAD0, /* 0x54 */ VK_NUMPAD1, /* 0x55 */ VK_NUMPAD2, /* 0x56 */ VK_NUMPAD3, /* 0x57 */ VK_NUMPAD4, /* 0x58 */ VK_NUMPAD5, /* 0x59 */ VK_NUMPAD6, /* 0x5a */ VK_NUMPAD7, /* 0x5b */ VK_NUMPAD8, /* 0x5c */ VK_NUMPAD9, /* 0x5d */ VK_MULTIPLY, /* 0x5e */ VK_DIVIDE, /* 0x5f */ VK_ADD,
    /* 0x60 */ VK_SUBTRACT, /* 0x61 */ VK_SEPARATOR, /* 0x62 */ VK_DECIMAL, /* 0x63 */ VK_F1, /* 0x64 */ VK_F2, /* 0x65 */ VK_F3, /* 0x66 */ VK_F4, /* 0x67 */ VK_F5, /* 0x68 */ VK_F6, /* 0x69 */ VK_F7, /* 0x6a */ VK_F8, /* 0x6b */ VK_F9, /* 0x6c */ VK_F10, /* 0x6d */ VK_F11, /* 0x6e */ VK_F12, /* 0x6f */ VK_F13,
    /* 0x70 */ VK_F14, /* 0x71 */ VK_F15, /* 0x72 */ VK_F16, /* 0x73 */ VK_F17, /* 0x74 */ VK_F18, /* 0x75 */ VK_F19, /* 0x76 */ VK_F20, /* 0x77 */ VK_F21, /* 0x78 */ VK_F22, /* 0x79 */ VK_F23, /* 0x7a */ VK_F24, /* 0x7b */ VK_NUMLOCK, /* 0x7c */ VK_SCROLL, /* 0x7d */ VK_LSHIFT, /* 0x7e */ VK_RSHIFT, /* 0x7f */ VK_LCONTROL,
    /* 0x80 */ VK_RCONTROL, /* 0x81 */ VK_LMENU, /* 0x82 */ VK_RMENU, /* 0x83 */ VK_BROWSER_BACK, /* 0x84 */ VK_BROWSER_FORWARD, /* 0x85 */ VK_BROWSER_REFRESH, /* 0x86 */ VK_BROWSER_STOP, /* 0x87 */ VK_BROWSER_SEARCH, /* 0x88 */ VK_BROWSER_FAVORITES, /* 0x89 */ VK_BROWSER_HOME, /* 0x8a */ VK_VOLUME_MUTE, /* 0x8b */ VK_VOLUME_DOWN, /* 0x8c */ VK_VOLUME_UP, /* 0x8d */ VK_MEDIA_NEXT_TRACK, /* 0x8e */ VK_MEDIA_PREV_TRACK, /* 0x8f */ VK_MEDIA_STOP,
    /* 0x90 */ VK_MEDIA_PLAY_PAUSE, /* 0x91 */ VK_LAUNCH_MAIL, /* 0x92 */ VK_LAUNCH_MEDIA_SELECT, /* 0x93 */ VK_LAUNCH_APP1, /* 0x94 */ VK_LAUNCH_APP2, /* 0x95 */ VK_OEM_PLUS, /* 0x96 */ VK_OEM_MINUS, /* 0x97 */ VK_OEM_COMMA, /* 0x98 */ VK_OEM_PERIOD, /* 0x99 */ VK_OEM_1, /* 0x9a */ VK_OEM_2, /* 0x9b */ VK_OEM_3, /* 0x9c */ VK_OEM_4, /* 0x9d */ VK_OEM_5, /* 0x9e */ VK_OEM_6, /* 0x9f */ VK_OEM_7,
    /* 0xa0 */ VK_OEM_8, /* 0xa1 */ VK_OEM_102, /* 0xa2 */ VK_OEM_CLEAR, /* 0xa3 */ VK_PROCESSKEY, /* 0xa4 */ VK_PACKET, /* 0xa5 */ VK_ATTN, /* 0xa6 */ VK_CRSEL, /* 0xa7 */ VK_EXSEL, /* 0xa8 */ VK_EREOF, /* 0xa9 */ VK_PLAY, /* 0xaa */ VK_ZOOM, /* 0xab */ VK_NONAME, /* 0xac */ VK_PA1,
    0
#endif
};

stu void internal_input_vk_key_check( ) {
#if WINDOWS
#define vk_key_mapping( vk, key ) _assert( input_key_to_vk[ key ] == vk ); _assert( input_vk_to_key[ vk ] == key );
    
    vk_key_mapping( VK_LBUTTON, vk_mouse_left );
    vk_key_mapping( VK_RBUTTON, vk_mouse_right );
    vk_key_mapping( VK_MBUTTON, vk_mouse_midle );
    vk_key_mapping( VK_XBUTTON1, vk_mouse_4 );
    vk_key_mapping( VK_XBUTTON2, vk_mouse_5 );
    vk_key_mapping( VK_CANCEL, vk_cancel );
    vk_key_mapping( VK_BACK, vk_backspace );
    vk_key_mapping( VK_TAB, vk_tab );
    vk_key_mapping( VK_CLEAR, vk_clear );
    vk_key_mapping( VK_RETURN, vk_return );
    vk_key_mapping( VK_SHIFT, vk_shift );
    vk_key_mapping( VK_CONTROL, vk_control );
    vk_key_mapping( VK_MENU, vk_alt );
    vk_key_mapping( VK_PAUSE, vk_pause );
    vk_key_mapping( VK_CAPITAL, vk_caps_lock );
    vk_key_mapping( VK_KANA, vk_kana );
    vk_key_mapping( VK_HANGUL, vk_hangul );
    vk_key_mapping( VK_JUNJA, vk_junja );
    vk_key_mapping( VK_FINAL, vk_final );
    vk_key_mapping( VK_HANJA, vk_hanja );
    vk_key_mapping( VK_KANJI, vk_kanji );
    vk_key_mapping( VK_ESCAPE, vk_escape );
    vk_key_mapping( VK_CONVERT, vk_convert );
    vk_key_mapping( VK_NONCONVERT, vk_non_convert );
    vk_key_mapping( VK_ACCEPT, vk_accept );
    vk_key_mapping( VK_MODECHANGE, vk_mode_change );
    vk_key_mapping( VK_SPACE, vk_space );
    vk_key_mapping( VK_PRIOR, vk_page_up  );
    vk_key_mapping( VK_NEXT, vk_page_down );
    vk_key_mapping( VK_END, vk_end );
    vk_key_mapping( VK_HOME, vk_home );
    vk_key_mapping( VK_LEFT, vk_left );
    vk_key_mapping( VK_UP, vk_up );
    vk_key_mapping( VK_RIGHT, vk_right );
    vk_key_mapping( VK_DOWN, vk_down );
    vk_key_mapping( VK_SELECT, vk_select );
    vk_key_mapping( VK_PRINT, vk_print );
    vk_key_mapping( VK_EXECUTE, vk_execute );
    vk_key_mapping( VK_SNAPSHOT, vk_print_screen );
    vk_key_mapping( VK_INSERT, vk_insert );
    vk_key_mapping( VK_DELETE, vk_delete );
    vk_key_mapping( VK_HELP, vk_help );
    vk_key_mapping( '0', vk_0 );
    vk_key_mapping( '1', vk_1 );
    vk_key_mapping( '2', vk_2 );
    vk_key_mapping( '3', vk_3 );
    vk_key_mapping( '4', vk_4 );
    vk_key_mapping( '5', vk_5 );
    vk_key_mapping( '6', vk_6 );
    vk_key_mapping( '7', vk_7 );
    vk_key_mapping( '8', vk_8 );
    vk_key_mapping( '9', vk_9 );
    vk_key_mapping( 'A', vk_a );
    vk_key_mapping( 'B', vk_b );
    vk_key_mapping( 'C', vk_c );
    vk_key_mapping( 'D', vk_d );
    vk_key_mapping( 'E', vk_e );
    vk_key_mapping( 'F', vk_f );
    vk_key_mapping( 'G', vk_g );
    vk_key_mapping( 'H', vk_h );
    vk_key_mapping( 'I', vk_i );
    vk_key_mapping( 'J', vk_j );
    vk_key_mapping( 'K', vk_k );
    vk_key_mapping( 'L', vk_l );
    vk_key_mapping( 'M', vk_m );
    vk_key_mapping( 'N', vk_n );
    vk_key_mapping( 'O', vk_o );
    vk_key_mapping( 'P', vk_p );
    vk_key_mapping( 'Q', vk_q );
    vk_key_mapping( 'R', vk_r );
    vk_key_mapping( 'S', vk_s );
    vk_key_mapping( 'T', vk_t );
    vk_key_mapping( 'U', vk_u );
    vk_key_mapping( 'V', vk_v );
    vk_key_mapping( 'W', vk_w );
    vk_key_mapping( 'X', vk_x );
    vk_key_mapping( 'Y', vk_y );
    vk_key_mapping( 'Z', vk_z );
    vk_key_mapping( VK_LWIN, vk_left_meta );
    vk_key_mapping( VK_RWIN, vk_right_meta );
    vk_key_mapping( VK_APPS, vk_application );
    vk_key_mapping( VK_SLEEP, vk_sleep );
    vk_key_mapping( VK_NUMPAD0, vk_numpad_0 );
    vk_key_mapping( VK_NUMPAD1, vk_numpad_1 );
    vk_key_mapping( VK_NUMPAD2, vk_numpad_2 );
    vk_key_mapping( VK_NUMPAD3, vk_numpad_3 );
    vk_key_mapping( VK_NUMPAD4, vk_numpad_4 );
    vk_key_mapping( VK_NUMPAD5, vk_numpad_5 );
    vk_key_mapping( VK_NUMPAD6, vk_numpad_6 );
    vk_key_mapping( VK_NUMPAD7, vk_numpad_7 );
    vk_key_mapping( VK_NUMPAD8, vk_numpad_8 );
    vk_key_mapping( VK_NUMPAD9, vk_numpad_9 );
    vk_key_mapping( VK_MULTIPLY, vk_multiply );
    vk_key_mapping( VK_DIVIDE, vk_divide );
    vk_key_mapping( VK_ADD, vk_add );
    vk_key_mapping( VK_SUBTRACT, vk_subtract );
    vk_key_mapping( VK_SEPARATOR, vk_separator );
    vk_key_mapping( VK_DECIMAL, vk_decimal );
    vk_key_mapping( VK_F1, vk_f1 );
    vk_key_mapping( VK_F2, vk_f2 );
    vk_key_mapping( VK_F3, vk_f3 );
    vk_key_mapping( VK_F4, vk_f4 );
    vk_key_mapping( VK_F5, vk_f5 );
    vk_key_mapping( VK_F6, vk_f6 );
    vk_key_mapping( VK_F7, vk_f7 );
    vk_key_mapping( VK_F8, vk_f8 );
    vk_key_mapping( VK_F9, vk_f9 );
    vk_key_mapping( VK_F10, vk_f10 );
    vk_key_mapping( VK_F11, vk_f11 );
    vk_key_mapping( VK_F12, vk_f12 );
    vk_key_mapping( VK_F13, vk_f13 );
    vk_key_mapping( VK_F14, vk_f14 );
    vk_key_mapping( VK_F15, vk_f15 );
    vk_key_mapping( VK_F16, vk_f16 );
    vk_key_mapping( VK_F17, vk_f17 );
    vk_key_mapping( VK_F18, vk_f18 );
    vk_key_mapping( VK_F19, vk_f19 );
    vk_key_mapping( VK_F20, vk_f20 );
    vk_key_mapping( VK_F21, vk_f21 );
    vk_key_mapping( VK_F22, vk_f22 );
    vk_key_mapping( VK_F23, vk_f23 );
    vk_key_mapping( VK_F24, vk_f24 );
    vk_key_mapping( VK_NUMLOCK, vk_num_lock );
    vk_key_mapping( VK_SCROLL, vk_scroll_lock );
    vk_key_mapping( VK_LSHIFT, vk_left_shift );
    vk_key_mapping( VK_RSHIFT, vk_right_shift );
    vk_key_mapping( VK_LCONTROL, vk_left_control );
    vk_key_mapping( VK_RCONTROL, vk_right_control );
    vk_key_mapping( VK_LMENU, vk_left_alt );
    vk_key_mapping( VK_RMENU, vk_right_alt );
    
    vk_key_mapping( VK_BROWSER_BACK, vk_browser_back );
    vk_key_mapping( VK_BROWSER_FORWARD, vk_browser_forward );
    vk_key_mapping( VK_BROWSER_REFRESH, vk_browser_refresh );
    vk_key_mapping( VK_BROWSER_STOP, vk_browser_stop );
    vk_key_mapping( VK_BROWSER_SEARCH, vk_browser_search );
    vk_key_mapping( VK_BROWSER_FAVORITES, vk_browser_favorites );
    vk_key_mapping( VK_BROWSER_HOME, vk_browser_home );
    vk_key_mapping( VK_VOLUME_MUTE, vk_volume_mute );
    vk_key_mapping( VK_VOLUME_DOWN, vk_volume_down );
    vk_key_mapping( VK_VOLUME_UP, vk_volume_up );
    vk_key_mapping( VK_MEDIA_NEXT_TRACK, vk_media_next );
    vk_key_mapping( VK_MEDIA_PREV_TRACK, vk_media_previous );
    vk_key_mapping( VK_MEDIA_STOP, vk_media_stop );
    vk_key_mapping( VK_MEDIA_PLAY_PAUSE, vk_media_play );
    vk_key_mapping( VK_LAUNCH_MAIL, vk_launch_mail );
    vk_key_mapping( VK_LAUNCH_MEDIA_SELECT, vk_launch_media );
    vk_key_mapping( VK_LAUNCH_APP1, vk_launch_application_1 );
    vk_key_mapping( VK_LAUNCH_APP2, vk_launch_application_2 );
    vk_key_mapping( VK_OEM_PLUS, vk_oem_plus );
    vk_key_mapping( VK_OEM_COMMA, vk_oem_comma );
    vk_key_mapping( VK_OEM_MINUS, vk_oem_minus );
    vk_key_mapping( VK_OEM_PERIOD, vk_oem_period );
    vk_key_mapping( VK_OEM_1, vk_oem_1 );
    vk_key_mapping( VK_OEM_2, vk_oem_2 );
    vk_key_mapping( VK_OEM_3, vk_oem_3 );
    vk_key_mapping( VK_OEM_4, vk_oem_4 );
    vk_key_mapping( VK_OEM_5, vk_oem_5 );
    vk_key_mapping( VK_OEM_6, vk_oem_6 );
    vk_key_mapping( VK_OEM_7, vk_oem_7 );
    vk_key_mapping( VK_OEM_8, vk_oem_8 );
    vk_key_mapping( VK_OEM_102, vk_oem_102 );
    vk_key_mapping( VK_OEM_CLEAR, vk_oem_clear );
    vk_key_mapping( VK_PROCESSKEY, vk_process_key );
    vk_key_mapping( VK_PACKET, vk_packet );
    vk_key_mapping( VK_ATTN, vk_attn );
    vk_key_mapping( VK_CRSEL, vk_crsel );
    vk_key_mapping( VK_EXSEL, vk_exsel );
    vk_key_mapping( VK_EREOF, vk_ereof );
    vk_key_mapping( VK_PLAY, vk_play );
    vk_key_mapping( VK_ZOOM, vk_zoom );
    vk_key_mapping( VK_NONAME, vk_noname );
    vk_key_mapping( VK_PA1, vk_pa1 );
    
#undef vk_key_mapping
    
#endif
}

stu umm input_get_scancode_offset( u32 scancode ) {
    
    _assert( scancode );
    
    umm result = scancode;
    umm group_0_end = sc_nonconvert;
    umm group_1_start = sc_media_previous;
    umm group_1_end = sc_launch_media;
    umm group_2_start = sc_pause;
    
    if ( scancode >= group_2_start ) {
        result = group_0_end + 1 + ( group_1_end - group_1_start ) + 1 + ( scancode - group_2_start );
    } else if ( scancode >= group_1_start ) {
        result = group_0_end + 1 + ( scancode - group_1_start );
    }
    
    _assert( result <= 0xff );
    
    return result;
}

enum input_key_state_flag_t {
    
    input_key_state_flag_pressed = 1 << 0,
    input_key_state_flag_repeated = 1 << 1,
    input_key_state_flag_transition_count_mask = 0xff & ~( input_key_state_flag_pressed | input_key_state_flag_repeated ),
    
} input_key_state_flag_t;

stu u8 input_get_transition_count( window_state_t* state, u8 key ) {
    u8 result = ( state->keys[ key ] & input_key_state_flag_transition_count_mask ) >> 2;
    return result;
}

stu b32 input_key_is_pressed( window_state_t* state, u8 key ) {
    b32 result = state->keys[ key ] & input_key_state_flag_pressed;
    return result;
}

stu b32 input_key_just_pressed( window_state_t* state, u8 key ) {
    b32 result = ( state->keys[ key ] & input_key_state_flag_pressed ) && ( state->keys[ key ] & input_key_state_flag_transition_count_mask );
    return result;
}

stu b32 input_key_just_released( window_state_t* state, u8 key ) {
    b32 result = !( state->keys[ key ] & input_key_state_flag_pressed ) && ( state->keys[ key ] & input_key_state_flag_transition_count_mask );
    return result;
}

stu b32 input_key_repeated( window_state_t* state, u8 key ) {
    b32 result = state->keys[ key ] & input_key_state_flag_repeated;
    return result;
}

stu b32 input_key_just_pressed_or_repeated( window_state_t* state, u8 key ) {
    b32 result = ( state->keys[ key ] & input_key_state_flag_pressed ) && ( state->keys[ key ] & input_key_state_flag_transition_count_mask );
    result = result || state->keys[ key ] & input_key_state_flag_repeated;
    return result;
}

stu b32 input_scancode_is_pressed( window_state_t* state, u32 scancode ) {
    umm offset = input_get_scancode_offset( scancode );
    b32 result = state->scancodes[ offset ] & input_key_state_flag_pressed;
    return result;
}

stu b32 input_scancode_just_pressed( window_state_t* state, u32 scancode ) {
    umm offset = input_get_scancode_offset( scancode );
    b32 result = ( state->scancodes[ offset ] & input_key_state_flag_pressed ) && ( state->scancodes[ offset ] & input_key_state_flag_transition_count_mask );
    return result;
}

stu b32 input_scancode_just_released( window_state_t* state, u32 scancode ) {
    umm offset = input_get_scancode_offset( scancode );
    b32 result = !( state->scancodes[ offset ] & input_key_state_flag_pressed ) && ( state->scancodes[ offset ] & input_key_state_flag_transition_count_mask );
    return result;
}

stu b32 input_scancode_repeated( window_state_t* state, u32 scancode ) {
    umm offset = input_get_scancode_offset( scancode );
    b32 result = state->scancodes[ offset ] & input_key_state_flag_repeated;
    return result;
}

stu b32 input_scancode_just_pressed_or_repeated( window_state_t* state, u32 scancode ) {
    umm offset = input_get_scancode_offset( scancode );
    b32 result = ( state->scancodes[ offset ] & input_key_state_flag_pressed ) && ( state->scancodes[ offset ] & input_key_state_flag_transition_count_mask );
    result = result || state->scancodes[ offset ] & input_key_state_flag_repeated;
    return result;
}

#if WINDOWS
stu void input_clear_keys( window_state_t* state ) {
    
    for ( umm key = 0; key < vk_count; key++ ) {
        state->keys[ key ] = ( state->keys[ key ] & input_key_state_flag_pressed ) << 2;
    }
    
    for ( umm scancode = 0; scancode < array_count( state->scancodes ); scancode++ ) {
        state->scancodes[ scancode ] = ( state->scancodes[ scancode ] & input_key_state_flag_pressed ) << 2;
    }
    
    state->codepoint_count = 0;
    state->codepoints_total_byte_count_utf8 = 0;
    
    POINT p;
    GetCursorPos( &p );
    state->mouse_x = p.x - state->x;
    state->mouse_y = state->screen_height - 1 - p.y - state->y;
}
#endif

#ifdef WINDOWS

// MARK windows

stu u32 input_key_to_scan_code( u32 key ) {
    /* TODO simon: make sure this works with all scancodes. */
    u32 result = MapVirtualKeyEx( input_key_to_vk[ key ], MAPVK_VK_TO_VSC_EX, 0 );
    
    if ( key == vk_home || key == vk_end || key == vk_insert || key == vk_delete || key == vk_page_up || key == vk_page_down
        || key == vk_up || key == vk_down || key == vk_left || key == vk_right ) {
        
        result += 0xe000;
    }
    
    return result;
}

stu u32 input_scan_code_to_key( u32 scancode ) {
    
    u32 result;
    
    if ( scancode == 0x45 ){
        scancode = 0xE045;
    } else if ( scancode == 0xE11D45 ) {
        scancode = 0x45;
    }
    
    result = MapVirtualKeyEx( scancode, MAPVK_VSC_TO_VK_EX, 0 );
    result = input_vk_to_key[ result ];
    return result;
}

stu u32 input_get_scancode_name( u32 scancode, char* buffer, u32 maxLength ) {
    
    u32 result = 0;
    u32 extended = scancode & 0xffff00;
    u32 lParam = 0;
    
    if ( extended ) {
        
        if ( extended == 0xE11D00 ) {
            lParam = 0x45 << 16;
        } else {
            lParam = ( 0x100 | ( scancode & 0xff ) ) << 16;
        }
        
    } else {
        
        lParam = scancode << 16;
        
        if ( scancode == 0x45 ) {
            lParam |= ( 0x1 << 24 );
        }
    }
    
    result = GetKeyNameText( lParam, buffer, maxLength );
    return result;
}

stu void internal_window_set_vk_or_sc_state( u8* key_state, b32 pressed, b32 repeated, umm transition_count ) {
    
    ( *key_state ) = pressed ? input_key_state_flag_pressed : 0;
    ( *key_state ) |= repeated ? input_key_state_flag_repeated : 0;
    ( *key_state ) |= ( ( u8 ) transition_count ) << 2;
}

stu void input_sync_keys( window_state_t* state ) {
    
    profiler_event_start( input_sync_keys );
    POINT p;
    GetCursorPos( &p );
    state->mouse_x = p.x - state->x;
    state->mouse_y = state->screen_height - 1 - p.y - state->y;
    
    state->keys[ vk_none ] = 0;
    
    b32 inside_client_rect =
        state->mouse_x >= 0
        && state->mouse_x < ( s32 ) state->width
        && state->mouse_y >= 0
        && state->mouse_y < ( s32 ) state->height;
    
    if ( inside_client_rect ) {
        
        for ( umm index = vk_mouse_left; index < vk_cancel; index++ ) {
            u16 key_state = GetAsyncKeyState( input_key_to_vk[ index ] );
            b32 pressed = ( key_state & ( 0x1 << 15 ) ) > 0;
            internal_window_set_vk_or_sc_state( state->keys + index, pressed, 0, 0 );
        }
    } else {
        
        for ( umm index = vk_mouse_left; index < vk_cancel; index++ ) {
            state->keys[ index ] = 0;
        }
    }
    
    for ( umm index = vk_cancel; index < vk_pa1; index++ ) {
        u16 key_state = GetAsyncKeyState( input_key_to_vk[ index ] );
        b32 pressed = ( key_state & ( 0x1 << 15 ) ) > 0;
        internal_window_set_vk_or_sc_state( state->keys + index, pressed, 0, 0 );
    }
    
    for ( umm index = 0; index < array_count( input_scancode_list ); index++ ) {
        
        u32 scancode = input_scancode_list[ index ];
        umm offset;
        u32 vk;
        u16 key_state;
        
        if ( scancode == 0x45 ) {
            scancode = 0xE045;
        } else if ( scancode == 0xE11D45 ) {
            scancode = 0x45;
        }
        
        offset = input_get_scancode_offset( scancode );
        vk = MapVirtualKeyEx( scancode, MAPVK_VSC_TO_VK_EX, 0 );
        key_state = GetAsyncKeyState( vk );
        
        b32 pressed = ( key_state & ( 0x1 << 15 ) ) > 0;
        internal_window_set_vk_or_sc_state( state->scancodes + offset, pressed, 0, 0 );
    }
    
    state->codepoint_count = 0;
    state->codepoints_total_byte_count_utf8 = 0;
    profiler_event_end( input_sync_keys );
}

stu LRESULT CALLBACK window_proc( HWND window, UINT message, WPARAM wParam, LPARAM lParam ) {
    
    profiler_event_start( window_proc );
    
    LRESULT result = 0;
    window_state_t* state = ( window_state_t* ) GetWindowLongPtrW( window, 0 );
    /* TODO simon: when the window style is popup, we get here before the state variable is set. Find why or a solution. */
    
    switch ( message ) {
        
        // NOTE simon: Not caught by PeekMessage.
        
        case WM_DESTROY :
        case WM_CLOSE : {
            if ( state ) {
                state->close_requested = true;
            }
        } break;
        
        // NOTE simon: WM_SIZE and WM_MOVE will not be called if we don't call DefWIndowProc with WM_WINDOWPOSCHANGED. WM_SIZE and  WM_MOVE will be called a startup only.
        case WM_WINDOWPOSCHANGED :
        case WM_MOVE :
        case WM_SIZE : {
            
            if ( state ) {
                RECT clientRect = { 0 };
                GetClientRect( window, &clientRect );
                state->width = clientRect.right;
                state->height = clientRect.bottom;
                state->resized = true;
                
                POINT p;
                p.x = 0;
                p.y = state->height;
                ClientToScreen( window, &p );
                
                state->x = p.x;
                state->y = state->screen_height - p.y;
            }
            
        } break;
        
        case WM_SYSCOMMAND: {
            
            // NOTE simon: Remove beeping sound when ALT + some key is pressed.
            if ( wParam == SC_KEYMENU ) {
                result = 0;
            } else {
                result = DefWindowProcW( window, message, wParam, lParam );
            }
        } break;
        
        case WM_SETFOCUS: {
            
            if ( state ) {
                input_sync_keys( state );
                state->has_focus = true;
            }
            
            result = 0;
            
        } break;
        
        case WM_KILLFOCUS: {
            
            if ( state ) {
                input_clear_keys( state );
                state->has_focus = true;
            }
            
            result = 0;
            
        } break;
        
        // NOTE simon: End of not caught by PeekMessage.
        
        default : {
            profiler_event_start( default_window_proc );
            result = DefWindowProcW( window, message, wParam, lParam );
            profiler_event_end( default_window_proc );
        } break;
    }
    
    profiler_event_end( window_proc );
    
    return result;
}

stu void window_clean_up( HGLRC* render_context, HWND* window_handle, HDC* device_context ) {
    
    if ( render_context != 0 && *render_context != 0 ) {
        wglMakeCurrent( 0, 0 );
        wglDeleteContext( *render_context );
        *render_context = 0;
    }
    
    if ( window_handle != 0 && *window_handle != 0 ) {
        
        if ( device_context != 0 && *device_context != 0 ){
            ReleaseDC( *window_handle, *device_context );
            *device_context = 0;
        }
        
        DestroyWindow( *window_handle );
        *window_handle = 0;
    }
}

stu void window_destroy( platform_specific_t* platform ) {
    window_clean_up( &platform->render_context, &platform->window_handle, &platform->device_context );
}

typedef enum window_create_flag_t {
    window_create_flag_none,
    window_create_flag_no_title_bar = 1 << 0,
    window_create_flag_maximized = 1 << 1,
} window_create_flag_t;

/* TODO simon: check this https://git.handmade.network/snippets/17 */
stu void window_create( window_state_t* state, void* window_name, u32 width, u32 height, s32 x, s32 y, u32 flags ) {
    
    profiler_event_start( window_create );
    
    WNDCLASSW window_class = { 0 };
    window_class.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
    window_class.lpfnWndProc = window_proc;
    window_class.hInstance = GetModuleHandle( 0 );
    window_class.lpszClassName = L"simple_window";
    window_class.hCursor = LoadCursor( 0, IDC_ARROW );
    window_class.cbWndExtra = sizeof( window_state_t* );
    
    u16 window_name_utf16[ 1024 ] = { 0 };
    u32 error = 0;
    /* umm u16_count = */ utf8_to_utf16( window_name, string_length( window_name ), window_name_utf16, sizeof( window_name_utf16 ), &error );
    
    profiler_event_start( register_class_w );
    ATOM atom = RegisterClassW( &window_class );
    profiler_event_end( register_class_w );
    
    if ( atom != 0 ) {
        
        HWND dummy_handle = 0;
        u32 window_style = WS_OVERLAPPEDWINDOW;
        // u32 window_style = WS_POPUP | WS_THICKFRAME;
        
        if ( flag_is_set_u32( flags, window_create_flag_no_title_bar ) ) {
            window_style = WS_POPUPWINDOW;
        }
        
        profiler_event_start( dummy_create_window_ex_w );
        dummy_handle = CreateWindowExW( 0, window_class.lpszClassName, ( LPCWSTR ) window_name_utf16, window_style, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, 0, 0, window_class.hInstance, 0
                                       );
        profiler_event_end( dummy_create_window_ex_w );
        
        SetWindowLongPtrW( dummy_handle, 0, ( LONG_PTR ) state );
        
        PIXELFORMATDESCRIPTOR pfd = {
            sizeof( PIXELFORMATDESCRIPTOR ),
            1,
            PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
            PFD_TYPE_RGBA,
            32,
            0, 0, 0, 0, 0, 0,
            0,
            0,
            0,
            0, 0, 0, 0,
            24,
            8,
            0,
            PFD_MAIN_PLANE,
            0,
            0, 0, 0
        };
        
        HDC dummy_device_context = GetDC( dummy_handle );
        
        state->screen_width = GetDeviceCaps( dummy_device_context, HORZRES );
        state->screen_height =  GetDeviceCaps( dummy_device_context, VERTRES );
        state->screen_refresh_rate = ( u16 ) GetDeviceCaps( dummy_device_context, VREFRESH );
        
        RECT window_rect = { 0 };
        
        if ( x == s32_max ) {
            window_rect.left = ( state->screen_width - width ) / 2;
            x = window_rect.left;
        } else {
            window_rect.left = x;
        }
        
        if ( y == s32_max ) {
            window_rect.top = ( state->screen_height - height ) / 2;
            y = window_rect.top;
        } else {
            window_rect.top = state->screen_height - height - y;
        }
        
        window_rect.right = window_rect.left + width;
        window_rect.bottom = window_rect.top + height;
        
        AdjustWindowRect( &window_rect, window_style, 0 );
        
        /* NOTE simon: window_state contains the client rect information.
        Do not use window_rect here since it contains window rect information. */
        state->x = x;
        state->y = y;
        state->width = width;
        state->height = height;
        
        profiler_event_start( dummy_choose_pixel_format );
        int default_pixel_format = ChoosePixelFormat( dummy_device_context, &pfd );
        profiler_event_end( dummy_choose_pixel_format );
        
        profiler_event_start( dummy_describe_pixel_format );
        DescribePixelFormat( dummy_device_context, default_pixel_format, sizeof( pfd ), &pfd );
        profiler_event_end( dummy_describe_pixel_format );
        
        profiler_event_start( dummy_set_pixel_format );
        SetPixelFormat( dummy_device_context, default_pixel_format, &pfd );
        profiler_event_end( dummy_set_pixel_format );
        
        profiler_event_start( dummy_wgl_create_context );
        HGLRC dummy_context = wglCreateContext( dummy_device_context );
        profiler_event_end( dummy_wgl_create_context );
        
        if ( dummy_context ==  0 ) {
            
            log_s( log_name_gl, "Failed to create a default OpenGL context\n" );
            window_clean_up( &dummy_context, &dummy_handle, &dummy_device_context );
            
        } else {
            
            profiler_event_start( dummy_wgl_make_current );
            wglMakeCurrent( dummy_device_context, dummy_context );
            profiler_event_end( dummy_wgl_make_current );
            
            profiler_event_start( dummy_get_wgl_functions );
            gl_get_platform_functions( );
            profiler_event_end( dummy_get_wgl_functions );
            
            profiler_event_start( dummy_get_gl_functions );
            gl_get_functions( );
            profiler_event_end( dummy_get_gl_functions );
            
            profiler_event_start( log_functions );
            log_s( log_name_gl, "Plateform extensions (WGL): " );
            log_s( log_name_gl, ( char* ) wglGetExtensionsStringARB( dummy_device_context ) );
            log_s( log_name_gl, "\n\n" );
            
            log_s( log_name_gl, "Dummy context\n" );
            
            gl_log_infos( );
            profiler_event_end( log_functions );
            
            /* NOTE simon: add WS_EX_TOPMOST as the first argument to have always on top window. */
            u32 ex_style = 0;
            // u32 ex_style = WS_EX_TOPMOST;
            
            profiler_event_start( create_window_ex_w );
            /* NOTE simon: I request a size bigger than necessary because of a potential bug with the amd driver or the windows api.
The correct size is set later with a call to SetWindowPos before displaying the window. */
            HWND window_handle = CreateWindowExW( ex_style, window_class.lpszClassName, ( LPCWSTR ) window_name_utf16, window_style, window_rect.left, window_rect.top, window_rect.right - window_rect.left + 1, window_rect.bottom - window_rect.top - 1, 0, 0, window_class.hInstance, 0 );
            profiler_event_end( create_window_ex_w );
            
            SetWindowLongPtrW( window_handle, 0, ( LONG_PTR ) state );
            
            HDC device_context = GetDC( window_handle );
            
            const int wgl_attrib_list[ ] = {
                WGL_DRAW_TO_WINDOW_ARB, GL_TRUE,
                WGL_SUPPORT_OPENGL_ARB, GL_TRUE,
                WGL_DOUBLE_BUFFER_ARB, GL_TRUE,
                WGL_PIXEL_TYPE_ARB, WGL_TYPE_RGBA_ARB,
                WGL_COLOR_BITS_ARB, 32,
                WGL_DEPTH_BITS_ARB, 24,
                WGL_STENCIL_BITS_ARB, 8,
                0,
            };
            
            int pixel_format;
            UINT num_formats;
            
            profiler_event_start( wgl_choose_pixel_format );
            int choose_pixel_format_result = wglChoosePixelFormatARB( device_context, wgl_attrib_list, 0, 1, &pixel_format, &num_formats );
            profiler_event_end( wgl_choose_pixel_format );
            
            if ( choose_pixel_format_result ) {
                profiler_event_start( describe_pixel_format );
                DescribePixelFormat( device_context, pixel_format, sizeof( pfd ), &pfd );
                profiler_event_end( describe_pixel_format );
                
                profiler_event_start( set_pixel_format );
                SetPixelFormat( device_context, pixel_format, &pfd );
                profiler_event_end( set_pixel_format );
            } else {
                profiler_event_start( set_pixel_format );
                SetPixelFormat( device_context, default_pixel_format, &pfd );
                profiler_event_end( set_pixel_format );
            }
            
            const int wgl_context_attrib_list[ ] = {
                
                WGL_CONTEXT_MAJOR_VERSION_ARB, 3,
                WGL_CONTEXT_MINOR_VERSION_ARB, 3,
#if DEBUG
                WGL_CONTEXT_FLAGS_ARB, WGL_CONTEXT_DEBUG_BIT_ARB,
#endif
                //WGL_CONTEXT_PROFILE_MASK_ARB, WGL_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB,
                WGL_CONTEXT_PROFILE_MASK_ARB, WGL_CONTEXT_CORE_PROFILE_BIT_ARB,
                0
            };
            
            profiler_event_start( wgl_create_context_attribs );
            HGLRC render_context = wglCreateContextAttribsARB( device_context, 0, wgl_context_attrib_list );
            profiler_event_end( wgl_create_context_attribs );
            
            if ( render_context != 0 ) {
                
                window_clean_up( &dummy_context, &dummy_handle, &dummy_device_context );
                
                profiler_event_start( wgl_make_current );
                wglMakeCurrent( device_context, render_context );
                profiler_event_end( wgl_make_current );
                wglSwapIntervalEXT( 1 );
                
                profiler_event_start( log_functions );
                log_s( log_name_gl, "\nActual context\n" );
                gl_get_functions( );
                gl_log_infos( );
                profiler_event_end( log_functions );
                
                /* NOTE simon: This line is to fix a bug (at least I think it is) with certain version of the amd driver or its interaciton with the windows api. */
                SetWindowPos( window_handle, 0, 0, 0, window_rect.right - window_rect.left, window_rect.bottom - window_rect.top, SWP_NOMOVE );
                
                int show_window_flag = SW_SHOW;
                
                if ( flag_is_set_u32( flags, window_create_flag_maximized ) ) {
                    show_window_flag = SW_MAXIMIZE;
                }
                
                profiler_event_start( show_window );
                ShowWindow( window_handle, show_window_flag );
                profiler_event_end( show_window );
                
                state->platform.window_handle = window_handle;
                state->platform.device_context = device_context;
                state->platform.render_context = render_context;
                state->close_requested = false;
                state->running = true;
                
                /* NOTE simon: We need to set the viewport size here because
the ShowWindow call can trigger the window_proc function and set the state->resized flag
(if SW_MAXIMIZE for example) but the window_handle_messages function will clear it
before it is processed in the main loop. */
                glViewport( 0, 0, state->width, state->height );
                
            } else {
                
                log_s( log_name_gl, "Failed to create an advanced render context.\n" );
                window_clean_up( &dummy_context, &dummy_handle, &dummy_device_context );
            }
        }
        
        RAWINPUTDEVICE raw_input_devices[ ] = {
#if 1
            {
                /* NOTE simon: keyboard. */
                .usUsagePage = 0x01,
                .usUsage = 0x06,
                .dwFlags = 0, /* RIDEV_NOLEGACY;   // adds HID keyboard and also ignores legacy keyboard messages */
                .hwndTarget = 0,
            },
#endif
#if 0
            /* NOTE simon: this was previously used to capture mouse movement when the mouse is outside the window.
            I use SetCapture and RleaseCapture instead now.*/
            {
                /* NOTE simon: mouse. */
                .usUsagePage = 0x01,
                .usUsage = 0x02,
                .dwFlags = 0,
                .hwndTarget = 0,
            }
#endif
        };
        
        profiler_event_start( register_raw_input_devices );
        b32 registered = RegisterRawInputDevices( raw_input_devices, array_count( raw_input_devices ), sizeof( raw_input_devices[ 0 ] ) );
        profiler_event_end( register_raw_input_devices );
    }
    
    if ( arguments_find_s( "-gfx_log", 0 ) ) {
        log_write_to_file( log_name_gl, string_arg_l( "win32_opengl.log" ) );
    }
    
    profiler_event_end( window_create );
}

stu void window_set_swap_interval( s32 interval ) {
    wglSwapIntervalEXT( interval );
}

stu void window_set_top_most( window_state_t* window, b32 top_most ) {
    
    HWND position = HWND_TOPMOST;
    
    if ( !top_most ) {
        position = HWND_NOTOPMOST;
    }
    
    SetWindowPos( window->platform.window_handle, position, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE | SWP_NOACTIVATE );
}

typedef enum window_handle_messages_flag_t {
    whm_none = 0,
    whm_wait_for_message = 1 << 0,
    whm_wait_skip_mouse_move = 1 << 1,
} window_handle_messages_flag_t;

/* NOTE simon: SetCapture and ReleaseCapture explanation here:
https://docs.microsoft.com/en-us/windows/win32/learnwin32/mouse-movement
It allows to receive mouse messages when the mouse goes outside the window and there was a button pressed. */
stu void internal_window_mouse_button_down( window_state_t* state, u8 virtual_key, u64 lparam ) {
    u8 value = input_get_transition_count( state, virtual_key );
    internal_window_set_vk_or_sc_state( state->keys + virtual_key, 1, 0, value + 1 );
    SetCapture( state->platform.window_handle );
    /* TODO simon: This is to make sure that clicks have the correct position, if frame time is high.
Hopefully at some point I will make that better since we need a event stream anyway. */
    state->mouse_x = ( ( s32 ) ( s16 ) ( lparam & 0xffff ) );
    state->mouse_y = state->height - 1 - ( ( s32 ) ( s16 ) ( ( lparam >> 16 ) & 0xffff ) );
}

stu void internal_window_mouse_button_up( window_state_t* state, u8 virtual_key, u64 lparam ) {
    u8 value = input_get_transition_count( state, virtual_key );
    internal_window_set_vk_or_sc_state( state->keys + virtual_key, 0, 0, value + 1 );
    ReleaseCapture( );
    /* TODO simon: This is to make sure that clicks have the correct position, if frame time is high.
    Hopefully at some point I will make that better since we need a event stream anyway. */
    state->mouse_x = ( ( s32 ) ( s16 ) ( lparam & 0xffff ) );
    state->mouse_y = state->height - 1 - ( ( s32 ) ( s16 ) ( ( lparam >> 16 ) & 0xffff ) );
}

/* NOTE simon: In application that need a guarantee on key input order, maybe we should run
the application loop after every input (without rendering) until there is no more message and
than render ? Or at least having some way to know the order of each key press so that if the
application hangs for a few seconds (as it's common when Windows has just started) we can
still process correctly any text the user inputed + possible arrow key movements etc... */
stu void window_handle_messages( window_state_t* state, u32 whm_flags ) {
    
    profiler_event_start( window_handle_messages );
    
    profiler_event_start( setup );
    state->resized = false;
    state->codepoint_count = 0;
    state->codepoints_total_byte_count_utf8 = 0;
    state->scroll_increment = 0;
    
    for ( umm key = 0; key < vk_count; key++ ) {
        state->keys[ key ] &= input_key_state_flag_pressed;
    }
    
    for ( umm scancode = 0; scancode < array_count( state->scancodes ); scancode++ ) {
        state->scancodes[ scancode ] &= input_key_state_flag_pressed;
    }
    
    b32 pause_scancode_read = false;
    
    MSG message;
    b32 has_message = false;
    
    profiler_event_end( setup );
    
    if ( whm_flags & whm_wait_for_message ) {
        
        profiler_event_start( GetMessage );
        
        has_message = GetMessageW( &message, 0, 0, 0 );
        
        while ( ( whm_flags & whm_wait_skip_mouse_move ) && ( message.message == WM_MOUSEMOVE ) ) {
            TranslateMessage( &message );
            DispatchMessageW( &message );
            has_message = GetMessageW( &message, 0, 0, 0 );
        }
        
        profiler_event_end( GetMessage );
        
    } else {
        profiler_event_start( PeekMessage );
        has_message = PeekMessageW( &message, 0, 0, 0, PM_REMOVE );
        profiler_event_end( PeekMessage );
    }
    
    profiler_event_start( loop );
    
    POINT p;
    GetCursorPos( &p );
    state->mouse_x = p.x - state->x;
    state->mouse_y = state->screen_height - 1 - p.y - state->y;
    
    while ( has_message ) {
        
        switch ( message.message ) {
            
#if 0
            case WM_MOUSEMOVE: {
                POINTS p = MAKEPOINTS( message.lParam );
                state->mouse_x = p.x;
                state->mouse_y = state->screen_height - 1 - p.y;
            } break;
#endif
            
            /* NOTE simon: WM_QUIT only happens if PostQuitMessage is called. */
            case WM_QUIT : {
                state->close_requested = true;
            } break;
            
            
            
            case WM_LBUTTONDOWN: {
                internal_window_mouse_button_down( state, vk_mouse_left, ( u64 ) message.lParam );
            } break;
            
            case WM_LBUTTONUP: {
                internal_window_mouse_button_up( state, vk_mouse_left, ( u64 ) message.lParam );
            } break;
            
            
            
            case WM_RBUTTONDOWN: {
                internal_window_mouse_button_down( state, vk_mouse_right, ( u64 ) message.lParam );
            } break;
            
            case WM_RBUTTONUP: {
                internal_window_mouse_button_up( state, vk_mouse_right, ( u64 ) message.lParam );
            } break;
            
            
            
            case WM_MBUTTONDOWN: {
                internal_window_mouse_button_down( state, vk_mouse_midle, ( u64 ) message.lParam );
            } break;
            
            case WM_MBUTTONUP: {
                internal_window_mouse_button_up( state, vk_mouse_midle, ( u64 ) message.lParam );
            } break;
            
            
            
            case WM_XBUTTONDOWN: {
                u8 virtual_key = ( u8 ) ( vk_mouse_midle + HIWORD( message.wParam ) );
                internal_window_mouse_button_down( state, virtual_key, ( u64 ) message.lParam );
            } break;
            
            case WM_XBUTTONUP: {
                u8 virtual_key = ( u8 ) ( vk_mouse_midle + HIWORD( message.wParam ) );
                internal_window_mouse_button_up( state, virtual_key, ( u64 ) message.lParam );
            } break;
            
            
            
            case WM_SYSKEYDOWN :
            case WM_KEYDOWN :
            case WM_SYSKEYUP :
            case WM_KEYUP: {
                
                u32 key = input_vk_to_key[ message.wParam ];
                
                u8 pressed = ( u8 ) ( ( ~message.lParam ) >> 31 ) & 0x1;
                u8 repeated = ( ( state->keys[ key ] & input_key_state_flag_pressed ) == input_key_state_flag_pressed ) && pressed;
                u8 transition = ( state->keys[ key ] & input_key_state_flag_pressed ) != pressed;
                u8 value = state->keys[ key ] & input_key_state_flag_transition_count_mask;
                
                b32 extended = ( message.lParam >> 24 ) & 0x1;
                u32 scancode = ( message.lParam >> 16 ) & 0xff;
                
                if ( extended ) {
                    
                    /* Num lock is extended in the lparam but not in the scancode list. */
                    if ( scancode != 0x45 ) {
                        scancode |= 0xE000;
                    }
                    
                } else {
                    
                    /* And Pause in not extended when it should be 0xE11D45. */
                    if ( scancode == 0x45 ) {
                        scancode = 0xE11D45;
                    } else if ( scancode == 0x54 ) { /* ALT + print screen */
                        scancode = 0xE037;
                    }
                }
                
                /* We need to force the transition because print screen doesn't sends keydown messages. */
                if ( scancode == 0xE037 ) {
                    transition = 1;
                }
                
                value += ( transition << 2 );
                value |= pressed;
                value |= ( repeated << 1 );
                state->keys[ key ] = value;
                
                /* NOTE simon: scancodes are set with raw input. */
                /* state->scancodes[ window_getScancodeOffset( scancode ) ] = value; */
                
                if ( message.wParam == VK_SHIFT ) {
                    
                    key = input_vk_to_key[ ( scancode == sc_shift_left ) ? VK_LSHIFT : VK_RSHIFT ];
                    state->keys[ key ] = value;
                    
                } else if ( message.wParam == VK_CONTROL) {
                    
                    key = input_vk_to_key[ ( extended ) ? VK_RCONTROL : VK_LCONTROL ];
                    state->keys[ key ] = value;
                    
                } else if ( message.wParam == VK_MENU ) {
                    
                    key = input_vk_to_key[ ( extended ) ? VK_RMENU : VK_LMENU ];
                    state->keys[ key ] = value;
                }
                
                if ( message.message == WM_SYSKEYDOWN && message.wParam == VK_F4 ) {
                    state->close_requested = true;
                }
                
                TranslateMessage( &message );
                
            } break;
            
            case WM_INPUT: {
                
                RAWINPUT raw_input;
                u32 raw_input_size = sizeof( raw_input );
                GetRawInputData( ( HRAWINPUT ) message.lParam, RID_INPUT, &raw_input, &raw_input_size, sizeof( RAWINPUTHEADER ) );
                
                if ( raw_input.header.dwType == RIM_TYPEKEYBOARD ) {
                    
                    b32 ignore = false;
                    u32 scancode = raw_input.data.keyboard.MakeCode;
                    u16 flags = raw_input.data.keyboard.Flags;
                    
                    if ( scancode ) {
                        
                        _assert( scancode <= 0xff );
                        
                        if ( ( flags & RI_KEY_E0 ) != 0 ) {
                            scancode |= 0xE000;
                        } else if ( ( flags & RI_KEY_E1 ) != 0 ) {
                            scancode |= 0xE100;
                        }
                        
                        if ( pause_scancode_read ) {
                            
                            if ( scancode == 0x45 ) {
                                scancode = 0xE11D45;
                            }
                            
                            pause_scancode_read = false;
                            
                        } else if ( scancode == 0xE11D ) {
                            
                            pause_scancode_read = true;
                            
                        } else if ( scancode == 0x54 ) {
                            /* 0x54 is generated when print screen and one or both alt are pressed. */
                            scancode = 0xE037;
                        }
                        
                        if ( scancode == 0xE11D || scancode == 0xE02A || scancode == 0xE0AA || scancode == 0xE0B6 || scancode == 0xE036 ) {
                            ignore = true;
                        }
                        
                    } else {
                        /* NOTE simon: when switching from a virtual machine to your program,
                        a numlock key press can be sent to your application with a scancode of 0.
                        Speculation: since it doesn't come from a hardware keypress, there is no
                        scancode in the raw input struct.
                        Hard to reproduce but quickly switching between the vm and the application
                        using the mouse on the taskbar icon seem to be best way to reproduce. */
                        u32 key = input_vk_to_key[ raw_input.data.keyboard.VKey ];
                        scancode = input_key_to_scan_code( key );
                        _assert( scancode && scancode < 0xe11d45 );
                    }
                    
                    u8 pressed = 0;
                    
                    if ( ( flags & RI_KEY_BREAK ) == 0 ) {
                        pressed = 1;
                    }
                    
                    if ( !ignore ) {
                        
                        umm offset = input_get_scancode_offset( scancode );
                        u8 value = state->scancodes[ offset ] & input_key_state_flag_transition_count_mask;
                        u8 repeated = ( ( state->scancodes[ offset ] & input_key_state_flag_pressed ) == input_key_state_flag_pressed ) && pressed;
                        u8 transition = ( state->scancodes[ offset ] & input_key_state_flag_pressed ) != pressed;
                        value += ( transition << 2 );
                        value |= pressed;
                        value |= ( repeated << 1 );
                        state->scancodes[ offset ] = value;
                    }
                } else if ( raw_input.header.dwType == RIM_TYPEMOUSE ) {
                    /* NOTE simon: we use raw input to get mouse movement messages when the mouse is outside the window and the window has the focus. */
                }
                
            } break;
            
            case WM_MOUSEWHEEL: {
                s16 delta = GET_WHEEL_DELTA_WPARAM( message.wParam );
                state->scroll_increment =  delta / ( r32 ) WHEEL_DELTA;
            } break;
            
            case WM_CHAR: {
                
                u32 c = ( u32 ) message.wParam;
                
                /* NOTE simon: Decode utf16. */
                u16 hi_surrogate_start = 0xD800;
                u16 lo_surrogate_start = 0xDC00;
                u16 surrogate_end = 0xDFFF;
                
                static u16 highSurrogate = 0;
                
                if ( c >= hi_surrogate_start && c < lo_surrogate_start ) {
                    
                    highSurrogate = ( u16 ) c;
                    
                } else {
                    
                    if ( c >= lo_surrogate_start && c <= surrogate_end ) {
                        
                        u16 lowSurrogate = ( u16 ) c;
                        c = ( highSurrogate - hi_surrogate_start ) << 10;
                        c |= ( lowSurrogate - lo_surrogate_start );
                        c += 0x10000;
                        
                        highSurrogate = 0;
                    }
                    
                    if ( state->codepoint_count < array_count( state->codepoints ) ) {
                        
                        u8 u8_count = 0;
                        u32 error = 0;
                        u8* start = state->codepoints_utf8;
                        u8* current = start + state->codepoints_total_byte_count_utf8;
                        u8* end = start + array_count( state->codepoints_utf8 );
                        utf8_encode( current, end, c, &u8_count, &error );
                        
                        state->codepoints[ state->codepoint_count ] = c;
                        state->codepoints_byte_count_utf8[ state->codepoint_count ] = u8_count;
                        state->codepoints_total_byte_count_utf8 += u8_count;
                        state->codepoint_count++;
                    }
                }
                
            } break;
            
            case WM_USER: {
                /* NOTE simon: This can be used to wakeup the application from a thread. */
            } break;
            
            default : {
                profiler_event_start( default_translate );
                TranslateMessage( &message );
                profiler_event_end( default_translate );
                profiler_event_start( default_dispatch );
                DispatchMessageW( &message );
                profiler_event_end( default_dispatch );
            } break;
        }
        profiler_event_start( PeekMessage );
        has_message = PeekMessageW( &message, 0, 0, 0, PM_REMOVE );
        profiler_event_end( PeekMessage );
    }
    
    profiler_event_end( loop );
    
    profiler_event_end( window_handle_messages );
}

stu void window_swap_buffers( window_state_t* window ) {
    profiler_event_start( window_swap_buffers );
    SwapBuffers( window->platform.device_context );
    profiler_event_end( window_swap_buffers );
}

// MARK audio

/* Core Audio APIs on msdn: https://msdn.microsoft.com/en-us/library/windows/desktop/dd316599(v=vs.85).aspx */
// #include <initguid.h>
#include "../lib/system_header_copy.h"

#define DEBUG_WASAPI 0

#ifndef window_WAVE_FORMAT_PCM
#define window_WAVE_FORMAT_PCM 0x0001
#endif

#ifndef window_WAVE_FORMAT_EXTENSIBLE
#define window_WAVE_FORMAT_EXTENSIBLE 0xFFFE
#endif

#define window_WAVE_EXTENSIBLE_GUID( type, guid ) \
*( ( u16* ) ( guid ) ) = type; \
( guid )[ 2 ] = 0x00;\
( guid )[ 3 ] = 0x00;\
( guid )[ 4 ] = 0x00;\
( guid )[ 5 ] = 0x00;\
( guid )[ 6 ] = 0x00;\
( guid )[ 7 ] = 0x10;\
( guid )[ 8 ] = 0x80;\
( guid )[ 9 ] = 0x00;\
( guid )[ 10 ] = 0x00;\
( guid )[ 11 ] = 0xAA;\
( guid )[ 12 ] = 0x00;\
( guid )[ 13 ] = 0x38;\
( guid )[ 14 ] = 0x9B;\
( guid )[ 15 ] = 0x71;

#if 0
/* was */
#define window_WAVE_EXTENSIBLE_GUID( type, guid ) \
*( ( u16* ) ( guid ) ) = type; \
( guid )[ 2 ] = 0x00;\
( guid )[ 3 ] = 0x00;\
( guid )[ 4 ] = 0x00;\
( guid )[ 5 ] = 0x00;\
( guid )[ 6 ] = 0x10;\
( guid )[ 7 ] = 0x00;\
( guid )[ 8 ] = 0x80;\
( guid )[ 9 ] = 0x00;\
( guid )[ 10 ] = 0x00;\
( guid )[ 11 ] = 0xAA;\
( guid )[ 12 ] = 0x00;\
( guid )[ 13 ] = 0x38;\
( guid )[ 14 ] = 0x9B;\
( guid )[ 15 ] = 0x71;
#endif

/* NOTE simon: The buffer length is assumed to be one second. Is there a downside or reason to let the user choose another length ? */
stu b32 sound_create( platform_specific_t* platform, sound_format_t* format, u32* out_buffer_size_in_samples ) {
    
    b32 result = false;
    IMMDeviceEnumerator* enumerator;
    
    /* TODO simon: make sure that Ole32.dll is present by default on windows. */
    HMODULE oleDLL = LoadLibrary( "Ole32.dll" ); // LPUNKNOWN pUnkOuter
    struct IUnknown;
    typedef HRESULT ( *CoCreateInstance_type )( const IID* rclsid, struct IUnknown* pUnkOuter, DWORD dwClsContext, const IID* riid, LPVOID *ppv );
    CoCreateInstance_type CoCreateInstance = ( CoCreateInstance_type ) GetProcAddress( oleDLL, "CoCreateInstance" );
    HRESULT error = CoCreateInstance( &CLSID_MMDeviceEnumerator, NULL, CLSCTX_ALL, &IID_IMMDeviceEnumerator, ( void** ) &enumerator );
    
    FreeLibrary( oleDLL );
    
    if ( error == S_OK ) {
        
        IMMDevice* device;
        error = IMMDeviceEnumerator_GetDefaultAudioEndpoint( enumerator, eRender, eConsole, &device );
        
        if ( error == S_OK ) {
            
            error = IMMDevice_Activate( device, &IID_IAudioClient, CLSCTX_ALL, NULL, ( void** ) &platform->audio_client );
            
            if ( error == S_OK ) {
                
                WAVEFORMATEXTENSIBLE bufferFormat;
                
                bufferFormat.Format.wFormatTag = window_WAVE_FORMAT_EXTENSIBLE;
                bufferFormat.Format.nChannels = format->channel_count;
                bufferFormat.Format.nSamplesPerSec = format->samples_per_second;
                bufferFormat.Format.wBitsPerSample = format->container_size * 8;
                bufferFormat.Format.nBlockAlign = bufferFormat.Format.nChannels * format->container_size;
                bufferFormat.Format.nAvgBytesPerSec = bufferFormat.Format.nSamplesPerSec * bufferFormat.Format.nBlockAlign;
                bufferFormat.Format.cbSize = sizeof( bufferFormat );
                bufferFormat.Samples.wValidBitsPerSample = format->bits_per_sample;
                bufferFormat.dwChannelMask = format->channel_mask;
                window_WAVE_EXTENSIBLE_GUID( window_WAVE_FORMAT_PCM, ( u8* ) &bufferFormat.SubFormat );
                
                /* NOTE simon: 1 reference_time = 100 nanoseconds = 1 sec / 10 000 000. */
                // REFERENCE_TIME bufferDuration = 10000000;
                u64 bufferDuration = 10000000;
                error = IAudioClient_Initialize( platform->audio_client, AUDCLNT_SHAREMODE_SHARED, 0, bufferDuration, 0, &bufferFormat.Format, 0 );
                
                if ( error == S_OK ) {
                    
                    error = IAudioClient_GetBufferSize( platform->audio_client, out_buffer_size_in_samples );
                    _assert( *out_buffer_size_in_samples == format->samples_per_second );
                    error = IAudioClient_GetService( platform->audio_client, &IID_IAudioRenderClient, ( void** ) &platform->audio_render_client );
                    
                    if ( error == S_OK ) {
                        
                        result = true;
                        
                    } else {
                        
                        IAudioClient_Release( platform->audio_client );
                        platform->audio_client = 0;
                        log_s( 0, "Unable to get the audio render client.\n" );
                    }
                    
                } else {
                    
                    IAudioClient_Release( platform->audio_client );
                    platform->audio_client = 0;
                    log_s( 0, "Unable to initialize the audio client.\n" );
                }
                
            } else {
                log_s( 0, "Unable to activate the audio client.\n" );
            }
            
            IMMDevice_Release( device );
            
        } else {
            log_s( 0, "Unable to get the default audio end point.\n" );
        }
        
        IMMDeviceEnumerator_Release( enumerator );
        
    } else {
        log_s( 0, "Unable to create MMDeviceEnumerator.\n" );
    }
    
    return result;
}

stu void sound_fill_buffer( platform_specific_t* platform, sound_format_t* format, u32 sampleCount, void* source ) {
    
    u8* buffer;
    u8* source_byte = ( u8* ) source;
    HRESULT error;
    
    error = IAudioRenderClient_GetBuffer( platform->audio_render_client, sampleCount, &buffer );
    
    if ( error == S_OK ) {
        
        umm byteTotal = sampleCount * format->container_size * format->channel_count;
        
        for ( umm byteIndex = 0; byteIndex < byteTotal; byteIndex++ ) {
            *buffer = *source_byte;
            buffer++;
            source_byte++;
        }
    }
    
    error = IAudioRenderClient_ReleaseBuffer( platform->audio_render_client, sampleCount, 0 );
}

stu u32 sound_get_sample_count_to_add( platform_specific_t* platform, u32 samples_per_second, r32 delta_time ) {
    
    u32 sample_count_to_add = 0;
    u32 pending_samples;
    u32 latency = ( u32 ) ceil( delta_time * 1.2f * samples_per_second );
    
    HRESULT error = IAudioClient_GetCurrentPadding( platform->audio_client, &pending_samples );
    
    if ( error == S_OK ) {
        
        if ( pending_samples < latency ) {
            sample_count_to_add = latency - pending_samples;
        }
        
#if DEBUG_WASAPI
        log_s( 0, "p: " );
        log_u32( pending_samples );
        log_s( 0; "\nl: " );
        log_u32( latency );
        log_s( 0, "\na: " );
        log_u32( sample_count_to_add );
        log_s( 0, "\n" );
#endif
    }
    
    return sample_count_to_add;
}

stu void sound_update( platform_specific_t* platform, sound_format_t* format, u32 sample_count_to_add, void* samples ) {
    
    static b32 first_call = true;
    
    sound_fill_buffer( platform, format, sample_count_to_add, samples );
    
    if( first_call ) {
        IAudioClient_Start( platform->audio_client );
        first_call = false;
    }
    
#if DEBUG_WASAPI
    u64 position_frequency;
    u64 position_units;
    
    IAudioClock* audio_clock;
    IAudioClient_GetService( platform->audio_client, &IID_IAudioClock, &audio_clock );
    IAudioClock_GetFrequency( audio_clock, &position_frequency );
    IAudioClock_GetPosition( audio_clock, &position_units, 0 );
    IAudioClock_Release( audio_clock );
    
    // Marker->PlayCursor = (DWORD)(SoundOutput.SamplesPerSecond * PositionUnits / PositionFrequency) % SoundOutput.SamplesPerSecond;
#endif
}

stu void sound_start( platform_specific_t* platform ) {
    IAudioClient_Start( platform->audio_client );
}

stu void sound_stop( platform_specific_t* platform ) {
    IAudioClient_Stop( platform->audio_client );
}

#elif defined LINUX

stu void window_create( window_state_t* state, void* window_name, u32 width, u32 height, s32 x, s32 y, b32 title_bar ) {
    
    state->platform.display = XOpenDisplay( 0 );
    
    if ( state->platform.display ){
        
        Window desktop = DefaultRootWindow( state->platform.display );
        GLint attributes[ ] = {
            GLX_RGBA,
            GLX_DEPTH_SIZE, 24,
            GLX_DOUBLEBUFFER,
            None
        };
        XVisualInfo* info = glXChooseVisual( state->platform.display, 0, attributes );
        
        if ( info ){
            Colormap colormap = XCreateColormap( state->platform.display, desktop, info->visual, AllocNone );
            XSetWindowAttributes window_attributes = {
                .colormap = colormap,
                .event_mask = ExposureMask | KeyPressMask,
            };
            state->platform.window = XCreateWindow( state->platform.display, desktop, 0, 0, width, height, 0, info->depth, InputOutput, info->visual, CWColormap | CWEventMask, &window_attributes );
            XMapWindow( state->platform.display, state->platform.window );
            XStoreName( state->platform.display, state->platform.window, window_name );
            
            state->platform.render_context = glXCreateContext( state->platform.display, info, 0, GL_TRUE );
            glXMakeCurrent( state->platform.display, state->platform.window, state->platform.render_context );
            
            char* version = ( char* ) glGetString( GL_VERSION );
            log_s( log_name_gl, version );
            state->close_requested = false;
            state->running = true;
            
        } else {
            log_l( log_name_gl, "Failed to choose visual.\n" );
        }
        
    } else {
        log_l( log_name_gl, "Failed to open the display.\n" );
    }
    
    if ( arguments_find_s( "-gfx_log", 0 ) ) {
        log_write_to_file( log_name_gl, string_arg_l( "linux_opengl.log" ) );
    }
}

stu void window_handle_messages( window_state_t* state, u32 whm_flags ) {
    
    state->resized = false;
    
    while ( 1 ) {
        
        XEvent event;
        XNextEvent( state->platform.display, &event );
        
        if ( event.type == Expose ) {
            
            XWindowAttributes window_attributes;
            XGetWindowAttributes( state->platform.display, state->platform.window, &window_attributes );
            state->width = window_attributes.width;
            state->height = window_attributes.height;
            break;
            
        } else if ( event.type == KeyPress ) {
            
            /* TODO simon: handle messages. */
            XKeyPressedEvent* e = ( XKeyPressedEvent* ) &event;
            
            if ( e->keycode == XKeysymToKeycode( state->platform.display, XK_Left ) ) {
                state->close_requested = true;
            }
            
            break;
            
        } else if ( event.type == ConfigureNotify ) {
            
            XConfigureEvent* e = ( XConfigureEvent* ) &event;
            state->width = e->width;
            state->height = e->height;
            state->resized = true;
            break;
        }
    }
}

stu void window_clean_up( window_state_t* state ) {
    
    glXMakeCurrent( state->platform.display, None, 0 );
    glXDestroyContext( state->platform.display, state->platform.render_context );
    XDestroyWindow( state->platform.display, state->platform.window );
    XCloseDisplay( state->platform.display );
}

stu void window_destroy( platform_specific_t* platform ) {
    
    glXMakeCurrent( platform->display, None, 0 );
    glXDestroyContext( platform->display, platform->render_context );
    XDestroyWindow( platform->display, platform->window );
    XCloseDisplay( platform->display );
}

stu void window_swap_buffers( window_state_t* state ) {
    glXSwapBuffers( state->platform.display, state->platform.window );
}

stu b32 sound_create( platform_specific_t* platform, sound_format_t* format, u32* out_buffer_size_in_samples ) {
    _assert( !"Not implemented." );
    return 0;
}

stu void sound_fill_buffer( platform_specific_t* platform, sound_format_t* format, u32 sampleCount, void* source ) {
    _assert( !"Not implemented." );
}

stu u32 sound_get_sample_count_to_add( platform_specific_t* platform, u32 samples_per_second, r32 delta_time ) {
    _assert( !"Not implemented." );
    return 0;
}

stu void sound_update( platform_specific_t* platform, sound_format_t* format, u32 sample_count_to_add, void* samples ) {
    _assert( !"Not implemented." );
}

stu void sound_start( platform_specific_t* platform ) {
    _assert( !"Not implemented." );
}

stu void sound_stop( platform_specific_t* platform ) {
    _assert( !"Not implemented." );
}

#else
#error Unsuported platform.
#endif

#define WINDOW_H

#endif
