#if !defined( COMMON_H )
#define COMMON_H

#if !defined( profiler_event_start )
#define profiler_event_start( x )
#endif

#if !defined( profiler_event_end )
#define profiler_event_end( x )
#endif

#if !defined( profiler_collapsable_start )
#define profiler_collapsable_start( x )
#endif

#if !defined( profiler_collapsable_end )
#define profiler_collapsable_end( x )
#endif

#if defined( NO_CRT )
#include "../lib/no_crt.h"
#endif

#include <stddef.h>
#include <stdint.h>
#include <limits.h>

#if defined( WINDOWS )

#define WIN32_LEAN_AND_MEAN 1
#include <windows.h>
#include <intrin.h> // NOTE simon: https://msdn.microsoft.com/en-us/library/26td21ds.aspx

#elif defined( LINUX )

#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/sendfile.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <sys/syscall.h>
#include <dirent.h>
#include <unistd.h>
#include <time.h>
#include <x86intrin.h>
#include <X11/Xlib.h>
#include <GL/glx.h>

#endif

typedef uint8_t u8;
typedef uint16_t u16;
typedef uint32_t u32;
typedef uint64_t u64;

typedef int8_t s8;
typedef int16_t s16;
typedef int32_t s32;
typedef int64_t s64;

typedef uint8_t b8;
typedef uint32_t b32;

typedef uintptr_t umm;
typedef intptr_t smm;

typedef float r32;
typedef double r64;

#define true 1
#define false 0

#define u8_alignment sizeof( u8 )
#define u16_alignment sizeof( u16 )
#define u32_alignment sizeof( u32 )
#define u64_alignment sizeof( u64 )

#define s8_alignment sizeof( s8 )
#define s16_alignment sizeof( s16 )
#define s32_alignment sizeof( s32 )
#define s64_alignment sizeof( s64 )

#define b8_alignment sizeof( b8 )
#define b32_alignment sizeof( b32 )

#define umm_alignment sizeof( umm )
#define smm_alignment sizeof( smm )

#define r32_alignment sizeof( r32 )
#define r64_alignment sizeof( r64 )

/* NOTE simon: single translation unit. */
#define stu static

// NOTE simon: 1##high so that a number starting with a 0 is not an octal number.
#define bit( high, low ) _bit( 1##high, 1##low )
#define _bit( high, low ) \
( ( ( ( high / 1000 ) & 0x01 ) << 7 ) \
| ( ( ( high / 100 ) & 0x01 ) << 6 ) \
| ( ( ( high / 10 ) & 0x01 ) << 5 ) \
| ( ( ( high / 1 ) & 0x01 ) << 4 ) \
| ( ( ( low / 1000 ) & 0x01 ) << 3 ) \
| ( ( ( low / 100 ) & 0x01 ) << 2 ) \
| ( ( ( low / 10 ) & 0x01 ) << 1 ) \
| ( ( ( low / 1 ) & 0x01 ) << 0 ) )

#define u8_max 0xff
#define u16_max 0xffff
#define u32_max 0xffffffff
#define u64_max 0xffffffffffffffff

#define s8_max 0x7f
#define s8_min ( -0x7f - 1 )
#define s16_max 0x7fff
#define s16_min ( -0x7fff - 1 )
#define s32_max 0x7fffffff
#define s32_min ( -0x7fffffff - 1 )
#define s64_max 0x7fffffffffffffff
#define s64_min ( -0x7fffffffffffffff - 1 )

#define r32_epsilon 1.192092896e-07F
#define r32_max 3.402823466e+38F
#define r32_min 1.175494351e-38F

#define r64_epsilon 2.2204460492503131e-016
#define r64_max 1.7976931348623158e+308
#define r64_min 2.2250738585072014e-308

#define umm_max ( ( ( umm ) 0 ) - 1 )
#if defined( TARGET_x64 )
#define smm_max ( ( smm ) 0x7fffffffffffffff )
#define smm_min ( ( smm ) 0x8000000000000000 )
#elif defined( TARGET_x32 )
#define smm_max ( ( smm ) 0x7fffffff )
#define smm_min ( ( smm ) 0x80000000 )
#else
#error Target architecture not defined.
#endif

#define array_count( array ) ( ( umm ) ( sizeof( array ) / sizeof( ( array )[ 0 ] ) ) )
#define array_count_field( type, name ) arrayCount( ( ( type* ) 0 )->name )
#define u32_will_wrap( value, addin ) ( ( u32_max - ( value ) ) >= ( addin ) )
#define u64_will_wrap( value, addin ) ( ( u64_max - ( value ) ) >= ( addin ) )
#define swap_type( a, b, type ) { type __backup__ = ( a ); ( a ) = ( b ); ( b ) = __backup__; }
#define sizeof_field( type, field ) sizeof( ( ( type* ) 0 )->field )
#define no_error( error_pointer ) ( !( *( error_pointer ) ) )
#define is_error( error_pointer ) ( *( error_pointer ) )
#define set_error( error_pointer, error_value ) ( ( *( error_pointer ) ) = error_value )
#define clear_error( error_pointer ) ( ( *( error_pointer ) ) = 0 )

#define kibioctet( value ) ( ( value ) * 0x400 )
#define mebioctet( value ) ( ( value ) * 0x100000 )
#define gibioctet( value ) ( ( value ) * 0x40000000 )
#define tebioctet( value ) ( ( value ) * 0x10000000000 )

#if defined( COMPILER_CL )
typedef union float_special_t {
    u32 i;
    r32 f;
} float_special_t;

/* NOTE simon: in C89 (and C++ ?) union can only initialize the first member.
https://www.ibm.com/support/knowledgecenter/SSLTBW_2.1.0/com.ibm.zos.v2r1.cbclx01/strin.htm
*/
#if 0

float_special_t IND0000 = { 0xffc00000 };
float_special_t QNANP = { 0x7f800000 };
float_special_t QNANN = { 0xff800000 };
float_special_t infP = { 0x7f800000 };
float_special_t infN = { 0xff800000 };

#else

float_special_t float_msvc_positive_indeterminate_form = { 0x7fc00000 }; /* NOTE simon: This is just NaN with only bit 23 set in the mantissa ? */
float_special_t float_msvc_negative_indeterminate_form = { 0xffc00000 }; /* NOTE simon: This is just NaN with only bit 23 set in the mantissa ? */
float_special_t float_quiet_positive_nan = { 0x7f800000 }; /* NOTE simon: First 23 need can't be zero, but can be anything. */
float_special_t float_quiet_negative_nan = { 0xff800000 }; /* NOTE simon: First 23 need can't be zero, but can be anything. */
float_special_t float_positive_infinity = { 0x7f800000 };
float_special_t float_negative_infinity = { 0xff800000 };

#define float_sign_mask 0x80000000
#define float_exponent_mask 0x7f800000
#define float_mantissa_mask 0x007fffff

b32 float_is_positive_nan( r32 f ) {
    
    float_special_t a;
    a.f = f;
    
    u32 sign_exponent = a.i & ( float_sign_mask | float_exponent_mask );
    u32 mantissa = a.i & float_mantissa_mask;
    
    b32 result = sign_exponent == float_quiet_positive_nan.i;
    result = result && ( mantissa != 0 );
    
    return result;
}

b32 float_is_negative_nan( r32 f ) {
    
    float_special_t a;
    a.f = f;
    
    u32 sign_exponent = a.i & ( float_sign_mask | float_exponent_mask );
    u32 mantissa = a.i & float_mantissa_mask;
    
    b32 result = sign_exponent == float_quiet_negative_nan.i;
    result = result && ( mantissa != 0 );
    
    return result;
}

b32 float_is_nan( r32 f ) {
    
    float_special_t a;
    a.f = f;
    
    u32 exponent = a.i & float_exponent_mask;
    u32 mantissa = a.i & float_mantissa_mask;
    
    b32 result = exponent == 0x7f800000;
    result = result && ( mantissa != 0 );
    
    return result;
}

b32 float_is_denormalized( r32 f ) {
    
    float_special_t a;
    a.f = f;
    
    u32 exponent = a.i & float_exponent_mask;
    u32 mantissa = a.i & float_mantissa_mask;
    
    b32 result = exponent == 0;
    result = result && ( mantissa != 0 );
    
    return result;
}

#endif

#endif

#if 0
#include <utility>
template <typename F>
struct Defer {
    Defer( F f ) : f( f ) {}
    ~Defer( ) { f( ); }
    F f;
};

template <typename F>
Defer<F> makeDefer( F f ) {
    return Defer<F>( f );
};

#define __defer( line ) defer_ ## line
#define _defer( line ) __defer( line )

struct defer_dummy { };
template<typename F>
Defer<F> operator+( defer_dummy, F&& f )
{
    return makeDefer<F>( std::forward<F>( f ) );
}

#define defer auto _defer( __LINE__ ) = defer_dummy( ) + [ & ]( )
#endif

#define safe_macro( macro ) do { macro; } while ( 0 )

#if defined( _ASSERT )

#define IF_ASSERT( expression ) expression

#if defined( WINDOWS )

#define _assert( expression ) safe_macro( if ( !( expression ) ) { __debugbreak( ); } )
#define debug_break( ) __debugbreak( )
#define invalid_code_path( ) __debugbreak( )

#elif defined( LINUX )

#if defined ( __clang__ )
#define _assert( expression ) safe_macro( if ( !( expression ) ) { __builtin_debugtrap( ); } )
#define debug_break( ) __builtin_debugtrap( )
#define invalid_code_path( ) __builtin_debugtrap( )
#endif

#endif

#if !defined( _assert )
#define _assert( expression ) safe_macro( if ( !( expression ) ) { __asm__( "int $3" ); } )
// #define _assert( expression ) safe_macro( if ( !( expression ) ) { *( ( uintptr_t* ) 0 ) = 0x00f00f00; } )
#endif

#if !defined( debug_break )
#define debug_break( ) safe_macro( *( ( uintptr_t* ) 0 ) = 0x00f00f00; )
#endif

#if !defined( invalid_code_path )
#define invalid_code_path( ) safe_macro( *( ( uintptr_t* ) 0 ) = 0x00f00f00; )
#endif


#define breakpoint_line( ) { int x = 0; x++; }
#define IF_DEBUG( expression ) expression

#else

#define IF_ASSERT( expression )
#define _assert( expression )
#define debug_break( )
#define invalid_code_path( )
#define breakpoint_line( )
#define IF_DEBUG( expression )

#endif

#define _ctassert3( expression, count ) struct Foo##count { int foo[ ( expression ) ? 1 : -1 ]; }
#define _ctassert2( expression, count ) _ctassert3( expression, count )
#define _ctassert( expression ) _ctassert2( expression, __COUNTER__ )

/* NOTE simon:
- Those "iterators" create a pointer "it" to the current value, and pointer "end" to one past the last value.
- The iterator finishing by "i" also create a "umm index" containing the current index.
- The iterator containing "n" create variables containing the given name (e.g if passed "abc", Type* abc, Type* abcEnd, umm abcIndex instead of it, end, index).
- The iterator containing "a" use array_count to get the number of iterations.
- The index version create the index before the loop, so it will be "leaked" in the current skope. If you don't want that you need to add extra { }.*/
#define for_p( type, pointer, count ) for ( type* it = ( pointer ), *end = ( pointer ) + ( count ); it < end; it++ )
#define for_pi( type, pointer, count ) umm index = 0; for ( type* it = ( pointer ), *end = ( pointer ) + ( count ); it < end; it++, index++ )
#define for_pn( type, pointer, count, name ) for ( type* name = ( pointer ), *name##End = ( pointer ) + ( count ); name < name##End; name++ )
#define for_pni( type, pointer, count, name ) umm name##Index = 0; for ( type* name = ( pointer ), *name##End = ( pointer ) + ( count ); name < name##End; name++, name##Index++ )

#define for_a( type, array ) for_p( type, ( array ), array_count( array ) )
#define for_ai( type, array ) for_pi( type, ( array ), array_count( array ) )
#define for_an( type, array, name ) for_pn( type, ( array ), array_count( array ), name )
#define for_ani( type, array, name ) for_pni( type, ( array ), array_count( array ), name )

#define equal_any( result, value, ... ) {\
umm internal__array_[] = { __VA_ARGS__ };\
internal_equal_any( result, value, internal__array_, arrayCount( internal__array_ ) );\
}

stu void internal_equal_any( b32* result, umm value, umm* array, umm count ) {
    
    ( *result ) = false;
    
    for ( umm i = 0; !( *result ) && i < count; i++ ) {
        ( *result ) = ( array[ i ] == value );
    }
}

/* MARK flags */

stu u8 flag_toggle_u8( u8* flags, u8 mask ) {
    u8 current = ( *flags ) & mask;
    u8 result = ( ~current ) & mask;
    ( *flags ) = ( ( *flags ) & ( ~mask ) ) | result;
    return result;
}

stu u16 flag_toggle_u16( u16* flags, u16 mask ) {
    u16 current = ( *flags ) & mask;
    u16 result = ( ~current ) & mask;
    ( *flags ) = ( ( *flags ) & ( ~mask ) ) | result;
    return result;
}

stu u32 flag_toggle_u32( u32* flags, u32 mask ) {
    u32 current = ( *flags ) & mask;
    u32 result = ( ~current ) & mask;
    ( *flags ) = ( ( *flags ) & ( ~mask ) ) | result;
    return result;
}

stu u64 flag_toggle_u64( u64* flags, u64 mask ) {
    u64 current = ( *flags ) & mask;
    u64 result = ( ~current ) & mask;
    ( *flags ) = ( ( *flags ) & ( ~mask ) ) | result;
    return result;
}

stu u8 flag_is_set_u8( u8 flags, u8 mask ) {
    u8 result = ( ( flags & mask ) == mask ) ? mask : 0;
    return result;
}

stu u16 flag_is_set_u16( u16 flags, u16 mask ) {
    u16 result = ( ( flags & mask ) == mask ) ? mask : 0;
    return result;
}

stu u32 flag_is_set_u32( u32 flags, u32 mask ) {
    u32 result = ( ( flags & mask ) == mask ) ? mask : 0;
    return result;
}

stu u64 flag_is_set_u64( u64 flags, u64 mask ) {
    u64 result = ( ( flags & mask ) == mask ) ? mask : 0;
    return result;
}

stu void flag_set_u8( u8* flags, u8 mask ) {
    ( *flags ) |= mask;
}

stu void flag_set_u16( u16* flags, u16 mask ) {
    ( *flags ) |= mask;
}

stu void flag_set_u32( u32* flags, u32 mask ) {
    ( *flags ) |= mask;
}

stu void flag_set_u64( u64* flags, u64 mask ) {
    ( *flags ) |= mask;
}

stu void flag_clear_u8( u8* flags, u8 mask ) {
    ( *flags ) &= ( ~mask );
}

stu void flag_clear_u16( u16* flags, u16 mask ) {
    ( *flags ) &= ( ~mask );
}

stu void flag_clear_u32( u32* flags, u32 mask ) {
    ( *flags ) &= ( ~mask );
}

stu void flag_clear_u64( u64* flags, u64 mask ) {
    ( *flags ) &= ( ~mask );
}

stu void flag_set_or_clear_u8( u8* flags, b32 set, u8 mask ) {
    if ( set ) {
        flag_set_u8( flags, mask );
    } else {
        flag_clear_u8( flags, mask );
    }
}

stu void flag_set_or_clear_u16( u16* flags, b32 set, u16 mask ) {
    if ( set ) {
        flag_set_u16( flags, mask );
    } else {
        flag_clear_u16( flags, mask );
    }
}

stu void flag_set_or_clear_u32( u32* flags, b32 set, u32 mask ) {
    if ( set ) {
        flag_set_u32( flags, mask );
    } else {
        flag_clear_u32( flags, mask );
    }
}

stu void flag_set_or_clear_u64( u64* flags, b32 set, u64 mask ) {
    if ( set ) {
        flag_set_u64( flags, mask );
    } else {
        flag_clear_u64( flags, mask );
    }
}

typedef struct data_t {
    
    umm reserved;
    umm used;
    
    union {
        u8* bytes;
        char* text;
        void* raw;
        umm offset;
    };
    
} data_t;

#define data_t_alignment sizeof_field( data_t, reserved )

stu data_t memory_construct( umm reserved, umm used, void* memory ) {
    
    data_t result;
    result.reserved = reserved;
    result.used = used;
    result.raw = memory;
    
    return result;
}

// MARK memory

u32 memory_page_size = 0x1000;
u32 memory_allocation_granularity = 0x10000;
void* memory_minimum_address = ( void* ) 0x10000; /* NOTE simon: windows reserve the first 16K to use as handles. */
void* memory_maximum_address = ( void* ) ( tebioctet( 8 ) - 1 ); /* NOTE simon: windows 8.1 and up should be terabytes( 128 ) - 1. */

/* NOTE simon: from https://en.wikipedia.org/wiki/Data_structure_alignment#Computing_padding */
stu umm memory_get_padding( void* address, umm alignment ) {
    
    umm a = ( umm ) address;
    umm padding = ( ( ~( a - 1 ) ) & ( alignment - 1 ) );
    return padding;
}

stu void* memory_get_aligned( void* address, umm alignment ) {
    
    umm a = ( umm ) address;
    void* result = ( void* ) ( ( a + ( alignment - 1 ) ) & ( ~( alignment - 1 ) ) );
    return result;
}

#if defined( WINDOWS )

stu void memory_get_info( u32* page_size, u32* allocation_granularity, void* minimum_address, void* maximum_address ) {
    
#if 0
    typedef struct _SYSTEM_INFO {
        union {
            DWORD  dwOemId;
            struct {
                WORD wProcessorArchitecture;
                WORD wReserved;
            };
        };
        DWORD     dwPageSize;
        LPVOID    lpMinimumApplicationAddress;
        LPVOID    lpMaximumApplicationAddress;
        DWORD_PTR dwActiveProcessorMask;
        DWORD     dwNumberOfProcessors;
        DWORD     dwProcessorType;
        DWORD     dwAllocationGranularity;
        WORD      wProcessorLevel;
        WORD      wProcessorRevision;
    } SYSTEM_INFO;
#endif
    
    SYSTEM_INFO info;
    
    GetSystemInfo( &info );
    
    if ( page_size ) {
        *page_size = info.dwPageSize;
    }
    
    if ( allocation_granularity ) {
        *allocation_granularity = info.dwAllocationGranularity;
    }
    
    if ( minimum_address ) {
        ( *( u64** ) minimum_address ) = ( u64* ) info.lpMinimumApplicationAddress;
    }
    
    if ( maximum_address ) {
        ( *( u64** ) maximum_address ) = ( u64* ) info.lpMaximumApplicationAddress;
    }
}

stu umm memory_get_offset( void* start, void* end ) {
    
    umm result = ( umm ) ( ( u8* ) end - ( u8* ) start );
    return result;
}

stu b32 memory_is_address_in_block( data_t* block, void* address ) {
    void* start = ( void* ) block->bytes;
    void* end = ( void* ) ( block->bytes + block->reserved );
    b32 result = ( address >= start && address < end );
    return result;
}

/* NOTE simon: Virtual alloc rounds to page boundaries. So it may return more space then requested. Also pages need to be 64 K aligned (allocation granularity) so requesting a 4k page wastes 60k of memory. */
#if defined( DEBUG )
umm debug_total_virtual_memory = 0;
#endif

stu data_t memory_get_virtual( umm size, umm address ) {
    
    data_t memory = { 0 };
    memory.reserved = size;
    memory.bytes = ( u8* ) VirtualAlloc( ( LPVOID ) address, size, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );
#if defined( DEBUG )
    debug_total_virtual_memory += size;
#endif
    return memory;
}

stu void memory_free_virtual( data_t* memory ) {
    
    if ( memory->bytes != 0 ) {
        VirtualFree( memory->bytes, 0, MEM_RELEASE );
        memory->bytes = 0;
        memory->reserved = memory->used = 0;
#if defined( DEBUG )
        debug_total_virtual_memory -= memory->reserved;
#endif
    }
}

/* NOTE simon: Address will be rounded down to match the allocation granularity. The size will be rounded up to allocation granularity. */
stu data_t memory_reserve( umm size, umm address ) {
    
    size += memory_get_padding( ( void* ) size, memory_allocation_granularity );
    
    data_t memory;
    memory.raw = VirtualAlloc( ( LPVOID ) address, size, MEM_RESERVE, PAGE_READWRITE );
    memory.reserved = size;
    memory.used = 0;
    
    _assert( memory.raw );
    
    return memory;
}

/* NOTE simon: Size will be rounded up to allocation granularity. */
stu void* memory_commit( data_t* memory, umm size ) {
    
    size += memory_get_padding( ( void* ) size, memory_page_size );
    _assert( memory->used + size <= memory->reserved );
    void* result = VirtualAlloc( memory->bytes + memory->used, size, MEM_COMMIT, PAGE_READWRITE );
    memory->used += size;
    
    _assert( result );
    return result;
}

/* NOTE simon: Size will be rounded up to allocation granularity. */
stu b32 memory_decommit( data_t* memory, umm size ) {
    
    size += memory_get_padding( ( void* ) size, memory_page_size );
    _assert( memory->used >= size );
    _assert( memory_get_padding( memory->bytes + memory->used, memory_page_size ) == 0 );
    b32 result = VirtualFree( memory->bytes + memory->used - size, size, MEM_DECOMMIT );
    memory->used -= size;
    
    _assert( result );
    return result;
}

#elif defined LINUX

stu data_t memory_get_virtual( umm size, umm address ) {
    
    data_t memory = { };
    memory.reserved = size;
    memory.bytes = ( u8* ) mmap( ( void* ) address, size, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, -1, 0 );
    return memory;
}

stu void memory_free_virtual( data_t* memory ) {
    
    if ( memory->bytes != 0 ) {
        munmap( memory->bytes, memory->reserved );
    }
}

stu data_t memory_reserve( umm size, umm address ) {
    
    size += memory_get_padding( ( void* ) size, memory_allocation_granularity );
    
    data_t memory;
    memory.raw = mmap( ( void* ) address, size, PROT_NONE, MAP_PRIVATE | MAP_ANONYMOUS, -1, 0 );
    memory.reserved = size;
    memory.used = 0;
    
    _assert( memory.raw );
    
    return memory;
}

stu void* memory_commit( data_t* memory, umm size ) {
    
    size += memory_get_padding( ( void* ) size, memory_page_size );
    _assert( memory->used + size <= memory->reserved );
    b32 failure = mprotect( memory->bytes + memory->used, size, PROT_READ | PROT_WRITE );
    void* result = memory->bytes + memory->used;
    memory->used += size;
    
    _assert( failure == 0 );
    
    return result;
}

/* TODO simon: This is not a proper implementation as it it keeps memory around. */
stu b32 memory_decommit( data_t* memory, umm size ) {
    
    size += memory_get_padding( ( void* ) size, memory_page_size );
    _assert( memory->used >= size );
    _assert( memory_get_padding( memory->bytes + memory->used, memory_page_size ) == 0 );
    b32 result = mprotect( memory->bytes + memory->used - size, size, PROT_NONE );
    memory->used -= size;
    
    _assert( result == 0 );
    return result;
}

#else
#error Unsupported platform.
#endif

#define memory_get_on_stack( name, size ) u8 name_##name##_stack_memory[ size ]; data_t name = memory_construct( size, 0, name_##name##_stack_memory )

#define string_l( string ) { sizeof( string ) /* reserved */, sizeof( string ) - 1 /* used */, ( u8* ) ( string ) /* text */ }
#define string_arg_l( string ) memory_construct( sizeof( string ), sizeof( string ) - 1, ( void* ) ( string ) )
stu void log_d( umm index, data_t string );

/* TODO simon: Test for alignment with the custom allocator because as far as I know everything could be miss-aligned. */
umm memory_default_alignment = sizeof( umm );

stu data_t memory_get_block( data_t* memory, umm size, umm alignment ) {
    
    data_t block = { 0 };
    umm use_alignment = alignment;
    
    _assert( alignment == 0 || alignment == 1 || alignment == 2 || alignment == 4 || alignment == 8 || alignment == 16 || alignment == 32 || alignment == 64 );
    
    if ( alignment == 0 ) {
        use_alignment = memory_default_alignment;
    }
    
    umm padding = memory_get_padding( memory->bytes + memory->used, use_alignment );
    
    _assert( memory->used + size + padding <= memory->reserved );
    
    if ( memory->used + size + padding <= memory->reserved ) {
        block.bytes = memory->bytes + padding + memory->used;
        memory->used += ( size + padding );
        block.reserved = size;
    } else {
#if defined( LOG )
        log_d( 0, string_arg_l( "***Out of memory in the data_t block.***\n " ) );
#endif
    }
    
    return block;
}

stu data_t memory_get_block_remaining( data_t* memory, umm alignment ) {
    
    data_t block = { 0 };
    
    umm use_alignment = alignment;
    
    _assert( alignment == 0 || alignment == 1 || alignment == 2 || alignment == 4 || alignment == 8 || alignment == 16 || alignment == 32 || alignment == 64 );
    
    if ( alignment == 0 ) {
        use_alignment = memory_default_alignment;
    }
    
    umm padding = memory_get_padding( memory->bytes + memory->used, use_alignment );
    
    if ( memory->used + padding <= memory->reserved ) {
        block.bytes = memory->bytes + memory->used + padding;
        block.reserved = memory->reserved - memory->used - padding;
        memory->used = memory->reserved;
    }
    
    return block;
}

stu data_t memory_get_temp( data_t memory ) {
    
    data_t result = { 0 };
    umm padding = memory_get_padding( memory.bytes + memory.used, 64 );
    
    if ( memory.used + padding <= memory.reserved ) {
        result.bytes = memory.bytes + memory.used + padding;
        result.reserved = memory.reserved - memory.used - padding;
    }
    
    return result;
}

stu data_t memory_get_temp_unaligned( data_t memory ) {
    
    data_t result = { 0 };
    
    if ( memory.used <= memory.reserved ) {
        result.bytes = memory.bytes + memory.used;
        result.reserved = memory.reserved - memory.used;
    }
    
    return result;
}

#define memory_from_array( array ) memory_construct( sizeof( array ), 0, array )

stu void memory_invalidate( data_t* memory ) {
    memory->used = memory->reserved = 0;
    memory->raw = 0;
}

stu void memory_free( data_t* memory ) {
    memory->used = 0;
}

stu umm memory_tight( data_t* memory ) {
    umm result = memory->reserved - memory->used;
    memory->reserved = memory->used;
    return result;
}

stu data_t memory_start_end( void* start, void* end ) {
    
    data_t result;
    result.reserved = result.used = ( u8* ) end - ( u8* ) start;
    result.bytes = ( u8* ) start;
    
    return result;
}

stu u8* memory_end( data_t memory ) {
    u8* result = memory.bytes + memory.used;
    return result;
}

/* NOTE simon: Alignment for a structure should be to the size of its biggest member. If it contains a "sub struct" the member of the sub struct must be considered. Source: http://www.catb.org/esr/structure-packing/. */

#define memory_push_type( memory, type ) ( type* ) memory_push_size( ( memory ), sizeof( type ), sizeof( type ) )
#define memory_push_struct( memory, type ) ( type* ) memory_push_size( ( memory ), sizeof( type ), type##_alignment )
#define memory_push_struct_aligned( memory, type, alignment ) ( type* ) memory_push_size( ( memory ), sizeof( type ), ( alignment ) )
#define memory_push_array( memory, type, count ) ( type* ) memory_push_size( ( memory ), ( count ) * sizeof( type ), type##_alignment )
#define memory_push_array_aligned( memory, type, count, alignment ) ( type* ) memory_push_size( ( memory ), ( count ) * sizeof( type ), ( alignment ) )

stu void* memory_push_size( data_t* memory, umm size, umm alignment ) {
    
    void* result = 0;
    
    umm use_alignment = alignment;
    
    _assert( alignment == 0 || alignment == 1 || alignment == 2 || alignment == 4 || alignment == 8 );
    
    if ( alignment == 0 ) {
        use_alignment = memory_default_alignment;
    }
    
    umm padding = memory_get_padding( memory->bytes + memory->used, use_alignment );
    
#if 0
    _assert( memory->used + size + padding <= memory->reserved );
#endif
    
    if ( memory->used + size + padding <= memory->reserved ) {
        result = memory->bytes + memory->used + padding;
        memory->used += size + padding;
    } else {
#if defined( LOG )
        log_d( 0, string_arg_l( "***Out of memory in the data_t block.***\n" ) );
#endif
    }
    
    return result;
}

#define memory_as( memory, type ) ( type* ) ( ( memory ).bytes )
#define memory_current( memory, type ) ( type* ) ( ( memory ).bytes + ( memory ).used )
#define memory_current_aligned( memory, type, alignment ) ( type* ) memory_get_aligned( ( memory )->bytes + ( memory )->used, alignment )

#define memory_left( memory ) ( ( memory ).reserved - ( memory ).used )

stu u8* memory_align( data_t* memory, umm alignment ) {
    
    umm padding = memory_get_padding( memory_current( *memory, u8 ), alignment );
    memory_push_size( memory, padding, 1 );
    
    return memory_current( *memory, u8 );
}

stu u8* memory_push_u8( data_t* memory, u8 number ) {
    u8* n = ( u8* ) memory_push_size( memory, sizeof( number ), sizeof( number ) );
    if ( n ) {
        *n = number;
    }
    return n;
}

stu s8* memory_push_s8( data_t* memory, s8 number ) {
    s8* n = ( s8* ) memory_push_size( memory, sizeof( number ), sizeof( number ) );
    if ( n ) {
        *n = number;
    }
    return n;
}

stu u16* memory_push_u16( data_t* memory, u16 number ) {
    u16* n = ( u16* ) memory_push_size( memory, sizeof( number ), sizeof( number ) );
    if ( n ) {
        *n = number;
    }
    return n;
}

stu s16* memory_push_s16( data_t* memory, s16 number ) {
    s16* n = ( s16* ) memory_push_size( memory, sizeof( number ), sizeof( number ) );
    if ( n ) {
        *n = number;
    }
    return n;
}

stu u32* memory_push_u32( data_t* memory, u32 number ) {
    u32* n = ( u32* ) memory_push_size( memory, sizeof( number ), sizeof( number ) );
    if ( n ) {
        *n = number;
    }
    return n;
}

stu s32* memory_push_s32( data_t* memory, s32 number ) {
    s32* n = ( s32* ) memory_push_size( memory, sizeof( number ), sizeof( number ) );
    if ( n ) {
        *n = number;
    }
    return n;
}

stu u64* memory_push_u64( data_t* memory, u64 number ) {
    u64* n = ( u64* ) memory_push_size( memory, sizeof( number ), sizeof( number ) );
    if ( n ) {
        *n = number;
    }
    return n;
}

stu s64* memory_push_s64( data_t* memory, s64 number ) {
    s64* n = ( s64* ) memory_push_size( memory, sizeof( number ), sizeof( number ) );
    if ( n ) {
        *n = number;
    }
    return n;
}

#define memory_pop_array( memory, type, count ) memory_pop_size( ( memory ), sizeof( type ) * ( count ) )
#define memory_pop_struct( memory, type ) memory_pop_size( ( memory ), sizeof( type ) )
#define memory_pop_type( memory, type ) memory_pop_size( ( memory ), sizeof( type ) )

stu void memory_pop_size( data_t* memory, umm size ) {
    memory->used -= size;
}

#define memory_zero_block( block ) memory_zero( ( block )->bytes, ( block )->reserved )
#define memory_zero_array( memory, type, count ) memory_zero( ( u8* ) ( memory ), ( count ) * sizeof( type ) )
#define memory_zero_struct( pointer ) memory_zero( ( u8* ) ( pointer ), sizeof( *( pointer ) ) )

stu void memory_zero( void* bytes, umm size ) {
    
    u8* b = ( u8* ) bytes;
    
    while ( size > 0 ) {
        size--;
        *b++ = 0;
    }
}

stu void memory_set( void* bytes, umm size, u8 value ) {
    
    u8* b = ( u8* ) bytes;
    
    while ( size > 0 ) {
        size--;
        *b++ = value;
    }
}

stu void memory_set_d( data_t* memory, u8 value ) {
    memory_set( memory->bytes, memory->reserved, value );
}

#if 0
// dst and src must be 16-byte aligned
// size must be multiple of 16*2 = 32 bytes
static void CopyWithSSENoCache(uint8_t* dst, uint8_t* src, size_t size)
{
    size_t stride = 2 * sizeof(__m128);
    while (size)
    {
        __m128 a = _mm_load_ps((float*)(src + 0*sizeof(__m128)));
        __m128 b = _mm_load_ps((float*)(src + 1*sizeof(__m128)));
        _mm_stream_ps((float*)(dst + 0*sizeof(__m128)), a);
        _mm_stream_ps((float*)(dst + 1*sizeof(__m128)), b);
        
        size -= stride;
        src += stride;
        dst += stride;
    }
}

#endif

stu void memory_copy( void* destination, void* source, umm size ) {
    
    _assert( destination || size == 0 );
    _assert( source || size == 0 );
    
    u8* d = ( u8* ) destination;
    u8* s = ( u8* ) source;
    u8* s_end = s + size;
    
    while ( s < s_end ) {
        *d++ = *s++;
    }
}

stu void memory_copy_dd( data_t* destination, data_t* source ) {
    _assert( destination->reserved < source->used );
    memory_copy( destination->bytes, source->bytes, source->used );
    destination->used = source->used;
}

stu void memory_replace_byte( data_t* memory, void* search, void* replace ) {
    
    u8* s = ( u8* ) search;
    u8* r = ( u8* ) replace;
    
    umm search_count = 0;
    umm replace_count = 0;
    
    while ( s[ search_count ] ) {
        search_count++;
    }
    
    while ( r[ replace_count ] ) {
        replace_count++;
    }
    
    _assert( search_count == replace_count );
    
    for ( umm i = 0; i < memory->used; i++ ) {
        
        for ( umm j = 0; j < search_count; j++ ) {
            if ( memory->bytes[ i ] == s[ j ] ) {
                memory->bytes[ i ] = r[ j ];
            }
        }
    }
}

stu void* memory_push_copy_p( data_t* memory, void* source, umm size, umm alignment ) {
    void* destination = memory_push_size( memory, size, alignment );
    if ( destination ) {
        memory_copy( destination, source, size );
    }
    
    return destination;
}

stu void* memory_push_copy( data_t* memory, data_t source, umm alignment ) {
    void* destination = memory_push_size( memory, source.used, alignment );
    if ( destination ) {
        memory_copy( destination, source.bytes, source.used );
    }
    
    return destination;
}

#define memory_push_copy_s( memory, string ) memory_push_copy_p( ( memory ), ( string ), string_length( string ), 1 )
#define memory_push_copy_l( memory, string ) memory_push_copy_p( ( memory ), ( string ), sizeof( string ) - 1, 1 )

stu umm memory_push_copy_region_d( data_t* memory, data_t* source, umm start_offset, umm size, b32 size_from_end ) {
    
    umm result = 0;
    
    if ( start_offset <= source->used ) {
        
        umm copy_size = source->used - start_offset;
        
        if ( size ) {
            
            if ( size_from_end ) {
                
                if ( size < copy_size ) {
                    size = copy_size - size;
                } else {
                    size = 0;
                }
                
            } else {
                size = ( size <= copy_size ) ? size : copy_size;
            }
            
        } else {
            size = copy_size;
        }
        
        umm left = memory_left( *memory );
        
        if ( size <= left ) {
            memory_push_copy_p( memory, source->bytes + start_offset, size, 1 );
            result = size;
        }
    }
    
    return result;
}

stu b32 memory_are_equal_p( void* a, umm a_length, void* b, umm b_length ) {
    
    _assert( ( ( a_length > 0 ) && a ) || a_length == 0 );
    _assert( ( ( b_length > 0 ) && b ) || b_length == 0 );
    
    b32 result = ( a_length == b_length );
    
    umm i = 0;
    u8* a_bytes = ( u8* ) a;
    u8* b_bytes = ( u8* ) b;
    
    while ( result && i < a_length ) {
        result = a_bytes[ i ] == b_bytes[ i ];
        i++;
    }
    
    return result;
}

stu b32 memory_are_equal( data_t a, data_t b ) {
    b32 result = memory_are_equal_p( a.bytes, a.used, b.bytes, b.used );
    return result;
}

stu b32 memory_start_with_p( void* start_arg, void* end_arg, void* what_arg, umm what_length ) {
    
    _assert( start_arg );
    _assert( what_arg );
    
    u8* current = ( u8* ) start_arg;
    u8* end = current + what_length;
    
    b32 result = end <= ( u8* ) end_arg;
    u8* what = ( u8* ) what_arg;
    
    while ( result && current < end ) {
        result = ( *current ) == ( *what );
        current++;
        what++;
    }
    
    return result;
}

stu b32 memory_start_with( data_t in, data_t what ) {
    b32 result = memory_start_with_p( in.bytes, memory_end( in ), what.bytes, what.used );
    return result;
}

/* TODO simon: this search algorithm https://en.wikipedia.org/wiki/Boyer%E2%80%93Moore_string-search_algorithm */
stu void*  memory_seek_p( void* start_param, void* end_param, void* what_param, umm what_size ) {
    
    u8* start = ( u8* ) start_param;
    u8* end = ( u8* ) end_param;
    umm in_size = end - start;
    u8* what = ( u8* ) what_param;
    u8* result = end;
    
    if ( start && end && what_size > 0 && what_size <= in_size ) {
        
        end -= ( what_size - 1 );
        
        while ( start < end ) {
            
            u8* bytes = start;
            umm size = 0;
            
            while ( size < what_size && bytes[ size ] == what[ size ] ) {
                size++;
            }
            
            if ( size == what_size ) {
                result = start;
                break;
            } else {
                start++;
            }
        }
    }
    
    return result;
}

stu void* memory_seek_backward_p( void* start_param, void* end_param, void* what_param, umm what_size ) {
    
    u8* start = ( u8* ) start_param;
    u8* end = ( u8* ) end_param;
    umm in_size = end - start;
    u8* what = ( u8* ) what_param;
    void* result = end;
    
    if ( start && end && what_size > 0 && what_size <= in_size ) {
        
        end -= what_size;
        
        while ( end >= start ) {
            
            u8* bytes = end;
            umm size = 0;
            
            while ( size < what_size && bytes[ size ] == what[ size ] ) {
                size++;
            }
            
            if ( size == what_size ) {
                result = end;
                break;
            } else {
                end--;
            }
        }
    }
    
    return result;
}

#define memory_seek( in, what ) memory_seek_p( ( in ).bytes, ( in ).bytes + ( in ).used, ( what ).bytes, ( what ).used )
#define memory_seek_backward( in, what ) memory_seek_backward_p( ( in ).bytes, ( in ).bytes + ( in ).used, ( what ).bytes, ( what ).used )

/* MARK reverse stack */

typedef struct reverse_stack_t {
    
    data_t* stack;
    umm used;
    
    union {
        void* data;
        u8* bytes;
        char* text;
        umm offset;
    };
    
} reverse_stack_t;

stu reverse_stack_t memory_reverse_stack_create( data_t* memory ) {
    
    reverse_stack_t reverse_stack = { 0 };
    reverse_stack.stack = memory;
    reverse_stack.bytes = memory->bytes + memory->reserved;
    reverse_stack.used = 0;
    
    return reverse_stack;
}

stu void* memory_reverse_stack_push( reverse_stack_t* reverse_stack, umm size ) {
    
    void* result = 0;
    _assert( reverse_stack->used + size < reverse_stack->stack->reserved - reverse_stack->stack->used );
    result = reverse_stack->bytes - reverse_stack->used - size;
    reverse_stack->used += size;
    
    return result;
}

// MARK array

#define array_reserved( array, type ) array_reserved_size( ( array ), sizeof( type ) )

stu umm array_reserved_size( data_t* array, umm elementSize ) {
    umm result = array->reserved / elementSize;
    return result;
}

#define array_used( array, type ) array_used_size( ( array ), sizeof( type ) )

stu umm array_used_size( data_t* array, umm elementSize ) {
    umm result = array->used / elementSize;
    return result;
}

#define array_get_pointer( array, type, index ) ( type* ) ( ( array ).bytes + ( sizeof( type ) * ( index ) ) )
#define array_get( array, type, index ) ( *array_get_pointer( ( array ), ( type ), ( index ) ) )
#define array_set( array, type, index, value ) ( *( array_get_pointer( ( array ), ( type ), ( index ) ) ) = ( value ) )

#define array_push( array, type, value ) array_push_size( ( array ), sizeof( type ), &( value ) )

stu umm array_push_size( data_t* array, umm size, void* value ) {
    
    umm result = umm_max;
    
    if ( array->used + size <= array->reserved ) {
        result = array_used_size( array, size );
        memory_copy( array->bytes + array->used, value, size );
        array->used += size;
    }
    
    return result;
}

// MARK char

stu b32 char_is_upper_case( u32 c ) {
    b32 result = ( c >= 'A' && c <= 'Z' );
    return result;
}

stu b32 char_is_lower_case( u32 c ) {
    b32 result = ( c >= 'a' && c <= 'z' );
    return result;
}

stu b32 char_is_letter( u32 c ) {
    b32 result = ( char_is_upper_case( c ) || char_is_lower_case( c ) );
    return result;
}

stu u32 char_to_lower_case( u32 c ) {
    u32 result = ( c >= 'A' && c <= 'Z' ) ? c + 0x20 : c;
    return result;
}

stu u32 char_to_upper_case( u32 c ) {
    u32 result = ( c >= 'a' && c <= 'z' ) ? c - 0x20 : c;
    return result;
}

stu b32 char_is_decimal( u32 c ) {
    b32 result = ( c >= '0' && c <= '9' );
    return result;
}

stu b32 char_is_hexadecimal( u32 c ) {
    b32 result = ( c >= '0' && c <= '9' ) || ( c >= 'a' && c <= 'f' ) || ( c >= 'A' && c <= 'F' );
    return result;
}

stu b32 char_is_alpha_numeric( u32 c ) {
    b32 result = ( char_is_decimal( c ) || char_is_letter( c ) );
    return result;
}

stu b32 char_is_alpha_numeric_underscore( u32 c ) {
    b32 result = ( ( c == '_' ) || char_is_alpha_numeric( c ) );
    return result;
}

stu b32 char_is_whitespace( u32 c ) {
    b32 result = ( c == ' ' || c == '\t' || c == '\v' || c == '\f' || c == '\n' || c == '\r' );
    return result;
}

stu b32 char_is_whitespace_not_new_line( u32 c ) {
    b32 result = ( c == ' ' || c == '\t' || c == '\v' || c == '\f' );
    return result;
}

stu b32 char_is_control( u32 c ) {
    b32 result = ( c < 0x20 || ( c >= 0x80 && c < 0xa0 ) );
    return result;
}

stu b32 char_is_any_u8( u32 c, void* list, umm count ) {
    
    b32 result = 0;
    u8* str = ( u8* ) list;
    u8* end = str + count;
    
    while ( str < end && !result ) {
        result = ( c == *str );
        str++;
    }
    
    return result;
}

stu b32 char_is_any_u16( u32 c, u16* list, umm count ) {
    
    b32 result = 0;
    u16* end = list + count;
    
    while ( list < end && !result ) {
        result = ( c == *list );
        list++;
    }
    
    return result;
}

stu b32 char_is_any_u32( u32 c, u32* list, umm count ) {
    
    profiler_event_start( char_is_any_u32 );
    
    b32 result = 0;
    u32* end = list + count;
    
    while ( list < end && !result ) {
        result = ( c == *list );
        list++;
    }
    
    profiler_event_end( char_is_any_u32 );
    
    return result;
}

stu u8 char_decimal_to_value( u32 c ) {
    u8 result = ( u8 ) ( c - '0' );
    return result;
}

stu u8 char_hexadecimal_to_value( u32 c ) {
    u8 result = 0;
    if ( c >= '0' && c <= '9' ) {
        result = ( u8 ) ( c - '0' );
    } else if ( c >= 'a' && c <= 'f' ) {
        result = ( u8 ) ( 10 + c - 'a' );
    } else {
        result = ( u8 ) ( 10 + c - 'A' );
    }
    return result;
}

// MARK string

#define string_zero_terminate( string ) if ( ( string ).used < ( string ).reserved ) { ( string ).bytes[ ( string ).used ] = 0; }
#define string_zero_terminate_on_stack( new_variable_name, size, source, source_length ) \
memory_get_on_stack( new_variable_name, size ); \
_assert( ( source_length ) + 1 < new_variable_name.reserved ); \
memory_push_copy_p( &new_variable_name, source, source_length, 1 ); \
memory_push_u8( &new_variable_name, 0 );

stu data_t string_p( void* text, umm length ) {
    
    data_t s;
    s.raw = text;
    s.used = s.reserved = length;
    
    return s;
}

stu u32 string_length( void* string ) {
    
    u32 result;
    u8* start = ( u8* ) string;
    u8* current = ( u8* ) string;
    
    if ( current ) {
        while ( *current != '\0' ) {
            current++;
        }
    }
    
    result = ( u32 ) ( current - start );
    
    return result;
}

stu data_t string_s( void* string ) {
    
    data_t result;
    result.raw = string;
    result.used = string_length( string );
    result.reserved = result.used + 1;
    
    return result;
}

stu void string_copy( void* destination, void* source, umm length ) {
    
    u8* d = ( u8* ) destination;
    u8* s = ( u8* ) source;
    
    if ( d && s ) {
        while ( length > 0 ) {
            length--;
            *d++ = *s++;
        }
    }
}

#define string_push_length( memory, source, length ) string_push_length_( memory, ( void* ) ( source ), length )

stu u8* string_push_length_( data_t* memory, void* source, umm length ) {
    
    void* destination = memory_push_size( memory, length, 1 );
    string_copy( destination, source, length );
    
    return ( u8* ) destination;
}

stu u8* string_push_s( data_t* memory, void* source ) {
    
    umm length = string_length( source );
    u8* result = string_push_length( memory, source, length );
    
    return result;
}

#define string_seek_ci( in, what ) string_seek_ci_p( ( in ).bytes, memory_end( in ), ( what ).bytes, ( what ).used )
#define string_seek_backward_ci( in, what ) string_seek_backward_ci_p( ( in ).bytes, memory_end( in ), ( what ).bytes, ( what ).used )

stu void* string_seek_ci_p( void* start_param, void* end_param, void* what_param, umm what_size ) {
    
    u8* start = ( u8* ) start_param;
    u8* end = ( u8* ) end_param;
    umm in_size = end - start;
    u8* what = ( u8* ) what_param;
    u8* result = end;
    
    if ( start && end && what_size > 0 && what_size <= in_size ) {
        
        end -= ( what_size - 1 );
        
        while ( start < end ) {
            
            u8* bytes = start;
            umm size = 0;
            
            while ( size < what_size ) {
                
                u8 a = ( u8 ) char_to_lower_case( bytes[ size ] );
                u8 b = ( u8 ) char_to_lower_case( what[ size ] );
                
                if ( a != b ) {
                    break;
                }
                
                size++;
            }
            
            if ( size == what_size ) {
                result = start;
                break;
            } else {
                start++;
            }
        }
    }
    
    return result;
}

stu void* string_seek_backward_ci_p( void* start_param, void* end_param, void* what_param, umm what_size ) {
    
    u8* start = ( u8* ) start_param;
    u8* end = ( u8* ) end_param;
    umm in_size = end - start;
    u8* what = ( u8* ) what_param;
    void* result = end;
    
    if ( start && end && what_size > 0 && what_size <= in_size ) {
        
        end -= what_size;
        
        while ( end >= start ) {
            
            u8* bytes = end;
            umm size = 0;
            
            while ( size < what_size ) {
                
                u8 a = ( u8 ) char_to_lower_case( bytes[ size ] );
                u8 b = ( u8 ) char_to_lower_case( what[ size ] );
                
                if ( a != b ) {
                    break;
                }
                
                size++;
            }
            
            if ( size == what_size ) {
                result = end;
                break;
            } else {
                end--;
            }
        }
    }
    
    return result;
}

stu void* string_seek_white_space_p( void* start_param, void* end_param ) {
    
    u8* start = ( u8* ) start_param;
    u8* end = ( u8* ) end_param;
    u8* result = 0;
    
    while ( start < end ) {
        
        if ( char_is_whitespace( *start ) ) {
            result = start;
            break;
        }
        
        start++;
    }
    
    return result;
}

stu void* string_seek_white_space_backward_p( void* start_param, void* end_param ) {
    
    u8* start = ( u8* ) start_param;
    u8* end = ( u8* ) end_param;
    u8* result = 0;
    
    while ( end > start ) {
        
        end--;
        
        if ( char_is_whitespace( *end ) ) {
            result = end;
            break;
        }
    }
    
    return result;
}

#define string_seek_white_space( in ) string_seek_white_space_p( ( in ).bytes, ( in ).bytes + ( in ).used )
#define string_seek_white_space_backward( in ) string_seek_white_space_backward_p( ( in ).bytes, ( in ).bytes + ( in ).used )

stu void* string_seek_non_white_space_p( void* start_param, void* end_param ) {
    
    u8* start = ( u8* ) start_param;
    u8* end = ( u8* ) end_param;
    u8* result = 0;
    
    while ( start < end ) {
        
        if ( !char_is_whitespace( *start ) ) {
            result = start;
            break;
        }
        
        start++;
    }
    
    return result;
}

stu void* string_seek_non_white_space_backward_p( void* start_param, void* end_param ) {
    
    u8* start = ( u8* ) start_param;
    u8* end = ( u8* ) end_param;
    u8* result = 0;
    
    while ( end > start ) {
        
        end--;
        
        if ( !char_is_whitespace( *end ) ) {
            result = end;
            break;
        }
    }
    
    return result;
}

#define string_seek_non_white_space( in ) string_seek_white_space_p( ( in ).bytes, ( in ).bytes + ( in ).used )
#define string_seek_non_white_space_backward( in ) string_seek_white_space_backward_p( ( in ).bytes, ( in ).bytes + ( in ).used )

stu u8* string_append_s( data_t* memory, data_t* destination, void* string ) {
    
    umm length = string_length( string );
    u8* result = string_push_length( memory, string, length );
    
    if ( !destination->used ) {
        destination->bytes = result;
        destination->reserved = destination->used = 0;
    }
    
    destination->used += length;
    destination->reserved += length;
    
    return result;
}

stu u8* string_append_p( data_t* memory, data_t* destination, void* string, umm length ) {
    
    u8* result = string_push_length( memory, string, length );
    
    if ( !destination->used ) {
        destination->bytes = result;
        destination->reserved = destination->used = 0;
    }
    
    destination->used += length;
    destination->reserved += length;
    
    return result;
}

stu u8* string_append( data_t* memory, data_t* destination, data_t* string ) {
    
    void* result = memory_push_size( memory, string->used, 1 );
    string_copy( result, string->text, string->used );
    
    if ( !destination->used ) {
        destination->raw = result;
        destination->reserved = destination->used = 0;
    }
    
    destination->used += string->used;
    destination->reserved += string->used;
    
    return ( u8* ) result;
}

stu s32 string_compare( void* string_1, void* string_2, umm length ) {
    
    s32 result = 0;
    umm index = 0;
    u8* s1 = ( u8* ) string_1;
    u8* s2 = ( u8* ) string_2;
    
    _assert( s1 != 0 );
    _assert( s2 != 0 );
    
    while ( index < length ) {
        
        result = *s1 - *s2;
        
        if ( *s1 == '\0' || *s2 == '\0' || result != 0 ) {
            break;
        }
        
        s1++;
        s2++;
        index++;
    }
    
    return result;
}

stu b32 string_are_equal_p( void* a, umm a_length, void* b, umm b_length ) {
    b32 result = memory_are_equal_p( a, a_length, b, b_length );
    return result;
}

stu b32 string_are_equal_s( void* a, void* b ) {
    
    b32 result = false;
    umm length_a = string_length( a );
    umm length_b = string_length( b );
    
    result = memory_are_equal_p( a, length_a, b, length_b );
    
    return result;
}

stu b32 string_are_equal_ci( data_t* string_a, data_t* string_b ) {
    
    b32 result = string_a->used == string_b->used;
    
    if ( result ) {
        
        u8 offset = 'a' - 'A';
        u8* ap = string_a->bytes;
        u8* bp = string_b->bytes;
        _assert( ap != 0 );
        _assert( bp != 0 );
        
        for ( umm i = 0; i < string_a->used; i++ ) {
            
            u8 a = *ap;
            u8 b = *bp;
            if ( a >= 'A' && a <= 'Z' ) {
                a = a + offset;
            }
            
            if ( b >= 'A' && b <= 'Z' ) {
                b = b + offset;
            }
            
            if ( a != b ) {
                result = false;
                break;
            }
            ap++;
            bp++;
        }
    }
    
    return result;
}

stu b32 string_start_with_ci_p( void* start_arg, void* end_arg, void* what_arg, umm what_length ) {
    
    _assert( start_arg );
    _assert( what_arg );
    
    u8* current = ( u8* ) start_arg;
    u8* end = current + what_length;
    
    b32 result = end <= ( u8* ) end_arg;
    u8* what = ( u8* ) what_arg;
    u8 offset = 'a' - 'A';
    
    while ( result && current < end ) {
        
        u8 in_c = *current;
        u8 what_c = *what;
        
        if ( in_c >= 'A' && in_c <= 'Z' ) {
            in_c += offset;
        }
        
        if ( what_c >= 'A' && what_c <= 'Z' ) {
            what_c += offset;
        }
        
        result = ( in_c == what_c );
        current++;
        what++;
    }
    
    return result;
}

stu b32 string_start_with_ci( data_t in, data_t what ) {
    b32 result = string_start_with_ci_p( in.bytes, memory_end( in ), what.bytes, what.used );
    return result;
}

stu void string_to_lower_case( data_t* string ) {
    
    for ( umm i = 0; i < string->used; i++ ) {
        if ( char_is_upper_case( string->bytes[ i ] ) ) {
            string->bytes[ i ] = ( u8 ) char_to_lower_case( string->bytes[ i ] );
        }
    }
}

stu void string_to_upper_case( data_t* string ) {
    
    for ( umm i = 0; i < string->used; i++ ) {
        if ( char_is_lower_case( string->bytes[ i ] ) ) {
            string->bytes[ i ] = ( u8 ) char_to_upper_case( string->bytes[ i ] );
        }
    }
}

stu s64 string_to_s64( u8* string, umm string_size ) {
    
    s64 result = 0;
    s32 sign = 1;
    umm at = 0;
    
    while ( at < string_size
           && ( string[ at ] == ' '
               || string[ at ] == '\t'
               || string[ at ] == '\v'
               || string[ at ] == '\f'
               || string[ at ] == '\n'
               || string[ at ] == '\r' ) ) {
        at++;
    }
    
    if ( at < string_size ) {
        
        b32 base = 10;
        
        if ( string[ at ] == '-' ) {
            sign = -1;
            at++;
        } else if ( string[ at ] == '+' ) {
            sign = 1;
            at++;
        }
        
        if ( at + 2 < string_size && string[ at ] == '0' && string[ at + 1 ] == 'x' ) {
            at += 2;
            base = 16;
        }
        
        while ( at < string_size ) {
            
            if ( string[ at ] >= '0' && string[ at ] <= '9' ) {
                result = result * base + ( string[ at ] - '0' );
            } else if ( base == 16 && string[ at ] >= 'a' && string[ at ] <= 'f' ) {
                result = result * base + 10 + ( string[ at ] - 'a' );
            } else if ( base == 16 && string[ at ] >= 'A' && string[ at ] <= 'F' ) {
                result = result * base + 10 + ( string[ at ] - 'A' );
            } else {
                break;
            }
            
            at++;
        }
    }
    
    result *= sign;
    
    return result;
}

stu u64 string_to_u64( void* numberString, umm string_size ) {
    
    u8* string = ( u8* ) numberString;
    u64 result = 0;
    umm at = 0;
    b32 complement = false;
    
    while ( at < string_size
           && ( string[ at ] == ' '
               || string[ at ] == '\t'
               || string[ at ] == '\v'
               || string[ at ] == '\f'
               || string[ at ] == '\n'
               || string[ at ] == '\r'
               || string[ at ] == '+' ) ) {
        at++;
    }
    
    if ( at < string_size ) {
        
        b32 base = 10;
        
        if ( string[ at ] == '-' ) {
            complement = true;
            at++;
        }
        
        if ( at + 2 < string_size && string[ at ] == '0' && string[ at + 1 ] == 'x' ) {
            at += 2;
            base = 16;
        }
        
        while ( at < string_size ) {
            
            if ( string[ at ] >= '0' && string[ at ] <= '9' ) {
                result = result * base + ( string[ at ] - '0' );
            } else if ( base == 16 && string[ at ] >= 'a' && string[ at ] <= 'f' ) {
                result = result * base + 10 + ( string[ at ] - 'a' );
            } else if ( base == 16 && string[ at ] >= 'A' && string[ at ] <= 'F' ) {
                result = result * base + 10 + ( string[ at ] - 'A' );
            } else {
                break;
            }
            
            at++;
        }
    }
    
    if ( complement ) {
        result = 0 - result;
    }
    
    return result;
}

typedef enum string_format_flag_t {
    
    string_format_flag_none = 0,
    string_format_flag_hexadecimal = 1 << 0,
    string_format_flag_hexadecimal_no_prefix = 1 << 1,
    string_format_flag_hexadecimal_prefix_befor_filling = 1 << 2,
    string_format_flag_hexadecimal_space_after_8_bits = 1 << 3,
    string_format_flag_hexadecimal_space_after_16_bits = 1 << 4,
    string_format_flag_hexadecimal_space_after_32_bits = 1 << 5,
    string_format_flag_hexadecimal_space_after_64_bits = 1 << 6,
    string_format_flag_hexadecimal_force_8_bits = 1 << 7,
    string_format_flag_hexadecimal_force_16_bits = 1 << 8,
    string_format_flag_hexadecimal_force_32_bits = 1 << 9,
    string_format_flag_hexadecimal_force_64_bits = 1 << 10,
    string_format_flag_left_align = 1 << 11,
    string_format_flag_sign_force = 1 << 12,
    string_format_flag_sign_before_filling = 1 << 13,
    string_format_flag_decimal_space_after_3_digits = 1 << 14,
    
} string_format_flag_t;

stu umm string_from_s64( void* output_buffer, umm buffer_size, s64 integer, umm minimum_character_count, u8 fill_with, u32 format ) {
    
    u8* buffer = ( u8* ) output_buffer;
    umm c = 0;
    b32 s64_min_fix = false;
    
    b32 hexadecimal = format & string_format_flag_hexadecimal;
    b32 hexadecimal_prefix = hexadecimal && !( format & string_format_flag_hexadecimal_no_prefix );
    b32 hexadecimal_force_8 = format & string_format_flag_hexadecimal_force_8_bits;
    b32 hexadecimal_force_16 = format & string_format_flag_hexadecimal_force_16_bits;
    b32 hexadecimal_force_32 = format & string_format_flag_hexadecimal_force_32_bits;
    b32 hexadecimal_force_64 = format & string_format_flag_hexadecimal_force_64_bits;
    b32 hexadecimal_space_8 = format & string_format_flag_hexadecimal_space_after_8_bits;
    b32 hexadecimal_space_16 = format & string_format_flag_hexadecimal_space_after_16_bits;
    b32 hexadecimal_space_32 = format & string_format_flag_hexadecimal_space_after_32_bits;
    b32 hexadecimal_space_64 = format & string_format_flag_hexadecimal_space_after_64_bits;
    b32 decimal_3 = format & string_format_flag_decimal_space_after_3_digits;
    
    u8 base = ( hexadecimal ) ? 16 : 10;
    b32 is_negative = false;
    
    if ( c < buffer_size && integer < 0 ) {
        
        is_negative = true;
        
        if ( integer == s64_min ) {
            integer = integer + 1;
            s64_min_fix = true;
        }
        
        integer = integer * -1;
    }
    
    s64 number = integer;
    u32 digits = 1;
    
    while ( number > ( u8 ) ( base - 1 ) ) {
        number /= base;
        digits++;
    }
    
    umm extra = ( is_negative || ( format & string_format_flag_sign_force ) ) ? 1 : 0;
    extra += ( hexadecimal_prefix ) ? 2 : 0;
    
    umm force = digits;
    
    if ( hexadecimal ) {
        
        if ( hexadecimal_force_8 ) {
            force = 2;
        } else if ( hexadecimal_force_16 ) {
            force = 4;
        } else if ( hexadecimal_force_32 ) {
            force = 8;
        } else if ( hexadecimal_force_64 ) {
            force = 16;
        }
        
        if ( force < digits ) {
            force = digits;
        }
        
        if ( hexadecimal_space_8 ) {
            extra += ( force - 1 ) / 2;
        } else if ( hexadecimal_space_16 ) {
            extra += ( force - 1 ) / 4;
        } else if ( hexadecimal_space_32 ) {
            extra += ( force - 1 ) / 8;
        } else if ( hexadecimal_space_64 ) {
            
        }
        
    } else if ( decimal_3 ) {
        extra += ( digits - 1 ) / 3;
    }
    
    umm filling = 0;
    
    if ( minimum_character_count > force + extra ) {
        filling = minimum_character_count - ( force + extra );
    }
    
    if ( format & string_format_flag_sign_before_filling ) {
        if ( is_negative ) {
            buffer[ c++ ] = '-';
        } else if ( format & string_format_flag_sign_force ) {
            buffer[ c++ ] = '+';
        }
    }
    
    if ( !( format & string_format_flag_left_align ) ) {
        for ( umm i = 0; i < filling && c < buffer_size; i++ ) {
            buffer[ c++ ] = fill_with;
        }
    }
    
    if ( !( format & string_format_flag_sign_before_filling ) ) {
        if ( is_negative ) {
            buffer[ c++ ] = '-';
        } else if ( format & string_format_flag_sign_force ) {
            buffer[ c++ ] = '+';
        }
    }
    
    if ( hexadecimal ) {
        
        if ( hexadecimal_prefix ) {
            
            if ( c < buffer_size ) {
                buffer[ c++ ] = '0';
            }
            
            if ( c < buffer_size ) {
                buffer[ c++ ] = 'x';
            }
        }
        
        if ( force > digits ) {
            
            while ( force > digits ) {
                
                if ( c < buffer_size ) {
                    buffer[ c++ ] = '0';
                }
                
                force--;
                
                if ( ( hexadecimal_space_8 && ( ( force % 2 ) == 0 ) )
                    || ( hexadecimal_space_16 && ( ( force % 4 ) == 0 ) )
                    || ( hexadecimal_space_32 && ( ( force % 8 ) == 0 ) )
                    || ( decimal_3 && ( ( force % 3 ) == 0 ) ) ) {
                    buffer[ c++ ] = ' ';
                }
            }
        }
    }
    
    u32 digit_count = digits;
    
    while ( digits > 0 ) {
        
        if ( buffer_size - c ) {
            
            if ( digits != digit_count && ( ( hexadecimal_space_8 && ( ( digits % 2 ) == 0 ) )
                                           || ( hexadecimal_space_16 && ( ( digits % 4 ) == 0 ) )
                                           || ( hexadecimal_space_32 && ( ( digits % 8 ) == 0 ) )
                                           || ( decimal_3 && ( ( digits % 3 ) == 0 ) ) ) ) {
                buffer[ c++ ] = ' ';
            }
        }
        
        s64 divider = 1;
        
        for ( u32 i = 0; i < ( digits - 1 ); i++ ) {
            divider *= base;
        }
        
        s64 digit = integer / divider;
        
        if ( digits <= ( buffer_size - c ) ) {
            
            if ( s64_min_fix ) {
                
                if ( hexadecimal ) {
                    if ( digits == digit_count ) {
                        digit += 1;
                    } else {
                        digit = 0;
                    }
                } else if( digits == 1 ) {
                    digit += 1;
                }
            }
            
            if ( digit < 10 ) {
                buffer[ c++ ] = ( char ) ( '0' + digit );
            } else {
                buffer[ c++ ] = ( char ) ( 'a' + ( digit - 10 ) );
            }
        }
        
        integer -= divider * digit;
        digits--;
    }
    
    if ( format & string_format_flag_left_align ) {
        for ( umm i = 0; i < filling && c < buffer_size; i++ ) {
            buffer[ c++ ] = fill_with;
        }
    }
    
    return c;
}

stu umm string_from_u64( void* output_buffer, umm buffer_size, u64 integer, umm minimum_character_count, u8 fill_with, u32 format ) {
    
    u8* buffer = ( u8* ) output_buffer;
    umm c = 0;
    
    b32 hexadecimal = format & string_format_flag_hexadecimal;
    b32 hexadecimal_prefix = hexadecimal && !( format & string_format_flag_hexadecimal_no_prefix );
    b32 hexadecimal_force_8 = format & string_format_flag_hexadecimal_force_8_bits;
    b32 hexadecimal_force_16 = format & string_format_flag_hexadecimal_force_16_bits;
    b32 hexadecimal_force_32 = format & string_format_flag_hexadecimal_force_32_bits;
    b32 hexadecimal_force_64 = format & string_format_flag_hexadecimal_force_64_bits;
    b32 hexadecimal_space_8 = format & string_format_flag_hexadecimal_space_after_8_bits;
    b32 hexadecimal_space_16 = format & string_format_flag_hexadecimal_space_after_16_bits;
    b32 hexadecimal_space_32 = format & string_format_flag_hexadecimal_space_after_32_bits;
    b32 hexadecimal_space_64 = format & string_format_flag_hexadecimal_space_after_64_bits;
    b32 decimal_3 = format & string_format_flag_decimal_space_after_3_digits;
    b32 force_sign = format & string_format_flag_sign_force;
    
    u8 base = ( hexadecimal ) ? 16 : 10;
    
    u64 number = integer;
    u32 digits = 1;
    
    while ( number > ( u8 ) ( base - 1 ) ) {
        number /= base;
        digits++;
    }
    
    umm extra = ( hexadecimal_prefix ) ? 2 : 0;
    extra += ( force_sign ) ? 1 : 0;
    
    umm force = digits;
    
    if ( hexadecimal ) {
        
        if ( hexadecimal_force_8 ) {
            force = 2;
        } else if ( hexadecimal_force_16 ) {
            force = 4;
        } else if ( hexadecimal_force_32 ) {
            force = 8;
        } else if ( hexadecimal_force_64 ) {
            force = 16;
        }
        
        if ( force < digits ) {
            force = digits;
        }
        
        if ( hexadecimal_space_8 ) {
            extra += ( force - 1 ) / 2;
        } else if ( hexadecimal_space_16 ) {
            extra += ( force - 1 ) / 4;
        } else if ( hexadecimal_space_32 ) {
            extra += ( force - 1 ) / 8;
        } else if ( hexadecimal_space_64 ) {
            
        }
        
    } else if ( decimal_3 ) {
        extra += ( digits - 1 ) / 3;
    }
    
    umm filling = 0;
    
    if ( minimum_character_count > force + extra ) {
        filling = minimum_character_count - ( force + extra );
    }
    
    if ( force_sign && ( format & string_format_flag_sign_before_filling ) ) {
        buffer[ c++ ] = '+';
    }
    
    if ( !( format & string_format_flag_left_align ) ) {
        for ( umm i = 0; i < filling && c < buffer_size; i++ ) {
            buffer[ c++ ] = fill_with;
        }
    }
    
    if ( force_sign && !( format & string_format_flag_sign_before_filling ) ) {
        buffer[ c++ ] = '+';
    }
    
    if ( hexadecimal ) {
        
        if ( hexadecimal_prefix ) {
            
            if ( c < buffer_size ) {
                buffer[ c++ ] = '0';
            }
            
            if ( c < buffer_size ) {
                buffer[ c++ ] = 'x';
            }
        }
        
        if ( force > digits ) {
            
            while ( force > digits ) {
                
                if ( c < buffer_size ) {
                    buffer[ c++ ] = '0';
                }
                
                force--;
                
                if ( ( hexadecimal_space_8 && ( ( force % 2 ) == 0 ) )
                    || ( hexadecimal_space_16 && ( ( force % 4 ) == 0 ) )
                    || ( hexadecimal_space_32 && ( ( force % 8 ) == 0 ) )
                    || ( decimal_3 && ( ( force % 3 ) == 0 ) ) ) {
                    buffer[ c++ ] = ' ';
                }
            }
        }
    }
    
    u32 digit_count = digits;
    
    while ( digits > 0 ) {
        
        if ( buffer_size - c ) {
            
            if ( digits != digit_count && ( ( hexadecimal_space_8 && ( ( digits % 2 ) == 0 ) )
                                           || ( hexadecimal_space_16 && ( ( digits % 4 ) == 0 ) )
                                           || ( hexadecimal_space_32 && ( ( digits % 8 ) == 0 ) )
                                           || ( decimal_3 && ( ( digits % 3 ) == 0 ) ) ) ) {
                buffer[ c++ ] = ' ';
            }
        }
        
        u64 divider = 1;
        
        for ( u32 i = 0; i < ( digits - 1 ); i++ ) {
            divider *= base;
        }
        
        u64 digit = integer / divider;
        
        if ( digits <= ( buffer_size - c ) ) {
            
            if ( digit < 10 ) {
                buffer[ c++ ] = ( char ) ( '0' + digit );
            } else {
                buffer[ c++ ] = ( char ) ( 'a' + ( digit - 10 ) );
            }
        }
        
        integer -= divider * digit;
        digits--;
    }
    
    if ( format & string_format_flag_left_align ) {
        for ( umm i = 0; i < filling && c < buffer_size; i++ ) {
            buffer[ c++ ] = fill_with;
        }
    }
    
    return c;
}

stu r64 string_to_r64( void* string_arg, umm string_size ) {
    
    u8* string = ( u8* ) string_arg;
    r64 result = 0;
    r32 sign = 1;
    u32 at = 0;
    
    while ( at < string_size
           && ( string[ at ] == ' '
               || string[ at ] == '\t'
               || string[ at ] == '\v'
               || string[ at ] == '\f'
               || string[ at ] == '\n'
               || string[ at ] == '\r' ) ) {
        at++;
    }
    
    if ( at < string_size ) {
        
        r32 decimal_divisor;
        b32 decimal;
        
        if ( string[ at ] == '-' ) {
            sign = -1;
            at++;
        } else if ( string[ at ] == '+' ) {
            sign = 1;
            at++;
        }
        
        decimal_divisor = 10;
        decimal = false;
        
        while ( at < string_size ) {
            
            if ( string[ at ] == '.' ) {
                
                decimal = true;
                
            } else if ( string[ at ] >= '0' && string[ at ] <= '9' ) {
                
                if ( !decimal ) {
                    result = result * 10 + ( string[ at ] - '0' );
                } else {
                    result = result + ( string[ at ] - '0' ) / decimal_divisor;
                    decimal_divisor *= 10;
                }
                
            } else {
                
                break;
            }
            
            at++;
        }
    }
    
    result *= sign;
    
    return result;
}

stu umm string_from_r64( void* output_buffer, umm buffer_size, r64 real, u32 not_decimal_character_count, u8 fill_with, u32 decimal_count, u32 format ) {
    
    u8* buffer = ( u8* ) output_buffer;
    umm result = 0;
    
    s64 integer_part = ( s64 ) real;
    umm integer_part_count = 0;
    
    if ( format & string_format_flag_left_align ) {
        integer_part_count = string_from_s64( buffer, buffer_size, integer_part, 0, 0, format & ( ~string_format_flag_left_align ) );
    } else {
        integer_part_count = string_from_s64( buffer, buffer_size, integer_part, not_decimal_character_count, fill_with, format );
    }
    
    result += integer_part_count;
    
    if ( real < 0.0f ) {
        real *= -1.0f;
    }
    
    if ( decimal_count && result < buffer_size ) {
        buffer[ result++ ] = '.';
    }
    
    r64 decimal = real - ( u64 ) real;
    
    while ( decimal_count > 0 ) {
        
        decimal *= 10;
        
        if ( decimal < 1.0f && result < buffer_size ) {
            *( buffer + result ) = '0';
            result++;
        }
        
        decimal_count--;
    }
    
    u64 decimal_part = ( u64 ) decimal;
    
    if ( decimal_part > 0 ) {
        result += string_from_u64( buffer + result, buffer_size - result, decimal_part, 0, 0, 0 );
    }
    
    if ( format & string_format_flag_left_align ) {
        
        smm left = not_decimal_character_count - integer_part_count;
        
        while ( result < buffer_size && left > 0 ) {
            buffer[ result++ ] = fill_with;
            left--;
        }
    }
    
    return result;
}

stu umm string_push_r64( data_t* buffer, r64 real, u32 not_decimal_character_count, u8 fill_with, u32 decimal_count, u32 format ) {
    
    umm length = string_from_r64( buffer->bytes + buffer->used, buffer->reserved - buffer->used, real, not_decimal_character_count, fill_with, decimal_count, format );
    buffer->used += length;
    
    return length;
}

stu umm string_push_u64( data_t* memory, u64 n, umm desired_digit_count, u8 fill_with, u32 hexadecimal_format ) {
    
    umm size = string_from_u64( memory_current( *memory, u8 ), memory_left( *memory ), n, desired_digit_count, fill_with, hexadecimal_format );
    memory->used += size;
    
    return size;
}

stu umm string_push_s64( data_t* memory, s64 n, umm desired_digit_count, u8 fill_with, u32 hexadecimal_format ) {
    
    umm size = string_from_s64( memory_current( *memory, u8 ), memory_left( *memory ), n, desired_digit_count, fill_with, hexadecimal_format );
    memory->used += size;
    
    return size;
}

umm string_printable_size_from_u64_p( umm number, u8* buffer, umm buffer_size ) {
    
    data_t output = { buffer_size, 0, buffer };
    
    u64 t = tebioctet( 1 );
    u64 g = gibioctet( 1 );
    umm m = mebioctet( 1 );
    umm k = kibioctet( 1 );
    
    data_t Tio = string_l( "Tio" );
    data_t Gio = string_l( "Gio" );
    data_t Mio = string_l( "Mio" );
    data_t Kio = string_l( "Kio" );
    data_t o = string_l( "o" );
    
    data_t* unit = &o;
    
    if ( number < k ) {
        output.used += string_from_u64( output.bytes, output.reserved, number, 0, 0, 0 );
    } else {
        
        r64 n = ( r64 ) number;
        
        if ( number >= t ) {
            n /= t;
            unit = &Tio;
        } else if ( number >= g ) {
            n /= g;
            unit = &Gio;
        } else if ( number >= m ) {
            n /= m;
            unit = &Mio;
        } else {
            n /= k;
            unit = &Kio;
        }
        
        output.used += string_from_r64( output.bytes, output.reserved, n, 0, 0, 3, 0 );
    }
    
    memory_push_copy( &output, string_arg_l( " " ), 1 );
    memory_push_copy( &output, *unit, 1 );
    
    return output.used;
}

data_t string_printable_size_from_u64( umm number, data_t* buffer ) {
    data_t result = memory_get_block( buffer, 16, 1 );
    result.used += string_printable_size_from_u64_p( number, result.bytes, result.reserved );
    memory_pop_size( buffer, result.reserved - result.used );
    result.reserved = result.used;
    return result;
}

#define string_format_3( memory, counter, ... ) { string_format_t internal_formats_ ## counter [ ] = { __VA_ARGS__ }; string_format_execute( ( memory ), internal_formats_ ## counter, array_count( internal_formats_ ## counter ) ); }
#define string_format_2( memory, counter, ... ) string_format_3( memory, counter, __VA_ARGS__ )
#define string_format( memory, ... ) string_format_2( memory, __COUNTER__, __VA_ARGS__ )

typedef enum string_format_type_t {
    
    string_format_type_text,
    
    string_format_type_unsigned,
    string_format_type_signed,
    
    string_format_type_floating_point,
    
} string_format_type_t;

typedef struct string_format_t {
    
    union {
        void* textp;
        
        u64 u;
        s64 s;
        
        r64 r;
    };
    
    union {
        umm text_length;
        struct {
            u8 desired_digit_count;
            u8 fill_with;
            u8 hexadecimal_format;
        };
        u8 decimal_count;
    };
    
    u8 type;
} string_format_t;

#define textfl( textp_arg ) { .textp = ( textp_arg ), sizeof( textp_arg ) - 1, string_format_type_text }
#define textfp( textp_arg, length ) { .textp = ( textp_arg ), length, string_format_type_text }
#define textfd( textd ) { .textp = ( textd ).bytes, ( textd ).used, string_format_type_text }
#define textfs( textp_arg ) { .textp = ( textp_arg ), 0, string_format_type_text }

#define uf( u_arg ) { .u = u_arg, 0, string_format_type_unsigned }
#define sf( s_arg ) { .s = s_arg, 0, string_format_type_signed }
#define rf( r_arg ) { .r = r_arg, 6, string_format_type_floating_point }

#define ufe( u_arg, desired_digit_count, fill_with, hexadecimal_format ) { .u = u_arg, .desired_digit_count = desired_digit_count, .fill_width = fill_with, .hexadecimal_format = hexadecimal_format, string_format_type_unsigned }
#define sfe( s_arg, desired_digit_count, fill_with, hexadecimal_format ) { .s = s_arg, .desired_digit_count = desired_digit_count, .fill_with = fill_with, .hexadecimal_format = hexadecimal_format, string_format_type_signed }
#define rfe( r_arg, decimal_count ) { .r = r_arg, decimal_count, string_format_type_floating_point }

stu data_t string_format_execute( data_t* memory, string_format_t* formats, umm count ) {
    
    data_t result = memory_construct( 0, 0, memory_current( *memory, u8 ) );
    
    for ( umm i = 0; i < count; i++ ) {
        
        string_format_t* format = formats + i;
        
        switch ( format->type ) {
            
            case string_format_type_text: {
                if ( format->text_length == 0 && format->textp ) {
                    format->text_length = string_length( format->textp );
                }
                
                memory_push_copy_p( memory, format->textp, format->text_length, 1 );
            } break;
            
            case string_format_type_unsigned: {
                memory->used += string_from_u64( memory_current( *memory, u8 ), memory_left( *memory ), format->u, 0, 0, format->hexadecimal_format );
            } break;
            
            case string_format_type_signed: {
                memory->used += string_from_s64( memory_current( *memory, u8 ), memory_left( *memory ), format->s, 0, 0, format->hexadecimal_format );
            } break;
            
            case string_format_type_floating_point: {
                memory->used += string_from_r64( memory_current( *memory, u8 ), memory_left( *memory ), format->r, 0, 0, format->decimal_count, 0 );
            } break;
        }
    }
    
    result.reserved = result.used = memory_current( *memory, u8 ) - result.bytes;
    
    return result;
}

stu data_t string_shrink_from_end( data_t string, data_t byte_list ) {
    
    u8* start = string.bytes;
    u8* end = start + string.used - 1;
    
    while ( end >= start ) {
        
        if ( char_is_any_u8( *end, byte_list.bytes, byte_list.used ) ) {
            string.used--;
            end--;
        } else {
            break;
        }
    }
    
    return string;
}

// MARK math

/* TODO simon: investigate
unlerp(a, x, b) = lerp(-a/(b-a), x, (1-a)/(b-a))
linear_remap(x, a,b,c,d) = lerp(c, unlerp(a,x,b), d) // == (x-a)/(b-a)*(d-c)+c
*/

#include <math.h>

r32 math_pi = 3.14159265358979323846f;
r32 math_tau = 6.28318530717958647692f;
r32 math_degree_to_radian = 0.01745329251994329576f; /* 0.01745329251994329576923690768489f = ( math_pi / 180.0f ) */
r32 math_radian_to_degree = 57.29577951308232087679f; /* 57.295779513082320876798154814105f = ( 180.0f / math_pi ) */;

#define math_min( a, b ) ( ( ( a ) < ( b ) ) ? ( a ) : ( b ) )
#define math_max( a, b ) ( ( ( a ) > ( b ) ) ? ( a ) : ( b ) )

#define math_clamp( min, value, max ) ( ( ( value ) < ( min ) ) ? ( min ) : ( ( value ) > ( max ) ) ? ( max ) : ( value ) )

stu umm math_rotate_right_umm( umm value, umm bit_count ) {
    umm result = ( value >> bit_count ) | ( value << ( sizeof( umm ) * 8 - bit_count ) );
    return result;
}

stu umm math_rotate_left_umm( umm value, umm bit_count ) {
    umm result = ( value << bit_count ) | ( value >> ( sizeof( umm ) * 8 - bit_count ) );
    return result;
}

stu u32 math_safe_truncate_to_u32( u64 value ) {
    
    u32 result;
    _assert( value <= u32_max );
    result = ( u32 ) value;
    
    return result;
}

stu s32 math_safe_truncate_to_s32( s64 value ) {
    
    u32 result;
    _assert( value >= s32_min && value <= s32_max);
    result = ( s32 ) value;
    
    return result;
}

stu u16 math_safe_truncate_to_u16( u64 value ) {
    
    u16 result;
    _assert( value <= u16_max );
    result = ( u16 ) value;
    
    return result;
}

stu u16 math_safe_truncate_to_s16( s64 value ) {
    
    u16 result;
    _assert( value >= s16_min && value <= s16_max );
    result = ( u16 ) value;
    
    return result;
}

u64 math_add_no_wrap_u64( u64 initial_value, u64 to_add ) {
    
    u64 after = u64_max - initial_value;
    u64 result = initial_value + math_min( after, to_add );
    return result;
}

u64 math_sub_no_wrap_u64( u64 initial_value, u64 to_sub ) {
    
    u64 before = initial_value;
    u64 result = initial_value - math_min( before, to_sub );
    return result;
}

stu r32 math_abs( r32 number ) {
    
    if ( number < 0 ) {
        number *= -1;
    }
    
    return number;
}

// NOTE simon: fast inverse square root (quake 3)
// https://en.wikipedia.org/wiki/Fast_inverse_square_root
// https://en.wikipedia.org/wiki/Methods_of_computing_square_roots#Approximations_that_depend_on_the_floating_point_representation
stu r32 math_rsqrt( r32 number ) {
    
    s32 i;
    r32 x2, y;
    const r32 threehalfs = 1.5f;
    
    x2 = number * 0.5f;
    y  = number;
    i  = * ( s32 * ) &y;                       // evil floating point bit level hacking
    i  = 0x5f3759df - ( i >> 1 );               // what the fuck?
    y  = * ( r32 * ) &i;
    y  = y * ( threehalfs - ( x2 * y * y ) );   // 1st iteration
    //	y  = y * ( threehalfs - ( x2 * y * y ) );   // 2nd iteration, this can be removed
    
    return y;
}

stu r32 math_sign( r32 number ) {
    
    r32 sign;
    
    if ( number < 0 ) {
        sign = -1;
    } else {
        sign = 1;
    }
    
    return sign;
}

stu r32 math_sign_or_zero( r32 number ) {
    
    r32 sign = 0;
    
    if ( number > 0.0f ) {
        sign = 1.0f;
    } else if ( number < 0.0f ) {
        sign = -1.0f;
    }
    
    return sign;
}

/* NOTE simon: This is not properly tested. */
stu s32 math_s24_to_s32( u8* s24 ) {
    
    s32 result;
    
    s32 rawValue = *( ( s32* ) s24 );
    s32 sign = ( rawValue & 0x00800000 ) >> 23;
    result = rawValue & 0x007fffff;
    result = result | ( 0xff800000 * sign );
    
    return result;
}

/* NOTE simon: This is not properly tested. */
stu s32 math_s32_to_s24( s32 s ) {
    
    s32 result;
    
    s32 sign = ( s & 0x80000000 ) >> 8;
    result = s & 0x00ffffff;
    result |= sign;
    
    return result;
}

r32 math_bezier_quadratic( r32 start, r32 end, r32 control, r32 x ) {
    
    r32 y = 1 - x;
    
#if 0
    r32 p1 = start * y + control * x;
    r32 p2 = control * y + end * x;
    r32 result = p1 * y + p2 * x;
#else
    /* 
    ( ( start * y ) + ( control * x ) ) * y + ( ( control * y ) + ( end * x ) ) * x;
    ( y * ( start * y ) + y * ( control * x ) ) + ( x * ( control * y ) + x ( end * x ) )
    start*y*y + y*x*control + x*y*control + x*x*end
    */
    r32 result = start * y * y + 2 * y * x * control + x * x * end;
#endif
    
    return result;
}

r32 math_bezier_cubic( r32 start, r32 end, r32 control_1, r32 control_2, r32 x ) {
    
    r32 y = 1 - x;
    
#if 0
    r32 p1 = start * y + control_1 * x;
    r32 p2 = control_1 * y + control_2 * x;
    r32 p3 = control_2 * y + end * x;
    r32 p4 = p1 * y + p2 * x;
    r32 p5 = p2 * y + p3 * x;
    r32 result = p4 * y + p5 * x;
#else
    
    /*
    ( p1 * y + p2 * x ) * y + ( p2 * y + p3 * x ) * x

      ( ( start * y + control_1 * x ) * y + ( control_1 * y + control_2 * x ) * x ) * y
    + ( ( control_1 * y + control_2 * x ) * y + ( control_2 * y + end * x ) * x ) * x

      ( start * y * y + control_1 * x * y + control_1 * y * x + control_2 * x * x ) * y
    + ( control_1 * y * y + control_2 * x * y + control_2 * y * x + end * x * x ) * x

      start * y * y * y + control_1 * x * y * y + control_1 * y * x * y + control_2 * x * x * y
    + control_1 * y * y * x + control_2 * x * y * x + control_2 * y * x * x + end * x * x * x 

    start * yyy + control_1 * xyy + control_1 * xyy + control_2 * xxy + control_1 * yyx + control_2 * xxy + control_2 * xxy + end * xxx

    start * yyy + control_1 * xyy * 2 + control_2 * xxy * 2 + control_1 * xyy + control_2 * xxy + end * xxx
    */
    
    r32 yyy = y * y * y;
    r32 xxx = x * x * x;
    r32 xyy = x * y * y;
    r32 xxy = x * x * y;
    r32 result = start * yyy + control_1 * xyy * 3 + control_2 * xxy * 3 + end * xxx; 
#endif
    
    return result;
}

#if defined( WINDOWS )

#if defined( COMPILER_CLANG ) || defined( COMPILER_GCC )
#include <xmmintrin.h>
#include <emmintrin.h>
#include <smmintrin.h>
#endif

#if defined( COMPILER_CL )
#define m128_f32( vector, index ) ( vector.m128_f32[ ( index ) ] )
#define m128_f64( vector, index ) ( vector.m128d_f64[ ( index ) ] )
#elif defined( COMPILER_CLANG ) || defined( COMPILER_GCC )
#define m128_f32( vector, index ) ( vector[ ( index ) ] )
#define m128_f64( vector, index ) ( vector[ ( index ) ] )
#else
# error Unsupported platform.
#endif


double math_fabs( double x ) {
    union { double f; uint64_t i; } u = { x };
    u.i &= 0x7fffffffffffffff;
    return u.f;
}

float math_fabsf( float x ) {
    union { float f; uint32_t i; } u = { x };
    u.i &= 0x7fffffff;
    return u.f;
}

double math_round( double x ) {
    
    __m128d a = _mm_set_pd1( x );
    __m128d result = _mm_round_pd( a, _MM_FROUND_CUR_DIRECTION );
    
    return m128_f64( result, 0 );
}

float math_roundf( float x ) {
    
    __m128 a = _mm_set_ps1( x );
    __m128 result = _mm_round_ps( a, _MM_FROUND_CUR_DIRECTION );
    
    return m128_f32( result, 0 );
}

double math_ceil( double x ) {
    
    __m128d a = _mm_set_pd1( x );
    __m128d result = _mm_ceil_pd( a );
    
    return m128_f64( result, 0 );
}

float math_ceilf( float x ) {
    
    __m128 a = _mm_set_ps1( x );
    __m128 result = _mm_ceil_ps( a );
    
    return m128_f32( result, 0 );
}

double math_floor( double x ) {
    
    __m128d a = _mm_set_pd1( x );
    __m128d result = _mm_floor_pd( a );
    
    return m128_f64( result, 0 );
}

float math_floorf( float x ) {
    
    __m128 a = _mm_set_ps1( x );
    __m128 result = _mm_floor_ps( a );
    
    return m128_f32( result, 0 );
}

double math_sqrt( double x ) {
    
    __m128d a = _mm_set_pd1( x );
    __m128d result = _mm_sqrt_pd( a );
    
    return m128_f64( result, 0 );
}

float math_sqrtf( float x ) {
    
    __m128 a = _mm_set_ps1( x );
    __m128 result = _mm_sqrt_ps( a );
    
    return m128_f32( result, 0 );
}
#endif

// MARK sort

void internal_quick_sort_u8_ascending( u8* array, umm start, umm end ) {
    
    if ( array && start < end ) {
        
        umm pivot_index = ( start + end ) / 2;
        u8 pivot_value = array[ pivot_index ];
        umm left = start;
        umm right = end;
        
        while ( 1 ) {
            
            while ( array[ left ] < pivot_value ) {
                left++;
            }
            
            while ( array[ right ] > pivot_value ) {
                right--;
            }
            
            if ( left < right ) {
                swap_type( array[ left ], array[ right ], u8 );
                left++;
                right--;
            } else {
                break;
            }
        }
        
        internal_quick_sort_u8_ascending( array, start, right );
        internal_quick_sort_u8_ascending( array, right + 1, end );
    }
}

void internal_quick_sort_u8_descending( u8* array, umm start, umm end ) {
    
    if ( array && start < end ) {
        
        umm pivot_index = ( start + end ) / 2;
        u8 pivot_value = array[ pivot_index ];
        umm left = start;
        umm right = end;
        
        while ( 1 ) {
            
            while ( array[ left ] > pivot_value ) {
                left++;
            }
            
            while ( array[ right ] < pivot_value ) {
                right--;
            }
            
            if ( left < right ) {
                swap_type( array[ left ], array[ right ], u8 );
                left++;
                right--;
            } else {
                break;
            }
        }
        
        internal_quick_sort_u8_descending( array, start, right );
        internal_quick_sort_u8_descending( array, right + 1, end );
    }
}

void quick_sort_u8_ascending( u8* array, umm count ) {
    if ( count ) {
        internal_quick_sort_u8_ascending( array, 0, count - 1 );
    }
}

void quick_sort_u8_descending( u8* array, umm count ) {
    if ( count ) {
        internal_quick_sort_u8_descending( array, 0, count - 1 );
    }
}

// MARK file

/* TODO simon: On Windows I should update those function to use the wide string version of the functions. */

typedef enum file_mode_t {
    file_mode_read = 0,
    file_mode_write = 1
} file_mode_t;

#if defined( WINDOWS )

typedef HANDLE file_handle_t;
#define file_invalid_handle INVALID_HANDLE_VALUE

#elif defined( LINUX )

typedef s32 file_handle_t;
#define file_invalid_handle -1

#endif

typedef enum file_error_t {
    
    file_error_none,
    file_error_open,
    file_error_close,
    file_error_invalid_handle,
    file_error_set_pointer,
    file_error_read_size, /* NOTE simon: The amount of data read is not equal to the amount requested. */
    file_error_write_size, /* NOTE simon: The amount of data written is not equal to the amount requested. */
    file_error_read,
    file_error_write,
    file_error_not_implemented,
    file_error_move,
    file_error_delete,
    file_error_end_of_file,
    
} file_error_t;

typedef struct file_list_t {
    
    data_t* memory;
    u8** names; /* NOTE simon: names are zero terminated. */
    u64* sizes;
    u16* name_lengths; /* NOTE simon: name_lengths doesn't include the null terminator. */
    u8* directories;
    u64* last_write_times;
    u32 count;
    u32 max_count;
    
} file_list_t;

typedef enum file_list_flags_t {
    file_list_recursive = 1 << 0,
    file_list_files = 1 << 1,
    file_list_directories = 1 << 2,
    file_list_add_self_and_parent = 1 << 3,
} file_list_flags_t;

stu file_list_t file_list_initialize( data_t* memory, u32 max_file_count ) {
    
    data_t backup = *memory;
    
    u8** names = memory_push_array_aligned( memory, u8*, max_file_count, 8 );
    u64* sizes = memory_push_array( memory, u64, max_file_count );
    u16* name_lengths = memory_push_array( memory, u16, max_file_count );
    u8* directories = memory_push_array( memory, u8, max_file_count );
    u64* last_write_times = memory_push_array( memory, u64, max_file_count );
    
    file_list_t result = { 0 };
    
    if ( names && sizes && name_lengths && directories && last_write_times ) {
        result.memory = memory;
        result.names = names;
        result.sizes = sizes;
        result.name_lengths = name_lengths;
        result.directories = directories;
        result.last_write_times = last_write_times;
        result.count = 0;
        result.max_count = max_file_count;
    } else {
        *memory = backup;
    }
    
    return result;
}


stu void file_free_memory( data_t* file ) {
    memory_free_virtual( file );
}

stu data_t file_get_extension_from_name_p( void* file_name_arg, umm length ) {
    
    data_t result = { 0 };
    u8* start = ( u8* ) file_name_arg;
    u8* end = start + length;
    u8* dot = ( u8* ) memory_seek_backward_p( start, end, ( void* ) ".", 1 );
    
    if ( dot < end ) {
        result.bytes = dot + 1;
        result.reserved = result.used = end - result.bytes;
    }
    
    return result;
}

stu data_t file_get_extension_from_name_s( void* file_name_arg ) {
    data_t result = file_get_extension_from_name_p( file_name_arg, string_length( file_name_arg ) );
    return result;
}

stu data_t file_get_extension_from_name( data_t file_name ) {
    data_t result = file_get_extension_from_name_p( file_name.bytes, file_name.used );
    return result;
}

stu data_t file_name_remove_extension_p( void* file_name_arg, umm length ) {
    data_t result = memory_construct( length, length, file_name_arg );
    u8* dot = ( u8* ) memory_seek_backward_p( result.bytes, memory_end( result ), ( void* ) ".", 1 );
    
    if ( dot < memory_end( result ) ) {
        result.reserved = result.used = dot - result.bytes;
    }
    
    return result;
}

stu data_t file_name_remove_extension_s( void* file_name_arg ) {
    data_t result = file_name_remove_extension_p( file_name_arg, string_length( file_name_arg ) );
    return result;
}

stu data_t file_name_remove_extension( data_t file_name ) {
    data_t result = file_name_remove_extension_p( file_name.bytes, file_name.used );
    return result;
}

stu data_t file_name_remove_path_p( void* file_name_arg, umm length ) {
    data_t result = memory_construct( length, length, file_name_arg );
    u8* separator = ( u8* ) memory_seek_backward_p( result.bytes, memory_end( result ), ( void* ) "\\", 1 );
    
    if ( separator >= memory_end( result ) ) {
        separator = ( u8* ) memory_seek_backward_p( result.bytes, memory_end( result ), ( void* ) "/", 1 );
    }
    
    if ( separator < memory_end( result ) ) {
        separator += 1;
        result.reserved = result.used = ( u8* ) memory_end( result ) - separator;
        result.bytes = separator;
    }
    
    return result;
}

stu data_t file_name_remove_path_s( void* file_name_arg ) {
    data_t result = file_name_remove_path_p( file_name_arg, string_length( file_name_arg ) );
    return result;
}

stu data_t file_name_remove_path( data_t file_name ) {
    data_t result = file_name_remove_path_p( file_name.bytes, file_name.used );
    return result;
}

stu u32 internal_split_comma_separated_into_list( data_t extensions, data_t* extension_list, u32 max_extension_count ) {
    
    u32 extension_count = 0;
    
    u8* current = extensions.bytes;
    u8* end = memory_end( extensions );
    
    while ( current < end && extension_count < max_extension_count ) {
        
        while ( current < end && char_is_whitespace( *current ) ) {
            current++;
        }
        
        extension_list[ extension_count ].bytes = current;
        extension_list[ extension_count ].used = 0;
        
        while ( current < end && ( *current != ',' ) ) {
            extension_list[ extension_count ].used++;
            current++;
        }
        
        current++;
        
        if ( extension_list[ extension_count ].used ) {
            extension_list[ extension_count ].reserved = extension_list[ extension_count ].used;
            extension_count++;
        }
    }
    
    return extension_count;
}

#if defined( WINDOWS )

stu void file_delete_s( void* file_name_arg, u32* error ) {
    
    if ( no_error( error ) ) {
        
        b32 result = DeleteFile( ( LPCSTR ) file_name_arg );
        
        if ( !result ) {
            set_error( error, file_error_delete );
        }
    }
}

#if 0
/* NOTE simon: This is working but require an include and a lib for the linker (shell32.lib). */
#include <Shellapi.h>

stu b32 file_delete_directory_s( void* folder_name_arg ) {
    
    u8* folder_name = ( u8* ) folder_name_arg;
    
    b32 result = false;
    umm length = string_length( folder_name );
    char folder[ 2048 ];
    
    if ( length < array_count( folder ) - 2 ) {
        
        string_copy( folder, folder_name, length );
        folder[ length ] = 0;
        folder[ length + 1 ] = 0;
        
        SHFILEOPSTRUCT fileop = {
            
            0,
            FO_DELETE,
            folder,
            0,
            FOF_NOCONFIRMATION | FOF_NOERRORUI | FOF_SILENT,
            0, 0, 0
        };
        
        result = !SHFileOperation( &fileop );
        result = result && fileop.fAnyOperationsAborted;
    }
    
    return result;
}
#endif

stu b32 file_is_handle_valid( file_handle_t file ) {
    b32 result = ( file != INVALID_HANDLE_VALUE );
    return result;
}

stu u64 file_get_size_s( void* file_name_arg ) {
    
    u64 result = 0;
    
    file_handle_t file = CreateFile( ( const char * ) file_name_arg, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, 0, 0 );
    
    if ( file_is_handle_valid( file ) ) {
        
        LARGE_INTEGER file_size;
        
        if ( GetFileSizeEx( file, &file_size ) ) {
            result = file_size.QuadPart;
        }
        
        CloseHandle( file );
    }
    
    return result;
}

stu u64 file_get_size_handle( file_handle_t handle ) {
    
    u64 result = 0;
    
    if ( file_is_handle_valid( handle ) ) {
        
        LARGE_INTEGER file_size;
        
        if ( GetFileSizeEx( handle, &file_size ) ) {
            result = file_size.QuadPart;
        }
    }
    
    return result;
}

stu file_handle_t file_open_s( void* file_name, u8 mode, b32 replace, u32* error ) {
    
    file_handle_t file = INVALID_HANDLE_VALUE;
    
    if ( no_error( error ) ) {
        
        file = CreateFile( ( LPCSTR ) file_name, ( mode == file_mode_read ) ? GENERIC_READ : GENERIC_WRITE, ( mode == file_mode_read ) ? FILE_SHARE_READ : 0, 0, ( mode == file_mode_read ) ? OPEN_EXISTING : ( replace ) ? CREATE_ALWAYS : CREATE_NEW, 0, 0 );
        
        if ( !file_is_handle_valid( file ) ) {
            set_error( error, file_error_open );
        }
    }
    
    return file;
}

stu void file_close( file_handle_t file, u32* error ) {
    
    if ( file_is_handle_valid( file ) ) {
        CloseHandle( file );
    } else {
        if ( no_error( error ) ) {
            set_error( error, file_error_close );
        }
    }
}

stu data_t file_read_entire_s( void* file_name_arg, b32 zero_terminated ) {
    
    u8* file_name = ( u8* ) file_name_arg;
    data_t file = { 0 };
    
    file_handle_t handle = CreateFile( ( LPCSTR ) file_name, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, 0, 0 );
    
    if ( file_is_handle_valid( handle ) ) {
        
        LARGE_INTEGER file_size;
        
        if ( GetFileSizeEx( handle, &file_size ) ) {
            
            u32 addin =( zero_terminated ? 1 : 0 );
            u64 size = file_size.QuadPart;
            file = memory_get_virtual( size + addin, 0 );
            file.used = file.reserved;
            
            if ( file.bytes != 0 ) {
                
                u64 total_read = 0;
                DWORD bytes_read = 0;
                u32 to_read = ( u32 ) ( ( size > u32_max ) ? u32_max : size );
                
                while ( size && ReadFile( handle, file.bytes + total_read, to_read, &bytes_read, 0 ) ) {
                    total_read += bytes_read;
                    size -= bytes_read;
                    to_read = ( u32 ) ( ( size > u32_max ) ? u32_max : size );
                }
                
                if ( total_read != ( file.reserved - addin ) ) {
                    
                    file_free_memory( &file );
                    file.bytes = 0;
                    file.reserved = file.used = 0;
                    
                } else if ( zero_terminated ) {
                    /* NOTE simon: this is un-necessary since virtual alloc return zeroed memory. */
                    file.bytes[ file.reserved - 1 ] = '\0';
                }
                
            } else {
                
                file.reserved = file.used = 0;
            }
        }
        
        CloseHandle( handle );
    }
    
    return file;
}

stu u64 file_read_handle( file_handle_t file, s64 position, u64 count, void* destination, u32* error ) {
    
    u64 total_read = 0;
    
    if ( no_error( error ) ) {
        
        if ( !file_is_handle_valid( file ) ) {
            set_error( error, file_error_invalid_handle );
        } else {
            
            LARGE_INTEGER offset;
            offset.QuadPart = position;
            b32 success = SetFilePointerEx( file, offset, 0, FILE_BEGIN );
            
            if ( !success ) {
                set_error( error, file_error_set_pointer );
            } else {
                
                DWORD bytes_read = 0;
                u64 size = count;
                u32 to_read = ( u32 ) ( ( size > u32_max ) ? u32_max : size );
                
                while ( no_error( error ) && total_read < count ) {
                    
                    success = ReadFile( file, ( u8* ) destination + total_read, to_read, &bytes_read, 0 );
                    
                    if ( !success ) {
                        set_error( error, file_error_read );
                    } else if ( bytes_read == 0 ) {
                        /* NOTE simon: if you request bytes but the file pointer is at the end of the file, ReadFile doesn't fail and bytes_read is set to 0. */
                        set_error( error, file_error_end_of_file );
                    } else {
                        total_read += bytes_read;
                        size -= bytes_read;
                        to_read = ( u32 ) ( ( size > u32_max ) ? u32_max : size );
                    }
                }
                
                if ( no_error( error ) && total_read != count ) {
                    set_error( error, file_error_read_size );
                }
            }
        }
    }
    
    return total_read;
}

stu u64 file_write_entire_s( void* file_name, void* bytes, umm size, b32 replace ) {
    
    u8* content = ( u8* ) bytes;
    u64 total_written = 0;
    
    file_handle_t handle = CreateFile( ( const char* ) file_name, GENERIC_WRITE, 0, 0, replace ? CREATE_ALWAYS : CREATE_NEW, 0, 0 );
    
    if ( file_is_handle_valid( handle ) ) {
        
        DWORD bytes_written = 0;
        u32 to_write = ( u32 ) ( ( size > u32_max ) ? u32_max : size );
        
        while ( size && WriteFile( handle, content + total_written, to_write, &bytes_written, 0 ) ) {
            total_written += bytes_written;
            size -= bytes_written;
            to_write = ( u32 ) ( ( size > u32_max ) ? u32_max : size );
        }
        
        CloseHandle( handle );
    }
    
    return total_written;
}

stu u64 file_write_handle( file_handle_t file, s64 position, u64 count, void* source, u32* error ) {
    
    u64 total_written = 0;
    
    if ( no_error( error ) ) {
        
        if ( !file_is_handle_valid( file ) ) {
            set_error( error, file_error_invalid_handle );
        } else {
            
            LARGE_INTEGER offset;
            offset.QuadPart = position;
            b32 success = SetFilePointerEx( file, offset, 0, FILE_BEGIN );
            
            if ( !success ) {
                set_error( error, file_error_set_pointer );
            } else {
                
                DWORD bytes_written = 0;
                u64 size = count;
                u32 to_write = ( u32 ) ( ( size > u32_max ) ? u32_max : size );
                
                while ( no_error( error ) && total_written < count ) {
                    
                    success = WriteFile( file, ( u8* ) source + total_written, to_write, &bytes_written, 0 );
                    
                    if ( !success ) {
                        set_error( error, file_error_write );
                    } else {
                        total_written += bytes_written;
                        size -= bytes_written;
                        to_write = ( u32 ) ( ( size > u32_max ) ? u32_max : size );
                    }
                }
                
                if ( no_error( error ) && count != total_written ) {
                    set_error( error, file_error_write_size );
                }
            }
        }
    }
    
    return total_written;
}

/* TODO simon: check for ascii VS wide. */
stu b32 file_create_directory_s( void* name ) {
    
    b32 result = false;
    result = CreateDirectory( ( LPCSTR ) name, 0 );
    
    if ( !result ) {
        u32 error = GetLastError( );
        result = ( error == ERROR_ALREADY_EXISTS );
    }
    
    return result;
}

stu void file_move_s( void* source_arg, void* destination_arg, b32 replace, u32* error ) {
    // NOTE simon: Can't copy a directory to a different drive/filesystem.
    // Need to copy when it's a different drive/filesystem.
    
    char* source = ( char* ) source_arg;
    char* destination = ( char* ) destination_arg;
    
    if ( *error == file_error_none ) {
        
        DWORD flags = MOVEFILE_COPY_ALLOWED;
        flags |= ( replace ) ? MOVEFILE_REPLACE_EXISTING : 0;
        b32 result = MoveFileEx( source, destination, flags );
        
        if ( !result ) {
            *error = file_error_move;
        }
    }
}

stu b32 file_copy_s( void* source_arg, void* destination_arg, b32 replace ) {
    
    b32 result = false;
    char* source = ( char* ) source_arg;
    char* destination = ( char* ) destination_arg;
    result = CopyFileEx( source, destination, 0, 0, 0, replace ? 0 : COPY_FILE_FAIL_IF_EXISTS );
    
    return result;
}

stu void file_list_get_internal( file_list_t* list, data_t directory, data_t* extensions, umm extension_count, data_t* exceptions, umm exception_count, u32 flags ) {
    
    data_t wildcard = string_l( "\\*\0" );
    data_t directory_wildcard = directory;
    memory_push_copy( &directory_wildcard, wildcard, 1 );
    
    /* TODO simon: Handle wide characters. */
    WIN32_FIND_DATA find_data;
    HANDLE search_handle = FindFirstFile( ( LPCSTR ) directory_wildcard.raw, &find_data );
    
    if ( search_handle != INVALID_HANDLE_VALUE ) {
        
        b32 has_file = true;
        
        while ( has_file && list->count < list->max_count ) {
            
            u8 is_directory = ( find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == FILE_ATTRIBUTE_DIRECTORY;
            umm length = string_length( find_data.cFileName );
            data_t name = memory_construct( length, length, find_data.cFileName );
            
            b32 is_exception = false;
            
            for ( umm i = 0; !is_exception && i < exception_count; i++ ) {
                is_exception = memory_are_equal( name, exceptions[ i ] );
            }
            
            if ( !is_exception ) {
                
                if ( is_directory && flag_is_set_u32( flags, file_list_directories ) ) {
                    
                    if ( flag_is_set_u32( flags, file_list_add_self_and_parent ) || ( !memory_are_equal( name, string_arg_l( "." ) ) && !memory_are_equal( name, string_arg_l( ".." ) ) ) ) {
                        
                        list->directories[ list->count ] = true;
                        list->names[ list->count ] = ( u8* ) memory_push_copy( list->memory, directory, 1 );
                        memory_push_u8( list->memory, '\\' );
                        memory_push_copy( list->memory, name, 1 );
                        memory_push_u8( list->memory, 0 );
                        list->name_lengths[ list->count ] = math_safe_truncate_to_u16( directory.used + 1 + length );
                        list->sizes[ list->count ] = ( ( u64 ) find_data.nFileSizeHigh << 32 ) | ( u64 ) find_data.nFileSizeLow;
                        list->last_write_times[ list->count ] = ( ( u64 ) find_data.ftLastWriteTime.dwLowDateTime ) | ( ( ( u64 ) find_data.ftLastWriteTime.dwHighDateTime ) << 32 );
                        list->count++;
                    }
                    
                }  else if ( !is_directory && flag_is_set_u32( flags, file_list_files ) ) {
                    
                    b32 valid = false;
                    
                    if ( extension_count ) {
                        
                        data_t extension = file_get_extension_from_name_p( find_data.cFileName, length );
                        
                        for ( umm index = 0; index < extension_count; index++ ) {
                            
                            if ( memory_are_equal( extensions[ index ], extension ) ) {
                                valid = true;
                                break;
                            }
                        }
                    } else {
                        valid = true;
                    }
                    
                    if ( valid ) {
                        list->directories[ list->count ] = false;
                        list->names[ list->count ] = ( u8* ) memory_push_copy( list->memory, directory, 1 );
                        memory_push_u8( list->memory, '\\' );
                        memory_push_copy( list->memory, name, 1 );
                        memory_push_u8( list->memory, 0 );
                        list->name_lengths[ list->count ] = math_safe_truncate_to_u16( directory.used + 1 + name.used );
                        list->sizes[ list->count ] = ( ( u64 ) find_data.nFileSizeHigh << 32 ) | ( u64 ) find_data.nFileSizeLow;
                        list->last_write_times[ list->count ] = ( ( u64 ) find_data.ftLastWriteTime.dwLowDateTime ) | ( ( ( u64 ) find_data.ftLastWriteTime.dwHighDateTime ) << 32 );
                        list->count++;
                    }
                }
                
                if ( is_directory && flag_is_set_u32( flags, file_list_recursive ) && list->count < list->max_count ) {
                    
                    if ( !memory_are_equal( name, string_arg_l( "." ) ) && !memory_are_equal( name, string_arg_l( ".." ) ) ) {
                        
                        data_t sub_dir = directory;
                        memory_push_u8( &sub_dir, '\\' );
                        memory_push_copy( &sub_dir, name, 1 );
                        memory_push_u8( list->memory, 0 );
                        
                        file_list_get_internal( list, sub_dir, extensions, extension_count, exceptions, exception_count, flags );
                    }
                }
            }
            
            has_file = FindNextFile( search_handle, &find_data );
        }
        
        FindClose( search_handle );
    }
}

stu void file_list_get_2( file_list_t* list, data_t directory, data_t comma_separated_extensions, data_t comma_separated_exceptions, u32 flags ) {
    
#if defined( TARGET_x32 )
    void* old_value = 0;
    Wow64DisableWow64FsRedirection( &old_value );
#endif
    
    data_t extensions[ 64 ] = { 0 };
    umm extension_count = internal_split_comma_separated_into_list( comma_separated_extensions, extensions, array_count( extensions ) );
    
    data_t exceptions[ 64 ] = { 0 };
    umm exception_count = internal_split_comma_separated_into_list( comma_separated_exceptions, exceptions, array_count( exceptions ) );
    
    memory_get_on_stack( directory_wildcard, 1024 );
    _assert( directory.used <= directory_wildcard.reserved );
    memory_push_copy( &directory_wildcard, directory, 1 );
    
    if ( !flag_is_set_u32( flags, file_list_files ) && !flag_is_set_u32( flags, file_list_directories ) ) {
        flags |= ( file_list_files | file_list_directories );
    }
    
    file_list_get_internal( list, directory_wildcard, extensions, extension_count, exceptions, exception_count, flags );
    
#if defined( TARGET_x32 )
    Wow64RevertWow64FsRedirection( old_value );
#endif
}


// NOTE simon: Directory backslash should be double backslash ( \\ ) and can contains wildcard.
// e.g.: ".\\*"
/* TODO simon: Make sure this function behaves the same on different platforms, by manually implementing wildcards or by preventing them. */
/* NOTE simon: the names are zero terminated. The lengths don't contain the null terminator. */
stu void file_list_get( file_list_t* list, data_t directory, data_t comma_separated_extensions, b32 recursive ) {
    
#if defined( TARGET_x32 )
#error This code will not work when calling this function recursively.
    // NOTE simon: for win32 applications on win64, we need to disable wow to search the right system directories.
    void* old_value = 0;
    Wow64DisableWow64FsRedirection( &old_value );
#endif
    
    data_t wildcard = string_l( "\\*\0" );
    memory_get_on_stack( directory_wildcard, 1024 );
    _assert( directory.used <= directory_wildcard.reserved - wildcard.used );
    memory_push_copy( &directory_wildcard, directory, 1 );
    memory_push_copy( &directory_wildcard, wildcard, 1 );
    
    /* TODO simon: This structure is a macro for ascii or wide char. Handle it. */
    WIN32_FIND_DATA find_data;
    HANDLE search_handle = FindFirstFile( ( LPCSTR ) directory_wildcard.raw, &find_data );
    
    if ( search_handle != INVALID_HANDLE_VALUE ) {
        
        b32 end = false;
        u8* extension_list[ 64 ] = { 0 };
        u32 extension_lengths[ array_count( extension_list ) ] = { 0 };
        // umm extension_count = internal_split_comma_separated_into_list( comma_separated_extensions, ( void** ) extension_list, extension_lengths, array_count( extension_list ) );
        umm extension_count = 0;
        
        while ( !end ) {
            
            umm length = string_length( find_data.cFileName );
            b32 valid = false;
            
            if ( extension_count ) {
                
                data_t extension = file_get_extension_from_name_p( find_data.cFileName, length );
                
                for ( umm index = 0; index < extension_count; index++ ) {
                    
                    if ( memory_are_equal_p( extension_list[ index ], extension_lengths[ index ], extension.bytes, extension.used ) ) {
                        valid = true;
                        break;
                    }
                }
                
            } else {
                valid = true;
            }
            
            u8 is_directory = ( find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == FILE_ATTRIBUTE_DIRECTORY;
            
            if ( valid && list->count < list->max_count ) {
                list->directories[ list->count ] = is_directory;
                list->names[ list->count ] = ( u8* ) memory_push_copy( list->memory, directory, 1 );
                memory_push_u8( list->memory, '\\' );
                memory_push_copy_p( list->memory, find_data.cFileName, length, 1 );
                memory_push_u8( list->memory, 0 );
                list->name_lengths[ list->count ] = math_safe_truncate_to_u16( directory.used + 1 + length );
                list->sizes[ list->count ] = ( ( u64 ) find_data.nFileSizeHigh << 32 ) | ( u64 ) find_data.nFileSizeLow;
                list->last_write_times[ list->count ] = ( ( u64 ) find_data.ftLastWriteTime.dwLowDateTime ) | ( ( ( u64 ) find_data.ftLastWriteTime.dwHighDateTime ) << 32 );
                list->count++;
            }
            
            if ( recursive && is_directory && list->count < list->max_count ) {
                
                if ( !memory_are_equal_p( ".", 1, find_data.cFileName, length ) && !memory_are_equal_p( "..", 2, find_data.cFileName, length ) ) {
                    
                    directory_wildcard.used = 0;
                    memory_push_copy( &directory_wildcard, directory, 1 );
                    memory_push_u8( &directory_wildcard, '\\' );
                    memory_push_copy_p( &directory_wildcard, find_data.cFileName, length, 1 );
                    memory_push_u8( list->memory, 0 );
                    
                    file_list_get( list, directory_wildcard, comma_separated_extensions, recursive );
                }
            }
            
            end = !FindNextFile( search_handle, &find_data );
        }
        
        FindClose( search_handle );
    }
    
#if defined( TARGET_x32 )
#error This code will not work when calling this function recursively.
    Wow64RevertWow64FsRedirection( old_value );
#endif
}

stu void internal_file_list_sort_date_ascending( file_list_t* list, umm start, umm end ) {
    
    if ( list && start < end ) {
        
        umm pivot_index = ( start + end ) / 2;
        u64 pivot = list->last_write_times[ pivot_index ];
        
        umm i = start, j = end;
        
        while ( 1 ) {
            
            while ( list->last_write_times[ i ] < pivot ) {
                i++;
            }
            
            while ( list->last_write_times[ j ] > pivot ) {
                j--;
            }
            
            if ( i < j ) {
                
                u8* name_temp = list->names[ i ];
                u64 size_temp = list->sizes[ i ];
                u16 name_length_temp = list->name_lengths[ i ];
                u8 directory_temp = list->directories[ i ];
                u64 last_write_time_temp = list->last_write_times[ i ];
                
                list->names[ i ] = list->names[ j ];
                list->sizes[ i ] = list->sizes[ j ];
                list->name_lengths[ i ] = list->name_lengths[ j ];
                list->directories[ i ] = list->directories[ j ];
                list->last_write_times[ i ] = list->last_write_times[ j ];
                
                list->names[ j ] = name_temp;
                list->sizes[ j ] = size_temp;
                list->name_lengths[ j ] = name_length_temp;
                list->directories[ j ] = directory_temp;
                list->last_write_times[ j ] = last_write_time_temp;
                
                i++;
                j--;
                
            } else {
                break;
            }
        }
        
        internal_file_list_sort_date_ascending( list, start, j );
        internal_file_list_sort_date_ascending( list, j + 1, end );
    }
}

stu void file_list_sort_date_ascending( file_list_t* list ) {
    internal_file_list_sort_date_ascending( list, 0, list->count - 1 );
}

stu void internal_file_list_sort_name_ascending( file_list_t* list, umm start, umm end ) {
    
    if ( list && start < end ) {
        
        umm pivot_index = ( start + end ) / 2;
        u8* pivot = list->names[ pivot_index ];
        u16 pivot_length = list->name_lengths[ pivot_index ];
        
        umm i = start, j = end;
        
        while ( 1 ) {
            
            while ( string_compare( list->names[ i ], pivot, math_min( list->name_lengths[ i ], pivot_length ) ) < 0 ) {
                i++;
            }
            
            while ( string_compare( list->names[ j ], pivot, math_min( list->name_lengths[ j ], pivot_length ) ) > 0 ) {
                j--;
            }
            
            if ( i < j ) {
                
                u8* name_temp = list->names[ i ];
                u64 size_temp = list->sizes[ i ];
                u16 name_length_temp = list->name_lengths[ i ];
                u8 directory_temp = list->directories[ i ];
                u64 last_write_time_temp = list->last_write_times[ i ];
                
                list->names[ i ] = list->names[ j ];
                list->sizes[ i ] = list->sizes[ j ];
                list->name_lengths[ i ] = list->name_lengths[ j ];
                list->directories[ i ] = list->directories[ j ];
                list->last_write_times[ i ] = list->last_write_times[ j ];
                
                list->names[ j ] = name_temp;
                list->sizes[ j ] = size_temp;
                list->name_lengths[ j ] = name_length_temp;
                list->directories[ j ] = directory_temp;
                list->last_write_times[ j ] = last_write_time_temp;
                
                i++;
                j--;
                
            } else {
                break;
            }
        }
        
        internal_file_list_sort_name_ascending( list, start, j );
        internal_file_list_sort_name_ascending( list, j + 1, end );
    }
}

stu void file_list_sort_name_ascending( file_list_t* list ) {
    internal_file_list_sort_name_ascending( list, 0, list->count - 1 );
}

stu b32 file_exists_s( void* file_name ) {
    
    u32 attributes = GetFileAttributes( ( LPCSTR ) file_name );
    b32 result = attributes != INVALID_FILE_ATTRIBUTES && !( attributes & FILE_ATTRIBUTE_DIRECTORY );
    
    return result;
}

stu b32 file_directory_exists_s( void* directory_name ) {
    
    u32 attributes = GetFileAttributes( ( LPCSTR ) directory_name );
    b32 result = attributes != INVALID_FILE_ATTRIBUTES && ( attributes & FILE_ATTRIBUTE_DIRECTORY );
    
    return result;
}

#elif defined( LINUX )

#include <stdio.h>

stu void file_delete_s( void* file_name, u32* error ) {
    
    if ( *error == file_error_none ) {
        
        if ( remove( ( const char* ) file_name ) != 0 ) {
            *error = file_error_delete;
        }
    }
}

stu u64 file_get_size_s( void* file_name ) {
    
    u64 result = 0;
    
    struct stat stats = {};
    
    if ( stat( ( const char* ) file_name, &stats ) >= 0 ) {
        result = stats.st_size;
    }
    
    return result;
}

stu u64 file_get_size( data_t* file_name );

stu b32 file_is_handle_valid( file_handle_t handle ) {
    b32 result = ( handle >= 0 );
    return result;
}

stu file_handle_t file_open_s( void* file_name, u8 mode, b32 replace, u32* error ) {
    
    file_handle_t handle = 0;
    
    if ( no_error( error ) ) {
        
        s32 flags;
        
        if ( mode == file_mode_read ) {
            flags = O_RDONLY;
        } else {
            flags = O_WRONLY | O_CREAT;
        }
        
        if ( mode == file_mode_write && !replace ) {
            flags |= O_EXCL;
        }
        
        mode_t linux_mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH;
        
        handle = open( ( const char* ) file_name, flags, linux_mode );
        
        if ( !file_is_handle_valid( handle ) ) {
            set_error( error, file_error_open );
        }
    }
    
    return handle;
}

stu void file_close( file_handle_t handle, u32* error ) {
    
    if ( file_is_handle_valid( handle ) ) {
        close( handle );
    } else if ( no_error( error ) ) {
        set_error( error, file_error_close );
    }
}

stu u64 file_write_handle( file_handle_t file, s64 position, u64 count, void* source, u32* error ) {
    
    u64 total_written = 0;
    
    if ( no_error( error ) ) {
        
        if ( !file_is_handle_valid( file ) ) {
            set_error( error, file_error_invalid_handle );
        } else {
            
            s64 result = lseek( file, position, SEEK_SET );
            
            if ( result == -1 ) {
                set_error( error, file_error_set_pointer );
            } else {
                
                size_t to_write = ( count > SSIZE_MAX ) ? SSIZE_MAX : count;
                
                while ( no_error( error ) && total_written < count ) {
                    
                    ssize_t bytes_written = write( file, ( ( u8* ) source ) + total_written, to_write );
                    
                    if ( bytes_written == -1 ) {
                        set_error( error, file_error_write );
                    } else {
                        total_written += bytes_written;
                        to_write = ( size_t ) ( count - total_written );
                        to_write = ( to_write > SSIZE_MAX ) ? SSIZE_MAX : to_write;
                    }
                }
                
                if ( no_error( error ) && count != total_written ) {
                    set_error( error, file_error_write_size );
                }
            }
        }
    }
    
    return total_written;
}

stu u64 file_read_handle( file_handle_t file, s64 position, u64 count, void* destination, u32* error ) {
    
    /* TODO simon: use pread / pwrite to avoid lseek ? */
    u64 total_read = 0;
    
    if ( no_error( error ) ) {
        
        if ( !file_is_handle_valid( file ) ) {
            set_error( error, file_error_invalid_handle );
        } else {
            
            s64 result = lseek( file, position, SEEK_SET );
            
            if ( result < 0 ) {
                set_error( error, file_error_set_pointer );
            } else {
                
                size_t to_read = ( count > SSIZE_MAX ) ? SSIZE_MAX : count;
                
                while ( no_error( error ) && total_read < count ) {
                    
                    ssize_t bytes_read = read( file, ( ( u8* ) destination ) + total_read, to_read );
                    
                    if ( bytes_read == -1 ) {
                        set_error( error, file_error_read );
                    } else if ( bytes_read == 0 ) {
                        /* NOTE simon: if you request bytes but the file pointer is at the end of the file, read returns 0. */
                        set_error( error, file_error_end_of_file );
                    } else {
                        total_read += bytes_read;
                        to_read = ( size_t ) ( count - total_read );
                        to_read = ( to_read > SSIZE_MAX ) ? SSIZE_MAX : to_read;
                    }
                }
                
                if ( no_error( error ) && total_read != count ) {
                    set_error( error, file_error_read_size );
                }
            }
        }
    }
    
    return total_read;
}

stu data_t file_read_entire_s( void* file_name, b32 zero_terminated ) {
    
    data_t file = { };
    
    file_handle_t handle = open( ( const char* ) file_name, O_RDONLY );
    
    if ( handle >= 0 ) {
        
        struct stat stats = {};
        
        if ( fstat( handle, &stats ) >= 0 ) {
            
            u32 addin =( zero_terminated ? 1 : 0 );
            u64 size = stats.st_size + addin;
            file = memory_get_virtual( size, 0 );
            file.used = file.reserved;
            
            if ( file.bytes != 0 ) {
                
                u64 total_read = 0;
                size_t to_read = ( file.reserved > SSIZE_MAX ) ? SSIZE_MAX : file.reserved;
                
                while ( total_read < ( file.reserved - addin ) ) {
                    
                    ssize_t bytes_read = read( handle, file.bytes + total_read, to_read );
                    
                    if ( bytes_read == -1 ) {
                        break;
                    } else {
                        
                        total_read += bytes_read;
                        to_read = ( size_t ) ( file.reserved - addin - total_read );
                        to_read = ( to_read > SSIZE_MAX ) ? SSIZE_MAX : to_read;
                        
                        if ( bytes_read == 0 ) {
                            break;
                        }
                    }
                }
                
                if ( total_read != ( file.reserved - addin ) ) {
                    
                    file_free_memory( &file );
                    file.bytes = 0;
                    file.reserved = file.used = 0;
                    
                } else if ( zero_terminated ) {
                    *( file.bytes + file.reserved - 1 ) = '\0';
                }
                
            } else {
                
                file.reserved = file.used = 0;
            }
        }
        
        close( handle );
    }
    
    return file;
}

stu u64 file_write_entire_s( void* file_name, void* content, umm size, b32 replace ) {
    
    u64 total_written = 0;
    s32 flags = O_WRONLY | O_CREAT;
    
    if ( !replace ) {
        flags |= O_EXCL;
    }
    
    mode_t mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH;
    file_handle_t handle = open( ( const char* ) file_name, flags, mode );
    
    if ( handle >= 0 ) {
        
        size_t to_write = ( size > SSIZE_MAX ) ? SSIZE_MAX : size;
        
        while ( total_written < size ) {
            
            ssize_t bytes_written = write( handle, ( ( u8* ) content ) + total_written, to_write );
            
            if ( bytes_written == -1 ) {
                break;
            }
            
            total_written += bytes_written;
            to_write = ( size_t ) ( size - total_written );
            to_write = ( to_write > SSIZE_MAX ) ? SSIZE_MAX : to_write;
        }
        
        close( handle );
    }
    
    return total_written;
}

stu b32 file_create_directory_s( void* name ) {
    
    b32 result = false;
    
    mode_t mode = S_IRWXU | S_IRWXG | S_IRGRP | S_IXOTH;
    
    if ( mkdir( ( const char* ) name, mode ) == 0 ) {
        result = true;
    }
    
    return result;
}

stu void file_move_s( void* source, void* destination, b32 replace, u32* error ) {
    
    if ( no_error( error ) ) {
        
        s32 failure = false;
        
        if ( replace ) {
            failure = rename( ( const char* ) source, ( const char* ) destination );
        } else {
            int local_RENAME_NOREPLACE = 1; /* NOTE simon: This is defined in <linux/fs.h> */
            failure = syscall( SYS_renameat2, AT_FDCWD, source, AT_FDCWD, destination, local_RENAME_NOREPLACE );
        }
        
        if ( !failure ) {
            set_error( error, file_error_move );
        }
    }
}

/* TODO simon: there is no check to see if the whole content of the file is copied. sendfile return
* the number of bytes copied which could be less then the input file. */
stu b32 file_copy_s( void* source_arg, void* destination_arg, b32 replace ) {
    
    b32 result = false;
    s32 in_file = open( ( const char* ) source_arg, O_RDONLY );
    
    if ( in_file >= 0 ) {
        
        s32 flags = O_WRONLY | O_CREAT;
        
        if ( !replace ) {
            flags |= O_EXCL;
        }
        
        struct stat stats = {};
        
        if ( fstat( in_file, &stats ) >= 0 ) {
            
            s32 out_file = open( ( const char* ) destination_arg, flags, stats.st_mode );
            
            if ( out_file >= 0 ) {
                
                _assert( stats.st_size <= 0x7ffff000 );
                s32 bytes_written = sendfile( out_file, in_file, 0, stats.st_size );
                
                if ( bytes_written != -1 ) {
                    result = true;
                }
            }
            
            close( out_file );
        }
        
        close( in_file );
    }
    
    return result;
}

stu void file_list_get_internal( file_list_t* list, data_t directory, data_t* extensions, umm extension_count, data_t* exceptions, umm exception_count, u32 flags ) {
    
    data_t directory_wildcard = directory;
    string_zero_terminate( directory_wildcard );
    
    DIR* dir = opendir( directory_wildcard.text );
    
    if ( dir != 0 ) {
        
        struct dirent* data_read = readdir( dir );
        
        while ( data_read && list->count < list->max_count ) {
            
            u8 is_directory = ( data_read->d_type == DT_DIR );
            umm length = string_length( data_read->d_name );
            data_t name = memory_construct( length, length, data_read->d_name );
            
            b32 is_exception = false;
            
            for ( umm i = 0; !is_exception && i < exception_count; i++ ) {
                is_exception = memory_are_equal( name, exceptions[ i ] );
            }
            
            if ( !is_exception ) {
                
                if ( is_directory && flag_is_set_u32( flags, file_list_directories ) ) {
                    
                    if ( flag_is_set_u32( flags, file_list_add_self_and_parent ) || ( !memory_are_equal( name, string_arg_l( "." ) ) && !memory_are_equal( name, string_arg_l( ".." ) ) ) ) {
                        
                        list->directories[ list->count ] = true;
                        list->names[ list->count ] = ( u8* ) memory_push_copy( list->memory, directory, 1 );
                        memory_push_u8( list->memory, '/' );
                        memory_push_copy( list->memory, name, 1 );
                        memory_push_u8( list->memory, 0 );
                        list->name_lengths[ list->count ] = math_safe_truncate_to_u16( directory.used + 1 + length );
                        
                        struct stat stats = { 0 };
                        int stat_result = stat( ( char* ) list->names[ list->count ], &stats );
                        _assert( stat_result >= 0 );
                        
                        list->sizes[ list->count ] = stats.st_size;
                        list->last_write_times[ list->count ] = stats.st_mtime;
                        list->count++;
                    }
                    
                }  else if ( !is_directory && flag_is_set_u32( flags, file_list_files ) ) {
                    
                    b32 valid = false;
                    
                    if ( extension_count ) {
                        
                        data_t extension = file_get_extension_from_name_p( data_read->d_name, length );
                        
                        for ( umm index = 0; index < extension_count; index++ ) {
                            
                            if ( memory_are_equal( extensions[ index ], extension ) ) {
                                valid = true;
                                break;
                            }
                        }
                    } else {
                        valid = true;
                    }
                    
                    if ( valid ) {
                        list->directories[ list->count ] = false;
                        list->names[ list->count ] = ( u8* ) memory_push_copy( list->memory, directory, 1 );
                        memory_push_u8( list->memory, '/' );
                        memory_push_copy( list->memory, name, 1 );
                        memory_push_u8( list->memory, 0 );
                        list->name_lengths[ list->count ] = math_safe_truncate_to_u16( directory.used + 1 + name.used );
                        
                        struct stat stats = { 0 };
                        int stat_result = stat( ( char* ) list->names[ list->count ], &stats );
                        _assert( stat_result >= 0 );
                        
                        list->sizes[ list->count ] = stats.st_size;
                        list->last_write_times[ list->count ] = stats.st_mtime;
                        list->count++;
                    }
                }
                
                if ( is_directory && flag_is_set_u32( flags, file_list_recursive ) && list->count < list->max_count ) {
                    
                    if ( !memory_are_equal( name, string_arg_l( "." ) ) && !memory_are_equal( name, string_arg_l( ".." ) ) ) {
                        
                        data_t sub_dir = directory;
                        memory_push_u8( &sub_dir, '/' );
                        memory_push_copy( &sub_dir, name, 1 );
                        memory_push_u8( list->memory, 0 );
                        
                        file_list_get_internal( list, sub_dir, extensions, extension_count, exceptions, exception_count, flags );
                    }
                }
            }
            
            data_read = readdir( dir );
        }
        
        closedir( dir );
    }
}

stu void file_list_get_2( file_list_t* list, data_t directory, data_t comma_separated_extensions, data_t comma_separated_exceptions, u32 flags ) {
    
    data_t extensions[ 64 ] = { 0 };
    umm extension_count = internal_split_comma_separated_into_list( comma_separated_extensions, extensions, array_count( extensions ) );
    
    data_t exceptions[ 64 ] = { 0 };
    umm exception_count = internal_split_comma_separated_into_list( comma_separated_exceptions, exceptions, array_count( exceptions ) );
    
    memory_get_on_stack( directory_wildcard, 1024 );
    _assert( directory.used <= directory_wildcard.reserved );
    memory_push_copy( &directory_wildcard, directory, 1 );
    
    if ( !flag_is_set_u32( flags, file_list_files ) && !flag_is_set_u32( flags, file_list_directories ) ) {
        flags |= ( file_list_files | file_list_directories );
    }
    
    file_list_get_internal( list, directory_wildcard, extensions, extension_count, exceptions, exception_count, flags );
}

stu void file_list_get( file_list_t* list, data_t directory, data_t comma_separated_extensions, b32 recursive ) {
    
    memory_get_on_stack( directory_wildcard, 1024 );
    _assert( directory.used <= directory_wildcard.reserved );
    memory_push_copy( &directory_wildcard, directory, 1 );
    memory_push_u8( &directory_wildcard, 0 );
    
    DIR* dir = opendir( directory_wildcard.text );
    
    if ( dir != 0 ) {
        
        struct dirent* data_read = 0;
        
        data_t extensions[ 64 ] = { 0 };
        umm extension_count = internal_split_comma_separated_into_list( comma_separated_extensions, extensions, array_count( extensions ) );
        
        while ( ( data_read = readdir( dir ) ) != 0 ) {
            
            u32 length = string_length( data_read->d_name );
            b32 valid = false;
            
            if ( extension_count ) {
                
                data_t extension = file_get_extension_from_name_p( data_read->d_name, length );
                
                for ( umm index = 0; index < extension_count; index++ ) {
                    
                    if ( memory_are_equal( extensions[ index ], extension ) ) {
                        valid = true;
                        break;
                    }
                }
                
            } else {
                valid = true;
            }
            
            u8 is_directory = ( data_read->d_type == DT_DIR );
            
            if ( valid && list->count < list->max_count ) {
                
                list->directories[ list->count ] = is_directory;
                list->names[ list->count ] = ( u8* ) memory_push_copy( list->memory, directory, 1 );
                memory_push_u8( list->memory, '/' );
                memory_push_copy_p( list->memory, data_read->d_name, length, 1 );
                memory_push_u8( list->memory, 0 );
                list->name_lengths[ list->count ] = math_safe_truncate_to_u16( directory.used + 1 + length );
                
                struct stat stats = { 0 };
                int stat_result = stat( ( char* ) list->names[ list->count ], &stats );
                _assert( stat_result >= 0 );
                
                list->sizes[ list->count ] = stats.st_size;
                list->last_write_times[ list->count ] = stats.st_mtime;
                list->count++;
            }
            
            if ( recursive && is_directory && list->count < list->max_count ) {
                
                if ( !memory_are_equal_p( ( void* ) ".", 1, data_read->d_name, length ) && !memory_are_equal_p( ( void* ) "..", 2, data_read->d_name, length ) ) {
                    
                    directory_wildcard.used = 0;
                    memory_push_copy( &directory_wildcard, directory, 1 );
                    memory_push_u8( &directory_wildcard, '/' );
                    memory_push_copy_p( &directory_wildcard, data_read->d_name, length, 1 );
                    memory_push_u8( list->memory, 0 );
                    file_list_get( list, directory_wildcard, comma_separated_extensions, recursive );
                }
            }
        }
        
        closedir( dir );
    }
}

stu b32 file_exists_s( void* file_name ) {
    
    b32 result = false;
    struct stat buffer;
    s32 status = stat( ( const char* ) file_name, &buffer );
    
    if ( status == 0 && ( buffer.st_mode & S_IFDIR ) == 0 ) {
        result = true;
    }
    
    return result;
}

stu b32 file_directory_exists_s( void* directory_name ) {
    
    b32 result = false;
    struct stat buffer;
    s32 status = stat( ( const char* ) directory_name, &buffer );
    
    if ( status == 0 && ( buffer.st_mode & S_IFDIR ) != 0 ) {
        result = true;
    }
    
    return result;
}

#else
# error Unsupported platform.
#endif

stu void file_delete_p( void* file_name_arg, umm file_name_length, u32* error ) {
    string_zero_terminate_on_stack( file_name, 512, file_name_arg, file_name_length );
    file_delete_s( file_name.bytes, error );
}

stu void file_delete( data_t file_name, u32* error ) {
    file_delete_p( file_name.bytes, file_name.used, error );
}

#if 1
stu u64 file_get_size_p( void* file_name_arg, umm file_name_length ) {
    string_zero_terminate_on_stack( file_name, 512, file_name_arg, file_name_length );
    u64 result = file_get_size_s( file_name.bytes );
    return result;
}

stu u64 file_get_size( data_t* file_name ) {
    u64 result = file_get_size_p( file_name->bytes, file_name->used );
    return result;
}

stu file_handle_t file_open_p( void* file_name_arg, umm file_name_length, u8 mode, b32 replace, u32* error ) {
    string_zero_terminate_on_stack( file_name, 512, file_name_arg, file_name_length );
    file_handle_t result = file_open_s( file_name.bytes, mode, replace, error );
    return result;
}

stu file_handle_t file_open( data_t file_name, u8 mode, b32 replace, u32* error ) {
    file_handle_t result = file_open_p( file_name.bytes, file_name.used, mode, replace, error );
    return result;
}

stu data_t file_read_entire_p( void* file_name_arg, umm file_name_length, b32 zero_terminated ) {
    string_zero_terminate_on_stack( file_name, 512, file_name_arg, file_name_length );
    data_t result = file_read_entire_s( file_name.bytes, zero_terminated );
    return result;
}

stu data_t file_read_entire( data_t file_name, b32 zero_terminated ) {
    data_t result = file_read_entire_p( file_name.bytes, file_name.used, zero_terminated );
    return result;
}

stu void file_read_s( void* file_name, s64 position, u64 count, void* destination, u32* error ) {
    
    file_handle_t file = file_open_s( file_name, file_mode_read, false, error );
    file_read_handle( file, position, count, destination, error );
    file_close( file, error );
}

stu void file_read_p( void* file_name_arg, umm file_name_length, s64 position, u64 count, void* destination, u32* error ) {
    string_zero_terminate_on_stack( file_name, 512, file_name_arg, file_name_length );
    file_read_s( file_name.bytes, position, count, destination, error );
}

stu void file_read( data_t* file_name, s64 position, u64 count, void* destination, u32* error ) {
    file_read_p( file_name->bytes, file_name->used, position, count, destination, error );
}

stu u64 file_write_entire_p( void* file_name_arg, umm file_name_length, void* bytes, umm size, b32 replace ) {
    string_zero_terminate_on_stack( file_name, 512, file_name_arg, file_name_length );
    u64 result = file_write_entire_s( file_name.bytes, bytes, size, replace );
    return result;
}

stu u64 file_write_entire( data_t file_name, void* bytes, umm size, b32 replace ) {
    u64 result = file_write_entire_p( file_name.bytes, file_name.used, bytes, size, replace );
    return result;
}

stu void file_write_s( void* file_name, u8* source, s64 position, u64 count, b32 replace, u32* error ) {
    
    file_handle_t file = file_open_s( file_name, file_mode_write, replace, error );
    file_write_handle( file, position, count, source, error );
    file_close( file, error );
}

stu void file_write_p( void* file_name_arg, umm file_name_length, u8* source, s64 position, u64 count, b32 replace, u32* error ) {
    string_zero_terminate_on_stack( file_name, 512, file_name_arg, file_name_length );
    file_write_s( file_name.bytes, source, position, count, replace, error );
}

stu void file_write( data_t file_name, u8* source, s64 position, u64 count, b32 replace, u32* error ) {
    file_write_p( file_name.bytes, file_name.used, source, position, count, replace, error );
}

stu b32 file_create_directory_p( void* directory_name_arg, umm directory_name_length ) {
    string_zero_terminate_on_stack( directory_name, 512, directory_name_arg, directory_name_length );
    b32 result = file_create_directory_s( directory_name.bytes );
    return result;
}

stu b32 file_create_directory( data_t* directory_name ) {
    b32 result = file_create_directory_p( directory_name->bytes, directory_name->used );
    return result;
}

#if 0
stu b32 file_delete_directory_p( void* directory_name_arg, umm directory_name_length ) {
    string_zero_terminate_on_stack( directory_name, 512, directory_name_arg, directory_name_length );
    b32 result = file_delete_directory_s( directory_name.bytes );
    return result;
}

stu b32 file_delete_directory( data_t* directory_name ) {
    b32 result = file_delete_directory_p( directory_name->bytes, directory_name->used );
    return result;
}
#endif

stu void file_move_p( void* source_arg, umm source_length, void* destination_arg, umm destination_length, b32 replace, u32* error ) {
    string_zero_terminate_on_stack( source, 512, source_arg, source_length );
    string_zero_terminate_on_stack( destination, 512, destination_arg, destination_length );
    file_move_s( source.bytes, destination.bytes, replace, error );
}

stu void file_move( data_t source, data_t destination, b32 replace, u32* error ) {
    file_move_p( source.bytes, source.used, destination.bytes, destination.used, replace, error );
}

stu b32 file_copy_p( void* source_arg, umm source_length, void* destination_arg, umm destination_length, b32 replace ) {
    string_zero_terminate_on_stack( source, 512, source_arg, source_length );
    string_zero_terminate_on_stack( destination, 512, destination_arg, destination_length );
    b32 result = file_copy_s( source.bytes, destination.bytes, replace );
    return result;
}

stu b32 file_copy( data_t* source, data_t* destination, b32 replace ) {
    b32 result = file_copy_p( source->bytes, source->used, destination->bytes, destination->used, replace );
    return result;
}

stu void file_list_get_p( file_list_t* list, void* directory_name_arg, umm directory_name_length, void* comma_separeted_extensions_arg, umm comma_separeted_extensions_length, b32 recursive ) {
    data_t directory_name = memory_construct( directory_name_length, directory_name_length, directory_name_arg );
    data_t comma_separeted_extensions = memory_construct( comma_separeted_extensions_length, comma_separeted_extensions_length, comma_separeted_extensions_arg );
    
    file_list_get( list, directory_name, comma_separeted_extensions, recursive );
}

stu void file_list_get_s( file_list_t* list, void* directory_name, void* comma_separeted_extensions, b32 recursive ) {
    file_list_get_p( list, directory_name, string_length( directory_name ), comma_separeted_extensions, string_length( comma_separeted_extensions ), recursive );
}

#define file_list_get_l( list, directory_name, comma_separeted_extensions, recursive ) \
file_list_get_p( ( list ), ( directory_name ), sizeof( directory_name ) - 1, ( comma_separeted_extensions ), sizeof( comma_separeted_extensions ) - 1, ( recursive ) )

stu b32 file_exists_p( void* file_name_arg, umm file_name_length ) {
    string_zero_terminate_on_stack( file_name, 512, file_name_arg, file_name_length );
    b32 result = file_exists_s( file_name.bytes );
    return result;
}

stu b32 file_exists( data_t file_name ) {
    b32 result = file_exists_p( file_name.bytes, file_name.used );
    return result;
}

stu b32 file_directory_exists_p( void* directory_name_arg, umm directory_name_length ) {
    string_zero_terminate_on_stack( directory_name, 512, directory_name_arg, directory_name_length );
    b32 result = file_directory_exists_s( directory_name.bytes );
    return result;
}

stu b32 file_directory_exists( data_t directory_name ) {
    b32 result = file_directory_exists_p( directory_name.bytes, directory_name.used );
    return result;
}
#endif

// MARK random
stu u64 random_u64( u64* series ) {
    
    // NOTE simon: list of RNG http://www.pcg-random.org/other-rngs.html
    // NOTE simon: xorshift64* https://en.wikipedia.org/wiki/Xorshift
    u64 x = *series; // The state must be seeded with a nonzero value.
    x ^= x >> 12; // a
    x ^= x << 25; // b
    x ^= x >> 27; // c
    x *= 2685821657736338717ULL;
    *series = x;
    
    return x;
}

/* NOTE simon: returns a number between min included and max excluded. */
stu u64 random_range_u64( u64* series, u64 min, u64 max ) {
    
    u64 n = random_u64( series );
    n %= ( max - min );
    n += min;
    
    return n;
}

stu r64 random_r64( u64* series ) {
    
    r64 n = random_u64( series ) / ( r64 ) u64_max;
    
    return n;
}

stu r64 random_range_r64( u64* series, r64 min, r64 max ) {
    
    r64 n = random_r64( series );
    n *= max - min;
    n += min;
    
    return n;
}

stu r32 random_r32( u64* series ) {
    
    r32 n = ( r32 ) random_r64( series );
    return n;
}

stu r32 random_range_r32( u64* series, r32 min, r32 max ) {
    
    r32 n = random_r32( series );
    n *= max - min;
    n += min;
    
    return n;
}

/* MARK linear algebra */

typedef union vec2 {
    
    struct {
        r32 x, y;
    };
    
    struct {
        r32 u, v;
    };
    
    struct {
        r32 left, right;
    };
    
    r32 e[ 2 ];
} vec2;

#define vec2_alignment sizeof( r32 )

typedef union vec2i {
    
    struct {
        s32 x, y;
    };
    
    s32 e[ 2 ];
} vec2i;

#define vec2i_alignment sizeof( s32 )

typedef union vec3 {
    
    struct {
        r32 x, y, z;
    };
    
    struct {
        vec2 xy;
    };
    
    struct {
        r32 r, g, b;
    };
    
    struct {
        r32 u, v, w;
    };
    
    struct {
        vec2 uv;
    };
    
    r32 e[ 3 ];
} vec3;

#define vec3_alignment sizeof( r32 )

/* TODO simon: Should I merge vec4 and rect2 ? */

typedef union vec4 {
    
    struct {
        r32 x, y, z, w;
    };
    
    struct {
        vec2 xy;
    };
    
    struct {
        vec3 xyz;
    };
    
    struct {
        r32 r, g, b, a;
    };
    
    struct {
        vec3 rgb;
    };
    
    struct {
        r32 top, bottom, left, right;
    };
    
    r32 e[ 4 ];
} vec4;

#define vec4_alignment sizeof( r32 )

stu vec2 v2( r32 x, r32 y ) {
    vec2 result;
    result.x = x;
    result.y = y;
    
    return result;
}

stu vec2 v2_s32( s32 x, s32 y ) {
    vec2 result;
    result.x = ( r32 ) x;
    result.y = ( r32 ) y;
    
    return result;
}

stu vec2 v2_u32( u32 x, u32 y ) {
    vec2 result;
    result.x = ( r32 ) x;
    result.y = ( r32 ) y;
    
    return result;
}

stu vec2i v2i( s32 x, s32 y ) {
    vec2i result;
    result.x = x;
    result.y = y;
    
    return result;
}

stu vec2i v2i_fromR32( r32 x, r32 y ) {
    vec2i result;
    result.x = ( s32 ) x;
    result.y = ( s32 ) y;
    
    return result;
}

stu vec3 v3( r32 x, r32 y, r32 z ) {
    vec3 result;
    result.x = x;
    result.y = y;
    result.z = z;
    
    return result;
}

stu vec3 v3_fromvec2( vec2 v )  {
    vec3 result;
    result.xy = v;
    result.z = 0;
    
    return result;
}

stu vec4 v4( r32 x, r32 y, r32 z, r32 w ) {
    vec4 result;
    result.x = x;
    result.y = y;
    result.z = z;
    result.w = w;
    
    return result;
}

stu vec4 v4_tblr( r32 t, r32 b, r32 l, r32 r ) {
    vec4 result;
    result.top = t;
    result.bottom = b;
    result.left = l;
    result.right = r;
    
    return result;
}

stu vec4 v4_add( vec4 a, vec4 b ) {
    
    a.x += b.x;
    a.y += b.y;
    a.z += b.z;
    a.w += b.w;
    
    return a;
}

#define c_white { 1.0f, 1.0f, 1.0f, 1.0f }
#define c_black { 0.0f, 0.0f, 0.0f, 1.0f }
#define c_red { 1.0f, 0.0f, 0.0f, 1.0f }
#define c_green { 0.0f, 1.0f, 0.0f, 1.0f }
#define c_blue { 0.0f, 0.0f, 1.0f, 1.0f }
#define c_yellow { 1.0f, 1.0f, 0.0f, 1.0f }
#define c_magenta { 1.0f, 0.0f, 1.0f, 1.0f }
#define c_cyan { 0.0f, 1.0f, 1.0f, 1.0f }
#define c_light_gray { 0.75f, 0.75f, 0.75f, 1.0f }
#define c_gray { 0.5f, 0.5f, 0.5f, 1.0f }
#define c_dark_gray { 0.25f, 0.25f, 0.25f, 1.0f }
#define c_orange { 1.0f, 0.5f, 0.0f, 1.0f }

vec4 white = c_white;
vec4 black = c_black;
vec4 red = c_red;
vec4 green = c_green;
vec4 blue = c_blue;
vec4 yellow = c_yellow;
vec4 magenta = c_magenta;
vec4 cyan = c_cyan;
vec4 light_gray = c_light_gray;
vec4 gray = c_gray;
vec4 dark_gray = c_dark_gray;
vec4 orange = c_orange;

stu u32 v4_to_u32_color( vec4 v ) {
    u32 result;
    result = ( ( u8 ) ( v.r * 255.0f ) ) << 0;
    result |= ( ( u8 ) ( v.g * 255.0f ) ) << 8;
    result |= ( ( u8 ) ( v.b * 255.0f ) ) << 16;
    result |= ( ( u8 ) ( v.a * 255.0f ) ) << 24;
    return result;
}

stu r32 v4_to_r32_color( vec4 v ) {
    u32 c = v4_to_u32_color( v );
    r32 result = * ( r32* ) &c;
    return result;
}

stu vec4 v4_from_u32_color( u32 c ) {
    vec4 result;
    r32 normalize = 1.0f / 255.0f;
    result.r = ( ( c & 0x000000ff ) >> 0 ) * normalize;
    result.g = ( ( c & 0x0000ff00 ) >> 8 ) * normalize;
    result.b = ( ( c & 0x00ff0000 ) >> 16 ) * normalize;
    result.a = ( ( c & 0xff000000 ) >> 24 ) * normalize;
    return result;
}

stu b32 v4_are_equal( vec4 a, vec4 b ) {
    b32 result = ( a.x == b.x ) && ( a.y == b.y ) && ( a.z == b.z ) && ( a.w == b.w );
    return result;
}

stu vec2 v2_negate( vec2 v ) {
    
    v.x = -v.x;
    v.y = -v.y;
    
    return v;
}

stu vec2 v2_add( vec2 a, vec2 b ) {
    
    a.x += b.x;
    a.y += b.y;
    
    return a;
}

stu vec2 v2_sub( vec2 a, vec2 b ) {
    
    a.x -= b.x;
    a.y -= b.y;
    
    return a;
}

stu vec2 v2_mul( vec2 v, r32 s ) {
    
    v.x *= s;
    v.y *= s;
    
    return v;
}

stu vec2 v2_div( vec2 v, r32 s ) {
    
    v.x /= s;
    v.y /= s;
    
    return v;
}

stu vec2 v2_hadamard( vec2 a, vec2 b ) {
    
    a.x *= b.x;
    a.y *= b.y;
    
    return a;
}

stu void v2_normalize( vec2* v ) {
    r32 length = sqrtf( v->x * v->x + v->y * v->y );
    v->x = v->x / length;
    v->y = v->y / length;
}

stu vec2 v2_normalized( vec2 v ) {
    vec2 result = v;
    v2_normalize( &result );
    return result;
}

stu void v2_normalize_fast( vec2* v ) {
    r32 rsqrt = math_rsqrt( v->x * v->x + v->y * v->y );
    v->x = v->x * rsqrt;
    v->y = v->y * rsqrt;
}

stu vec2 v2_normalized_fast( vec2 v ) {
    vec2 result = v;
    v2_normalize_fast( &result );
    return result;
}

stu r32 v2_length( vec2 v ) {
    
    r32 result = sqrtf( v.x * v.x + v.y * v.y );
    
    return result;
}

stu r32 v2_lengthSquared( vec2 v ) {
    
    r32 result = v.x * v.x + v.y * v.y;
    
    return result;
}

stu r32 v2_dot( vec2 a, vec2 b ) {
    
    r32 result = a.x * b.x + a.y * b.y;
    
    return result;
}

stu vec2 v2_perp( vec2 v ) {
    
    vec2 result;
    result.x = -v.y;
    result.y = v.x;
    
    return result;
}

stu vec2 v2_perp2( vec2 v, vec2 direction ) {
    
    vec2 result;
    result.x = -v.y;
    result.y = v.x;
    
    if ( v2_dot( result, direction ) < 0 ) {
        result.x = -result.x;
        result.y = -result.y;
    }
    
    return result;
}

stu b32 v2_equals( vec2 a, vec2 b, r32 epsilon ) {
    
    r32 x = ( a.x - b.x );
    r32 y = ( a.y - b.y );
    r32 e = epsilon * epsilon;
    b32 result = ( x * x  <= e ) && ( y * y <= e );
    
    return result;
}

/* NOTE simon: The 2d cross product between a and b is equal to the dot product of the perpendicular of a with b.
The result indicate if b ( 0 -> b ) is on the left (positive) or on the right (negative) of a ( 0 -> a ). */
stu r32 v2_cross( vec2 a, vec2 b ) {
    r32 result = -a.y * b.x + a.x * b.y;
    return result;
}

/* NOTE simon: The range test for the right part was changed from >= to >. I should test that is
all code... r2_contains does a similar thing so it make more sens the way it is now. */
stu b32 v2_contains( vec2 range, r32 p ) {
    b32 result = ( range.left <= p && range.right > p );
    return result;
}

stu b32 v2_contains_v2( vec2 contains, vec2 is_contained ) {
    b32 result = ( contains.left <= is_contained.left && contains.right > is_contained.right );
    return result;
}

stu b32 v2_overlap( vec2 a, r32 left, r32 right ) {
    b32 result = ( a.left < right && a.right > left );
    return result;
}

stu b32 v2_overlap_v2( vec2 a, vec2 b ) {
    b32 result = ( a.left < b.right && a.right > b.left );
    return result;
}

stu vec2 v2_lerp( vec2 a, vec2 b, r32 c ) {
    
    vec2 result = v2_add( v2_mul( a, 1.0f - c ), v2_mul( b, c ) );
    
    return result;
}

stu vec2 v2_min( vec2 a, vec2 b ) {
    
    vec2 result;
    result.x = ( ( a.x < b.x ) ? a.x : b.x );
    result.y = ( ( a.y < b.y ) ? a.y : b.y );
    
    return result;
}

stu vec2 v2_max( vec2 a, vec2 b ) {
    
    vec2 result;
    result.x = ( ( a.x > b.x ) ? a.x : b.x );
    result.y = ( ( a.y > b.y ) ? a.y : b.y );
    
    return result;
}

stu vec3 v3_sub( vec3 a, vec3 b ) {
    
    vec3 result;
    result.x = a.x - b.x;
    result.y = a.x - b.y;
    result.z = a.z - b.z;
    
    return result;
}

stu vec3 v3_normalized( vec3 v ) {
    
    r32 length = sqrtf( v.x * v.x + v.y * v.y + v.z * v.z );
    v.x = v.x / length;
    v.y = v.y / length;
    v.z = v.z / length;
    
    return v;
}

stu vec3 v3_normalized_fast( vec3 v ) {
    
    r32 rsqrt = math_rsqrt( v.x * v.x + v.y * v.y + v.z * v.z );
    v.x = v.x * rsqrt;
    v.y = v.y * rsqrt;
    v.z = v.z * rsqrt;
    
    return v;
}

stu vec3 v3_cross( vec3 a, vec3 b )  {
    
    vec3 result;
    result.x = a.y * b.z - a.z * b.y;
    result.y = a.z * b.x - a.x * b.z;
    result.z = a.x * b.y - a.y * b.x;
    
    return result;
}

stu vec4 v4_hadamard( vec4 a, vec4 b ) {
    
    a.x *= b.x;
    a.y *= b.y;
    a.z *= b.z;
    a.w *= b.w;
    
    return a;
}

stu vec4 v4_mul( vec4 a, r32 s ) {
    
    a.x *= s;
    a.y *= s;
    a.z *= s;
    a.w *= s;
    
    return a;
}

stu vec4 v4_lerp( vec4 a, vec4 b, r32 c ) {
    
    vec4 result = v4_add( v4_mul( a, 1.0f - c ), v4_mul( b, c ) );
    
    return result;
}

typedef union rect2 {
    
    struct {
        vec2 min;
        vec2 max;
    };
    
    struct {
        r32 left;
        r32 bottom;
        r32 right;
        r32 top;
    };
    
    r32 e[ 4 ];
} rect2;

stu void r2_zero( rect2* rect ) {
    
    for ( umm i = 0; i < array_count( rect->e ); i++ ) {
        rect->e[ i ] = 0.0f;
    }
}

stu rect2 r2_min_size( r32 x, r32 y, r32 w, r32 h ) {
    
    rect2 result;
    result.min.x = x;
    result.min.y = y;
    result.max.x = x + w;
    result.max.y = y + h;
    
    return result;
}

stu rect2 r2_min_size_v( vec2 min, vec2 size ) {
    
    rect2 result;
    result.min = min;
    result.max.x = min.x + size.x;
    result.max.y = min.y + size.y;
    
    return result;
}

stu rect2 r2_min_max( r32 minX, r32 minY, r32 maxX, r32 maxY ) {
    
    rect2 result;
    result.min.x = minX;
    result.min.y = minY;
    result.max.x = maxX;
    result.max.y = maxY;
    
    return result;
}

stu rect2 r2_min_max_v( vec2 min, vec2 max ) {
    
    rect2 result;
    result.min = min;
    result.max = max;
    
    return result;
}

stu rect2 r2_top_left_size( r32 left, r32 top, r32 width, r32 height ) {
    
    rect2 result;
    result.left = left;
    result.top = top;
    result.right = left + width;
    result.bottom = top - height;
    
    return result;
}

stu rect2 r2_center_size( r32 cx, r32 cy, r32 w, r32 h ) {
    
    rect2 result;
    result.min.x = cx - w * 0.5f;
    result.min.y = cy - h * 0.5f;
    result.max.x = cx + w * 0.5f;
    result.max.y = cy + h * 0.5f;
    
    return result;
}

stu rect2 r2_center_size_v( vec2 center, vec2 size ) {
    rect2 result = r2_center_size( center.x, center.y, size.x, size.y );
    return result;
}

stu rect2 r2_center_radius( r32 cx, r32 cy, r32 rw, r32 rh ) {
    
    rect2 result;
    result.min.x = cx - rw;
    result.min.y = cy - rh;
    result.max.x = cx + rw;
    result.max.y = cy + rh;
    
    return result;
}

stu rect2 r2_center_radius_v( vec2 center, vec2 radius ) {
    rect2 result = r2_center_radius( center.x, center.y, radius.x, radius.y );
    return result;
}

stu vec2 r2_get_size( rect2 rect ) {
    
    vec2 result = v2_sub( rect.max, rect.min );
    
    return result;
}

stu r32 r2_get_width( rect2 rect ) {
    
    r32 result = rect.max.x - rect.min.x;
    return result;
}

stu void r2_set_width( rect2* rect, r32 width ) {
    rect->max.x = rect->min.x + width;
}

stu void r2_add_width( rect2* rect, r32 width ) {
    rect->max.x += width;
}

stu r32 r2_get_height( rect2 rect ) {
    
    r32 result = rect.max.y - rect.min.y;
    return result;
}

stu void r2_set_height( rect2* rect, r32 height ) {
    rect->max.y = rect->min.y + height;
}

stu void r2_add_height( rect2* rect, r32 height ) {
    rect->max.y += height;
}

stu void r2_sub_radius( rect2* rect, r32 width, r32 height ) {
    rect->min.x += width;
    rect->min.y += height;
    rect->max.x -= width;
    rect->max.y -= height;
}

stu rect2 r2_copy_sub_radius( rect2 rect, r32 width, r32 height ) {
    rect2 result = rect;
    r2_sub_radius( &result, width, height );
    return result;
}

stu void r2_add_radius( rect2* rect, r32 width, r32 height ) {
    rect->min.x -= width;
    rect->min.y -= height;
    rect->max.x += width;
    rect->max.y += height;
}

stu rect2 r2_copy_add_radius( rect2 rect, r32 width, r32 height ) {
    rect2 result = rect;
    r2_add_radius( &result, width, height );
    return result;
}

stu void r2_add_border( rect2* rect, vec4 border ) {
    rect->min.x -= border.left;
    rect->min.y -= border.bottom;
    rect->max.x += border.right;
    rect->max.y += border.top;
}

stu rect2 r2_copy_add_border( rect2 rect, vec4 border ) {
    rect2 result = rect;
    r2_add_border( &result, border );
    return result;
}

stu void r2_sub_border( rect2* rect, vec4 border ) {
    rect->min.x += border.left;
    rect->min.y += border.bottom;
    rect->max.x -= border.right;
    rect->max.y -= border.top;
}

stu rect2 r2_copy_sub_border( rect2 rect, vec4 border ) {
    rect2 result = rect;
    r2_sub_border( &result, border );
    return result;
}

stu void r2_translate( rect2* rect, r32 x_offset, r32 y_offset ) {
    rect->min.x += x_offset;
    rect->max.x += x_offset;
    rect->min.y += y_offset;
    rect->max.y += y_offset;
}

stu rect2 r2_translated( rect2 rect, r32 x_offset, r32 y_offset ) {
    rect2 result = rect;
    r2_translate( &result, x_offset, y_offset );
    return result;
}

stu void r2_translate_x( rect2* rect, r32 x_offset ) {
    rect->min.x += x_offset;
    rect->max.x += x_offset;
}

stu void r2_translate_y( rect2* rect, r32 y_offset ) {
    rect->min.y += y_offset;
    rect->max.y += y_offset;
}

stu b32 r2_contains_v2( rect2 rect, vec2 point ) {
    
    b32 result = point.x >= rect.min.x && point.x < rect.max.x && point.y >= rect.min.y && point.y < rect.max.y;
    return result;
}

stu b32 r2_contains_r2( rect2 container, rect2 contained ) {
    
    b32 result = ( container.min.x <= contained.min.x
                  && container.max.x >= contained.max.x
                  && container.min.y <= contained.min.y
                  && container.max.y >= contained.max.y );
    return result;
}

stu b32 r2_overlap( rect2 a, rect2 b ) {
    
    b32 vertical = ( b.min.x < a.max.x && b.max.x > a.min.x );
    b32 horizontal = ( b.min.y < a.max.y && b.max.y > a.min.y );
    b32 result = vertical && horizontal;
    
    return result;
}

stu rect2 r2_intersection( rect2 a, rect2 b ) {
    
    rect2 result = { 0 };
    
    if ( r2_overlap( a, b ) ) {
        result.left = math_max( a.left, b.left );
        result.right = math_min( a.right, b.right );
        result.bottom = math_max( a.bottom, b.bottom );
        result.top = math_min( a.top, b.top );
    }
    
    return result;
}

stu rect2 r2_centered_size( rect2 container, vec2 size ) {
    r32 x_offset = ( ( container.right - container.left ) - size.x ) * 0.5f;
    r32 y_offset = ( ( container.top - container.bottom ) - size.y ) * 0.5f;
    rect2 result = r2_min_size( container.left + x_offset, container.bottom + y_offset, size.x, size.y );
    return result;
}

typedef union rect2i {
    
    struct {
        vec2i min;
        vec2i max;
    };
    
    struct {
        s32 left;
        s32 bottom;
        s32 right;
        s32 top;
    };
    
    s32 e[ 4 ];
    
} rect2i;

stu rect2i r2i_min_size( s32 x, s32 y, s32 width, s32 height ) {
    rect2i rect = { x, y, x + width, y + height };
    return rect;
}

stu b32 r2i_overlap( rect2i a, rect2i b ) {
    
    b32 horizontal = a.right > b.left && a.left < b.right;
    b32 vertical = a.top > b.bottom && a.bottom < b.top;
    b32 result = horizontal && vertical;
    
    return result;
}

stu rect2i r2i_intersection( rect2i a, rect2i b ) {
    
    rect2i result = { 0 };
    
    if ( r2i_overlap( a, b ) ) {
        result.left = math_max( a.left, b.left );
        result.right = math_min( a.right, b.right );
        result.bottom = math_max( a.bottom, b.bottom );
        result.top = math_min( a.top, b.top );
    }
    
    return result;
}

// MARK matrix

/* NOTE simon: OpenGL matrix should be column first in memory (column major).
OpenGL expect things to be transformed into Normalized Device Coordinates (NDC).
NDC a -1 to 1 left handed cube (z -1 is toward the viewer, z +1 is toward the screen).
But it can be confgured to use 0 to 1 on z using ARB_clip_control. */
/* TODO simon: look at https://www.khronos.org/registry/OpenGL/extensions/ARB/ARB_clip_control.txt */

/* NOTE simon: Matrix order is important. When calling multiply, you need to put the new
transformation in the lhs and the result of previous operation in the rhs.

For example to have a camera looking at X, from a point P and transforming into clip space:
- ( C ) transforming in clip space is the last transformation that needs to happen;
- ( T ) moving the camera at P is the first things that needs to happen;
- ( R ) rotating the camera to look at the X needs to happen after T;

so the desired order is TRC it should be executed in the oposite order CRT * P:
- mul( R, T, RT )
- mul( C, RT, CRT )
- mul( CRT, P )

Matrix multiplication is associative, (AB)C = A(BC). In our case it allows to combine
transformations together and multiply them with clip space matrix at the end.

Matrix A(m,n) : m rows, n colmn.
To be able to multiply two matrices, the number of column from the lhs matrix needs to be equal to
the number of row from the rhs matrix.

== Properties of matrix multiplication ==
In this table, A, B, and C are n*n matrices, I is the n*n identity matrix, and O is the n*n zero matrix

The commutative property of multiplication does not hold!
--- AB!=BA

Associative property of multiplication
--- (AB)C=A(BC)

Distributive properties
--- A(B+C)=AB+AC
--- (B+C)A=BA+CA

Multiplicative identity property
--- IA=A
--- AI=A

Multiplicative property of zero
--- OA=O
--- AO=O

Dimension property: the product of an m*n matrix and an n*k matrix is an m*k matrix.
*/

stu void matrix_create_identity3( r32* mat ) {
    mat[ 0 ] = 1.0f; mat[ 3 ] = 0.0f; mat[ 6 ] = 0.0f;
    mat[ 1 ] = 0.0f; mat[ 4 ] = 1.0f; mat[ 7 ] = 0.0f;
    mat[ 2 ] = 0.0f; mat[ 5 ] = 0.0f; mat[ 8 ] = 1.0f;
}

stu void matrix_create_identity( r32* mat ) {
    
    mat[ 0 ] = 1.0f; mat[ 4 ] = 0.0f; mat[ 8 ] = 0.0f; mat[ 12 ] = 0.0f;
    mat[ 1 ] = 0.0f; mat[ 5 ] = 1.0f; mat[ 9 ] = 0.0f; mat[ 13 ] = 0.0f;
    mat[ 2 ] = 0.0f; mat[ 6 ] = 0.0f; mat[ 10 ] = 1.0f; mat[ 14 ] = 0.0f;
    mat[ 3 ] = 0.0f; mat[ 7 ] = 0.0f; mat[ 11 ] = 0.0f; mat[ 15 ] = 1.0f;
}

stu void matrix_create_screen_space3( r32* mat, s32 screenWidth, s32 screenHeight ) {
    
    mat[ 0 ] = 2.0f / screenWidth;
    mat[ 1 ] = 0.0f;
    mat[ 2 ] = -1.0f;
    
    mat[ 3 ] = 0.0f;
    mat[ 4 ] = 2.0f / screenHeight;
    mat[ 5 ] = -1.0f;
    
    mat[ 6 ] = 0.0f;
    mat[ 7 ] = 0.0f;
    mat[ 8 ] = 1.0f;
}

stu void matrix_create_screen_space( r32* mat, s32 screen_width, s32 screen_height ) {
    
    mat[ 0 ] = 2.0f / screen_width;
    mat[ 1 ] = 0.0f;
    mat[ 2 ] = 0.0f;
    mat[ 3 ] = 0.0f;
    
    mat[ 4 ] = 0.0f;
    mat[ 5 ] = 2.0f / screen_height;
    mat[ 6 ] = 0.0f;
    mat[ 7 ] = 0.0f;
    
    mat[ 8 ] = 0.0f;
    mat[ 9 ] = 0.0f;
    mat[ 10 ] = 1.0f;
    mat[ 11 ] = 0.0f;
    
    mat[ 12 ] = -1.0f;
    mat[ 13 ] = -1.0f;
    mat[ 14 ] = 0.0f;
    mat[ 15 ] = 1.0f;
}

stu void matrix_create_screen_space_top_down3( r32* mat, s32 screenWidth, s32 screenHeight ) {
    
    mat[ 0 ] = 2.0f / screenWidth;
    mat[ 1 ] = 0.0f;
    mat[ 2 ] = -1.0f;
    
    mat[ 3 ] = 0.0f;
    mat[ 4 ] = -2.0f / screenHeight;
    mat[ 5 ] = 1.0f;
    
    mat[ 6 ] = 0.0f;
    mat[ 7 ] = 0.0f;
    mat[ 8 ] = 1.0f;
}

stu void matrix_create_screen_space_top_down( r32* mat, s32 screenWidth, s32 screenHeight ) {
    
    mat[ 0 ] = 2.0f / screenWidth;
    mat[ 1 ] = 0.0f;
    mat[ 2 ] = 0.0f;
    mat[ 3 ] = -1.0f;
    
    mat[ 4 ] = 0.0f;
    mat[ 5 ] = -2.0f / screenHeight;
    mat[ 6 ] = 0.0f;
    mat[ 7 ] = 1.0f;
    
    mat[ 8 ] = 0.0f;
    mat[ 9 ] = 0.0f;
    mat[ 10 ] = 1.0f;
    mat[ 11 ] = 0.0f;
    
    mat[ 12 ] = 0.0f;
    mat[ 13 ] = 0.0f;
    mat[ 14 ] = 0.0f;
    mat[ 15 ] = 1.0f;
}

stu void matrix_projection_orthographic( void* matrix, s32 width, s32 height, r32 near_plane, r32 far_plane ) {
    
    r32* mat = ( r32* ) matrix;
    
    mat[ 0 ] = 2.0f / width;
    mat[ 1 ] = 0.0f;
    mat[ 2 ] = 0.0f;
    mat[ 3 ] = 0.0f;
    
    mat[ 4 ] = 0.0f;
    mat[ 5 ] = 2.0f / height;
    mat[ 6 ] = 0.0f;
    mat[ 7 ] = 0.0f;
    
    mat[ 8 ] = 0.0f;
    mat[ 9 ] = 0.0f;
    mat[ 10 ] = -2.0f / ( far_plane - near_plane );
    mat[ 11 ] = 0.0f;
    
    mat[ 12 ] = 0.0f;
    mat[ 13 ] = 0.0f;
    mat[ 14 ] = -( ( far_plane + near_plane ) / ( far_plane - near_plane ) );
    mat[ 15 ] = 1.0f;
}

/* NOTE simon: look at https://developer.nvidia.com/content/depth-precision-visualized */
/* NOTE simon: http://www.humus.name/index.php?ID=255 */
/* TODO simon: use ARB_clip_control, floating point depth buffer and reverse Z order ( 1 is near and 0 is far ). */
stu void matrix_projection_perspective( void* matrix, r32 horizontal_fov, r32 aspect_ratio, r32 near_plane, r32 far_plane ) {
    
    r32* mat = ( r32* ) matrix;
    r32 width = tanf( horizontal_fov * 0.5f );
    
    mat[ 0 ] = 1 / width;
    mat[ 1 ] = 0.0f;
    mat[ 2 ] = 0.0f;
    mat[ 3 ] = 0.0f;
    
    mat[ 4 ] = 0.0f;
    mat[ 5 ] = aspect_ratio / width;
    mat[ 6 ] = 0.0f;
    mat[ 7 ] = 0.0f;
    
    mat[ 8 ] = 0.0f;
    mat[ 9 ] = 0.0f;
    mat[ 10 ] = - ( far_plane + near_plane ) / ( far_plane - near_plane );
    mat[ 11 ] = -1.0f;
    
    mat[ 12 ] = 0.0f;
    mat[ 13 ] = 0.0f;
    mat[ 14 ] = -( 2 * far_plane * near_plane ) / ( far_plane - near_plane );
    mat[ 15 ] = 0.0f;
}

stu void matrix_create_translation3( r32* mat, r32 x, r32 y ) {
    
    mat[ 0 ] = 1.0f;
    mat[ 1 ] = 0.0f;
    mat[ 2 ] = x;
    
    mat[ 3 ] = 0.0f;
    mat[ 4 ] = 1.0f;
    mat[ 5 ] = y;
    
    mat[ 6 ] = 0.0f;
    mat[ 7 ] = 0.0f;
    mat[ 8 ] = 1.0f;
}

stu void matrix_create_translation3_vec2( r32* mat, r32* v ) {
    matrix_create_translation3( mat, v[ 0 ], v[ 1 ] );
}

stu void matrix_create_translation( void* matrix, r32 x, r32 y, r32 z ) {
    
    r32* mat = ( r32* ) matrix;
    
    mat[ 0 ] = 1.0f;
    mat[ 1 ] = 0.0f;
    mat[ 2 ] = 0.0f;
    mat[ 3 ] = 0.0f;
    
    mat[ 4 ] = 0.0f;
    mat[ 5 ] = 1.0f;
    mat[ 6 ] = 0.0f;
    mat[ 7 ] = 0.0f;
    
    mat[ 8 ] = 0.0f;
    mat[ 9 ] = 0.0f;
    mat[ 10 ] = 1.0f;
    mat[ 11 ] = 0.0f;
    
    mat[ 12 ] = x;
    mat[ 13 ] = y;
    mat[ 14 ] = z;
    mat[ 15 ] = 1.0f;
}

stu void matrix_create_translation_vec3( r32* mat, r32* v ) {
    matrix_create_translation( mat, v[ 0 ], v[ 1 ], v[ 2 ] );
}

stu void matrix_create_scale3( r32* mat, r32 x, r32 y ) {
    
    mat[ 0 ] = x;
    mat[ 1 ] = 0.0f;
    mat[ 2 ] = 0.0f;
    
    mat[ 3 ] = 0.0f;
    mat[ 4 ] = y;
    mat[ 5 ] = 0.0f;
    
    mat[ 6 ] = 0.0f;
    mat[ 7 ] = 0.0f;
    mat[ 8 ] = 1.0f;
}

stu void matrix_create_scale3_vec2( r32* mat, r32* v ) {
    matrix_create_scale3( mat, v[ 0 ], v[ 1 ] );
}

stu void matrix_create_scale( r32* mat, r32 x, r32 y, r32 z ) {
    
    mat[ 0 ] = x;
    mat[ 1 ] = 0.0f;
    mat[ 2 ] = 0.0f;
    mat[ 3 ] = 0.0f;
    
    mat[ 4 ] = 0.0f;
    mat[ 5 ] = y;
    mat[ 6 ] = 0.0f;
    mat[ 7 ] = 0.0f;
    
    mat[ 8 ] = 0.0f;
    mat[ 9 ] = 0.0f;
    mat[ 10 ] = z;
    mat[ 11 ] = 0.0f;
    
    mat[ 12 ] = 0.0f;
    mat[ 13 ] = 0.0f;
    mat[ 14 ] = 0.0f;
    mat[ 15 ] = 1.0f;
}

stu void matrix_create_scale_vec3( r32* mat, r32* v ) {
    matrix_create_scale( mat, v[ 0 ], v[ 1 ], v[ 2 ] );
}

stu void matrix_create( r32* mat, vec3 x, vec3 y, vec3 z, vec3 t ) {
    
    mat[ 0 ] = x.x;
    mat[ 1 ] = y.x;
    mat[ 2 ] = z.x;
    mat[ 3 ] = t.x;
    
    mat[ 4 ] = x.y;
    mat[ 5 ] = y.y;
    mat[ 6 ] = z.y;
    mat[ 7 ] = t.y;
    
    mat[ 8 ] = x.z;
    mat[ 9 ] = x.z;
    mat[ 10 ] = x.z;
    mat[ 11 ] = t.z;
    
    mat[ 12 ] = 0.0f;
    mat[ 13 ] = 0.0f;
    mat[ 14 ] = 0.0f;
    mat[ 15 ] = 1.0f;
}

stu void matrix_look_at( r32* mat, vec3 position, vec3 target, vec3 up_vector ) {
    
    vec3 z = v3_normalized( v3_sub( target, position ) );
    vec3 x = v3_normalized( v3_cross( up_vector, z ) );
    vec3 y = v3_cross( z, x );
    
    matrix_create( mat, x, y, z, v3( 0, 0, 0 ) );
}

stu void matrix_rotation_x( r32* mat, r32 degree ) {
    
    r32 radian = degree * math_degree_to_radian;
    r32 c = cosf( radian );
    r32 s = sinf( radian );
    
    mat[ 0 ] = 1; mat[ 4 ] = 0; mat[ 8 ] = 0; mat[ 12 ] = 0;
    mat[ 1 ] = 0; mat[ 5 ] = c; mat[ 9 ] = -s; mat[ 13 ] = 0;
    mat[ 2 ] = 0; mat[ 6 ] = s; mat[ 10 ] = c; mat[ 14 ] = 0;
    mat[ 3 ] = 0; mat[ 7 ] = 0; mat[ 11 ] = 0; mat[ 15 ] = 1;
}

stu void matrix_rotation_y( r32* mat, r32 degree ) {
    
    r32 radian = degree * math_degree_to_radian;
    r32 c = cosf( radian );
    r32 s = sinf( radian );
    
    mat[ 0 ] = c; mat[ 4 ] = 0; mat[ 8 ] = s; mat[ 12 ] = 0;
    mat[ 1 ] = 0; mat[ 5 ] = 1; mat[ 9 ] = 0; mat[ 13 ] = 0;
    mat[ 2 ] = -s; mat[ 6 ] = 0; mat[ 10 ] = c; mat[ 14 ] = 0;
    mat[ 3 ] = 0; mat[ 7 ] = 0; mat[ 11 ] = 0; mat[ 15 ] = 1;
}

stu void matrix_rotation_z( r32* mat, r32 degree ) {
    
    r32 radian = degree * math_degree_to_radian;
    r32 c = cosf( radian );
    r32 s = sinf( radian );
    
    mat[ 0 ] = c; mat[ 4 ] = -s; mat[ 8 ] = 0; mat[ 12 ] = 0;
    mat[ 1 ] = s; mat[ 5 ] = c; mat[ 9 ] = 0; mat[ 13 ] = 0;
    mat[ 2 ] = 0; mat[ 6 ] = 0; mat[ 10 ] = 1; mat[ 14 ] = 0;
    mat[ 3 ] = 0; mat[ 7 ] = 0; mat[ 11 ] = 0; mat[ 15 ] = 1;
}

stu void matrix_multiply_square( umm dimention, r32* lhs, r32* rhs, r32* result ) {
    
#define lhs( x, y ) ( lhs[ x * dimention + y ] )
#define rhs( x, y ) ( rhs[ x * dimention + y ] )
#define r( x, y ) ( result[ x * dimention + y ] )
    
    
    for ( umm x = 0; x < dimention; x++ ) {
        for ( umm y = 0; y < dimention; y++ ) {
            
            r( x, y ) = 0;
            
            for ( umm i = 0; i < dimention; i++ ) {
                r( x, y ) += lhs( i, y ) * rhs( x, i );
            }
        }
    }
    
#undef lhs
#undef rhs
#undef r
}

stu void matrix_multiply3( r32* lhs, r32* rhs, r32* result ) {
    matrix_multiply_square( 3, lhs, rhs, result );
}

stu void matrix_multiply( r32* lhs, r32* rhs, r32* result ) {
    matrix_multiply_square( 4, lhs, rhs, result );
}

stu void matrix_multiply4x4( r32* lhs, r32* rhs, r32* result ) {
    
#define lhs( y, x ) ( lhs[ y * 4 + x ] )
#define rhs( y, x ) ( rhs[ y * 4 + x ] )
    result[ 0 ] = lhs( 0, 0 ) * rhs( 0, 0 ) + lhs( 1, 0 ) * rhs( 0, 1 ) + lhs( 2, 0 ) * rhs( 0, 2 ) + lhs( 3, 0 ) * rhs( 0, 3 );
    result[ 1 ] = lhs( 0, 1 ) * rhs( 0, 0 ) + lhs( 1, 1 ) * rhs( 0, 1 ) + lhs( 2, 1 ) * rhs( 0, 2 ) + lhs( 3, 1 ) * rhs( 0, 3 );
    result[ 2 ] = lhs( 0, 2 ) * rhs( 0, 0 ) + lhs( 1, 2 ) * rhs( 0, 1 ) + lhs( 2, 2 ) * rhs( 0, 2 ) + lhs( 3, 2 ) * rhs( 0, 3 );
    result[ 3 ] = lhs( 0, 3 ) * rhs( 0, 0 ) + lhs( 1, 3 ) * rhs( 0, 1 ) + lhs( 2, 3 ) * rhs( 0, 2 ) + lhs( 3, 3 ) * rhs( 0, 3 );
    
    result[ 4 ] = lhs( 0, 0 ) * rhs( 1, 0 ) + lhs( 1, 0 ) * rhs( 1, 1 ) + lhs( 2, 0 ) * rhs( 1, 2 ) + lhs( 3, 0 ) * rhs( 1, 3 );
    result[ 5 ] = lhs( 0, 1 ) * rhs( 1, 0 ) + lhs( 1, 1 ) * rhs( 1, 1 ) + lhs( 2, 1 ) * rhs( 1, 2 ) + lhs( 3, 1 ) * rhs( 1, 3 );
    result[ 6 ] = lhs( 0, 2 ) * rhs( 1, 0 ) + lhs( 1, 2 ) * rhs( 1, 1 ) + lhs( 2, 2 ) * rhs( 1, 2 ) + lhs( 3, 2 ) * rhs( 1, 3 );
    result[ 7 ] = lhs( 0, 3 ) * rhs( 1, 0 ) + lhs( 1, 3 ) * rhs( 1, 1 ) + lhs( 2, 3 ) * rhs( 1, 2 ) + lhs( 3, 3 ) * rhs( 1, 3 );
    
    result[ 8 ] = lhs( 0, 0 ) * rhs( 2, 0 ) + lhs( 1, 0 ) * rhs( 2, 1 ) + lhs( 2, 0 ) * rhs( 2, 2 ) + lhs( 3, 0 ) * rhs( 2, 3 );
    result[ 9 ] = lhs( 0, 1 ) * rhs( 2, 0 ) + lhs( 1, 1 ) * rhs( 2, 1 ) + lhs( 2, 1 ) * rhs( 2, 2 ) + lhs( 3, 1 ) * rhs( 2, 3 );
    result[ 10 ] = lhs( 0, 2 ) * rhs( 2, 0 ) + lhs( 1, 2 ) * rhs( 2, 1 ) + lhs( 2, 2 ) * rhs( 2, 2 ) + lhs( 3, 2 ) * rhs( 2, 3 );
    result[ 11 ] = lhs( 0, 3 ) * rhs( 2, 0 ) + lhs( 1, 3 ) * rhs( 2, 1 ) + lhs( 2, 3 ) * rhs( 2, 2 ) + lhs( 3, 3 ) * rhs( 2, 3 );
    
    result[ 12 ] = lhs( 0, 0 ) * rhs( 3, 0 ) + lhs( 1, 0 ) * rhs( 3, 1 ) + lhs( 2, 0 ) * rhs( 3, 2 ) + lhs( 3, 0 ) * rhs( 3, 3 );
    result[ 13 ] = lhs( 0, 1 ) * rhs( 3, 0 ) + lhs( 1, 1 ) * rhs( 3, 1 ) + lhs( 2, 1 ) * rhs( 3, 2 ) + lhs( 3, 1 ) * rhs( 3, 3 );
    result[ 14 ] = lhs( 0, 2 ) * rhs( 3, 0 ) + lhs( 1, 2 ) * rhs( 3, 1 ) + lhs( 2, 2 ) * rhs( 3, 2 ) + lhs( 3, 2 ) * rhs( 3, 3 );
    result[ 15 ] = lhs( 0, 3 ) * rhs( 3, 0 ) + lhs( 1, 3 ) * rhs( 3, 1 ) + lhs( 2, 3 ) * rhs( 3, 2 ) + lhs( 3, 3 ) * rhs( 3, 3 );
#undef lhs
#undef rhs
}

stu void matrix_transpose3( r32* a ) {
#define matrix_a( x, y ) ( a[ x * 3 + y ] )
    swap_type( matrix_a( 1, 0 ), matrix_a( 0, 1 ), r32 );
    swap_type( matrix_a( 2, 0 ), matrix_a( 0, 2 ), r32 );
    swap_type( matrix_a( 2, 1 ), matrix_a( 1, 2 ), r32 );
#undef matrix_a
}

stu void matrix_transpose( r32* a ) {
#define matrix_a( x, y ) ( a[ x * 4 + y ] )
    swap_type( matrix_a( 1, 0 ), matrix_a( 0, 1 ), r32 );
    swap_type( matrix_a( 2, 0 ), matrix_a( 0, 2 ), r32 );
    swap_type( matrix_a( 3, 0 ), matrix_a( 0, 3 ), r32 );
    swap_type( matrix_a( 2, 1 ), matrix_a( 1, 2 ), r32 );
    swap_type( matrix_a( 3, 1 ), matrix_a( 1, 3 ), r32 );
    swap_type( matrix_a( 3, 2 ), matrix_a( 2, 3 ), r32 );
#undef matrix_a
}

stu vec2 m4_mul_v2( r32* lhs, r32* rhs ) {
    vec2 result;
    result.x = lhs[ 0 ] * rhs[ 0 ] + lhs[ 4 ] * rhs[ 1 ] + lhs[ 12 ];
    result.y = lhs[ 1 ] * rhs[ 0 ] + lhs[ 5 ] * rhs[ 1 ] + lhs[ 13 ];
    return result;
}

stu vec3 m4_mul_v3( r32* lhs, r32* rhs ) {
    vec3 result;
    result.x = lhs[ 0 ] * rhs[ 0 ] + lhs[ 4 ] * rhs[ 1 ] + lhs[ 8 ] * rhs[ 2 ] + lhs[ 12 ];
    result.y = lhs[ 1 ] * rhs[ 0 ] + lhs[ 5 ] * rhs[ 1 ] + lhs[ 9 ] * rhs[ 2 ] + lhs[ 13 ];
    result.z = lhs[ 2 ] * rhs[ 0 ] + lhs[ 6 ] * rhs[ 1 ] + lhs[ 10 ] * rhs[ 2 ] + lhs[ 14 ];
    return result;
}

stu vec4 m4_mul_v4( r32* lhs, r32* rhs ) {
    vec4 result;
    result.x = lhs[ 0 ] * rhs[ 0 ] + lhs[ 4 ] * rhs[ 1 ] + lhs[ 8 ] * rhs[ 2 ] + lhs[ 12 ] * rhs[ 3 ];
    result.y = lhs[ 1 ] * rhs[ 0 ] + lhs[ 5 ] * rhs[ 1 ] + lhs[ 9 ] * rhs[ 2 ] + lhs[ 13 ] * rhs[ 3 ];
    result.z = lhs[ 2 ] * rhs[ 0 ] + lhs[ 6 ] * rhs[ 1 ] + lhs[ 10 ] * rhs[ 2 ] + lhs[ 14 ] * rhs[ 3 ];
    result.w = lhs[ 3 ] * rhs[ 0 ] + lhs[ 7 ] * rhs[ 1 ] + lhs[ 11 ] * rhs[ 2 ] + lhs[ 15 ] * rhs[ 3 ];
    return result;
}

stu void matrix_transform_xy_mat3( r32* x, r32* y, r32* m ) {
    r32 xv = *x, yv = *y;
    ( *x ) = xv * m[ 0 ] + yv * m[ 1 ] + m[ 2 ];
    ( *y ) = xv * m[ 3 ] + yv * m[ 4 ] + m[ 5 ];
}

stu void matrix_transform_vec2_mat3( r32* v, r32* m ) {
    matrix_transform_xy_mat3( v, v + 1, m );
}

stu void matrix_multiply_vec2_mat3( r32* v, r32* m, r32* result ) {
    result[ 0 ] = v[ 0 ] * m[ 0 ] + v[ 1 ] * m[ 1 ] + m[ 2 ];
    result[ 1 ] = v[ 0 ] * m[ 3 ] + v[ 1 ] * m[ 4 ] + m[ 5 ];
}

stu void matrix_transform_xyz_mat3( r32* x, r32* y, r32* z, r32* m ) {
    r32 xv = *x, yv = *y, zv = *z;
    ( *x ) = xv * m[ 0 ] + yv * m[ 1 ] + zv * m[ 2 ];
    ( *y ) = xv * m[ 3 ] + yv * m[ 4 ] + zv * m[ 5 ];
    ( *z ) = xv * m[ 6 ] + yv * m[ 7 ] + zv * m[ 8 ];
}

stu void matrix_transform_vec3_mat3( r32* v, r32* m ) {
    matrix_transform_xyz_mat3( v, v + 1, v + 2, m );
}

stu void matrix_multiply_vec3_mat3( r32* v, r32* m, r32* result ) {
    result[ 0 ] = v[ 0 ] * m[ 0 ] + v[ 1 ] * m[ 1 ] + v[ 2 ] * m[ 2 ];
    result[ 1 ] = v[ 0 ] * m[ 3 ] + v[ 1 ] * m[ 4 ] + v[ 2 ] * m[ 5 ];
    result[ 2 ] = v[ 0 ] * m[ 6 ] + v[ 1 ] * m[ 7 ] + v[ 2 ] * m[ 8 ];
}

stu void matrix_transform_xyz( r32* x, r32* y, r32* z, r32* m ) {
    r32 xv = *x, yv = *y, zv = *z;
    ( *x ) = xv * m[ 0 ] + yv * m[ 1 ] + zv * m[ 2 ] + m[ 3 ];
    ( *y ) = xv * m[ 4 ] + yv * m[ 5 ] + zv * m[ 6 ] + m[ 7 ];
    ( *z ) = xv * m[ 8 ] + yv * m[ 9 ] + zv * m[ 10 ] + m[ 11 ];
}

stu void matrix_transform_vec3( r32* v, r32* m ) {
    matrix_transform_xyz( v, v + 1, v + 2, m );
}

stu void matrix_multiply_vec3( r32* v, r32* m, r32* result ) {
    result[ 0 ] = v[ 0 ] * m[ 0 ] + v[ 1 ] * m[ 1 ] + v[ 2 ] * m[ 2 ] + m[ 3 ];
    result[ 1 ] = v[ 0 ] * m[ 4 ] + v[ 1 ] * m[ 5 ] + v[ 2 ] * m[ 6 ] + m[ 7 ];
    result[ 2 ] = v[ 0 ] * m[ 8 ] + v[ 1 ] * m[ 9 ] + v[ 2 ] * m[ 10 ] + m[ 11 ];
}

stu void matrix_transform_xyzw( r32* x, r32* y, r32* z, r32* w, r32* m ) {
    r32 xv = *x, yv = *y, zv = *z, wv = *w;
    ( *x ) = xv * m[ 0 ] + yv * m[ 1 ] + zv * m[ 2 ] + wv * m[ 3 ];
    ( *y ) = xv * m[ 4 ] + yv * m[ 5 ] + zv * m[ 6 ] + wv * m[ 7 ];
    ( *z ) = xv * m[ 8 ] + yv * m[ 9 ] + zv * m[ 10 ] + wv * m[ 11 ];
    ( *w ) = xv * m[ 12 ] + yv * m[ 13 ] + zv * m[ 14 ] + wv * m[ 15 ];
}

stu void matrix_transform_vec4( r32* v, r32* m ) {
    matrix_transform_xyzw( v, v + 1, v + 2, v + 3, m );
}

stu void matrix_multiply_vec4( r32* matrix_lhs, r32* vector_rhs, r32* result ) {
    
    r32* m = matrix_lhs;
    r32* v = vector_rhs;
    
    result[ 0 ] = m[ 0 ] * v[ 0 ] + m[ 4 ] * v[ 1 ] + m[ 8 ] * v[ 2 ] + m[ 12 ] * v[ 3 ];
    result[ 1 ] = m[ 1 ] * v[ 0 ] + m[ 5 ] * v[ 1 ] + m[ 9 ] * v[ 2 ] + m[ 13 ] * v[ 3 ];
    result[ 2 ] = m[ 2 ] * v[ 0 ] + m[ 6 ] * v[ 1 ] + m[ 10 ] * v[ 2 ] + m[ 14 ] * v[ 3 ];
    result[ 3 ] = m[ 3 ] * v[ 0 ] + m[ 7 ] * v[ 1 ] + m[ 11 ] * v[ 2 ] + m[ 15 ] * v[ 3 ];
}

stu b32 matrix_areEqual( r32* a, r32* b, umm elementCount ) {
    
    b32 result = true;
    for ( umm i = 0; result && i < elementCount; i++ ) {
        result = a[ i ] == b[ i ];
    }
    
    return result;
}

stu b32 matrix_areEqualSquare( r32* a, r32* b, umm dimention ) {
    
    b32 result = matrix_areEqual( a, b, dimention * dimention );
    return result;
}

stu void matrix_copy( r32* dest, r32* source, umm elementCount ) {
    for ( umm i = 0; i < elementCount; i++ ) {
        dest[ i ] = source[ i ];
    }
}

stu void matrix_copySquare( r32* dest, r32* source, umm dimention ) {
    matrix_copy( dest, source, dimention * dimention );
}

// MARK geometry

/* NOTE simon: Area is negative if the polygon is counter clock wise. */
stu r32 geometry_polygonArea( vec2* points, umm pointCount ) {
    
    r32 area = 0.0f;
    umm pointIndex;
    
    for ( pointIndex = 0; pointIndex < pointCount; pointIndex++ ) {
        
        vec2 a, b;
        umm secondIndex = pointIndex + 1;
        
        if ( secondIndex == pointCount ) {
            secondIndex = 0;
        }
        
        a = points[ pointIndex ];
        b = points[ secondIndex ];
        
        area += ( b.x - a.x ) * ( b.y + a.y );
        /* NOTE: alternative:
        area += ( a.x * b.y ) - ( b.x * a.y );
        but this one would be negative for clockwise polygons.
        */
    }
    
    return area / 2.0f;
}

stu vec2 geometry_findCentroid( vec2* points, umm pointCount ) {
    
    vec2 result = { 0 };
    r32 area = 0.0f;
    r32 _6A;
    umm index;
    
    for ( index = 0; index < pointCount; index++ ) {
        
        vec2 a, b;
        r32 aPart;
        umm secondIndex = index + 1;
        
        if ( secondIndex == pointCount ) {
            secondIndex = 0;
        }
        
        a = points[ index ];
        b = points[ secondIndex ];
        
        aPart = ( a.x * b.y ) - ( a.y * b.x );
        area += aPart;
        result.x += ( a.x + b.x ) * aPart;
        result.y += ( a.y + b.y ) * aPart;
    }
    
    _6A = 3 * area;
    result.x /= _6A;
    result.y /= _6A;
    
    return result;
}

stu vec2 geometry_findTriangleCentroid( vec2* points )  {
    
    vec2 result = { 0 };
    
    for ( umm i = 0; i < 3; i++ ) {
        result = v2_add( result, points[ i ] );
    }
    
    result = v2_div( result, 3 );
    
    return result;
}

/* NOTE simon:
- 0 colinear
- less then 0 if angle is CCW
- more then 0 if angle is CW */
stu r32 geometry_angleOrientation( vec2 a, vec2 b, vec2 c ) {
    
    r32 result = ( b.y - a.y ) * ( c.x - b.x ) - ( c.y - b.y ) * ( b.x - a.x );
    return result;
}

stu b32 geometry_isColinearPointOnSegment( vec2 a, vec2 b, vec2 p ) {
    
    vec2 bRel = v2_sub( b, a );
    vec2 pRel = v2_sub( p, a );
    r32 one = bRel.x * bRel.x + bRel.y * bRel.y;
    r32 two = pRel.x * pRel.x + pRel.y * pRel.y;
    
    b32 result = two >= 0.0f && two < one;
    return result;
}

stu b32 geometry_areSegmentInteresecting( vec2 aStart, vec2 aEnd, vec2 bStart, vec2 bEnd ) {
    
    r32 A1 = math_sign_or_zero( geometry_angleOrientation( aStart, aEnd, bStart ) );
    r32 A2 = math_sign_or_zero( geometry_angleOrientation( aStart, aEnd, bEnd ) );
    r32 B1 = math_sign_or_zero( geometry_angleOrientation( bStart, bEnd, aStart ) );
    r32 B2 = math_sign_or_zero( geometry_angleOrientation( bStart, bEnd, aEnd ) );
    
    b32 result =
        ( A1 != A2 && B1 != B2 )
        || ( A1 == 0.0f && geometry_isColinearPointOnSegment( aStart, aEnd, bStart ) )
        || ( A2 == 0.0f && geometry_isColinearPointOnSegment( aStart, aEnd, bEnd ) )
        || ( B1 == 0.0f && geometry_isColinearPointOnSegment( bStart, bEnd, aStart ) )
        || ( B2 == 0.0f && geometry_isColinearPointOnSegment( bStart, bEnd, aEnd ) );
    
    return result;
}

/* NOTE simon: the return value is the interpolation factor between bStart (f = 0) and bEnd (f=1). */
stu r32 geometry_intersectionPoint( vec2 aStart, vec2 aEnd, vec2 bStart, vec2 bEnd ) {
    
    vec2 a = v2_sub( aEnd, aStart );
    
    vec2 up = v2_sub( bStart, aStart );
    vec2 down = v2_sub( bEnd, aStart );
    
    r32 fUp = fabsf( v2_cross( a, up ) );
    r32 fDown = fabsf( v2_cross( a, down ) );
    r32 f = fUp / ( fUp + fDown );
    
    return f;
}

stu b32 geometry_isPolygonValid( vec2* points, u32 pointCount ) {
    
    b32 result = false;
    umm checkingIndex = 0;
    
    while ( !result && checkingIndex < pointCount - 1 ) {
        
        umm currentIndex = checkingIndex + 1;
        vec2 aStart = points[ checkingIndex ];
        vec2 aEnd = points[ checkingIndex + 1 ];
        
        while ( !result && currentIndex < pointCount ) {
            
            vec2 bStart, bEnd;
            r32 A1, A2, B1, B2;
            umm secondIndex = currentIndex + 1;
            
            if ( secondIndex == pointCount ) {
                secondIndex = 0;
            }
            
            bStart = points[ currentIndex ];
            bEnd = points[ secondIndex ];
            
            A1 = math_sign_or_zero( geometry_angleOrientation( aStart, aEnd, bStart ) );
            A2 = math_sign_or_zero( geometry_angleOrientation( aStart, aEnd, bEnd ) );
            B1 = math_sign_or_zero( geometry_angleOrientation( bStart, bEnd, aStart ) );
            B2 = math_sign_or_zero( geometry_angleOrientation( bStart, bEnd, aEnd ) );
            
            if ( currentIndex == checkingIndex + 1 ) { /* NOTE simon: aEnd == bStart */
                result = ( A2 == 0.0f && geometry_isColinearPointOnSegment( aStart, aEnd, bEnd ) );
            } else if ( checkingIndex == 0 && secondIndex == 0 ) { /* NOTE simon: aStart == bEnd */
                result = ( A1 == 0.0f && geometry_isColinearPointOnSegment( aStart, aEnd, bStart ) );
            } else {
                result = ( A1 != A2 && B1 != B2 );
                result = result || ( A1 == 0.0f && geometry_isColinearPointOnSegment( aStart, aEnd, bStart ) );
                result = result || ( B1 == 0.0f && geometry_isColinearPointOnSegment( bStart, bEnd, aStart ) );
                result = result || ( A2 == 0.0f && geometry_isColinearPointOnSegment( aStart, aEnd, bEnd ) );
                result = result || ( B2 == 0.0f && geometry_isColinearPointOnSegment( bStart, bEnd, aEnd ) );
            }
            
            currentIndex++;
        }
        
        checkingIndex++;
    }
    
    return result == false;
}

stu b32 geometry_isPolygonCCW( vec2* points, umm pointCount ) {
    
    r32 sum = 0;
    umm pointIndex;
    
    for ( pointIndex = 0; pointIndex < pointCount; pointIndex++ ) {
        
        r32 x0, x1, y0, y1;
        umm secondIndex = pointIndex + 1;
        
        if ( pointIndex == pointCount - 1 ) {
            secondIndex = 0;
        }
        
        x0 = points[ pointIndex ].x;
        x1 = points[ secondIndex ].x;
        y0 = points[ pointIndex ].y;
        y1 = points[ secondIndex ].y;
        
        sum += ( x1 - x0 ) * ( y1 + y0 );
    }
    
    return sum < 0;
}

stu void geometry_inversePolygonWinding( vec2* points, u32 pointCount ) {
    
    umm pointIndex;
    
    for ( pointIndex = 0; pointIndex < pointCount / 2; pointIndex++ ) {
        
        umm secondIndex = pointCount - pointIndex - 1;
        vec2 temp = points[ pointIndex ];
        points[ pointIndex ] = points[ secondIndex ];
        points[ secondIndex ] = temp;
    }
}

/* NOTE simon: http://lab.polygonal.de/2006/08/17/calculating-the-moment-of-inertia-of-a-convex-polygon/ */
/* NOTE simon: I have no idea if those function are correct. */
stu r32 geometry_findTriangleMomentOfInertia( vec2* points )  {
    
    b32 isCCW = geometry_isPolygonCCW( points, 3 );
    vec2 A = points[ 0 ];
    vec2 B = points[ 1 ];
    vec2 C = points[ 2 ];
    
    if ( !isCCW )  {
        vec2 temp = A;
        A = B;
        B = temp;
    }
    
    vec2 AB = v2_sub( B, A );
    vec2 AC = v2_sub( C, A );
    
    vec2 ABN = v2_normalized( AB );
    
    r32 b = v2_length( AB );
    r32 h = v2_dot( v2_perp( ABN ), AC );
    r32 a = v2_dot( ABN, AC );
    
    r32 momentOfInertia = ( ( b * b * b * h ) - ( b * b * h * a ) + ( b * h * a * a ) + ( b * h * h * h ) ) / 36;
    
    return momentOfInertia;
}

stu r32 geometry_findConvexMomentOfInertia( vec2* points, umm pointCount, vec2 centroid ) {
    
    r32 MOI = 0.0f;
    b32 isCCW = geometry_isPolygonCCW( points, pointCount );
    
    vec2 triangle[ 3 ];
    triangle[ 0 ] = centroid;
    
    for ( umm pointIndex = 0; pointIndex < pointCount; pointIndex++ ) {
        
        umm firstIndex = pointIndex;
        umm secondIndex = pointIndex + 1;
        
        if ( !isCCW ) {
            firstIndex = pointCount - pointIndex;
            secondIndex = pointCount - ( pointIndex + 1 );
        }
        
        if ( firstIndex >= pointCount )  {
            firstIndex = 0;
        }
        
        if ( secondIndex >= pointCount ) {
            secondIndex = 0;
        }
        
        triangle[ 1 ] = points[ firstIndex ];
        triangle[ 2 ] = points[ secondIndex ];
        
        vec2 triangleCentroid = geometry_findTriangleCentroid( triangle );
        
#if 1
        vec2 AB = v2_sub( triangle[ 1 ], triangle[ 0 ] );
        vec2 AC = v2_sub( triangle[ 2 ], triangle[ 0 ] );
        
        vec2 ABN = v2_normalized( AB );
        
        r32 b = v2_length( AB );
        r32 h = v2_dot( v2_perp( ABN ), AC );
        r32 a = v2_dot( ABN, AC );
        
        r32 triangleMOI = ( ( b * b * b * h ) - ( b * b * h * a ) + ( b * h * a * a ) + ( b * h * h * h ) ) / 36;
        r32 triangleMass = ( h * b ) * 0.5f;
#else
        r32 triangleMOI = geometry_findTriangleMomentOfInertia( triangle );
        r32 triangleMass = abs( geometry_polygonArea( triangle, 3 ) ); /* NOTE simon: This way of calculating mass is little more precise but slower. */
#endif
        
        MOI += ( triangleMOI + triangleMass * v2_length( v2_sub( centroid, triangleCentroid ) ) );
    }
    
    return MOI;
}

/* NOTE simon:
Expected:
- Line don't cross;
- The last point != the first point;
- Points are counter clockwise.

Output in memory:
{
---- u32 convexCount;
---- { Convex 1
-------- u32 indexCount;
-------- u32* indexes;
---- }
---- { Convex 2 }
---- { ... }
}
*/
/* TODO simon: verify memory_push* alignments. */
// #define DEBUG_PARTITION_POLYGON 1

/* NOTE simon: This could help find a better solution: https://top.handmade.network/blogs/p/2802-writing_a_custom_vector_graphics_renderer_-_part_2#13812 */

stu void geometry_partitionPolygon( data_t* memory, vec2* points, u32 pointCount ) {
    
#if defined( DEBUG_PARTITION_POLYGON )
    vec2* pointList = memory_pushArray( &debugGl.memory, vec2, pointCount + 1 );
    
    for ( umm i = 0; i < pointCount; i++ ) {
        pointList[ i ] = v2_add(  points[ i ], v2( 300.0f, 300.0f ) );
    }
    
    pointList[ pointCount ] = pointList[ 0 ];
    
    debug_frameInit( 0 );
    debug_clear( 0 );
    debug_drawLines( 0, pointList, pointCount + 1, white );
    displayDebugRender = true;
    
    memory_popArray( &debugGl.memory, vec2, pointCount + 1 );
#endif
    
    _assert( geometry_isPolygonCCW( points, pointCount ) );
    _assert( geometry_isPolygonValid( points, pointCount ) );
    
    u32* resultCount = memory_push_type( memory, u32 );
    *resultCount = 0;
    
    reverse_stack_t reverse_stack = memory_reverse_stack_create( memory );
    u32* polygonCount = ( u32* ) memory_reverse_stack_push( &reverse_stack, sizeof( u32 ) );
    *polygonCount = 1;
    u32* workingPointCount = ( u32* ) memory_reverse_stack_push( &reverse_stack, sizeof( u32 ) );
    *workingPointCount = pointCount;
    u32* workingPoints = ( u32* ) memory_reverse_stack_push( &reverse_stack, pointCount * sizeof( u32 ) );
    u32 currentPolygon = 0;
    
    for ( u32 i = 0; i < *workingPointCount; i++ ) {
        workingPoints[ i ] = i;
    }
    
    while ( currentPolygon < *polygonCount ) {
        
        /* NOTE simon: workingPoints is a polygon part (non convex) that as been pushed on the reverse stack. */
        workingPoints = workingPointCount - *workingPointCount;
        
#if defined( DEBUG_PARTITION_POLYGON )
        vec2* pointList = memory_pushArray( &debugGl.memory, vec2, *workingPointCount );
        
        for ( umm i = 0; i < *workingPointCount; i++ ) {
            pointList[ i ] = v2_add(  points[ *( workingPoints + i ) ], v2( 300.0f, 300.0f ) );
        }
        
        debug_frameInit( 1 );
        debug_drawLines( 1, pointList, *workingPointCount, yellow );
        
        memory_popArray( &debugGl.memory, vec2, *workingPointCount );
#endif
        
        ( *resultCount )++;
        u32* polygonPointCount = ( u32* ) memory_push_size( memory, sizeof( u32 ) * 3, 4 );
        /* NOTE simon: polygon contains the new convex that will be pushed in the output memory.*/
        u32* polygon = polygonPointCount + 1;
        
        polygon[ 0 ] = workingPoints[ 0 ];
        polygon[ 1 ] = workingPoints[ 1 ];
        *polygonPointCount = 2;
        
        u32 workingPointIndex = *polygonPointCount;
        s32 startPoint = -1;
        
        while ( workingPointIndex < *workingPointCount ) {
            
            vec2 a = points[ polygon[ *polygonPointCount - 2 ] ];
            vec2 b = points[ polygon[ *polygonPointCount - 1 ] ];
            vec2 c = points[ workingPoints[ workingPointIndex ] ];
            vec2 d = points[ workingPoints[ 0 ] ];
            vec2 e = points[ workingPoints[ 1 ] ];
            
#if defined( DEBUG_PARTITION_POLYGON )
            debug_frameInit( 3 );
            debug_drawPoint( 3, v2_add( c, v2( 300, 300 ) ), 3, blue );
            debug_render( 0, false );
            debug_render( 1, false );
            debug_render( 3, true );
#endif
            
            memory_push_type( memory, u32 );
            polygon[ *polygonPointCount ] = workingPoints[ workingPointIndex ];
            
            b32 valid = geometry_angleOrientation( a, b, c ) <= 0.0f;
            valid = valid && geometry_angleOrientation( b, c, d ) <= 0.0f;
            valid = valid && geometry_angleOrientation( c, d, e ) <= 0.0f;
            
            if ( valid ) {
                
                /* NOTE simon: Should I test the convex agains every shapes that have been generated ? */
                /* NOTE simon: Validate that segments of convex don't intersect with segments of shape */
                /* NOTE simon: this if was !isConvexIntersectingShape; */
                
                u32* convexIndexes = polygon;
                u32 convexPointCount = ( *polygonPointCount ) + 1;
                u32* shapeIndexes = workingPoints;
                u32 shapePointCount = *workingPointCount;
                
                b32 intersects = false;
                umm convexIndex = 0;
                
                while ( !intersects && convexIndex < convexPointCount ) {
                    
                    umm secondIndex = convexIndex + 1;
                    
                    if ( secondIndex == convexPointCount ) {
                        secondIndex = 0;
                    }
                    
                    umm aStartIndex = convexIndexes[ convexIndex ];
                    umm aEndIndex = convexIndexes[ secondIndex ];
                    vec2 aStart = points[ aStartIndex ];
                    vec2 aEnd = points[ aEndIndex ];
                    
                    umm shapeIndex = 0;
                    
                    while ( !intersects && shapeIndex < shapePointCount ) {
                        
                        umm shapeSecondIndex = shapeIndex + 1;
                        
                        if ( shapeSecondIndex == shapePointCount ) {
                            shapeSecondIndex = 0;
                        }
                        
                        umm bStartIndex = shapeIndexes[ shapeIndex ];
                        umm bEndIndex = shapeIndexes[ shapeSecondIndex ];
                        vec2 bStart = points[ bStartIndex ];
                        vec2 bEnd = points[ bEndIndex ];
                        
                        if ( aStartIndex != bStartIndex && aEndIndex != bEndIndex ) {
                            
                            r32 A1 = math_sign_or_zero( geometry_angleOrientation( aStart, aEnd, bStart ) );
                            r32 A2 = math_sign_or_zero( geometry_angleOrientation( aStart, aEnd, bEnd ) );
                            r32 B1 = math_sign_or_zero( geometry_angleOrientation( bStart, bEnd, aStart ) );
                            r32 B2 = math_sign_or_zero( geometry_angleOrientation( bStart, bEnd, aEnd ) );
                            
                            if ( aEndIndex == bStartIndex ) {
                                intersects = ( A2 == 0.0f && geometry_isColinearPointOnSegment( aStart, aEnd, bEnd ) );
                            } else if ( aStartIndex == bEndIndex ) {
                                intersects = ( A1 == 0.0f && geometry_isColinearPointOnSegment( aStart, aEnd, bStart ) );
                            } else {
                                intersects = ( A1 != A2 && B1 != B2 );
                                intersects = intersects || ( A1 == 0.0f && geometry_isColinearPointOnSegment( aStart, aEnd, bStart ) );
                                intersects = intersects || ( B1 == 0.0f && geometry_isColinearPointOnSegment( bStart, bEnd, aStart ) );
                                intersects = intersects || ( A2 == 0.0f && geometry_isColinearPointOnSegment( aStart, aEnd, bEnd ) );
                                intersects = intersects || ( B2 == 0.0f && geometry_isColinearPointOnSegment( bStart, bEnd, aEnd ) );
                            }
                        }
                        
                        shapeIndex++;
                    }
                    
                    convexIndex++;
                }
                
                valid = !intersects;
            }
            
            if ( valid ) {
                
                /* NOTE simon: this if was isAnyPointInsideConvex. */
                u32* pointIndexes = workingPoints + workingPointIndex + 1;
                u32 indexCount = ( *workingPointCount ) - workingPointIndex - 1;
                u32* convexIndexes = polygon;
                u32 convexPointCount = *polygonPointCount + 1;
                
                r32 dotProduct = 1.0f;
                
                if ( indexCount == 0 ) {
                    dotProduct = -1.0f;
                }
                
                for ( umm i = 0; i < convexPointCount && dotProduct >= 0.0f; i++ ) {
                    
                    vec2 aPoint = points[ convexIndexes[ i ] ];
                    vec2 bPoint = ( i == convexPointCount - 1 ) ? points[ convexIndexes[ 0 ] ] : points[ convexIndexes[ i + 1 ] ];
                    vec2 side = v2_sub( bPoint, aPoint );
                    vec2 normal = v2_perp( side );
                    
                    for ( umm j = 0; j < indexCount && dotProduct >= 0.0f; j++ ) {
                        vec2 test = v2_sub( points[ pointIndexes[ j ] ], aPoint );
                        dotProduct = v2_dot( normal, test );
                    }
                }
                
                valid = dotProduct < 0.0f;
            }
            
            if ( valid ) {
                
                if ( startPoint != -1 ) {
                    
                    u32 count = workingPointIndex - startPoint + 1;
                    u32* newPolygon = ( u32* ) memory_reverse_stack_push( &reverse_stack, sizeof( u32 ) * ( count + 1 ) );
                    u32* newPolygonPointCount = newPolygon + count;
                    *newPolygonPointCount = count;
                    
                    for ( umm i = 0; i < *newPolygonPointCount; i++ ) {
                        newPolygon[ i ] = workingPoints[ startPoint + i ];
                    }
                    
                    ( *polygonCount )++;
                    startPoint = -1;
                }
                
                ( *polygonPointCount )++;
                
            } else {
                
                if ( startPoint == -1 ) {
                    startPoint = workingPointIndex - 1;
                }
                
                memory_pop_size( memory, sizeof( u32 ) );
            }
            
            workingPointIndex++;
        }
        
        if ( startPoint != -1 ) {
            
            u32 count = ( *workingPointCount ) - startPoint + 1;
            u32* newPolygon = ( u32* ) memory_reverse_stack_push( &reverse_stack, sizeof( u32 ) * ( count + 1 ) );
            u32* newPolygonPointCount = newPolygon + count;
            *newPolygonPointCount = count;
            
            for ( umm i = 0; i < ( *newPolygonPointCount ) - 1; i++ ) {
                newPolygon[ i ] = workingPoints[ startPoint + i ];
            }
            
            newPolygon[ ( *newPolygonPointCount ) - 1 ] = workingPoints[ 0 ];
            
            ( *polygonCount )++;
            /* startPoint = -1; // NOTE simon: clang static code analyser suggest to remove this as the result is never used. */
        }
        
        workingPointCount = workingPoints - 1;
        currentPolygon++;
        
#if defined( DEBUG_PARTITION_POLYGON )
        vec2* pointListAdded = memory_pushArray( &debugGl.memory, vec2, *polygonPointCount );
        
        for ( umm i = 0; i < *polygonPointCount; i++ ) {
            pointListAdded[ i ] = v2_add(  points[ *( polygon + i ) ], v2( 300.0f, 300.0f ) );
        }
        
        debug_frameInit( 2 );
        debug_drawLines( 2, pointList, *polygonPointCount, red );
        
        memory_popArray( &debugGl.memory, vec2, *polygonPointCount );
        
        debug_render( 0, false );
        debug_render( 1, false );
        debug_render( 2, false );
        debug_render( 3, true );
        break_here;
#endif
    }
}

/* MARK text */

/* NOTE simon:
utf-8 spec:
http://www.unicode.org/versions/Unicode11.0.0/ch02.pdf#G13708 chapter 2.5
http://www.unicode.org/versions/Unicode11.0.0/ch03.pdf#G7404 chapter 3.9
https://tools.ietf.org/html/rfc3629
*/

/*
https://www.unicode.org/faq/utf_bom.html
https://www.unicode.org/faq/private_use.html
https://unicodebook.readthedocs.io/guess_encoding.html#is-utf8
*/

/* NOTE:
codepoint: number that represent a character.
codeunit: group of bit that represent one part of the encoding of a codepoint. 8 bits in utf-8, 16 bits in utf-16.
== From the Unicode spec V11 ==
http://www.unicode.org/versions/Unicode11.0.0/ch03.pdf#G7404
codeunit: The minimal bit combination that can represent a unit of encoded text for processing or interchange.*/

u32 ucs_invalid_codepoint = 0xffffffff;
u32 ucs_last_codepoint = 0x10ffff;
u32 ucs_replacement_codepoint = 0xfffd;

u16 utf16_hi_surrogate_start = 0xD800;
u16 utf16_lo_surrogate_start = 0xDC00;
u16 utf16_surrogate_end = 0xDFFF;

typedef enum utf_error {
    
    utf_error_none,
    
    utf_error_end_of_bytes,
    utf_error_not_enough_bytes_to_decode,
    utf_error_not_enough_bytes_to_encode,
    utf_error_invalid_codepoint_to_encode,
    
    utf_error_utf8_invalid_lead_byte,
    utf_error_utf8_invalid_continuation_byte,
    
    utf_error_utf16_invalid_surrogate_pair,
    
    utf_error_count
} utf_error;

stu uint32_t ucs_is_codepoint_valid( uint32_t codepoint ) {
    uint32_t result = codepoint < utf16_hi_surrogate_start;
    result = result || ( codepoint > utf16_surrogate_end && codepoint <= ucs_last_codepoint );
    return result;
}

/* MARK utf-8 */

/* When error is set, you can't use codeunit_count to advance the bytes pointer, you should advance only 1 byte. */
stu void utf8_decode( void* bytes, void* bytes_end, uint32_t* codepoint, uint8_t* codeunit_count, uint32_t* error ) {
    
    if ( no_error( error ) ) {
        
        *codepoint = ucs_invalid_codepoint;
        *codeunit_count = 0;
        
        if ( bytes < bytes_end ) {
            
            uint8_t bytes_seen = 0;
            uint8_t bytes_needed = 0;
            uint32_t lower_boundary = 0x80;
            uint32_t upper_boundary = 0xbf;
            
            uint8_t* current = ( uint8_t* ) bytes;
            uint8_t* end = ( uint8_t* ) bytes_end;
            uint8_t byte = *current;
            
            if ( byte <= 0x7f ) {
                
                *codeunit_count = 1;
                *codepoint = byte;
                
            } else if ( byte >= 0xc2 && byte <= 0xdf ) {
                
                bytes_needed = 1;
                *codeunit_count = 2;
                *codepoint = byte & 0x1f;
                
            } else if ( byte >= 0xe0 && byte <= 0xef ) {
                
                if ( byte == 0xe0 ) {
                    lower_boundary = 0xa0;
                } else if ( byte == 0xed ) {
                    upper_boundary = 0x9f;
                }
                
                bytes_needed = 2;
                *codeunit_count = 3;
                *codepoint = byte & 0xf;
                
            } else if ( byte >= 0xf0 && byte <= 0xf4 ) {
                
                if ( byte == 0xf0 ) {
                    lower_boundary = 0x90;
                } else if ( byte == 0xf4 ) {
                    upper_boundary = 0x8f;
                }
                
                bytes_needed = 3;
                *codeunit_count = 4;
                *codepoint = byte & 0x7;
                
            } else {
                set_error( error, utf_error_utf8_invalid_lead_byte );
            }
            
            if ( no_error( error ) ) {
                
                if ( current + bytes_needed < end ) {
                    
                    while ( no_error( error ) && bytes_seen < bytes_needed ) {
                        
                        current++;
                        byte = *current;
                        
                        if ( byte < lower_boundary || byte > upper_boundary ) {
                            
                            set_error( error, utf_error_utf8_invalid_continuation_byte );
                            
                        } else {
                            
                            ( *codepoint ) = ( ( *codepoint ) << 6 ) | ( byte & 0x3f );
                            bytes_seen++;
                            lower_boundary = 0x80;
                            upper_boundary = 0xbf;
                        }
                    }
                    
                } else {
                    set_error( error, utf_error_not_enough_bytes_to_decode );
                }
            }
            
            if ( is_error( error ) ) {
                *codepoint = ucs_invalid_codepoint;
            }
        } else {
            set_error( error, utf_error_end_of_bytes );
        }
    }
}

stu void utf8_encode( void* bytes, void* bytes_end, uint32_t codepoint, uint8_t* codeunit_count, uint32_t* error ) {
    
    if ( no_error( error ) ) {
        
        *codeunit_count = 0;
        
        if ( ucs_is_codepoint_valid( codepoint ) ) {
            
            uint8_t temp[ 4 ] = { 0 };
            
            if ( codepoint <= 0x7f ) {
                *codeunit_count = 1;
                temp[ 0 ] = ( uint8_t ) codepoint;
            } else if ( codepoint < 0x7ff ) {
                *codeunit_count = 2;
                temp[ 0 ] = ( uint8_t ) ( ( ( codepoint >> 6 ) & 0x1f ) | 0xc0 );
                temp[ 1 ] = ( uint8_t ) ( ( ( codepoint >> 0 ) & 0x3f ) | 0x80 );
            } else if ( codepoint < 0xffff ) {
                *codeunit_count = 3;
                temp[ 0 ] = ( uint8_t ) ( ( ( codepoint >> 12 ) & 0x0f ) | 0xe0 );
                temp[ 1 ] = ( uint8_t ) ( ( ( codepoint >> 6 ) & 0x3f ) | 0x80 );
                temp[ 2 ] = ( uint8_t ) ( ( ( codepoint >> 0 ) & 0x3f ) | 0x80 );
            } else if ( codepoint < 0x10ffff ) {
                *codeunit_count = 4;
                temp[ 0 ] = ( uint8_t ) ( ( ( codepoint >> 18 ) & 0x07 ) | 0xf0 );
                temp[ 1 ] = ( uint8_t ) ( ( ( codepoint >> 12 ) & 0x3f ) | 0x80 );
                temp[ 2 ] = ( uint8_t ) ( ( ( codepoint >> 6 ) & 0x3f ) | 0x80 );
                temp[ 3 ] = ( uint8_t ) ( ( ( codepoint >> 0 ) & 0x3f ) | 0x80 );
            }
            
            if ( bytes ) {
                
                uint8_t* current = ( uint8_t* ) bytes;
                uint8_t* end = ( uint8_t* ) bytes_end;
                uintptr_t length = end - current;
                
                if ( length >= *codeunit_count ) {
                    
                    for ( uintptr_t i = 0; i < *codeunit_count; i++ ) {
                        current[ i ] = temp[ i ];
                    }
                } else {
                    set_error( error, utf_error_not_enough_bytes_to_encode );
                }
            }
        } else {
            set_error( error, utf_error_invalid_codepoint_to_encode );
        }
    }
}

stu umm utf8_get_codepoint_count( void* codeunits, void* codeunits_end ) {
    
    umm codepoint_count = 0;
    u8* current = ( u8* ) codeunits;
    u8* end = ( u8* ) codeunits_end;
    
    while ( current < end ) {
        
        u32 error = 0;
        u32 codepoint = ucs_invalid_codepoint;
        u8 codeunit_count = 0;
        utf8_decode( current, end, &codepoint, &codeunit_count, &error );
        codepoint_count++;
        
        if ( no_error( &error ) ) {
            current += codeunit_count;
        } else {
            current += 1;
        }
    }
    
    return codepoint_count;
}

/* MARK utf-16 */

stu uint16_t utf16_get_codeunit( uint8_t* bytes, uint32_t big_endian ) {
    
    uint16_t byte_1 = bytes[ 0 ];
    uint16_t byte_2 = bytes[ 1 ];
    uint16_t codeunit;
    
    if ( big_endian ) {
        codeunit = ( byte_1 << 8 ) + byte_2;
    } else {
        codeunit = ( byte_2 << 8 ) + byte_1;
    }
    
    return codeunit;
}

/* NOTE simon: on error don't use codeunit_count to advance the bytes pointer, advance 1 codeunit (2 bytes). */
stu void utf16_decode( void* bytes, void* bytes_end, uint32_t* codepoint, uint8_t* codeunit_count, uint32_t big_endian, uint32_t* error ) {
    
    if ( no_error( error ) ) {
        
        uint8_t* current = ( uint8_t* ) bytes;
        uint8_t* end = ( uint8_t* ) bytes_end;
        *codepoint = ucs_invalid_codepoint;
        *codeunit_count = 0;
        
        if ( current + 1 < end ) {
            
            uint16_t codeunit = utf16_get_codeunit( current, big_endian );
            current += 2;
            
            if ( codeunit < utf16_hi_surrogate_start || codeunit > utf16_surrogate_end ) {
                
                *codepoint = codeunit;
                *codeunit_count = 1;
                
            } else if ( current + 1 < end ) {
                
                uint16_t hi_surrogate = codeunit;
                uint16_t lo_surrogate = utf16_get_codeunit( current, big_endian );
                *codeunit_count = 2;
                
                uint32_t hi_is_valid = hi_surrogate < utf16_lo_surrogate_start;
                uint32_t lo_is_valid = lo_surrogate >= utf16_lo_surrogate_start && lo_surrogate <= utf16_surrogate_end;
                
                if ( hi_is_valid && lo_is_valid ) {
                    *codepoint = ( hi_surrogate - utf16_hi_surrogate_start ) << 10;
                    *codepoint |= ( lo_surrogate - utf16_lo_surrogate_start );
                    *codepoint += 0x10000;
                } else {
                    set_error( error, utf_error_utf16_invalid_surrogate_pair );
                }
                
            } else {
                *codeunit_count = 2;
                set_error( error, utf_error_not_enough_bytes_to_decode );
            }
            
        } else {
            *codeunit_count = 1;
            set_error( error, utf_error_end_of_bytes );
        }
    }
}

stu void utf16_encode( void* bytes, void* bytes_end, uint32_t codepoint, uint8_t* codeunit_count, uint32_t big_endian, uint32_t* error ) {
    
    if ( no_error( error ) ) {
        
        *codeunit_count = 0;
        
        if ( ucs_is_codepoint_valid( codepoint ) ) {
            
            uint16_t temp[ 2 ] = { 0 };
            
            if ( codepoint <= 0xffff ) {
                *codeunit_count = 1;
                temp[ 0 ] = ( uint16_t ) codepoint;
            } else {
                *codeunit_count = 2;
                codepoint -= 0x10000;
                temp[ 0 ] = ( uint16_t ) ( ( codepoint >> 10 ) + utf16_hi_surrogate_start );
                temp[ 1 ] = ( uint16_t ) ( ( codepoint & 0x3ff ) + utf16_lo_surrogate_start );
            }
            
            if ( bytes ) {
                
                uint8_t* current = ( uint8_t* ) bytes;
                uint8_t* end = ( uint8_t* ) bytes_end;
                uintptr_t length = end - current;
                
                if ( length >= ( *codeunit_count ) * 2 ) {
                    
                    for ( uintptr_t i = 0; i < *codeunit_count; i++ ) {
                        
                        if ( big_endian ) {
                            *current = ( uint8_t ) ( temp[ i ] >> 8 );
                            current++;
                            *current = ( uint8_t ) ( temp[ i ] & 0xff );
                        } else {
                            *current = ( uint8_t ) ( temp[ i ] & 0xff );
                            current++;
                            *current = ( uint8_t ) ( temp[ i ] >> 8 );
                        }
                        
                        current++;
                    }
                    
                } else {
                    set_error( error, utf_error_not_enough_bytes_to_encode );
                }
            }
        } else {
            set_error( error, utf_error_invalid_codepoint_to_encode );
        }
    }
}

stu umm utf16_get_codepoint_count( void* codeunits, void* codeunits_end ) {
    
    umm codepoint_count = 0;
    u16* current = ( u16*) codeunits;
    u16* end = ( u16* ) codeunits_end;
    
    while ( current < end ) {
        
        u32 error = 0;
        u32 codepoint = ucs_invalid_codepoint;
        u8 codeunit_count = 0;
        utf16_decode( current, end, &codepoint, &codeunit_count, 0, &error );
        codepoint_count++;
        
        if ( no_error( &error ) ) {
            current += codeunit_count;
        } else {
            current += 1;
        }
    }
    
    return codepoint_count;
}

stu umm utf16_to_utf8( void* source_codeunits, umm source_codeunit_count, void* destination_codeunits, umm output_buffer_byte_count, u32* error ) {
    
    umm result = 0;
    
    u16* u16_current = ( u16* ) source_codeunits;
    u16* u16_end = u16_current + source_codeunit_count;
    u8* u8_current = ( u8* ) destination_codeunits;
    u8* u8_end = u8_current + output_buffer_byte_count;
    
    while ( no_error( error ) && u16_current < u16_end ) {
        
        u32 codepoint = 0;
        u8 u8_codeunit_count = 0;
        u8 u16_codeunit_count = 0;
        
        utf16_decode( u16_current, u16_end, &codepoint, &u16_codeunit_count, 0, error );
        
        if ( is_error( error ) ) {
            codepoint = ucs_replacement_codepoint;
            u16_codeunit_count = 1;
            clear_error( error );
        }
        
        utf8_encode( u8_current, u8_end, codepoint, &u8_codeunit_count, error );
        
        if ( no_error( error ) ) {
            u8_current += u8_codeunit_count;
            u16_current += u16_codeunit_count;
            result += u8_codeunit_count;
        }
    }
    
    return result;
}

stu umm utf8_to_utf16( void* source_codeunits, umm source_codeunit_count, void* destination_codeunits, umm output_buffer_byte_count, u32* error ) {
    
    umm result = 0;
    
    u8* u8_current = ( u8* ) source_codeunits;
    u8* u8_end = u8_current + source_codeunit_count;
    u16* u16_current = ( u16* ) destination_codeunits;
    u16* u16_end = ( u16* ) ( ( ( u8* ) u16_current ) + output_buffer_byte_count );
    
    while ( no_error( error ) && u8_current < u8_end ) {
        
        u32 codepoint = 0;
        u8 u8_codeunit_count = 0;
        u8 u16_codeunit_count = 0;
        
        utf8_decode( u8_current, u8_end, &codepoint, &u8_codeunit_count, error );
        
        if ( is_error( error ) ) {
            codepoint = ucs_replacement_codepoint;
            u8_codeunit_count= 1;
            clear_error( error );
        }
        
        utf16_encode( u16_current, u16_end, codepoint, &u16_codeunit_count, 0, error );
        
        if ( no_error( error ) ) {
            u8_current += u8_codeunit_count;
            u16_current += u16_codeunit_count;
            result += u16_codeunit_count;
        }
    }
    
    return result;
}

/* TODO simon: Should I return the new pointer instead of the number of bytes to back off ? */
stu u8 utf8_back_one_codepoint( void* bytes, void* bytes_end ) {
    
    u8 result = 0;
    
    u8* start = ( u8* ) bytes;
    u8* end = ( u8* ) bytes_end;
    
    if ( end - start ) {
        result = 1;
    }
    
    u8* current = end - 1;
    
    while ( current >= start ) {
        
        u32 error = 0;
        u32 codepoint = ucs_invalid_codepoint;
        u8 codeunit_count = 0;
        utf8_decode( current, end, &codepoint, &codeunit_count, &error );
        
        if ( no_error( &error ) ) {
            result = codeunit_count;
            break;
        } else {
            current--;
            if ( end - current > 4 ) {
                break;
            }
        }
    }
    
    if ( current < start ) {
        result = ( u8 ) ( end - start );
    }
    
    return result;
}

/* MARK arguments */

/* NOTE simon: win32 writing console attributes https://msdn.microsoft.com/en-us/library/windows/desktop/ms685032(v=vs.85).aspx */

#if !defined ARGUMENTS_MAX_COUNT
#define ARGUMENTS_MAX_COUNT 128
#endif

u32 arguments_count = 0;
u8** arguments_s = 0;
data_t arguments_d[ ARGUMENTS_MAX_COUNT ] = { 0 };
u8 executable_path_buffer[ 2048 ] = { 0 };
data_t executable_path;
data_t executable_name;

#if defined( WINDOWS )
/* NOTE simon: contains the argv strings (separated by \0) AND the argv array itself, aligned on the pointer size. */
u8 arguments_utf8_command_line[ 2048 ] = { 0 };

/* NOTE simon: https://docs.microsoft.com/en-us/previous-versions/17w5ykft(v=vs.85)
Tests:
"abc" d e => abc | d | e
a\\\b d"e f"g h => a\\\b | de fg | h
a\\\"b c d => a\"b | c | d
a\\\\"b c" d e => a\\b c | d | e
*/
char** arguments_parse_command_line( int* argc ) {
    
    u16* current = ( u16* ) GetCommandLineW( );
    u16* end = current;
    
    while ( *end ) {
        end++;
    }
    
    u8* utf8_start = arguments_utf8_command_line;
    u8* utf8_current = utf8_start;
    u8* utf8_end = utf8_start + array_count( arguments_utf8_command_line );
    
    b32 in_quotes = false;
    b32 previous_was_space = false;
    u32 backslash_count = 0;
    
    while ( current < end && utf8_current < utf8_end ) {
        
        u32 codepoint = ucs_invalid_codepoint, error = 0;
        u8 codeunit_count = 0;
        utf16_decode( current, end, &codepoint, &codeunit_count, 0, &error );
        current += codeunit_count;
        _assert( no_error( &error ) );
        
        if ( codepoint == '\\' ) {
            
            backslash_count++;
            previous_was_space = false;
            
        } else {
            
            if ( codepoint == '"' ) {
                
                for ( umm i = 0; i < backslash_count / 2; i++ ) {
                    if ( utf8_current < utf8_end ) {
                        *utf8_current = '\\';
                        utf8_current++;
                    }
                }
                
                if ( ( backslash_count % 2 ) == 0 ) {
                    in_quotes = !in_quotes;
                } else {
                    if ( utf8_current < utf8_end ) {
                        *utf8_current = '"';
                        utf8_current++;
                    }
                }
                
                backslash_count = 0;
                previous_was_space = false;
                
            } else {
                
                for ( umm i = 0; i < backslash_count; i++ ) {
                    if ( utf8_current < utf8_end ) {
                        *utf8_current = '\\';
                        utf8_current++;
                    }
                }
                
                backslash_count = 0;
                
                if ( codepoint == ' ' || codepoint == '\t' ) {
                    
                    if ( in_quotes ) {
                        if ( utf8_current < utf8_end ) {
                            *utf8_current = ( u8 ) codepoint;
                            utf8_current++;
                        }
                    } else {
                        
                        if ( !previous_was_space ) {
                            if ( utf8_current < utf8_end ) {
                                *utf8_current = 0;
                                utf8_current++;
                            }
                        }
                        
                        previous_was_space = true;
                    }
                    
                } else {
                    
                    previous_was_space = false;
                    u8 utf8_codeunit_count = 0;
                    utf8_encode( utf8_current, utf8_end, codepoint, &utf8_codeunit_count, &error );
                    if ( no_error( &error ) ) {
                        utf8_current += utf8_codeunit_count;
                    }
                }
            }
        }
    }
    
    if ( backslash_count ) {
        for ( umm i = 0; i < backslash_count; i++ ) {
            if ( utf8_current < utf8_end ) {
                *utf8_current = '\\';
                utf8_current++;
            }
        }
    }
    
    /* NOTE simon: If the command line ended with a space or tab there is already a 0 at the end */
    if ( utf8_current > utf8_start && utf8_current < utf8_end ) {
        
        if ( utf8_current[ -1 ] ) {
            *utf8_current = 0;
            utf8_current++;
        }
        
    } else {
        _assert( !"Empty command line or tool long command line." );
    }
    
    umm command_line_length = utf8_current - utf8_start;
    
    u8** argv = ( u8** ) utf8_current + memory_get_padding( utf8_current, sizeof( u8** ) );
    *argc = 0;
    b32 add = true;
    
    for ( umm i = 0; i < command_line_length && ( argv + i + 1 ) <= ( u8** ) utf8_end; i++ ) {
        
        if ( add && utf8_start[ i ] != 0 ) {
            argv[ *argc ] = utf8_start + i;
            ( *argc )++;
            add = false;
        } else if ( utf8_start[ i ] == 0 ) {
            add = true;
        }
    }
    
    return ( char** ) argv;
}

#endif

stu void arguments_initialize( s32 argc, char** argv ) {
    
    _assert( argc <= array_count( arguments_d ) );
    arguments_count = argc;
    arguments_s = ( u8** ) argv;
    
    for ( umm i = 0; i < argc && i < array_count( arguments_d ); i++ ) {
        arguments_d[ i ].used = string_length( argv[ i ] );
        arguments_d[ i ].reserved = arguments_d[ i ].used + 1;
        arguments_d[ i ].bytes = arguments_s[ i ];
    }
    
#if defined( WINDOWS )
    u32 length = GetModuleFileNameA( 0, ( LPSTR ) executable_path_buffer, array_count( executable_path_buffer ) );
    _assert( length != array_count( executable_path_buffer ) );
    data_t path = memory_construct( 0, 0, executable_path_buffer );
#elif defined( LINUX )
    s32 length = readlink( "/proc/self/exe", ( char* ) executable_path_buffer, array_count( executable_path_buffer ) - 1 );
    _assert( length != -1 );
    executable_path_buffer[ length ] = 0;
    data_t path = memory_construct( 0, 0, executable_path_buffer );
#else
# error Unsupported platform.
#endif
    
    u8* c = path.bytes;
    
    while ( *c != 0 ) {
        if ( *c == '/' || *c == '\\' ) {
            path.used = ( c - path.bytes ) + 1;
        }
        c++;
    }
    
    path.reserved = path.used;
    executable_path = path;
    
    u8* name_bytes = path.bytes + path.used;
    executable_name = memory_construct( c - name_bytes, c - name_bytes, name_bytes );
}

#define arguments_find_l( literal, start_index ) arguments_find_p( literal, sizeof( literal ) - 1, start_index )

stu u32 arguments_find_p( void* name_arg, umm length, u32 start_index ) {
    
    u8* name = ( u8* ) name_arg;
    u32 result = 0;
    data_t search = memory_construct( length, length, name );
    
    for ( u32 arg_index = start_index; arg_index < arguments_count; arg_index++ ) {
        
        data_t* arg = arguments_d + arg_index;
        
        if ( memory_are_equal( search, *arg ) ) {
            result = arg_index;
            break;
        }
    }
    
    return result;
}

stu u32 arguments_find_s( void* name, u32 start_index ) {
    
    umm length = string_length( ( u8* ) name );
    u32 result = arguments_find_p( name, length, start_index );
    return result;
}

stu u32 arguments_find_d( data_t name, u32 start_index ) {
    u32 result = arguments_find_p( name.text, name.used, start_index );
    return result;
}

#define arguments_get_value_l( literal, start_index ) arguments_get_value_p( literal, sizeof( literal ) - 1, start_index )

stu data_t arguments_get_value_p( void* name, umm length, u32* start_index ) {
    
    data_t result = { 0 };
    u32 index = 0;
    
    if ( start_index ) {
        index = *start_index;
    }
    
    u32 argument_index = arguments_find_p( name, length, index );
    
    if ( argument_index > 0 && argument_index + 1 < arguments_count ) {
        
        if ( start_index ) {
            *start_index = ( argument_index + 1 );
        }
        
        result = arguments_d[ ( argument_index + 1 ) ];
    }
    
    return result;
}

stu data_t arguments_get_value_d( data_t name, u32* start_index ) {
    data_t result = arguments_get_value_p( name.text, name.used, start_index );
    return result;
}

/* MARK terminal */

#if 0
#include <sys/ioctl.h>
#include <stdio.h>

int main (void)
{
    struct winsize w;
    ioctl(0, TIOCGWINSZ, &w);
    
    printf ("lines %d\n", w.ws_row);
    printf ("columns %d\n", w.ws_col);
    return 0;
}
#endif


#if defined( WINDOWS )
HANDLE terminal_stdin = 0;
HANDLE terminal_stdout = 0;
HANDLE terminal_stderror = 0;
#elif defined( LINUX )
u32 terminal_stdin = STDIN_FILENO;
u32 terminal_stdout = STDOUT_FILENO;
u32 terminal_stderror = STDERR_FILENO;
#else
#error Unsupported platform.
#endif

#if defined( WINDOWS )

stu void terminal_write_unbuffered_p( void* string, umm string_length ) {
    
    _assert( string_length <= u32_max );
    
    if ( !terminal_stdout ) {
        terminal_stdout = GetStdHandle( STD_OUTPUT_HANDLE );
    }
    
    DWORD written;
    
    if ( terminal_stdout && terminal_stdout != INVALID_HANDLE_VALUE ) {
        WriteFile( terminal_stdout, string, ( u32 ) string_length, &written, 0 );
    }
}

#define terminal_write_unbuffered_l( string ) terminal_write_unbuffered_p( ( string ), sizeof( string ) - 1 )

stu void terminal_write_unbuffered_s( void* string ) {
    terminal_write_unbuffered_p( string, string_length( string ) );
}

stu void terminal_write_unbuffered( data_t buffer ) {
    terminal_write_unbuffered_p( buffer.bytes, buffer.used );
}

stu void terminal_error_unbuffered_p( void* string, umm string_length ) {
    
    if ( !terminal_stderror ) {
        terminal_stderror = GetStdHandle( STD_ERROR_HANDLE );
    }
    
    DWORD written;
    
    if ( terminal_stderror && terminal_stderror != INVALID_HANDLE_VALUE ) {
        WriteFile( terminal_stderror, string, ( u32 ) string_length, &written, 0 );
    }
}

#define terminal_error_unbuffered_l( string ) terminal_error_unbuffered_p( ( string ), sizeof( string ) - 1 )

stu void terminal_error_unbuffered_s( void* string ) {
    terminal_error_unbuffered_p( string, string_length( string ) );
}

stu void terminal_error_unbuffered( data_t buffer ) {
    terminal_error_unbuffered_p( buffer.bytes, buffer.used );
}

#elif defined( LINUX )

stu void terminal_write_unbuffered_p( void* string, umm string_length ) {
    write( terminal_stdout, string, string_length );
}

#define terminal_write_unbuffered_l( string ) terminal_write_unbuffered_p( ( string ), sizeof( string ) - 1 )

stu void terminal_write_unbuffered_s( void* string ) {
    write( terminal_stdout, string, string_length( string ) );
}

stu void terminal_write_unbuffered( data_t buffer ) {
    write( terminal_stdout, buffer.bytes, buffer.used );
}

stu void terminal_error_unbuffered_p( void* string, umm string_length ) {
    write( terminal_stderror, string, string_length );
}

#define terminal_error_unbuffered_l( string ) terminal_error_unbuffered_p( ( string ), sizeof( string ) - 1 )

stu void terminal_error_unbuffered_s( void* string ) {
    write( terminal_stderror, string, string_length( string ) );
}

stu void terminal_error_unbuffered( data_t buffer ) {
    write( terminal_stderror, buffer.bytes, buffer.used );
}

#else
# error Unsupported platform.
#endif

/* MARK process */

stu void log_windows_error_message( umm index, u32 error );

#if defined( WINDOWS )

/* TODO simon: better return code to know what is happening without debugging ? Done on linux ? */
/* NOTE simon: to execute dir, rmdir or other command, you need to use "cmd /c dir" as a command. */
stu b32 process_execute( void* command, void* workingDirectory, b32 shouldWait, u32* returnValue ) {
    
    b32 result = 0;
    
    PROCESS_INFORMATION processInformation = { 0 };
    STARTUPINFO startupInformation = { 0 };
    startupInformation.cb = sizeof( startupInformation );
    /* TODO simon: use the CreateProcessW version. */
    if ( CreateProcessA( 0, ( LPSTR ) command, 0, 0, 0, 0, 0, ( LPCSTR ) workingDirectory, &startupInformation, &processInformation ) ) {
        
        result = 1;
        
        if ( shouldWait && processInformation.hProcess != 0 ) {
            
            DWORD waitStatus = WaitForSingleObject( processInformation.hProcess, INFINITE );
            
            if ( waitStatus != WAIT_OBJECT_0 ) {
                result = 0;
            }
        }
        
        if ( processInformation.hProcess ) {
            
            if ( returnValue ) {
                
                b32 success = GetExitCodeProcess( processInformation.hProcess, ( LPDWORD ) returnValue );
                
                if ( !success || *returnValue == STILL_ACTIVE ) {
                    result = 0;
                }
            }
            
            CloseHandle( processInformation.hProcess );
        }
        
        if ( processInformation.hThread ) {
            CloseHandle( processInformation.hThread );
        }
    } else {
        log_windows_error_message( 0, GetLastError( ) );
    }
    
    return result;
}

#elif defined( LINUX )

stu b32 process_execute( void* command, void* working_directory, b32 should_wait, u32* return_value ) {
    
#define TERMINAL_EXECUTE_ERROR_FINALCOMMAND_LENGTH -1
#define TERMINAL_EXECUTE_ERROR_GETCWD -2
#define TERMINAL_EXECUTE_ERROR_CHDIR -3
#define TERMINAL_EXECUTE_ERROR_ARGUMENT_COUNT -4
#define TERMINAL_EXECUTE_ERROR_EXEC -5
    
    b32 result = 0;
    s32 pid = fork( );
    
    if ( pid == 0 ) {
        
        char final_command[ 2048 ] = { 0 };
        char* cwd_error = getcwd( final_command, array_count( final_command ) );
        
        if ( cwd_error ) {
            
            umm final_command_length;
            final_command_length = string_length( final_command );
            final_command[ final_command_length++ ] = '/';
            
            char* command_end = ( char* ) command;
            umm command_length = string_length( command );
            
            for ( umm i = 0; i < command_length; i++ ) {
                
                if ( *command_end == 0 || *command_end == ' ' ) {
                    break;
                }
                
                if ( final_command_length < array_count( final_command ) ) {
                    
                    final_command[ final_command_length++ ] = *command_end;
                    command_end++;
                    
                } else {
                    
                    data_t error = string_l( "Final command buffer doesn't contain enough space.\n" );
                    terminal_error_unbuffered_p( error.text, error.used );
                    _exit( TERMINAL_EXECUTE_ERROR_FINALCOMMAND_LENGTH );
                }
            }
            
            umm path_command_length = ( u8* ) command_end - ( u8* ) command;
            
            if ( !file_exists_s( final_command ) && path_command_length < array_count( final_command ) ) {
                string_copy( final_command, command, path_command_length );
                final_command[ path_command_length ] = 0;
            }
            
            if ( ( !working_directory || chdir( ( const char* ) working_directory ) == 0 ) ) {
                
                char* c = command_end;
                char* end = ( ( char* ) command ) + command_length;
                
                char* arguments[ 128 ] = { 0 };
                umm argument_count = 0;
                b32 in_quote = false;
                arguments[ argument_count++ ] = ( char* ) command;
                
                while ( c < end && *c != 0 && argument_count < array_count( arguments ) - 1 ) {
                    
                    if ( *c == '\"' ) {
                        
                        in_quote = !in_quote;
                        
                    } else {
                        
                        if ( !in_quote && *c == ' ' ) {
                            
                            *c = 0;
                            char* n = c + 1;
                            
                            if ( n < end && *n != 0 ) {
                                arguments[ argument_count++ ] = n;
                            }
                        }
                    }
                    
                    c++;
                }
                
                if ( argument_count < array_count( arguments ) ) {
                    
                    arguments[ argument_count++ ] = 0;
                    
                    if ( execvp( final_command, arguments ) == -1 ) {
                        data_t error = string_l( "exec returned -1. Which probably means that the command doesn't exists. Check errno.\n" );
                        terminal_error_unbuffered_p( error.text, error.used );
                        _exit( TERMINAL_EXECUTE_ERROR_EXEC );
                    }
                    
                } else {
                    
                    data_t error = string_l( "Not enough space for arguments.\n" );
                    terminal_error_unbuffered_p( error.text, error.used );
                    _exit( TERMINAL_EXECUTE_ERROR_ARGUMENT_COUNT );
                }
                
            } else {
                
                data_t error = string_l( "chdir returned an error.\n" );
                terminal_error_unbuffered_p( error.text, error.used );
                _exit( TERMINAL_EXECUTE_ERROR_CHDIR );
            }
            
        } else {
            
            data_t error = string_l( "getcwd returned NULL. Does the buffer contains enough space ?\n" );
            terminal_error_unbuffered_p( error.text, error.used );
            _exit( TERMINAL_EXECUTE_ERROR_GETCWD );
        }
        
    } else if ( pid > 0 ) {
        
        if ( should_wait ) {
            
            s32 status;
            waitpid( pid, &status, 0 );
            
            if ( WIFEXITED( status ) ) {
                
                result = 1;
                
                if ( return_value ) {
                    *return_value = WEXITSTATUS( status );
                }
            }
            
        } else {
            
            result = 1;
        }
    }
    
    return result;
    
#undef TERMINAL_EXECUTE_ERROR_FINALCOMMAND_LENGTH
#undef TERMINAL_EXECUTE_ERROR_GETCWD
#undef TERMINAL_EXECUTE_ERROR_CHDIR
#undef TERMINAL_EXECUTE_ERROR_ARGUMENT_COUNT
#undef TERMINAL_EXECUTE_ERROR_EXEC
}

#else
# error Unsupported platfrom.
#endif

// MARK encoding

/* NOTE simon:
http://www.unicode.org/reports/tr44/tr44-22.html#General_Category_Values
ftp://ftp.unicode.org/Public/UNIDATA/
*/

/* NOTE simon: from UnicodeData.txt with Cc as the general type. */

#define raw_unicode_control_characters \
0x0000, 0x0001, 0x0002, 0x0003, 0x0004, 0x0005, 0x0006, 0x0007, 0x0008, 0x0009, 0x000a, 0x000b, 0x000c, 0x000d, 0x000e, 0x000f, \
0x0010, 0x0011, 0x0012, 0x0013, 0x0014, 0x0015, 0x0016, 0x0017, 0x0018, 0x0019, 0x001a, 0x001b, 0x001c, 0x001d, 0x001e, 0x001f, \
0x007f, \
0x0080, 0x0081, 0x0082, 0x0083, 0x0084, 0x0085, 0x0086, 0x0087, 0x0088, 0x0089, 0x008a, 0x008b, 0x008c, 0x008d, 0x008e, 0x008f, \
0x0090, 0x0091, 0x0092, 0x0093, 0x0094, 0x0095, 0x0096, 0x0097, 0x0098, 0x0099, 0x009a, 0x009b, 0x009c, 0x009d, 0x009e, 0x009f

u32 unicode_control_characters[ ] = { raw_unicode_control_characters };
_ctassert( array_count( unicode_control_characters ) == 65 );

/* NOTE simon: from PropList.txt with Cn && Noncharacter_Code_Point */

#define raw_unicode_non_characters \
/* NOTE simon: non characters in the BMP. */ \
0xfdd0, 0xfdd1, 0xfdd2, 0xfdd3,0xfdd4, 0xfdd5, 0xfdd6, 0xfdd7, 0xfdd8, 0xfdd9, 0xfdda, 0xfddb, 0xfddc, 0xfddd, 0xfdde, 0xfddf, \
0xfde0, 0xfde1, 0xfde2, 0xfde3,0xfde4, 0xfde5, 0xfde6, 0xfde7, 0xfde8, 0xfde9, 0xfdea, 0xfdeb, 0xfdec, 0xfded, 0xfdee, 0xfdef, \
/* NOTE simon: last two codepoints of each plane. */ \
0xfffe, 0xffff, \
0x1fffe, 0x1ffff, \
0x2fffe, 0x2ffff, \
0x3fffe, 0x3ffff, \
0x4fffe, 0x4ffff, \
0x5fffe, 0x5ffff, \
0x6fffe, 0x6ffff, \
0x7fffe, 0x7ffff, \
0x8fffe, 0x8ffff, \
0x9fffe, 0x9ffff, \
0xafffe, 0xaffff, \
0xbfffe, 0xbffff, \
0xcfffe, 0xcffff, \
0xdfffe, 0xdffff, \
0xefffe, 0xeffff, \
0xffffe, 0xfffff, \
0x10fffe, 0x10ffff

u32 unicode_non_characters[ ] = { raw_unicode_non_characters };
_ctassert( array_count( unicode_non_characters ) == 66 );
#if 0
u32 encoding_windows_1252_to_unicode[ ] = {
    0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf,
    0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
    0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
    0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
    0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
    0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
    0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
    0x20ac, 0x0, 0x201a, 0x192, 0x201e, 0x2026, 0x2020, 0x2021, 0x2c6, 0x2030, 0x160, 0x2039, 0x152, 0x0, 0x17d, 0x0,
    0x0, 0x2018, 0x2019, 0x201c, 0x201d, 0x2022, 0x2013, 0x2014, 0x02dc, 0x2122, 0x161, 0x203a, 0x153, 0x0, 0x17e, 0x178,
    0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
    0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
    0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
    0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
    0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
    0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff
};

u32 encoding_iso_8859_15_to_unicode[ ] = {
    0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
    0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
    0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
    0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
    0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
    0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
    0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x0,
    0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
    0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
    0xa0, 0xa1, 0xa2, 0xa3, 0x20ac, 0xa5, 0x160, 0xa7, 0x161, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
    0xb0, 0xb1, 0xb2, 0xb3, 0x17d, 0xb5, 0xb6, 0xb7, 0x17e, 0xb9, 0xba, 0xbb, 0x152, 0x153, 0x178, 0xbf,
    0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
    0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
    0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
    0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff
};

_ctassert( array_count( encoding_windows_1252_to_unicode ) == 256 );
_ctassert( array_count( encoding_iso_8859_15_to_unicode ) == 256 );
#endif

// MARK performance

#if defined( WINDOWS )

#define perf_get_frequency( u64_ptr ) QueryPerformanceFrequency( ( LARGE_INTEGER* ) u64_ptr )
#define perf_get_time( u64_ptr ) QueryPerformanceCounter( ( LARGE_INTEGER* ) u64_ptr )
#define perf_get_cycle_count( ) __rdtsc( )

#elif defined( LINUX )

#define perf_get_frequency( u64_ptr ) ( *( u64_ptr ) ) = 1000000000
#define perf_get_time( u64_ptr ) { struct timespec spec; clock_gettime( CLOCK_MONOTONIC, &spec ); ( *( u64_ptr ) ) = ( spec.tv_sec * ( u64 ) 1000000000 + spec.tv_nsec ); }
#define perf_get_cycle_count( ) __rdtsc( )

#else
#error Unsupported platform.
#endif

stu r64 perf_get_elapsed_time( u64 start, u64 end, u64 frequency ) {
    r64 result = ( ( r64 ) ( end - start ) ) / ( r64 ) frequency;
    return result;
}

/* MARK date */

typedef struct date_time_t {
    
    u16 year; /* NOTE simon: full year xxxx */
    u8 month; /* NOTE simon: 1 -> 12 */
    u8 day; /* NOTE simon: 1 -> 31 */
    
    u8 hour; /* NOTE simon: 0 -> 23 */
    u8 minute; /* NOTE simon: 0 -> 59 */
    u8 second; /* NOTE simon: 0 -> 59 */
    
} date_time_t;

#define date_time_t_alignment sizeof_field( date_time_t, year )

stu b32 date_is_leap_year( umm year ) {
    b32 result = ( ( year % 4 ) == 0 && ( ( year % 100 ) != 0 || ( year % 400 ) == 0 ) );
    return result;
}

stu u64 date_to_u64( date_time_t* d ) {
    
    u64 offset = 1;
    u64 result = d->second * offset;
    offset *= 100;
    result += d->minute * offset;
    offset *= 100;
    result += d->hour * offset;
    offset *= 100;
    result += d->day * offset;
    offset *= 100;
    result += d->month * offset;
    offset *= 100;
    result += d->year * offset;
    
    return result;
}

stu b32 date_is_after( date_time_t* a, date_time_t* b ) {
    
    u64 au = date_to_u64( a );
    u64 bu = date_to_u64( b );
    b32 result = au > bu;
    
    return result;
}

stu b32 date_is_before( date_time_t* a, date_time_t* b ) {
    
    u64 au = date_to_u64( a );
    u64 bu = date_to_u64( b );
    b32 result = au < bu;
    
    return result;
}

stu b32 date_is_zero( date_time_t* a ) {
    
    b32 result = !a->year && !a->month && !a->day && !a->hour && !a->minute && !a->second;
    return result;
}

stu b32 date_is_same_day( date_time_t* a, date_time_t* b ) {
    
    b32 result = a->year == b->year && a->month == b->month && a->day == b->day;
    return result;
}

stu date_time_t date_add_offset( date_time_t date, smm years, smm months, smm days, smm hours, smm minutes, smm seconds ) {
    
    date_time_t result;
    
    smm s = date.second + seconds;
    
    if ( s < 0 ) {
        
        while ( s < 0 ) {
            minutes -= 1;
            s += 60;
        }
    } else {
        
        while ( s >= 60 ) {
            minutes += 1;
            s -= 60;
        }
    }
    
    smm m = date.minute + minutes;
    
    if ( m < 0 ) {
        
        while ( m < 0 ) {
            hours -= 1;
            m += 60;
        }
    } else {
        
        while ( m >= 60 ) {
            hours += 1;
            m -= 60;
        }
    }
    
    smm h = date.hour + hours;
    
    if ( h < 0 ) {
        
        while ( h < 0 ) {
            days -= 1;
            h += 24;
        }
    } else {
        
        while ( h >= 24 ) {
            days += 1;
            h -= 24;
        }
    }
    
    smm d = date.day;
    
    if ( days < 0 ) {
        
        days *= -1;
        
        while ( days ) {
            
            if ( days < d ) {
                d -= days;
                days = 0;
            } else {
                
                days -= d;
                
                if ( date.month == 1 ) {
                    date.month = 12;
                    date.year -= 1;
                } else {
                    date.month -= 1;
                }
                
                u8 days_in_month[ 12 ] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
                b32 leap_year = date_is_leap_year( date.year );
                days_in_month[ 1 ] += ( u8 ) leap_year;
                d = days_in_month[ date.month - 1 ];
            }
        }
    } else {
        
        while ( days ) {
            
            u8 days_in_month[ 12 ] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
            b32 leap_year = date_is_leap_year( date.year );
            days_in_month[ 1 ] += ( u8 ) leap_year;
            
            if ( d + days < days_in_month[ date.month - 1 ] ) {
                d += days;
                days = 0;
            } else {
                
                days -= days_in_month[ date.month - 1 ] - d;
                d = 1;
                
                if ( date.month == 12 ) {
                    date.month = 1;
                    date.year += 1;
                } else {
                    date.month += 1;
                }
            }
        }
    }
    
    smm mo = date.month;
    
    if ( months < 0 ) {
        
        months *= -1;
        
        while ( months ) {
            
            if ( months < mo ) {
                mo -= months;
                months = 0;
                
                u8 days_in_month[ 12 ] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
                b32 leap_year = date_is_leap_year( date.year );
                days_in_month[ 1 ] += ( u8 ) leap_year;
                
                if ( days_in_month[ mo - 1 ] < d ) {
                    d = days_in_month[ mo - 1 ];
                }
            } else {
                
                months -= mo;
                mo = 12;
                date.year -= 1;
            }
        }
    } else {
        
        while ( months ) {
            
            if ( mo + months <= 12 ) {
                
                mo += months;
                months = 0;
                
                u8 days_in_month[ 12 ] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
                b32 leap_year = date_is_leap_year( date.year );
                days_in_month[ 1 ] += ( u8 ) leap_year;
                
                if ( days_in_month[ mo - 1 ] < d ) {
                    d = days_in_month[ mo - 1 ];
                }
            } else {
                
                months -= 13 - mo;
                mo = 1;
                date.year += 1;
            }
        }
    }
    
    smm y = date.year + years;
    
    if ( y < 0 ) {
        y = 0;
    } else if ( y > u16_max ) {
        y = u16_max;
    }
    
    b32 leap_year = date_is_leap_year( y );
    
    if ( mo == 2 && d == 29 && !leap_year ) {
        d = 28;
    }
    
    result.year = ( u16 ) y;
    result.month = ( u8 ) mo;
    result.day = ( u8 ) d;
    result.hour = ( u8 ) h;
    result.minute = ( u8 ) m;
    result.second = ( u8 ) s;
    
    return result;
}

stu data_t date_to_string_prefix( date_time_t date, data_t* memory ) {
    
    data_t result = memory_construct( memory_left( *memory ), 0, memory_current( *memory, u8 ) );
    
    string_push_u64( memory, date.year, 4, '0', 0 );
    string_push_u64( memory, date.month, 2, '0', 0 );
    string_push_u64( memory, date.day, 2, '0', 0 );
    memory_push_u8( memory, '_' );
    string_push_u64( memory, date.hour, 2, '0', 0 );
    string_push_u64( memory, date.minute, 2, '0', 0 );
    string_push_u64( memory, date.second, 2, '0', 0 );
    
    return result;
}

#if defined( WINDOWS )

/* NOTE simon: GetLocaleInfoEx with LOCALE_SLONGDATE to get the date format. https://msdn.microsoft.com/en-us/library/dd318103(v=vs.85).aspx */
stu data_t date_to_string( date_time_t date, data_t* memory, b32 use_date, b32 use_time, data_t separator ) {
    profiler_event_start( date_to_string );
    
    /* NOTE simon: daylight time is from spring to winter, the clock is advanced by 1 hour. */
    data_t result = { 0 };
    
    SYSTEMTIME utc_date = { 0 };
    utc_date.wYear = date.year;
    utc_date.wMonth = date.month;
    utc_date.wDay = date.day;
    utc_date.wHour = date.hour;
    utc_date.wMinute = date.minute;
    utc_date.wSecond = date.second;
    
    SYSTEMTIME local_time = { 0 };
    
    DYNAMIC_TIME_ZONE_INFORMATION time_zone_info;
    u32 zone_id = GetDynamicTimeZoneInformation( &time_zone_info );
    _assert( zone_id == TIME_ZONE_ID_STANDARD || zone_id == TIME_ZONE_ID_DAYLIGHT );
    b32 success = SystemTimeToTzSpecificLocalTimeEx( &time_zone_info, &utc_date, &local_time );
    
    u16* buffer[ 1024 ] = { 0 };
    _ctassert( sizeof( u16 ) == sizeof( wchar_t ) );
    wchar_t* utf16 = ( LPWSTR ) buffer;
    
    u32 error = 0;
    umm codeunit_count = 0;
    
    if ( use_date ) {
        codeunit_count = GetDateFormatEx( LOCALE_NAME_USER_DEFAULT, DATE_LONGDATE, &local_time, 0, utf16, array_count( buffer ), 0 );
    }
    
    if ( separator.used ) {
        codeunit_count--;
        umm buffer_bytes_left = sizeof( buffer ) - codeunit_count * sizeof( u16 );
        codeunit_count += utf8_to_utf16( separator.bytes, separator.used, utf16 + codeunit_count, buffer_bytes_left, &error );
        utf16 += codeunit_count;
    }
    
    if ( use_time ) {
        codeunit_count += GetTimeFormatEx( LOCALE_NAME_USER_DEFAULT, 0, &local_time, 0, utf16, array_count( buffer ) - ( s32 ) codeunit_count );
    }
    
    if ( codeunit_count ) {
        codeunit_count--;
    }
    
    result.bytes = memory_current( *memory, u8 );
    result.used = utf16_to_utf8( buffer, codeunit_count, result.bytes, memory_left( *memory ), &error );
    result.reserved = result.used;
    memory->used += result.used;
    
    profiler_event_end( date_to_string );
    
    return result;
}

stu date_time_t date_get_current( void ) {
    
    date_time_t result = { 0 };
    
    SYSTEMTIME utc_time;
    GetSystemTime( &utc_time );
    result.year = utc_time.wYear;
    result.month = ( u8 ) utc_time.wMonth;
    result.day = ( u8 ) utc_time.wDay;
    result.hour = ( u8 ) utc_time.wHour;
    result.minute = ( u8 ) utc_time.wMinute;
    result.second = ( u8 ) utc_time.wSecond;
    
    return result;
}

#elif defined( LINUX )

stu data_t date_to_string( date_time_t date, data_t* memory, b32 use_date, b32 use_time, data_t separator ) {
    
    /* NOTE simon: daylight time is from spring to winter, the clock is advanced by 1 hour. */
    data_t result = { 0 };
    
    time_t timestamp = time( 0 );
    struct tm a = *localtime( &timestamp );
    struct tm b = *gmtime( &timestamp );
    time_t at = mktime( &a );
    time_t bt = mktime( &b );
    
    struct tm utc_date = { 0 };
    utc_date.tm_year = date.year - 1900;
    utc_date.tm_mon = date.month - 1;
    utc_date.tm_mday = date.day;
    utc_date.tm_hour = date.hour;
    utc_date.tm_min = date.minute;
    utc_date.tm_sec = date.second + ( s32 ) difftime( at, bt );
    
    /* NOTE simon: instead of adding a difftime we could maybe use timegm on linux, but it's not a standard function. */
    // timegm( &utc_date );
    
    /* NOTE simon: mktime updates the given struct tm members.
    e.g. a value of 65 seconds will add a minute and set seconds to 5. tm_yday and tm_wday will be set to the correct values... */
    mktime( &utc_date );
    
    memory_get_on_stack( date_format, 1024 );
    
    if ( use_date ) {
#define DATE_FORMAT_TEMP "%A %B %e %Y"
        memory_push_copy_p( &date_format, ( void* ) DATE_FORMAT_TEMP, sizeof( DATE_FORMAT_TEMP ) - 1, 1 );
#undef DATE_FORMAT_TEMP
    }
    
    if ( separator.used ) {
        memory_push_copy( &date_format, separator, 1 );
    }
    
    if ( use_time ) {
#define TIME_FORMAT_TEMP "%T"
        memory_push_copy_p( &date_format, ( void* ) TIME_FORMAT_TEMP, sizeof( TIME_FORMAT_TEMP ) - 1, 1 );
#undef TIME_FORMAT_TEMP
    }
    
    string_zero_terminate( date_format );
    
    result.bytes = memory_current( *memory, u8 );
    umm length = strftime( result.text, memory_left( *memory ), date_format.text, &utc_date );
    result.used = length;
    result.reserved = length;
    
    memory->used += length;
    
    return result;
}

stu date_time_t date_get_current( void ) {
    
    date_time_t result = { 0 };
    
    time_t utc_timestamp = time( 0 );
    struct tm* utc_time = gmtime( &utc_timestamp );
    result.year = ( u16 ) ( 1900 + utc_time->tm_year );
    result.month = 1 + ( u8 ) utc_time->tm_mon;
    result.day = ( u8 ) utc_time->tm_mday;
    result.hour = ( u8 ) utc_time->tm_hour;
    result.minute = ( u8 ) utc_time->tm_min;
    result.second = ( u8 ) utc_time->tm_sec;
    
    return result;
}

#else
# error Unsupported platform.
#endif

// MARK log

#if defined( LOG )

typedef struct log_t {
    data_t memory;
    u32 indent_level;
    b32 overflowed;
} log_t;

typedef enum log_name_t {
    log_name_default,
    log_name_gl,
    
    log_name_user
} log_name_t;

log_t log_logs[ 16 ] = { 0 };

_ctassert( log_name_user < array_count( log_logs ) );

// data_t* log_memory = &log_logs[ 0 ].memory;
data_t log_end_message = string_l( "## Log end ##\n" );
data_t log_overflowed_message = string_l( "## Log overflowed ##\n" );

stu log_t* log_get( umm index ) {
    _assert( index < array_count( log_logs ) );
    log_t* result = log_logs + index;
    return result;
}

stu void log_initialize( umm index ) {
    
    log_t* log = log_logs + index;
    
    if ( log->memory.reserved == 0 ) {
        log->memory = memory_get_virtual( mebioctet( 5 ), 0 );
    }
}

stu void log_handle_overflow( umm index, umm length ) {
    
    log_t* log = log_logs + index;
    log_initialize( index );
    
    if ( log->memory.used + length + log_end_message.used >= log->memory.reserved ) {
        log->overflowed = true;
        log->memory.used = 0;
        string_push_length( &log->memory, log_overflowed_message.text, log_overflowed_message.used );
    }
}

stu void log_end( umm index ) {
    log_t* log = log_logs + index;
    string_push_length( &log->memory, log_end_message.text, log_end_message.used );
    log->memory.used -= log_end_message.used;
}

stu void log_increment_indent( umm index ) {
    log_t* log = log_logs + index;
    log->indent_level++;
}

stu void log_decrement_indent( umm index ) {
    log_t* log = log_logs + index;
    log->indent_level--;
}

stu void log_indent( umm index ) {
    
    log_t* log = log_logs + index;
    
    log_handle_overflow( index, log->indent_level * 4 );
    u8* bytes = memory_current( log->memory, u8 );
    u32 space_count = log->indent_level * 4;
    
    for ( umm i = 0; i < space_count; i++ ) {
        *bytes++ = ' ';
    }
    
    log->memory.used += space_count;
    log_end( index );
}

stu umm log_get_position( umm index ) {
    log_t* log = log_logs + index;
    return log->memory.used;
}

stu void log_set_position( umm index, umm position ) {
    
    log_t* log = log_logs + index;
    
    if ( position < log->memory.reserved ) {
        log->memory.used = position;
    } else {
        log->memory.used = log->memory.reserved;
    }
}

stu void log_blank_region( umm index, umm start_position, umm end_position ) {
    
    log_t* log = log_logs + index;
    _assert( start_position <= end_position );
    _assert( end_position < log->memory.reserved );
    u8* current = log->memory.bytes + start_position;
    u8* end = log->memory.bytes + end_position;
    
    while ( current < end ) {
        *current = ' ';
        current++;
    }
}

#define log_l( index, literal ) log_p( index, ( literal ), sizeof( literal ) - 1 )

stu void log_s( umm index, void* string ) {
    log_t* log = log_logs + index;
    log_handle_overflow( index, string_length( string ) );
    string_push_s( &log->memory, string );
    log_end( index );
}

stu void log_p( umm index, void* string, umm length ) {
    log_t* log = log_logs + index;
    log_handle_overflow( index, length );
    string_push_length( &log->memory, string, length );
    log_end( index );
}

stu void log_d( umm index, data_t memory ) {
    log_t* log = log_logs + index;
    log_handle_overflow( index, memory.used );
    string_push_length( &log->memory, memory.text, memory.used );
    log_end( index );
}

stu void log_u64( umm index, u64 number, umm desired_digit_count, u8 fill_with, u32 hexadecimal_format ) {
    
    log_t* log = log_logs + index;
    
    umm length = 20;
    log_handle_overflow( index, length );
    u8* bytes = memory_current( log->memory, u8 );
    length = string_from_u64( bytes, length, number, desired_digit_count, fill_with, hexadecimal_format );
    memory_push_size( &log->memory, length, 1 );
    log_end( index );
}

stu void log_u64_params( umm index, u64 number, b32 indented, void* label, b32 new_line, umm desired_digit_count, u8 fill_with, u32 hexadecimal_format ) {
    
    if ( indented ) {
        log_indent( index );
    }
    
    if ( label ) {
        log_s( index, label );
    }
    
    log_u64( index, number, desired_digit_count, fill_with, hexadecimal_format );
    
    if ( new_line ) {
        log_d( index, string_arg_l( "\n" ) );
    }
}

stu void log_s64( umm index, s64 number, umm desired_digit_count, u8 fill_with, u32 hexadecimal_format ) {
    
    log_t* log = log_logs + index;
    
    umm length = 20;
    log_handle_overflow( index, length );
    u8* bytes = memory_current( log->memory, u8 );
    length = string_from_s64( bytes, length, number, desired_digit_count, fill_with, hexadecimal_format );
    memory_push_size( &log->memory, length, 1 );
    log_end( index );
}

stu void log_s64_params( umm index, s64 number, b32 indented, void* label, b32 new_line, umm desired_digit_count, u8 fill_with, u32 hexadecimal_format ) {
    
    if ( indented ) {
        log_indent( index );
    }
    
    if ( label ) {
        log_s( index, label );
    }
    
    log_s64( index, number, desired_digit_count, fill_with, hexadecimal_format );
    
    if ( new_line ) {
        log_d( index, string_arg_l( "\n" ) );
    }
}

stu void log_r64( umm index, r64 number ) {
    
    log_t* log = log_logs + index;
    
    umm length = 32;
    log_handle_overflow( index, length );
    u8* bytes = memory_current( log->memory, u8 );
    length = string_from_r64( bytes, length, number, 0, 0, 6, 0 );
    memory_push_size( &log->memory, length, 1 );
    log_end( index );
}

stu void log_r64_params( umm index, r64 number, b32 indented, void* label, b32 new_line ) {
    
    if ( indented ) {
        log_indent( index );
    }
    
    if ( label ) {
        log_s( index, label );
    }
    
    log_r64( index, number );
    
    if ( new_line ) {
        log_d( index, string_arg_l( "\n" ) );
    }
}

stu void log_date( umm index, date_time_t* date ) {
    
    date_time_t current;
    
    if ( !date ) {
        current = date_get_current( );
        date = &current;
    }
    
    log_u64( index, date->year, 4, ' ', 0 );
    log_d( index, string_arg_l( "/" ) );
    log_u64( index, date->month, 2, ' ', 0 );
    log_d( index, string_arg_l( "/" ) );
    log_u64( index, date->day, 2, ' ', 0 );
    log_d( index, string_arg_l( " - " ) );
    log_u64( index, date->hour, 2, '0', 0 );
    log_d( index, string_arg_l( ":" ) );
    log_u64( index, date->minute, 2, '0', 0 );
    log_d( index, string_arg_l( ":" ) );
    log_u64( index, date->second, 2, '0', 0 );
    
    log_end( index );
}

stu void log_reset( umm index ) {
    log_t* log = log_logs + index;
    log->memory.used = 0;
}

stu void log_write_to_file( umm index, data_t name ) {
    log_t* log = log_logs + index;
    umm to_write = ( log->overflowed ) ? log->memory.reserved : log->memory.used;
    file_write_entire( name, log->memory.bytes, to_write, true );
}

data_t log_get_buffer( umm index, umm size ) {
    log_t* log = log_logs + index;
    data_t result = memory_get_block( &log->memory, size, 1 );
    return result;
}

/* NOTE simon: This doesn't handle overflow.*/
#define log_format( index, format, ... ) { log_initialize( index ); log_t* log = log_logs + index; string_format( &log->memory, __VA_ARGS__ ); }

#if defined( WINDOWS )

stu void log_windows_error_message( umm index, u32 error ) {
    
    char buffer[ 1024 ];
    u32 length = array_count( buffer ) / sizeof( TCHAR );
    u32 flags = FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS;
    u32 result = FormatMessage( flags, 0, error, 0, buffer, length, 0 ) * sizeof( TCHAR );
    log_p( index, buffer, result );
}

stu void debug_string( data_t string ) {
    
    string_zero_terminate_on_stack( out, 1024, string.bytes, string.used );
    OutputDebugStringA( ( LPCSTR ) out.bytes );
}

#define debug_string_l( string ) debug_string( string_arg_l( string ) )

stu void debug_u64( u64 n, b32 new_line ) {
    
    memory_get_on_stack( out, 128 );
    string_push_u64( &out, n, 0, 0, 0 );
    
    if ( new_line ) {
        memory_push_u8( &out, '\n' );
    }
    
    memory_push_u8( &out, 0 );
    OutputDebugStringA( ( LPCSTR ) out.bytes );
}

stu void debug_s64( s64 n, b32 new_line ) {
    
    memory_get_on_stack( out, 128 );
    string_push_s64( &out, n, 0, 0, 0 );
    
    if ( new_line ) {
        memory_push_u8( &out, '\n' );
    }
    
    memory_push_u8( &out, 0 );
    OutputDebugStringA( ( LPCSTR ) out.bytes );
}

stu void debug_r64( r64 n, b32 new_line ) {
    
    memory_get_on_stack( out, 128 );
    string_push_r64( &out, n, 0, 0, 8, 0 );
    
    if ( new_line ) {
        memory_push_u8( &out, '\n' );
    }
    
    memory_push_u8( &out, 0 );
    OutputDebugStringA( ( LPCSTR ) out.bytes );
}

#else
# define log_windows_error_message( ... )
#endif

#else

#define log_initialize( ... )
#define log_handle_overflow( ... )
#define log_end( ... )
#define log_increment_indent( ... )
#define log_decrement_indent( ... )
#define log_indent( ... )
#define log_get_position( ... ) ( umm ) 0
#define log_set_position( ... )
#define log_blank_region( ... )
#define log_string( ... )
#define log_string_length( ... )
#define log_u32( ... )
#define log_u64( ... )
#define log_s32( ... )
#define log_s64( ... )
#define log_r32( ... )
#define log_reset( ... )
#define log_write_to_file( ... )
#define log_get_buffer( ... ) { 0 }
#define log_floats( ... )
#define log_windows_error_message( ... )

#endif

#endif
